Package: Chromatograms
Title: Infrastructure for Chromatographic Mass Spectrometry Data
Version: 1.2.0
Description: The Chromatograms packages defines an efficient infrastructure
   for storing and handling of chromatographic mass spectrometry data. It
   provides different implementations of *backends* to store and represent the
   data. Such backends can be optimized for small memory footprint or fast
   data access/processing. A lazy evaluation queue and chunk-wise processing
   capabilities ensure efficient analysis of also very large data sets.
Authors@R: c(person(given = "Johannes", family = "Rainer",
		    email = "Johannes.Rainer@eurac.edu",
		    role = c("aut"),
		    comment = c(ORCID = "0000-0002-6977-7147")),
		person(given = "Laurent", family = "Gatto",
		    comment = c(ORCID = "0000-0002-1520-2268"),
		    email = "laurent.gatto@uclouvain.be",
		    role = c("aut")),
	     person(given = "Philippine", family = "Louail",
		    email = "philippine.louail@outlook.com",
		    role = c("aut", "cre"),
		    comment = c(ORCID = "0009-0007-5429-6846",
		    fnd = "European Union HORIZON-MSCA-2021 project
		    Grant No. 101073062: HUMAN – Harmonising and Unifying Blood
		    Metabolic Analysis Networks")))
Depends: BiocParallel, ProtGenerics (>= 1.39.2), R (>= 4.5.0)
Imports: data.table, methods, S4Vectors, MsCoreUtils (>= 1.7.5),
        Spectra
Suggests: msdata (>= 0.19.3), roxygen2, BiocStyle (>= 2.5.19),
        testthat, knitr (>= 1.1.0), rmarkdown, mzR (>= 2.41.4),
        MsBackendMetaboLights (>= 1.3.1), pheatmap, vdiffr, IRanges,
        RColorBrewer
License: Artistic-2.0
Encoding: UTF-8
VignetteBuilder: knitr
BugReports: https://github.com/RforMassSpectrometry/Chromatograms/issues
URL: https://github.com/RforMassSpectrometry/Chromatograms
biocViews: Infrastructure, Metabolomics, MassSpectrometry, Proteomics
Roxygen: list(markdown=TRUE)
RoxygenNote: 7.3.3
Collate: 'AllGenerics.R' 'ChromBackend-functions.R' 'ChromBackend.R'
        'hidden_aliases.R' 'helpers.R' 'ChromBackendMemory.R'
        'ChromBackendMzR.R' 'ChromBackendSpectra.R' 'Chromatograms.R'
        'Chromatograms-chromData.R' 'Chromatograms-peaksData.R'
        'Chromatograms-plot.R' 'Chromatograms-processingQueue.R'
git_url: https://git.bioconductor.org/packages/Chromatograms
git_branch: RELEASE_3_23
git_last_commit: afed21c
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-28 21:13:27 UTC; biocbuild
Author: Johannes Rainer [aut] (ORCID: <https://orcid.org/0000-0002-6977-7147>),
  Laurent Gatto [aut] (ORCID: <https://orcid.org/0000-0002-1520-2268>),
  Philippine Louail [aut, cre] (ORCID:
    <https://orcid.org/0009-0007-5429-6846>, fnd: European Union
    HORIZON-MSCA-2021 project Grant No. 101073062: HUMAN – Harmonising
    and Unifying Blood Metabolic Analysis Networks)
Maintainer: Philippine Louail <philippine.louail@outlook.com>

# Generated by roxygen2: do not edit by hand

export(ChromBackendMemory)
export(ChromBackendMzR)
export(ChromBackendSpectra)
export(chromSpectraIndex)
export(concatenateChromatograms)
export(coreChromVariables)
export(corePeaksVariables)
export(fillCoreChromVariables)
export(matchRtime)
export(plotChromatograms)
export(plotChromatogramsOverlay)
export(validChromData)
export(validPeaksData)
exportClasses(ChromBackend)
exportClasses(ChromBackendMemory)
exportClasses(ChromBackendMzR)
exportClasses(ChromBackendSpectra)
exportClasses(Chromatograms)
exportMethods("$")
exportMethods("$<-")
exportMethods("[")
exportMethods("[[")
exportMethods("[[<-")
exportMethods("chromData<-")
exportMethods("chromIndex<-")
exportMethods("collisionEnergy<-")
exportMethods("dataOrigin<-")
exportMethods("intensity<-")
exportMethods("msLevel<-")
exportMethods("mz<-")
exportMethods("mzMax<-")
exportMethods("mzMin<-")
exportMethods("peaksData<-")
exportMethods("precursorMz<-")
exportMethods("precursorMzMax<-")
exportMethods("precursorMzMin<-")
exportMethods("processingChunkSize<-")
exportMethods("productMz<-")
exportMethods("productMzMax<-")
exportMethods("productMzMin<-")
exportMethods("rtime<-")
exportMethods(Chromatograms)
exportMethods(applyProcessing)
exportMethods(backendBpparam)
exportMethods(backendInitialize)
exportMethods(backendMerge)
exportMethods(backendParallelFactor)
exportMethods(c)
exportMethods(chromData)
exportMethods(chromExtract)
exportMethods(chromIndex)
exportMethods(chromVariables)
exportMethods(collisionEnergy)
exportMethods(compareChromatograms)
exportMethods(dataOrigin)
exportMethods(extractByIndex)
exportMethods(factorize)
exportMethods(filterChromData)
exportMethods(filterEmptyChromatograms)
exportMethods(filterPeaksData)
exportMethods(imputePeaksData)
exportMethods(intensity)
exportMethods(isEmpty)
exportMethods(isReadOnly)
exportMethods(length)
exportMethods(lengths)
exportMethods(msLevel)
exportMethods(mz)
exportMethods(mzMax)
exportMethods(mzMin)
exportMethods(peakBoundary)
exportMethods(peaksData)
exportMethods(precursorMz)
exportMethods(precursorMzMax)
exportMethods(precursorMzMin)
exportMethods(processingChunkFactor)
exportMethods(processingChunkSize)
exportMethods(productMz)
exportMethods(productMzMax)
exportMethods(productMzMin)
exportMethods(reset)
exportMethods(rtime)
exportMethods(setBackend)
exportMethods(show)
exportMethods(split)
exportMethods(supportsSetBackend)
importClassesFrom(ProtGenerics,ProcessingStep)
importClassesFrom(Spectra,Spectra)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpparam)
importFrom(MsCoreUtils,between)
importFrom(MsCoreUtils,i2index)
importFrom(MsCoreUtils,maxi)
importFrom(MsCoreUtils,rbindFill)
importFrom(MsCoreUtils,reduce)
importFrom(MsCoreUtils,sumi)
importFrom(MsCoreUtils,valleys)
importFrom(MsCoreUtils,vapply1c)
importFrom(MsCoreUtils,vapply1l)
importFrom(ProtGenerics,ProcessingStep)
importFrom(Spectra,Spectra)
importFrom(Spectra,concatenateSpectra)
importFrom(Spectra,filterRanges)
importFrom(Spectra,filterRt)
importFrom(Spectra,peaksData)
importFrom(Spectra,rtime)
importFrom(Spectra,spectraData)
importFrom(Spectra,spectraVariables)
importFrom(data.table,rbindlist)
importFrom(grDevices,dev.flush)
importFrom(grDevices,dev.hold)
importFrom(grDevices,n2mfrow)
importFrom(grDevices,xy.coords)
importFrom(graphics,axis)
importFrom(graphics,box)
importFrom(graphics,par)
importFrom(graphics,plot.new)
importFrom(graphics,plot.window)
importFrom(graphics,plot.xy)
importFrom(graphics,title)
importFrom(methods,"slot<-")
importFrom(methods,.valueClassTest)
importFrom(methods,as)
importFrom(methods,callNextMethod)
importFrom(methods,existsMethod)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,validObject)
importFrom(stats,approx)
importFrom(stats,cor)
importFrom(stats,dnorm)
importFrom(stats,filter)
importFrom(stats,loess)
importFrom(stats,predict)
importFrom(stats,quantile)
importFrom(stats,sd)
importFrom(stats,setNames)
importFrom(stats,spline)
importFrom(utils,capture.output)
importFrom(utils,head)
importMethodsFrom(ProtGenerics,"collisionEnergy<-")
importMethodsFrom(ProtGenerics,"dataOrigin<-")
importMethodsFrom(ProtGenerics,"intensity<-")
importMethodsFrom(ProtGenerics,"msLevel<-")
importMethodsFrom(ProtGenerics,"mz<-")
importMethodsFrom(ProtGenerics,"peaksData<-")
importMethodsFrom(ProtGenerics,"precursorMz<-")
importMethodsFrom(ProtGenerics,"processingChunkSize<-")
importMethodsFrom(ProtGenerics,"productMz<-")
importMethodsFrom(ProtGenerics,"rtime<-")
importMethodsFrom(ProtGenerics,addProcessing)
importMethodsFrom(ProtGenerics,applyProcessing)
importMethodsFrom(ProtGenerics,backendBpparam)
importMethodsFrom(ProtGenerics,backendInitialize)
importMethodsFrom(ProtGenerics,backendMerge)
importMethodsFrom(ProtGenerics,backendParallelFactor)
importMethodsFrom(ProtGenerics,collisionEnergy)
importMethodsFrom(ProtGenerics,dataOrigin)
importMethodsFrom(ProtGenerics,extractByIndex)
importMethodsFrom(ProtGenerics,intensity)
importMethodsFrom(ProtGenerics,isReadOnly)
importMethodsFrom(ProtGenerics,msLevel)
importMethodsFrom(ProtGenerics,mz)
importMethodsFrom(ProtGenerics,peaksData)
importMethodsFrom(ProtGenerics,peaksVariables)
importMethodsFrom(ProtGenerics,precursorMz)
importMethodsFrom(ProtGenerics,processingChunkFactor)
importMethodsFrom(ProtGenerics,processingChunkSize)
importMethodsFrom(ProtGenerics,productMz)
importMethodsFrom(ProtGenerics,rtime)
importMethodsFrom(ProtGenerics,setBackend)
importMethodsFrom(ProtGenerics,supportsSetBackend)
importMethodsFrom(S4Vectors,"$")
importMethodsFrom(S4Vectors,"$<-")
importMethodsFrom(S4Vectors,"[")
importMethodsFrom(S4Vectors,"[<-")
importMethodsFrom(S4Vectors,"[[")
importMethodsFrom(S4Vectors,"[[<-")
importMethodsFrom(S4Vectors,isEmpty)
importMethodsFrom(S4Vectors,split)
importMethodsFrom(methods,show)

# Low level infrastructure to handle chromatographic data

[![Project Status: Active – The project has reached a stable, usable state and is being actively developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![R-CMD-check-bioc](https://github.com/RforMassSpectrometry/Chromatograms/workflows/R-CMD-check-bioc/badge.svg)](https://github.com/RforMassSpectrometry/Chromatograms/actions?query=workflow%3AR-CMD-check-bioc)
[![codecov](https://codecov.io/gh/rformassspectrometry/Chromatograms/branch/main/graph/badge.svg?token=jy0Mid9gKn)](https://codecov.io/gh/rformassspectrometry/Chromatograms)
[![license](https://img.shields.io/badge/license-Artistic--2.0-brightgreen.svg)](https://opensource.org/licenses/Artistic-2.0)


This package, part of the [*R for Mass
Spectrometry*](https://www.rformassspectrometry.org/) initiative,
provides support for handling and analysing chromatographic mass spectrometry
(MS) data.

It follows the API that is currenlty available in the
[`Chromatogram`](http://lgatto.github.io/MSnbase/reference/Chromatogram-class.html)
class in the [`MSnbase`](http://lgatto.github.io/MSnbase/index.html) package,
with support for different *backends* to provide/represent the data, similar to
the [`Spectra`](https://rformassspectrometry.github.io/Spectra/) package.


## General concept

A `Chromatograms` object is designed to contain multiple chromatographic data
(i.e. chromatogram entities). The data will be stored linearly, i.e. as a long
list of chromatograms. The `Chromatograms` object will be the main object for
the end user, providing functionality to access, filter or process
chromatographic data, with the actual chromatographic MS data being stored
within *backend* classes. Different implementations of backend classes can be
designed for high performance or low memory footprint.

The existing backend classes are: 

- `ChromBackendMemory`: a memory-based backend, storing the data in
  memory. This is the default backend and is used for testing purposes.
- The `ChromBackendMzR` inherits all slots and methods from the base
  `ChromBackendMemory` backend, providing additional functionality for reading 
  chromatographic data from mzML files.

- `ChromBackendSpectra`: The `ChromBackendSpectra` inherits all slots and
   methods from the base `ChromBackendMemory` backend, providing additional
   functionality for reading chromatographic data from `Spectra` objects.
   
These backend are then handled on a user level by the `Chromatograms` class,
which provides a unified interface to access and manipulate the chromatographic
data. 

## Installation

```
install.packages("BiocManager")
BiocManager::install("Chromatograms")
```

## Contributions

Please the *R for Mass Spectrometry* [code of conduct](https://rformassspectrometry.github.io/RforMassSpectrometry/articles/RforMassSpectrometry.html#code-of-conduct) and [contribution guidelines](https://rformassspectrometry.github.io/RforMassSpectrometry/articles/RforMassSpectrometry.html#contributions).

# `testthat` unit test suites for `ChromBackend`

This folder contains unit test suites that can be called by other packages
defining `ChromBackend` instances to ensure that they are compliant with the
`ChromBackend` definition and that they provide all necessary
methods/functionality.

Different test suites are available within sub-folders of this directory, each
of them having one or multiple *.R* files defining the unit tests. A description
of the unit tests and which variables need to be defined by the calling package
is provided within these.

The unit tests from one test suite can be called from within another package (in
their `testthat.R` file) like shown in the example below that will call all unit
tests from the *test_ChromBackend* test suite/directory. This tests perform all
tests on a backend class assigned to a variable `be`, thus developers should
ensure to have create an instance of their `ChromBackend` implementation and assign
it to a variable `be` before the code lines below.

```
test_suite <- system.file("test_backends", "test_ChromBackend",
    package = "Chromatograms")
test_dir(test_suite, stop_on_failure = TRUE)
```

Note that setting `stop_on_failure = TRUE` is required since the tests would
otherwise silently fail.  Alternatively, it is also possible to run single test
files, e.g. the test file that checks that all spectra variables can be
correctly accessed or set:

```
test_file(file.path(test_suite, "test_peaksData.R"),
    stop_on_failure = TRUE)
```

## Authors/contributors

- Philippine Louail (@philouail)
- Johannes Rainer (@jorainer)

## ----style, echo = FALSE, results = 'asis', message=FALSE---------------------
BiocStyle::markdown()

## ----echo = FALSE, message = FALSE--------------------------------------------
library(Chromatograms)
library(BiocStyle)

## ----message = FALSE----------------------------------------------------------
library(Chromatograms)

#' Definition of the backend class extending ChromBackend
setClass("ChromBackendTest",
    contains = "ChromBackend",
    slots = c(
        chromData = "data.frame",
        peaksData = "list"
    ),
    prototype = prototype(
        chromData = data.frame(),
        peaksData = list()
    )
)

#' Simple constructor function
ChromBackendTest <- function() {
    new("ChromBackendTest")
}

## ----message = FALSE----------------------------------------------------------
#' Basic validation function
setValidity("ChromBackendTest", function(object) {
    if (length(object@peaksData) != nrow(object@chromData)) {
        return(
            "length of 'peaksData' has to match the number of rows of ",
            "'chromData'"
        )
    }
    NULL
})

## -----------------------------------------------------------------------------
#' Create an empty instance of ChromBackendTest
be <- ChromBackendTest()
be

## -----------------------------------------------------------------------------
#' implementation of show for ChromBackendTest
setMethod("show", "ChromBackendTest", function(object) {
    cd <- object@chromData
    cat(class(object), "with", nrow(cd), "chromatograms\n")
})
be

## -----------------------------------------------------------------------------
#' dataStorage method to provide information *where* data is stored
setMethod("dataStorage", "ChromBackendTest", function(object) {
    as.character(object@chromData$dataStorage)
})

## -----------------------------------------------------------------------------
dataStorage(be)

## -----------------------------------------------------------------------------
#' length to provide information on the number of chromatograms
setMethod("length", "ChromBackendTest", function(x) {
    nrow(x@chromData)
})
length(be)

## -----------------------------------------------------------------------------
#' backendInitialize method to fill the backend with data.
setMethod(
    "backendInitialize", "ChromBackendTest",
    function(object, chromData, peaksData) {
        if (!is.data.frame(chromData)) {
            stop(
                "'chromData' needs to be a 'data.frame' with the general",
                "chromatogram variables"
            )
        }
        ## Defining dataStorage and dataOrigin, if not available
        if (is.null(chromData$dataOrigin)) {
            chromData$dataOrigin <- NA_character_
        }
        ## Validate the provided data
        validChromData(chromData)
        validPeaksData(peaksData)
        ## Fill the object with data
        object@chromData <- chromData
        object@peaksData <- peaksData
        object
    }
)

## -----------------------------------------------------------------------------
# A data.frame with chromatogram variables.
cdata <- data.frame(
    msLevel = c(1L, 1L),
    mz = c(112.2, 123.3)
)

# Retention time and intensity values for each chromatogram.
pdata <- list(
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    ),
    data.frame(
        rtime = c(45.1, 46.2),
        intensity = c(100, 80.1)
    )
)

#' Create and initialize the backend
be <- backendInitialize(ChromBackendTest(),
    chromData = cdata, peaksData = pdata
)
be

## -----------------------------------------------------------------------------
#' List core chromatogram variables along with data types.
coreChromVariables()

## -----------------------------------------------------------------------------
#' Accessor for available chromatogram variables
setMethod("chromVariables", "ChromBackendTest", function(object) {
    union(names(object@chromData), names(coreChromVariables()))
})

chromVariables(be)

## -----------------------------------------------------------------------------
#' Get the data.frame with the available chrom variables
be@chromData
#' Complete this data.frame with missing core variables
fillCoreChromVariables(be@chromData)

## -----------------------------------------------------------------------------
#' function to extract the full chromData
setMethod(
    "chromData", "ChromBackendTest",
    function(object, columns = chromVariables(object),
    drop = FALSE) {
        if (!any(chromVariables(object) %in% columns)) {
            stop(
                "Some of the requested Chromatogram variables are not ",
                "available"
            )
        }
        res <- fillCoreChromVariables(object@chromData)
        res <- res[, columns, drop = drop]
        res
    }
)

## -----------------------------------------------------------------------------
#' Extract the full data
chromData(be)

#' Selected variables
chromData(be, c("mz", "msLevel"))

#' Only missing core chromatograms variables
chromData(be, c("collisionEnergy", "mzMin"))

## -----------------------------------------------------------------------------
setMethod("peaksVariables", "ChromBackendTest", function(object) {
    union(names(corePeaksVariables()), names(object@peaksData[[1]]))
})

## -----------------------------------------------------------------------------
peaksVariables(be)

## -----------------------------------------------------------------------------
corePeaksVariables()

## -----------------------------------------------------------------------------
#' method to extract the full chromatographic data as list of arrays
setMethod(
    "peaksData", "ChromBackendTest",
    function(object, columns = peaksVariables(object), drop = FALSE) {
        if (!all(columns %in% peaksVariables(object))) {
            stop("Some of the requested peaks variables are not available")
        }
        res <- lapply(object@peaksData, function(x) x[, columns, drop = drop])
        res
    }
)

## -----------------------------------------------------------------------------
#' Extract the *peaks* data (i.e. intensity and retention times)
peaksData(be)

## -----------------------------------------------------------------------------
#' Main subset method.
setMethod("[", "ChromBackendTest", function(x, i, j, ..., drop = FALSE) {
    i <- MsCoreUtils::i2index(i, length = length(x))
    x@chromData <- x@chromData[i, ]
    x@peaksData <- x@peaksData[i]
    x
})

## -----------------------------------------------------------------------------
a <- be[1]
chromData(a)

## -----------------------------------------------------------------------------
a <- be[c(1, 1, 1)]
chromData(a)

## -----------------------------------------------------------------------------
#' Access a single chromatogram variable
setMethod("$", "ChromBackendTest", function(x, name) {
    if (name %in% union(chromVariables(x), names(coreChromVariables()))) {
        res <- chromData(x, columns = name, drop = TRUE)
    } else if (name %in% peaksVariables(x)) {
        res <- peaksData(x, columns = name, drop = TRUE)
    } else {
        stop("The requested variable '", name, "' is not available")
    }
    res
})

## -----------------------------------------------------------------------------
be$msLevel

## -----------------------------------------------------------------------------
be$precursorMz

## -----------------------------------------------------------------------------
be$intensity

## -----------------------------------------------------------------------------
#' Method allowing to join (concatenate) backends
setMethod("backendMerge", "ChromBackendTest", function(object, ...) {
    res <- object
    object <- unname(c(list(object), list(...)))
    res@peaksData <- do.call(c, lapply(object, function(z) z@peaksData))
    res@chromData <- do.call(
        MsCoreUtils::rbindFill,
        lapply(object, function(z) z@chromData)
    )
    validObject(res)
    res
})

## -----------------------------------------------------------------------------
a <- backendMerge(be, be[2], be)
a

## -----------------------------------------------------------------------------
#' Default for backends:
isReadOnly(be)

## -----------------------------------------------------------------------------
#' Implementation of isReadOnly for ChromBackendTest
setMethod("isReadOnly", "ChromBackendTest", function(object) FALSE)
isReadOnly(be)

## -----------------------------------------------------------------------------
#' Replacement method for the full chromatogram data
setReplaceMethod("chromData", "ChromBackendTest", function(object, value) {
    if (is(value, "DataFrame")) {
        value <- as(value, "data.frame")
    }
    if (!inherits(value, "data.frame")) {
        stop("'value' is expected to be a 'data.frame'")
    }
    if (length(object) && length(object) != nrow(value)) {
        stop("'value' has to be a 'data.frame' with ", length(object), " rows")
    }
    validChromData(value)
    object@chromData <- value
    object
})

## -----------------------------------------------------------------------------
d <- chromData(be)
d$new_col <- c("a", "b")

chromData(be) <- d

## -----------------------------------------------------------------------------
be$new_col

## -----------------------------------------------------------------------------
#' Replace or add a single chromatogram variable.
setReplaceMethod("$", "ChromBackendTest", function(x, name, value) {
    if (length(x) && length(value) != length(x)) {
        stop(
            "length of 'value' needs to match the number of chromatograms ",
            "in object."
        )
    }
    if (name %in% peaksVariables(x)) {
        if (!is.list(value)) {
            stop("The value for peaksData should be a list")
        }
        for (i in seq_along(value)) {
            x@peaksData[[i]][[name]] <- value[[i]]
            validPeaksData(x@peaksData)
        }
    } else {
        x@chromData[, name] <- value
        validChromData(x@chromData)
    }
    x
})

## -----------------------------------------------------------------------------
#' Values before replacement
be$msLevel

#' Replace MS levels
be$msLevel <- c(3L, 2L)

#' Values after replacement
be$msLevel

## -----------------------------------------------------------------------------
#' Add a new chromatogram variable
be$name <- c("A", "B")
be$name

## -----------------------------------------------------------------------------
#' Replace intensity values
be$msLevel3 <- be$msLevel + 3
be$msLevel3

## -----------------------------------------------------------------------------
#' replacement method for peaks data
setReplaceMethod("peaksData", "ChromBackendTest", function(object, value) {
    if (!is.list(value)) {
        stop("'value' is expected to be a list")
    }
    if (length(object) && length(object) != length(value)) {
        stop("'value' has to be a list with ", length(object), " elements")
    }
    validPeaksData(value)
    object@peaksData <- value
    object
})

## -----------------------------------------------------------------------------
#' Create a list with peaks matrices; our backend has 3 chromatograms
#' thus our `list` has to be of length 3
tmp <- list(
    data.frame(
        rtime = c(12.3, 14.4, 15.4, 16.4),
        intensity = c(200, 312, 354.1, 232)
    ),
    data.frame(
        rtime = c(14.4),
        intensity = c(13.4)
    )
)

be_2 <- be
#' Assign this peaks data to one of our test backends
peaksData(be_2) <- tmp

#' Evaluate that we properly added the peaks data
peaksData(be_2)

## ----eval = FALSE-------------------------------------------------------------
# #' Is there a specific way how the object could be best split for
# #' parallel processing?
# setMethod("backendParallelFactor", "ChromBackend", function(object, ...) {
#     factor()
# })

## -----------------------------------------------------------------------------
backendParallelFactor(be)

## ----eval = FALSE-------------------------------------------------------------
# #' get the values for the chromIndex chromatogram variable
# setMethod(
#     "chromIndex", "ChromBackend",
#     function(object, columns = chromVariables(object)) {
#         chromData(object, columns = "chromIndex", drop = TRUE)
#     }
# )

## -----------------------------------------------------------------------------
chromIndex(be)

## ----eval = FALSE-------------------------------------------------------------
# #' get the values for the collisionEnergy chromatogram variable
# setMethod("collisionEnergy", "ChromBackend", function(object) {
#     chromData(object, columns = "collisionEnergy", drop = TRUE)
# })

## -----------------------------------------------------------------------------
collisionEnergy(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default replacement method for collisionEnergy
# setReplaceMethod(
#     "collisionEnergy", "ChromBackend", function(object, value) {
#         object$collisionEnergy <- value
#         object
#     }
# )

## -----------------------------------------------------------------------------
#' Replace the collision energy
collisionEnergy(be) <- c(20, 30)
collisionEnergy(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementation to access dataOrigin
# setMethod("dataOrigin", "ChromBackend", function(object) {
#     chromData(object, columns = "dataOrigin", drop = TRUE)
# })

## -----------------------------------------------------------------------------
#' Access the dataOrigin values
dataOrigin(be)

## -----------------------------------------------------------------------------
#' Default implementation of the `dataOrigin<-` replacement method
setReplaceMethod("dataOrigin", "ChromBackend", function(object, value) {
    object$dataOrigin <- value
    object
})

## -----------------------------------------------------------------------------
#' Replace the backend's dataOrigin values
dataOrigin(be) <- rep("from somewhere", 2)
dataOrigin(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default method to extract intensity values
# setMethod("intensity", "ChromBackend", function(object) {
#     if (length(object)) {
#         peaksData(object, column = "intensity", drop = TRUE)
#     } else {
#         list()
#     }
# })

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementation of the replacement method for intensity values
# setReplaceMethod("intensity", "ChromBackend", function(object, value) {
#     pd <- peaksData(object)
#     if (!is.list(value) || length(pd) != length(value)) {
#         stop("'value' should be a list of the same length as 'object'")
#     }
#     for (i in seq_along(pd)) {
#         if (length(value[[i]]) != nrow(pd[[i]])) {
#             stop(paste0(
#                 "Length of 'value[[", i, "]]' does not match ",
#                 "the number of rows in the intensity of chromatogram: ",
#                 i, "'"
#             ))
#         }
#     }
#     peaksData(object) <- lapply(seq_along(pd), function(i) {
#         pd[[i]]$intensity <- value[[i]]
#         return(pd[[i]])
#     })
#     object
# })

## -----------------------------------------------------------------------------
#' Replace intensity values
intensity(be)[[1]] <- intensity(be)[[1]] + 10
intensity(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementation for `isEmpty()`
# setMethod("isEmpty", "ChromBackend", function(x) {
#     lengths(x) == 0L
# })

## -----------------------------------------------------------------------------
isEmpty(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementation of `isReadOnly()`
# setMethod("isReadOnly", "ChromBackend", function(object) {
#     TRUE
# })

## -----------------------------------------------------------------------------
isReadOnly(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementation for `length()`
# setMethod("length", "ChromBackend", function(x) {
#     nrow(chromData(x, columns = "dataStorage"))
# })

## -----------------------------------------------------------------------------
length(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementation for `lengths()`
# setMethod("lengths", "ChromBackend", function(x) {
#     lengths(intensity(x))
# })

## -----------------------------------------------------------------------------
lengths(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default methods to get or set MS levels
# setMethod("msLevel", "ChromBackend", function(object) {
#     chromData(object, columns = "msLevel", drop = TRUE)
# })
# setReplaceMethod("msLevel", "ChromBackend", function(object, value) {
#     object$msLevel <- value
#     object
# })

## -----------------------------------------------------------------------------
msLevel(be) <- c(1L, 2L)
msLevel(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementations to get or set m/z value(s)
# setMethod("mz", "ChromBackend", function(object) {
#     chromData(object, columns = "mz", drop = TRUE)
# })
# setReplaceMethod("mz", "ChromBackend", function(object, value) {
#     object$mz <- value
#     object
# })

## -----------------------------------------------------------------------------
mz(be) <- c(314.3, 312.5)
mz(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementations to get or set upper m/z limits
# setMethod("mzMax", "ChromBackend", function(object) {
#     chromData(object, columns = "mzMax", drop = TRUE)
# })
# setReplaceMethod("mzMax", "ChromBackend", function(object, value) {
#     object$mzMax <- value
#     object
# })

## -----------------------------------------------------------------------------
mzMax(be) <- mz(be) + 0.01
mzMax(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default methods to get or set the lower m/z boundary
# setMethod("mzMin", "ChromBackend", function(object) {
#     chromData(object, columns = "mzMin", drop = TRUE)
# })
# 
# setReplaceMethod("mzMin", "ChromBackend", function(object, value) {
#     object$mzMin <- value
#     object
# })

## -----------------------------------------------------------------------------
mzMin(be) <- mz(be) - 0.01
mzMin(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementations to get or set the precursorMz chrom variable
# setMethod("precursorMz", "ChromBackend", function(object) {
#     chromData(object, columns = "precursorMz", drop = TRUE)
# })
# setReplaceMethod("precursorMz", "ChromBackend", function(object, value) {
#     object$precursorMz <- value
#     object
# })

## -----------------------------------------------------------------------------
precursorMz(be) <- c(NA_real_, 123.3)
precursorMz(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementations for `precursorMzMax`
# setMethod("precursorMzMax", "ChromBackend", function(object) {
#     chromData(object, columns = "precursorMzMax", drop = FALSE)
# })
# setReplaceMethod("precursorMzMax", "ChromBackend", function(object, value) {
#     object$precursorMzMax <- value
#     object
# })

## -----------------------------------------------------------------------------
precursorMzMax(be) <- precursorMz(be) + 0.1
precursorMzMax(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementations for `precursorMzMin`
# setMethod("precursorMzMin", "ChromBackend", function(object) {
#     chromData(object, columns = "precursorMzMin", drop = FALSE)
# })
# setReplaceMethod("precursorMzMin", "ChromBackend", function(object, value) {
#     object$precursorMzMin <- value
#     object
# })

## -----------------------------------------------------------------------------
precursorMzMin(be) <- precursorMz(be) - 0.1
precursorMzMin(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementations for `productMz`
# setMethod("productMz", "ChromBackend", function(object) {
#     chromData(object, columns = "productMz", drop = TRUE)
# })
# setReplaceMethod("productMz", "ChromBackend", function(object, value) {
#     object$productMz <- value
#     object
# })

## -----------------------------------------------------------------------------
productMz(be) <- c(123.2, NA_real_)
productMz(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementations for `productMzMax`
# setMethod("productMzMax", "ChromBackend", function(object) {
#     chromData(object, columns = "productMzMax", drop = FALSE)
# })
# setReplaceMethod("productMzMax", "ChromBackend", function(object, value) {
#     object$productMzMax <- value
#     object
# })

## -----------------------------------------------------------------------------
productMzMax(be) <- productMz(be) + 0.02
productMzMax(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default implementations for `productMzMin`
# setMethod("productMzMin", "ChromBackend", function(object) {
#     chromData(object, columns = "productMzMin", drop = FALSE)
# })
# setReplaceMethod("productMzMin", "ChromBackend", function(object, value) {
#     object$productMzMin <- value
#     object
# })

## -----------------------------------------------------------------------------
productMzMin(be) <- productMz(be) - 0.2
productMzMin(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default methods for `rtime()` and `rtime<-`
# setMethod("rtime", "ChromBackend", function(object) {
#     if (length(object)) {
#         peaksData(object, column = "rtime", drop = TRUE)
#     } else {
#         list()
#     }
# })
# 
# setReplaceMethod("rtime", "ChromBackend", function(object, value) {
#     pd <- peaksData(object)
#     if (!is.list(value) || length(pd) != length(value)) {
#         stop("'value' should be a list of the same length as 'object'")
#     }
#     for (i in seq_along(pd)) {
#         if (length(value[[i]]) != nrow(pd[[i]])) {
#             stop(paste0(
#                 "Length of 'value[[", i, "]]' does not match ",
#                 "the number of rows in 'the rtime of chromatogram: ", i, "'"
#             ))
#         }
#     }
#     peaksData(object) <- lapply(seq_along(pd), function(i) {
#         pd[[i]]$rtime <- value[[i]]
#         return(pd[[i]])
#     })
#     object
# })

## -----------------------------------------------------------------------------
rtime(be)[[1]] <- rtime(be)[[1]] + 2
rtime(be)

## ----eval = FALSE-------------------------------------------------------------
# #' Default method to split a backend
# setMethod("split", "ChromBackend", function(x, f, drop = FALSE, ...) {
#     split.default(x, f, drop = drop, ...)
# })

## -----------------------------------------------------------------------------
split(be, f = c(1, 2, 1))

## ----si-----------------------------------------------------------------------
sessionInfo()

#' @include Chromatograms.R hidden_aliases.R ChromBackendMzR.R

#' @title Chromatographic peaks data
#'
#' @name peaksData
#'
#' @aliases peaksData
#' @aliases peaksVariables
#' @aliases filterPeaksData
#' @aliases imputePeaksData
#' @aliases peakBoundary
#' @aliases compareChromatograms
#' @aliases matchRtime
#'
#' @description
#'
#' As explained in the [Chromatograms] class documentation, the `Chromatograms`
#' object is a container for chromatographic data that includes chromatographic
#' peaks data (*retention time* and related intensity values, also referred to
#' as *peaks data variables* in the context of `Chromatograms`) and metadata of
#' individual chromatograms (so called *chromatograms variables*).
#'
#' The *peaks data variables* information can be accessed using the
#' `peaksData()` function. It is also possible to access specific peaks
#' variables using `$`.
#'
#' The peaks data can be accessed, replaced but also filtered/subsetted. Refer
#' to the sections below for more details.
#'
#' @param BPPARAM Parallel setup configuration. See [BiocParallel::bpparam()]
#'        for more information.
#'
#' @param columns For `peaksData()`: optional `character` with column
#'        names (peaks variables) that should be included in the
#'        returned `list` of `data.frame`. By default, all columns are returned.
#'        Available variables can be found by calling `peaksVariables()` on the
#'        object.
#'
#' @param drop `logical(1)` For `peaksData()`, default to `FALSE`. If `TRUE`,
#'        and one column is called by the user, the method returns a list of
#'        vector of the single column requested.
#'
#' @param f `factor` defining the grouping to split the `Chromatograms` object.
#'
#' @param keep For `filterPeaksData()`: `logical(1)` defining whether to
#'        keep (`keep = TRUE`) or remove (`keep = FALSE`) the chromatographic
#'        peaks data that match the condition.
#'
#' @param match For `filterPeaksData()` : `character(1) ` defining whether the
#'        condition has to match for all provided `ranges` (`match = "all"`;
#'        the default), or for any of them (`match = "any"`).
#'
#' @param method For `imputePeaksData()`: `character(1)`: Imputation
#'        method ("linear", "spline", "gaussian", "loess").
#'
#' @param FUN For `compareChromatograms()`: `function` to compute the
#'        similarity between two chromatograms from their aligned intensity
#'        vectors (as returned by `MAPFUN`). Must accept two numeric vectors
#'        as the first two arguments and return a single numeric value.
#'        Defaults to [stats::cor()] (Pearson correlation). Additional
#'        arguments can be passed via `...` (e.g., `method = "spearman"` for
#'        [stats::cor()]).
#'
#' @param MAPFUN For `compareChromatograms()`: `function` to align the
#'        retention times of two chromatograms before computing similarity.
#'        Must accept two `data.frame`s (with columns `rtime` and `intensity`)
#'        and return a `list` with elements `x` and `y`: numeric vectors of
#'        equal length containing the aligned intensities of the first and
#'        second chromatogram respectively, interpolated onto a common
#'        retention-time grid. Defaults to [matchRtime()]. Additional
#'        arguments can be passed via `...`.
#'
#' @param minPeaks For `compareChromatograms()`: `integer(1)` (default `4L`).
#'        Minimum number of overlapping retention-time points (as returned by
#'        `MAPFUN`) required to compute a similarity score. Pairs whose
#'        retention-time overlap contains fewer than `minPeaks` points return
#'        `NA` in the score layer; the actual overlap count is still recorded
#'        in the `n_peaks` layer. Setting `minPeaks = 2L` recovers the
#'        behaviour of always computing a score whenever at least two points
#'        overlap.
#'
#' @param labelsColumn For `compareChromatograms()`: optional `character(1)`
#'        giving the name of a chromatogram variable (column in `chromData()`)
#'        whose values should be used as row and column names of the returned
#'        array. The column must contain unique values. If `NULL` (the
#'        default), the array dimensions are unnamed. Only used when `y` is
#'        missing.
#'
#' @param object A [Chromatograms] object.
#'
#' @param ranges For `filterPeaksData()` : a `numeric` vector of paired values
#'        (upper and lower boundary) that define the ranges to filter the
#'        `object`. These paired values need to be in the same order as the
#'        `variables` parameter (see below).
#'
#' @param value For `rtime()` and `intensity()`: `numeric` vector with the
#'        values to replace the current values. The length of the vector must
#'        match the number of peaks data pairs in the `Chromatograms` object.
#'
#' @param variables For `filterPeaksData()`: `character` vector with the names
#'        of the peaks data variables to filter for. The list of available
#'        peaks data variables can be obtained with `peaksVariables()`.
#'
#' @param extrapolate For `imputePeaksData`: `logical(1)` (default `FALSE`).
#'        If `TRUE`, missing values at the beginning and end of a chromatogram
#'        (outside the range of observed values) will be extrapolated. If
#'        `FALSE`, only interpolation is performed and leading/trailing `NA`
#'        values remain `NA`.
#'
#' @param sd For `imputePeaksData`: `numeric(1)`, for the gaussian method:
#'        Standard deviation for Gaussian kernel
#'        (only used if method == "gaussian")
#'
#' @param span For `imputePeaksData`: `numeric(1)`, for the loess method:
#'        Smoothing parameter (only used if method == "loess")
#'
#' @param threshold For `peakBoundary()`: `numeric(1)` (default `0.1`).
#'        Fraction of the peak height above baseline used as a fallback
#'        cut-off when valley-based boundaries are not suitable. Must be
#'        `>= 0` and `< 1`.
#'
#' @param baselineThreshold For `peakBoundary()`: `numeric(1)` (default
#'        `0.1`). Fraction of the peak height above the baseline. Valley
#'        positions returned by `MsCoreUtils::valleys()` are accepted only if
#'        the intensity at the valley is at or below
#'        `baseline + peak_height * baselineThreshold`. Must be `>= 0` and
#'        `< 1`.
#'
#' @param baselineQuantile For `peakBoundary()`: `numeric(1)` (default
#'        `0.1`). Quantile of the intensity distribution used as the
#'        baseline estimate. Must be `>= 0` and `<= 1`.
#'
#' @param window For `imputePeaksData`: `integer`, for the gaussian method:
#'        Half-width of Gaussian kernel window (e.g., 2 gives window size 5)
#'
#' @param tolerance For `matchRtime()`: `numeric(1)` (default `Inf`). Maximum
#'        RT difference between two measured points to be considered a match.
#'        Controls both the overlap detection and the shared RT grid. Lower
#'        values prevent a peak from being compared against a long interpolated
#'        gap in the other chromatogram. Use `Inf` (the default) to consider
#'        all RT points as matching. Can be forwarded via `...` in
#'        `compareChromatograms()`.
#'
#' @param x For `lengths()` and `compareChromatograms()`: A [Chromatograms]
#'        object. For `matchRtime()`: a `data.frame` with columns `rtime` and
#'        `intensity` representing the first chromatogram.
#'
#' @param y For `compareChromatograms()`: A [Chromatograms] object against
#'        which `x` is compared. If missing, each chromatogram in `x` is
#'        compared with each other chromatogram in `x`. For `matchRtime()`:
#'        a `data.frame` with columns `rtime` and `intensity` representing
#'        the second chromatogram.
#'
#' @param ... Additional arguments passed to the method.
#'
#'
#' @section Filter Peaks Variables:
#'
#' Functions that filter a `Chromatograms`'s peaks data (i.e., `@peaksData`).
#' These functions remove peaks data that do not meet the
#' specified conditions. If a chromatogram in a `Chromatograms` object is
#' filtered, only the corresponding peaks variable pairs (i.e., rows) in the
#' `peaksData` are removed, while the chromatogram itself remains in the object.
#'
#' The available functions to filter chromatographic peaks data include:
#'
#' - `filterPeaksData()`: Filters numerical peaks data variables based on the
#'   specified numerical `ranges` parameter. This method returns the same input
#'   `Chromatograms` object, but the filtering step is added to the processing
#'   queue. The filtered data will be reflected when the user accesses
#'   `peaksData`. This function does *not* reduce the number of chromatograms
#'   in the object, but it removes the specified peaks data (e.g., "rtime" and
#'   "intensity" pairs) from the `peaksData`.
#'
#' In the case of a read-only backend, (such as the [ChromBackendMzR]), the
#' replacement of the peaks data is not possible. The peaks data can be
#' filtered, but the filtered data will not be saved in the backend. This means
#' the original mzML files will not be affected by computations performed on
#' the [Chromatograms].
#'
#' @section Impute Peaks Variables:
#'
#' `imputePeaksData` will impute missing values in a `Chromatograms`'s peaks data
#' (i.e., `@peaksData`). This functions replace missing peaks data values with
#' specified imputation methods using various methods such as linear
#' interpolation, spline interpolation, Gaussian kernel smoothing, or LOESS
#' smoothing. This method modifies the peaks data in place and returns the
#' same `Chromatograms` object with imputed values.
#'
#'
#'
#' @section Peak Boundary Detection:
#'
#' `peakBoundary()` determines the retention time boundaries of the tallest
#' peak in each chromatogram. The function uses `MsCoreUtils::valleys()` to
#' locate the valleys (local minima) flanking the apex. If the valley
#' intensities exceed a baseline-relative threshold (controlled by
#' `baselineThreshold`), it falls back to a threshold-based boundary search
#' using `threshold`. The baseline is estimated as the `baselineQuantile`
#' quantile of the chromatogram's intensity values.
#' The result is a `matrix` with one row per
#' chromatogram and columns `left_boundary` and `right_boundary`
#' (retention times). Chromatograms that are empty, have fewer than 3 data
#' points, contain only `NA` or all-zero intensities return `NA` for both
#' boundaries.
#'
#'
#' @section Compare Chromatograms:
#'
#' `compareChromatograms()` compares chromatograms in two steps:
#'
#' 1. **Align** – `MAPFUN` (default [matchRtime()]) maps two chromatograms
#'    onto a common retention-time grid and returns `list(x, y)`, where
#'    `x` and `y` are numeric vectors of equal length containing the aligned
#'    intensities of the first and second chromatogram respectively.
#' 2. **Score** – `FUN` (default [stats::cor()], Pearson correlation)
#'    computes a single similarity value from those aligned intensity vectors.
#'
#' If `y` is missing, each chromatogram in `x` is compared against every
#' other chromatogram in `x`; otherwise, each in `x` is compared with
#' each in `y`.
#'
#' The result is a 3-dimensional numeric array with dimensions
#' `length(x)` x `length(y)` x 2 (or symmetric `n x n x 2` for
#' self-comparison). Layer `[, , 1]` (named `"score"`) contains pairwise
#' similarity scores; layer `[, , 2]` (named `"n_peaks"`) contains the
#' number of overlapping retention-time points used to compute each score.
#' Pairs with fewer overlapping retention-time points than `minPeaks` (default
#' 4) return `NA` in the score layer; the actual overlap count is still
#' recorded in the `n_peaks` layer. The diagonal of a self-comparison is
#' always 1 (score) and the number of data points in that chromatogram
#' (count).
#'
#' `matchRtime()` is the default `MAPFUN`. Given two chromatograms as
#' `data.frame`s with `rtime` and `intensity` columns, it aligns their RT axes
#' and returns a named list with elements `x` and `y`: equal-length intensity
#' vectors evaluated on a shared RT grid, ready for similarity scoring.
#'
#' The alignment works as follows: `matchRtime()` first identifies the RT range
#' where both chromatograms have measured points within `tolerance` of each
#' other (the *overlap*). Within that range, it builds a shared RT grid from
#' all of `x`'s RT points, adding any RT points from `y` that have no close
#' match in `x` (within `tolerance`). Both intensity vectors are then linearly
#' interpolated at grid positions they do not natively cover, using
#' [stats::approx()]. If either chromatogram has fewer than 2 data points, or
#' the two chromatograms do not overlap, empty vectors are returned.
#'
#' The `tolerance` parameter (default `Inf`, meaning all RT points are
#' considered matching) controls the strictness of the matching. Lowering it
#' prevents comparing a measured peak against a long interpolated gap in the
#' other chromatogram. Pass `tolerance` via `...` in `compareChromatograms()`.
#'
#' When `y` is missing, the `labelsColumn` parameter assigns meaningful
#' row/column names to the output from a `chromData()` column (e.g., `"mz"`
#' or a user-defined feature identifier). The column must contain unique
#' values. To compare groups of chromatograms separately, split the object
#' with `split()` beforehand and apply `compareChromatograms()` to each subset.
#'
#' @seealso [Chromatograms] for a general description of the `Chromatograms`
#'          object, and [chromData] for accessing,substituting and filtering
#'          chromatographic variables. For more information on the queuing
#'          of processings and parallelization for larger dataset processing
#'          see [processingQueue].
#'
#' @md
#'
#' @importFrom BiocParallel SerialParam bpparam
#'
#' @author Philippine Louail
#'
#' @return Refer to the individual function description for information on the
#' return value.
#'
#' @examples
#'
#' # Create a Chromatograms object
#' cdata <- data.frame(
#'     msLevel = c(1L, 1L, 1L),
#'     mz = c(112.2, 123.3, 134.4),
#'     dataOrigin = c("mem1", "mem2", "mem3")
#' )
#'
#' pdata <- list(
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     ),
#'     data.frame(
#'         rtime = c(45.1, 46.2),
#'         intensity = c(100, 80.1)
#'     ),
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     )
#' )
#'
#' be <- backendInitialize(new("ChromBackendMemory"),
#'     chromData = cdata,
#'     peaksData = pdata
#' )
#'
#' chr <- Chromatograms(be)
#'
#' # Access peaks data
#' peaksData(chr)
#'
#' # Access specific peaks data variables
#' peaksData(chr, columns = "rtime")
#' rtime(chr)
#'
#' # Replace peaks data
#' rtime(chr)[[1]] <- c(1, 2, 3, 4)
#'
#' # Filter peaks data
#' filterPeaksData(chr, variables = "rtime", ranges = c(12.5, 13.5))
#'
#' # Pairwise similarity: returns a 3D array [i, j, layer]
#' res <- compareChromatograms(chr)
#' res[, , "score"]   ## similarity scores
#' res[, , "n_peaks"] ## number of overlapping RT points
#'
#' ## Use Spearman correlation (passed to cor() via ...)
#' compareChromatograms(chr, method = "spearman")[, , "score"]
#'
#' # Use a chromData column as row/column labels
#' compareChromatograms(chr, labelsColumn = "mz")[, , "score"]
#'
#' # Compare two Chromatograms objects
#' compareChromatograms(chr[1:2], chr[3])
#'
NULL

#' @rdname peaksData
setMethod(
    "imputePeaksData", signature = "Chromatograms",
    function(object, method = c("linear", "spline", "gaussian", "loess"),
             span = 0.3, sd = 1, window = 2, extrapolate = FALSE, ...) {
        method <- match.arg(method)
        object <- addProcessing(object, imputePeaksData,
            method = method, span = span, sd = sd,
            window = window, extrapolate = extrapolate)
        object@processing <- .logging(
            .processing(object),
            "Impute: replace missing peaks data ",
            "using the '", method, "' method",
            if (extrapolate) " (with extrapolation)" else "")
        object
    }
)

#' @rdname peaksData
setMethod(
    "filterPeaksData", signature = "Chromatograms",
    function(object, variables = character(), ranges = numeric(),
             match = c("any", "all"), keep = TRUE) {
        object <- addProcessing(object, filterPeaksData,
            variables = variables, ranges = ranges,
            match = match, keep = keep)
        object@processing <- .logging(
            .processing(object),
            "Filter: remove peaks based ",
            "on the variables: ",
            paste(variables, collapse = ", "),
            "the ranges: ",
            paste(ranges, collapse = ", "),
            "and the match condition: ", match)
        object
    }
)

#' @rdname peaksData
setMethod("intensity", signature = "Chromatograms", function(object, ...) {
  if (!length(.processingQueue(object)))
    return(intensity(.backend(object)))
  peaksData(object, columns = "intensity", drop = TRUE)
})

#' @rdname peaksData
setReplaceMethod(
    "intensity", signature = "Chromatograms",
    function(object, value) {
        if (isReadOnly(.backend(object)))
            stop("Cannot replace peaks data in a read-only backend")
        intensity(object@backend) <- value
        object
    }
)

#' @rdname peaksData
#' @importFrom MsCoreUtils between
setMethod(
    "peaksData", signature = "Chromatograms",
    function(object, columns = peaksVariables(object),
             f = processingChunkFactor(object),
             BPPARAM = bpparam(), drop = FALSE, ...) {
        queue <- .processingQueue(object)
        if (length(queue)) {
            bd <- .run_process_queue(.backend(object),
                queue = queue, f = f, BPPARAM = BPPARAM)
            return(peaksData(bd, columns = columns, drop = drop))
        }
        peaksData(.backend(object), columns = columns, drop = drop)
    }
)

#' @rdname peaksData
setReplaceMethod(
    "peaksData", signature = "Chromatograms",
    function(object, value) {
        if (isReadOnly(.backend(object)))
            stop("Cannot replace peaks data in a read-only backend")
        peaksData(object@backend) <- value
        object
    }
)

#' @rdname peaksData
setMethod("peaksVariables", signature = "Chromatograms", function(object, ...) {
  peaksVariables(.backend(object))
})

#' @rdname peaksData
setMethod("rtime", signature = "Chromatograms", function(object, ...) {
  if (!length(.processingQueue(object)))
    return(rtime(.backend(object)))
  peaksData(object, columns = "rtime", drop = TRUE)
})

#' @rdname peaksData
setReplaceMethod("rtime", signature = "Chromatograms", function(object, value) {
  if (isReadOnly(.backend(object))) {
    stop("Cannot replace peaks data in a read-only backend")
  }
  rtime(object@backend) <- value
  object
})

#' @rdname peaksData
setMethod("lengths", signature = "Chromatograms", function(x) {
    queue <- .processingQueue(x)
    if (length(queue)) {
        f <- processingChunkFactor(x)
        bd <- .run_process_queue(.backend(x),
            queue = queue, f = f, BPPARAM = bpparam())
        return(lengths(bd))
    }
    lengths(.backend(x))
})

#' @rdname peaksData
#' @importFrom stats cor
#' @export
matchRtime <- function(x, y, tolerance = Inf, ...) {
    if (nrow(x) < 2L || nrow(y) < 2L)
        return(list(x = numeric(), y = numeric()))
    x_rt <- x$rtime
    y_rt <- y$rtime
    rx <- range(x_rt)
    ry <- range(y_rt)

    if (is.infinite(tolerance)) {
        keep_x <- x_rt >= max(rx[1L], ry[1L]) & x_rt <= min(rx[2L], ry[2L])
        if (!any(keep_x))
            return(list(x = numeric(), y = numeric()))
        y_out <- approx(y_rt, y$intensity, xout = x_rt[keep_x], rule = 1)$y
        ok <- !is.na(y_out)
        return(list(x = x$intensity[keep_x][ok], y = y_out[ok]))
    }

    if (rx[2L] + tolerance < ry[1L] || ry[2L] + tolerance < rx[1L])
        return(list(x = numeric(), y = numeric()))
    close <- outer(x_rt, y_rt, function(a, b) abs(a - b) <= tolerance)
    x_has_match <- rowSums(close) > 0
    y_has_match <- colSums(close) > 0
    if (!any(x_has_match) || !any(y_has_match))
        return(list(x = numeric(), y = numeric()))
    rt_min <- min(x_rt[x_has_match], y_rt[y_has_match])
    rt_max <- max(x_rt[x_has_match], y_rt[y_has_match])
    if (rt_min >= rt_max)
        return(list(x = numeric(), y = numeric()))
    x_in_mask <- x_rt >= rt_min & x_rt <= rt_max
    y_in_mask <- y_rt >= rt_min & y_rt <= rt_max
    target_rt <- sort(c(x_rt[x_in_mask],
                        y_rt[y_in_mask][colSums(close[x_in_mask, y_in_mask, drop = FALSE]) == 0]))
    x_out <- approx(x_rt, x$intensity, xout = target_rt, rule = 1)$y
    y_out <- approx(y_rt, y$intensity, xout = target_rt, rule = 1)$y
    keep <- !is.na(x_out) & !is.na(y_out)
    list(x = x_out[keep], y = y_out[keep])
}

#' @rdname peaksData
#' @export
setMethod("compareChromatograms",
          signature(x = "Chromatograms", y = "Chromatograms"),
    function(x, y, MAPFUN = matchRtime, FUN = cor, ...,
             minPeaks = 4L, BPPARAM = SerialParam()) {
        nx <- length(x)
        ny <- length(y)
        if (nx == 0L || ny == 0L)
            return(array(numeric(0), dim = c(nx, ny, 2L),
                         dimnames = list(NULL, NULL, c("score", "n_peaks"))))
        .compare_chromatograms(peaksData(x), peaksData(y),
                               MAPFUN = MAPFUN, FUN = FUN,
                               minPeaks = minPeaks, BPPARAM = BPPARAM,
                               self = FALSE, ...)
    }
)

#' @rdname peaksData
#' @export
setMethod("compareChromatograms", signature(x = "Chromatograms", y = "missing"),
    function(x, y = NULL, MAPFUN = matchRtime, FUN = cor, ...,
             minPeaks = 4L, labelsColumn = NULL, BPPARAM = SerialParam()) {
        n <- length(x)
        if (n == 0L)
            return(array(numeric(0), dim = c(0L, 0L, 2L),
                         dimnames = list(NULL, NULL, c("score", "n_peaks"))))
        if (n == 1L) {
            arr1 <- array(NA_real_, dim = c(1L, 1L, 2L),
                          dimnames = list(NULL, NULL, c("score", "n_peaks")))
            arr1[1L, 1L, 1L] <- 1
            arr1[1L, 1L, 2L] <- nrow(peaksData(x)[[1L]])
            return(arr1)
        }
        labs <- .resolve_labels(x, labelsColumn)
        .compare_chromatograms(peaksData(x), MAPFUN = MAPFUN, FUN = FUN,
                               minPeaks = minPeaks, labelsColumn = labs,
                               BPPARAM = BPPARAM, self = TRUE, ...)
    }
)

#' @rdname peaksData
#' @export
setMethod("peakBoundary", signature = "Chromatograms",
    function(object, threshold = 0.1, baselineThreshold = 0.1,
             baselineQuantile = 0.1, ...) {
        if (!is.numeric(threshold) || length(threshold) != 1L ||
            is.na(threshold))
            stop("'threshold' must be a single non-missing numeric value.")
        if (threshold < 0 || threshold >= 1)
            stop("'threshold' must be >= 0 and < 1.")
        if (!is.numeric(baselineThreshold) ||
            length(baselineThreshold) != 1L || is.na(baselineThreshold))
            stop("'baselineThreshold' must be a single non-missing ",
                 "numeric value.")
        if (baselineThreshold < 0 || baselineThreshold >= 1)
            stop("'baselineThreshold' must be >= 0 and < 1.")
        if (!is.numeric(baselineQuantile) ||
            length(baselineQuantile) != 1L || is.na(baselineQuantile))
            stop("'baselineQuantile' must be a single non-missing ",
                 "numeric value.")
        if (baselineQuantile < 0 || baselineQuantile > 1)
            stop("'baselineQuantile' must be >= 0 and <= 1.")
        pd <- peaksData(object)
        res <- vapply(pd, function(p)
            .peak_boundary_one(p$rtime, p$intensity,
                               threshold = threshold,
                               baselineThreshold = baselineThreshold,
                               baselineQuantile = baselineQuantile),
            numeric(2))
        res <- t(res)
        colnames(res) <- c("left_boundary", "right_boundary")
        res
    }
)

#' @title Plot chromatograms
#'
#' @name plotChromatograms
#'
#' @description
#'
#' [Chromatograms()] can be plotted with the following functions:
#'
#' The `plotChromatograms()`: plots each chromatogram in its separate plot by
#' splitting the plot area into as many panels as there are spectra.
#'
#' @param x A [Chromatograms] object.
#'
#' @param xlab `character(1)` with the label for the x-axis (by default
#'        `xlab = "rtime (s)"`).
#'
#' @param ylab `character(1)` with the label for the y-axis (by default
#'        `ylab = "intensity"`).
#'
#' @param type `character(1)` specifying the type of plot. See [plot.default()]
#'        for details. Defaults to `type = "l"` which draws each peak as a line.
#'
#' @param xlim `numeric(2)` defining the x-axis limits. The range of m/z values
#'        are used by default.
#'
#' @param ylim `numeric(2)` defining the y-axis limits. The range of intensity
#'        values are used by default.
#'
#' @param main `character(1)` with the title for the plot. By default the
#'        spectrum's MS level and retention time (in seconds) is used.
#'
#' @param col color to be used to draw the peaks. Should be either of length 1,
#'        or equal to the number of chromatograms (to plot each chromatograms
#'        in a different color) or be a `list` with colors for each individual
#'        peak in each spectrum.
#'
#' @param axes `logical(1)` whether (x and y) axes should be drawn.
#'
#' @param asp `numeric(1)` the aspect ratio of the plot, i.e. the ratio of
#'        the y-axis to the x-axis. Defaults to 1.
#'
#' @param pch `numeric(1)` specifying the symbol to be used for the peaks.
#'        Defaults to 20, a filled circle. See [points()] for details.
#'
#' @param cex `numeric(1)` specifying the character expansion factor for the
#'        peaks. Defaults to 0.6.
#'
#' @param lwd `numeric(1)` specifying the line width for the peaks. Defaults to
#'        1.5.
#'
#' @param frame.plot `logical(1)` whether a box should be drawn around the
#'        plotting area.
#'
#' @param pch `integer(1)` or `character(1)` specifying the plotting symbol
#'       (see [plot.default()]).
#'
#' @param cex `numeric(1)` specifying the size of the plotting symbol (see
#'       [plot.default()]).
#'
#' @param lwd `numeric(1)` specifying the line width (see [plot.default()]).
#'
#' @param ... Additional arguments to be passed to [plot.default()].
#'
#' @return These functions create a plot.
#'
#' @author Philippine Louail, Johannes Rainer.
#'
#' @return Refer to the individual function description for information on the
#' return value.
#'
#' @examples
#'
#' ## Create a Chromatograms object
#' cdata <- data.frame(
#'     msLevel = c(1L, 1L, 1L),
#'     mz = c(112.2, 123.3, 134.4),
#'     chromIndex = c(1L, 2L, 3L)
#' )
#' pdata <- list(
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     ),
#'     data.frame(
#'         rtime = c(45.1, 46.2),
#'         intensity = c(100, 80.1)
#'     ),
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     )
#' )
#' chr <- backendInitialize(ChromBackendMemory(),
#'     chromData = cdata,
#'     peaksData = pdata
#' ) |> Chromatograms()
#'
#' ## Plot one chromatogram
#' plotChromatograms(chr[1])
#'
#' ## Plot the full Chromatograms object
#' plotChromatograms(chr)
#'
#' ## Define a color for each peak in each chromatogram
#' plotChromatograms(chr[1:2], col = c("green", "blue"))
#'
#' ## Overlay all chromatograms
#' plotChromatogramsOverlay(chr[1:2], col = c("green", "blue"))
#'
NULL


#' @rdname plotChromatograms
#' @importFrom graphics par
#' @importFrom grDevices n2mfrow
#' @export
plotChromatograms <- function(x, xlab = "rtime (s)", ylab = "intensity",
    type = "o",
    pch = 20, cex = 0.6, lwd = 1.5,
    xlim = numeric(), ylim = numeric(),
    main = character(), col = "#00000080",
    asp = 1, ...) {
    if (!length(main)) {
        main <- paste0("m/z: ", round(mz(x), 1))
    } # maybe range is better
    nsp <- length(x)
    if (nsp == 1) {
        col <- list(col)
    }
    if (length(col) != nsp) {
        col <- rep(col[1], nsp)
    }
    if (length(main) != nsp) {
        main <- rep(main[1], nsp)
    }
    if (nsp > 1) {
        par(mfrow = n2mfrow(nsp, asp = asp))
    }
    for (i in seq_len(nsp)) {
        .plot_single_chromatogram(x[i],
            xlab = xlab, ylab = ylab, type = type,
            xlim = xlim, ylim = ylim, main = main[i],
            col = col[[i]], pch = pch, cex = cex,
            lwd = lwd, ...
        )
    }
}

#' @rdname plotChromatograms
#' @export
plotChromatogramsOverlay <- function(x,
    xlab = "rtime (s)", ylab = "intensity",
    type = "o",
    pch = 20, cex = 0.6, lwd = 1.5,
    xlim = numeric(),
    ylim = numeric(),
    main = paste(length(x), "chromatograms"),
    col = "#00000080",
    axes = TRUE, frame.plot = axes, ...) {
    nsp <- length(x)
    if (nsp == 1) {
        col <- list(col)
    }
    if (length(col) != nsp) {
        col <- rep(col[1], nsp)
    }
    if (!length(xlim)) {
        xlim <- range(unlist(rtime(x)), na.rm = TRUE)
    }
    if (!length(ylim)) {
        ylim <- c(0, max(unlist(intensity(x)), na.rm = TRUE))
    }
    dev.hold()
    on.exit(dev.flush())
    plot.new()
    plot.window(xlim = xlim, ylim = ylim)
    if (axes) {
        axis(side = 1, ...)
        axis(side = 2, ...)
    }
    if (frame.plot) {
        box(...)
    }
    title(main = main, xlab = xlab, ylab = ylab, ...)
    for (i in seq_len(nsp)) {
        .plot_single_chromatogram(x[i],
            add = TRUE, type = type,
            col = col[[i]], pch = pch, cex = cex,
            lwd = lwd,
            ...
        )
    }
}

#' Run devtools::test_active_file(file = "tests/testthat/test_Chromatograms-plot.R")
#' if make changes to this file.

test_that("plotChromatograms works", {
    vdiffr::expect_doppelganger(
        "plotChromatograms-single",
        function() plotChromatograms(c_full[1])
    )

    vdiffr::expect_doppelganger(
        "plotChromatograms-multiple",
        function() plotChromatograms(c_full)
    )

    vdiffr::expect_doppelganger(
        "plotChromatograms-color",
        function() plotChromatograms(c_full[1:2], col = c("green", "blue"))
    )

    vdiffr::expect_doppelganger(
        "plotChromatograms-one-color",
        function() plotChromatograms(c_full[1:2], col = c("green"))
    )

    vdiffr::expect_doppelganger(
        "plotChromatograms-toomany-color",
        function() plotChromatograms(c_full[1:3], col = c("green", "blue"))
    )

    vdiffr::expect_doppelganger(
        "plotChromatograms-toomany-main",
        function() plotChromatograms(c_full[1:3], main = c("test1", "test2"))
    )

    vdiffr::expect_doppelganger(
        "plotChromatograms-one-title",
        function() plotChromatograms(c_full[1:2], main = "Test Title")
    )

    vdiffr::expect_doppelganger(
        "plotChromatograms-asp05",
        function() plotChromatograms(c_full, asp = 1 / 2)
    )

    vdiffr::expect_doppelganger(
        "plotChromatograms-asp2",
        function() plotChromatograms(c_full, asp = 2)
    )
})

test_that("plotChromatogramsOverlay works", {
    vdiffr::expect_doppelganger(
        "plotChromatogramsOverlay-basic",
        function() plotChromatogramsOverlay(c_full, col = c("red", "blue"))
    )

    vdiffr::expect_doppelganger(
        "plotChromatogramsOverlay-xlim",
        function() plotChromatogramsOverlay(c_full, xlim = c(10, 50))
    )

    vdiffr::expect_doppelganger(
        "plotChromatogramsOverlay-no-axes",
        function() plotChromatogramsOverlay(c_full, axes = FALSE)
    )

    vdiffr::expect_doppelganger(
        "plotChromatogramsOverlay-main-title",
        function() plotChromatogramsOverlay(c_full, main = "Overlay Test")
    )

    vdiffr::expect_doppelganger(
        "plotChromatogramsOverlay-one-sample",
        function() plotChromatogramsOverlay(c_full[1], col = c("red"))
    )
})

test_that(".plot_single_chromatogram works", {
    vdiffr::expect_doppelganger(
        "plot_single_chromatogram-basic",
        function() .plot_single_chromatogram(c_full[1])
    )

    vdiffr::expect_doppelganger(
        "plot_single_chromatogram-xlim",
        function() {
            .plot_single_chromatogram(c_full[1],
                xlim = c(10, 50),
                ylim = c(0, 3000)
            )
        }
    )

    vdiffr::expect_doppelganger(
        "plot_single_chromatogram-color",
        function() .plot_single_chromatogram(c_full[1], col = "purple")
    )

    vdiffr::expect_doppelganger(
        "plot_single_chromatogram-infinite-xlim",
        function() .plot_single_chromatogram(c_full[1], xlim = c(-Inf, Inf))
    )

    vdiffr::expect_doppelganger(
        "plot_single_chromatogram-infinite-ylim",
        function() .plot_single_chromatogram(c_full[1], ylim = c(-Inf, Inf))
    )
})

library(testthat)
library(Chromatograms)
library(Spectra)
library(MsBackendMetaboLights) ## remove MsbackendMetaboLights dependency for test and examples
library(IRanges)
library(S4Vectors)

### Test ChromBackendSpectra
be <- backendInitialize(MsBackendMetaboLights(),
    mtblsId = "MTBLS39",
    filePattern = c("63B.cdf")
)
s <- Spectra(be)
s <- setBackend(s, MsBackendMemory())
be_empty <- new("ChromBackendSpectra")
be <- backendInitialize(be_empty, s)
test_suite <- system.file("test_backends", "test_ChromBackend",
    package = "Chromatograms"
)
test_dir(test_suite, stop_on_failure = TRUE)
be_sp <- be
c_sp <- Chromatograms(be)

### Test ChrombackendMzR
# fetch files
MRM_file <- system.file("proteomics", "MRM-standmix-5.mzML.gz",
    package = "msdata"
)
be_empty <- ChromBackendMzR()
be <- backendInitialize(be_empty, files = MRM_file, BPPARAM = SerialParam())

test_suite <- system.file("test_backends", "test_ChromBackend",
    package = "Chromatograms"
)
test_dir(test_suite, stop_on_failure = TRUE)

be_mzr <- be
c_mzr <- Chromatograms(be)

### Test ChrombackendMemory
cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    mz = c(112.2, 123.3, 134.4),
    dataOrigin = c("mem1", "mem2", "mem3")
)
pdata <- list(
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    ),
    data.frame(
        rtime = c(45.1, 46.2),
        intensity = c(100, 80.1)
    ),
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    )
)

be_empty <- new("ChromBackendMemory")
be_cd <- backendInitialize(be_empty, chromData = cdata)
be <- backendInitialize(be_empty, chromData = cdata, peaksData = pdata)
test_dir(test_suite, stop_on_failure = TRUE)

c_empty <- Chromatograms()
c_full <- Chromatograms(be)

test_check("Chromatograms")

#' @rdname ChromBackend
#'
#' @export
coreChromVariables <- function() .CORE_CHROM_VARIABLES

#' *core* chromatogram variables with expected data type: `integer`, `numeric`,
#' and `character`. Must be a single value.
#'
#' @noRd
.CORE_CHROM_VARIABLES <- c(
    chromIndex = "integer",
    collisionEnergy = "numeric",
    dataOrigin = "character",
    msLevel = "integer",
    mz = "numeric",
    mzMin = "numeric",
    mzMax = "numeric",
    precursorMz = "numeric",
    precursorMzMin = "numeric",
    precursorMzMax = "numeric",
    productMz = "numeric",
    productMzMin = "numeric",
    productMzMax = "numeric"
)

#' @title Fill data.frame with columns for missing core chromatogram variables.
#'
#' @description
#'
#' `fillCoreChromVariables()` fills a provided `data.frame`
#' with columns for eventually missing *core* chromatogram variables.
#' The missing core variables are added as new columns with missing values
#' (`NA`) of the correct data type.
#' Use [coreChromVariables()] to list the set of core variables and their data
#' types.
#'
#' @param x `data.frame` with potentially present core chrom variable columns
#'
#' @return input data frame `x` with missing core variables added (with the
#'     correct data type).
#'
#' @importFrom methods as
#'
#' @export
#'
#' @examples
#'
#' ## Define a data frame
#' a <- data.frame(msLevel = c(1L, 1L, 2L), other_column = "b")
#'
#' ## Add missing core chromatogram variables to this data frame
#' fillCoreChromVariables(a)
#'
#' ## The data.frame thus contains columns for all core chromatogram
#' ## variables in the respective expected data type (but filled with
#' ## missing values).
#'
#' @rdname hidden_aliases
fillCoreChromVariables <- function(x = data.frame()) {
    nr <- nrow(x)
    cv <- .CORE_CHROM_VARIABLES
    miss <- cv[setdiff(names(cv), colnames(x))]
    if (!length(miss)) {
        return(x)
    }
    cbind(x, lapply(miss, function(z, n) rep(as(NA, z), n), nr))
}

#' @title Check core chromatogram variables for correct data types
#'
#' @description
#'
#' `validChromData()` checks that columns, representing *core* chromatogram
#' variables are of the correct data type.
#'
#' @param x `data.frame` representing metadata of a `Chromatograms`
#'
#' @param error `logical(1)` whether an error should be thrown (the default)
#'     if one or more columns don't have the correct data type.
#'
#' @return
#'
#' If the core variables have all the correct data type: an empty character.
#' If one or more core variables (columns) have the wrong data type the
#' function either throws an error (with `error = TRUE`) or returns a
#' `character` specifying which variables/columns don't have the correct
#' type (for `error = FALSE`).
#'
#' @importFrom methods is
#'
#' @export
#' @rdname hidden_aliases
validChromData <- function(x = data.frame(), error = TRUE) {
    cn <- intersect(colnames(x), names(.CORE_CHROM_VARIABLES))
    msg <- unlist(lapply(cn, function(z) {
        if (!is(x[, z], .CORE_CHROM_VARIABLES[z])) {
            paste0("Column \"", z, "\" has the wrong data type. ")
        } else {
            NULL
        }
    }), use.names = FALSE)
    if (length(msg) && error) {
        stop(msg)
    } else {
        msg
    }
}

#' *core* peaks variables with expected data type:`numeric`, and
#' must be plural.
#'
#' @noRd
.CORE_PEAKS_VARIABLES <- c(
    rtime = "numeric",
    intensity = "numeric"
)

#' An empty peaks data.frame
#' @noRd
.EMPTY_PEAKS_DATA <- as.data.frame(lapply(
    .CORE_PEAKS_VARIABLES,
    function(x) vector(x, 0)
))

#' @rdname ChromBackend
#'
#' @export
corePeaksVariables <- function() .CORE_PEAKS_VARIABLES

#' `validPeaksData()` checks that the names of the input peaksData list,
#' representing *core* peaks variables are of the correct data type.
#'
#' @param x `list` representing the peaks data of a `Chromatograms`
#'
#' @importFrom methods is
#'
#' @export

#' @rdname hidden_aliases
validPeaksData <- function(x = list(), error = TRUE) {
    if (!is.list(x)) stop("'peaksData' must be a 'list'")
    if (!length(x)) {
        return(NULL)
    }
    first_cols <- colnames(x[[1]])
    expected_cols <- names(.CORE_PEAKS_VARIABLES)
    expected_types <- .CORE_PEAKS_VARIABLES
    msgs <- unlist(lapply(seq_along(x), function(i) {
        df <- x[[i]]
        if (!identical(colnames(df), first_cols)) {
            return(paste(
                "All data.frames must have the same columns in the",
                " same order. Issue found in entry", i
            ))
        }
        .validate_entry(x[[i]], i, expected_cols, expected_types)
    }))
    if (length(msgs) && error) {
        stop(msgs)
    } else {
        msgs
    }
}

test_that("ChromBackendMemory function works", {
    expect_true(is.function(ChromBackendMemory))
    expect_true(is(ChromBackendMemory(), "ChromBackendMemory"))
    expect_true(supportsSetBackend(ChromBackendMemory()))
})

test_that("backendInitialize, ChromBackendMemory works", {
    expect_false(isReadOnly(be_empty))
    expect_error(
        backendInitialize(be_empty, chromData = matrix()),
        "needs to be a"
    )
    fill_test <- backendInitialize(be_empty, chromData = data.frame())
    expect_equal(.chromData(fill_test), fillCoreChromVariables(data.frame()))

    ### empty backend
    expect_true(is(be_empty, "ChromBackendMemory"))
    expect_true(is(.chromData(be_empty), "data.frame"))
    expect_true(nrow(.chromData(be_empty)) == 0)
    expect_true(all(names(.chromData(be_empty)) ==
                        names(.CORE_CHROM_VARIABLES)))
    expect_true(is.null(validChromData(.chromData(be_empty))))
    expect_true(is(.peaksData(be_empty), "list"))
    expect_true(length(.peaksData(be_empty)) == 1)
    expect_true(all(names(.peaksData(be_empty)[[1]]) ==
                        names(.CORE_PEAKS_VARIABLES)))
    expect_true(is.null(unlist(validChromData(.peaksData(be_empty)))))

    ### empty peaksData
    expect_true(is(be_cd, "ChromBackendMemory"))
    expect_true(is(.chromData(be_cd), "data.frame"))
    expect_true(nrow(.chromData(be_cd)) == nrow(cdata))
    expect_true(all(names(.chromData(be_cd)) %in% c(names(cdata),
                                                    "dataOrigin")))
    expect_true(is.null(validChromData(.chromData(be_cd))))
    expect_true(all(be_cd$dataOrigin == c("mem1", "mem2", "mem3")))

    expect_true(is(.peaksData(be_cd), "list"))
    expect_true(length(.peaksData(be_cd)) == nrow(.chromData(be_cd)))
    expect_true(all(vapply(.peaksData(be_cd), nrow, integer(1)) == 0))
    expect_true(all(vapply(.peaksData(be_cd), is.data.frame, logical(1))))
    expect_true(all(names(.peaksData(be_cd)[[1]]) %in%
                        names(.CORE_PEAKS_VARIABLES)))
    expect_true(is.null(unlist(validPeaksData(.peaksData(be_cd)))))

    ### full backend
    expect_true(is(be, "ChromBackendMemory"))
    expect_true(is(.chromData(be), "data.frame"))
    expect_true(nrow(.chromData(be)) == nrow(cdata))
    expect_true(all(names(.chromData(be)) %in% c(names(cdata), "dataOrigin")))
    expect_true(is.null(validChromData(.chromData(be))))

    expect_true(is(.peaksData(be), "list"))
    expect_true(length(.peaksData(be)) == nrow(.chromData(be)))
    expect_true(all(vapply(.peaksData(be), nrow, integer(1)) == nrow(pdata)))
    expect_true(all(vapply(.peaksData(be), is.data.frame, logical(1))))
    expect_true(all(names(.peaksData(be)[[1]]) %in%
                        names(.CORE_PEAKS_VARIABLES)))
    expect_true(is.null(unlist(validPeaksData(.peaksData(be)))))
})

test_that("backend removes NA columns", {
    chromData <- chromData(be)
    chromData$NAcol <- NA
    be2 <- backendInitialize(be, chromData)
    expect_false("NAcol" %in% names(.chromData(be2)))
})

test_that("backendMerge, ChromBackendMemory works", {
    be_merge <- backendMerge(c(be_cd, be))
    expect_true(is(be_merge, "ChromBackendMemory"))
    expect_true(is(.chromData(be_merge), "data.frame"))
    expect_true(is(.peaksData(be_merge), "list"))
    expect_true(nrow(.chromData(be_merge)) ==
                    nrow(.chromData(be_cd)) + nrow(.chromData(be)))
    expect_true(length(.peaksData(be_merge))
                == length(.peaksData(be_cd)) + length(.peaksData(be)))
    expect_equal(backendMerge(c(be_empty, be_empty)), be_empty)
})

test_that("show,ChromBackendMemory works", {
    expect_output(show(be_empty), "ChromBackendMemory")
    expect_output(show(be_cd), "3 chromatograms")
    expect_output(show(be_cd), "chromIndex")
    expect_output(show(be_cd), "msLevel")
    expect_output(show(be_cd), "mz")
})

test_that("chromData", {
    expect_error(chromData(be, "not a column"), "variables are not")
    expect_error(chromData(be) <- data.frame(), " rows")
    expect_error(chromData(be) <- matrix(), "is expected to be a 'data.frame'")
})

test_that("peaksData", {
    expect_error(peaksData(be, "not a column"), "variables are not")
    expect_error(peaksData(be) <- list(), " elements")
    expect_error(peaksData(be) <- matrix(), "is expected to be a list")
})

test_that("$<-,ChromBackendMemory works", {
    expect_error(
        be_cd$intensity <- c(100, 200),
        "length of 'value' needs to match the number of chromatograms in object."
    )

    new_peaks_intensity <- list(
        c(100, 200, 300, 400),
        c(50, 60),
        c(150, 250, 350, 450)
    )
    be$intensity <- new_peaks_intensity
    expect_equal(.peaksData(be)[[1]]$intensity, new_peaks_intensity[[1]])
    expect_equal(.peaksData(be)[[2]]$intensity, new_peaks_intensity[[2]])
    expect_equal(.peaksData(be)[[3]]$intensity, new_peaks_intensity[[3]])

    expect_error(
        be$intensity <- c(100, 200, 300),
        "The value for peaksData should be a list"
    )

    be_cd$new_var <- c("A", "B", "C")
    expect_equal(.chromData(be_cd)$new_var, c("A", "B", "C"))

    be_cd$mz <- c(111.1, 222.2, 333.3)
    expect_equal(.chromData(be_cd)$mz, c(111.1, 222.2, 333.3))
})

test_that("filterChromData works", {
    tmp <- be_cd
    tmp$new_var <- c(1, 2, 3)
    res <- filterChromData(tmp,
        variables = c("mz", "new_var"),
        ranges = c(134, 150, 1, 1),
        match = "any",
        keep = FALSE
    )

    expect_equal(nrow(chromData(res)), 1)
    expect_equal(chromData(res)$mz, 123.3)
    expect_error(
        filterChromData(tmp,
            variables = c("mz"),
            ranges = c(134), # Wrong length
            match = "any",
            keep = TRUE
        ),
        "needs to be twice the length"
    )
    expect_error(
        filterChromData(tmp,
            variables = c("invalid_var"),
            ranges = c(134, 150, 1, 1),
            match = "any",
            keep = TRUE
        ),
        "One or more values passed"
    )
    res <- filterChromData(tmp,
        variables = c("mz"),
        ranges = c(134, 150),
        match = "all",
        keep = FALSE
    )

    expect_equal(nrow(chromData(res)), 2)
    expect_equal(chromData(res)$mz, c(112.2, 123.3))

    expect_identical(
        filterChromData(tmp,
            match = "all",
            keep = TRUE
        ),
        tmp
    )
    expect_error(
        filterChromData(tmp,
            variables = c("mz"),
            ranges = c("a", "b"),
            keep = TRUE
        ),
        "only support filtering for numerical"
    )
    expect_error(
        filterChromData(tmp,
            variables = c(200),
            ranges = c(134, 150),
            keep = TRUE
        ),
        "parameter needs to be of type"
    )
    res <- filterChromData(tmp,
        variables = c("mz", "new_var"),
        ranges = c(134, 150, 1, 1),
        match = "any",
        keep = TRUE
    )
    expect_equal(nrow(chromData(res)), 2)
    expect_equal(chromData(res)$mz, c(112.2, 134.4))
    expect_equal(chromData(res)$new_var, c(1, 3))

    res <- filterChromData(tmp,
        variables = c("mz"),
        ranges = c(999, 1000), # No matching range
        match = "any",
        keep = FALSE
    )

    expect_identical(res, tmp)
})


test_that("split,ChrombackendMemory works", {
    be_split <- be
    be_split$new_vars <- c("a", "b", "b")
    tmp <- be
    tmp$newvar2s <- c(1, 2, 3)
    f <- factor(be_split$new_vars)
    res <- split(tmp, f)
    expect_true(is.list(res))
    expect_true(length(res) == 2)
    expect_s4_class(res[[1L]], "ChromBackendMemory")
    expect_s4_class(res[[2L]], "ChromBackendMemory")
    expect_equal(res[[1L]]$newvar2s, c(1))
    expect_equal(res[[2L]]$newvar2s, c(2, 3))

    res <- split(tmp, factor(be_split$new_vars, levels = c("b", "a")))
    expect_true(is.list(res))
    expect_true(length(res) == 2)
    expect_s4_class(res[[1L]], "ChromBackendMemory")
    expect_s4_class(res[[2L]], "ChromBackendMemory")
    expect_equal(res[[2L]]$newvar2s, c(1))
    expect_equal(res[[1L]]$newvar2s, c(2, 3))
})


test_that("chromExtract works correctly for ChromBackendMemory", {
    peak_tbl <- data.frame(
        rtMin = c(12.3, 45.0, 12.5),
        rtMax = c(13.5, 46.3, 14.8),
        msLevel = c(1L, 1L, 1L),
        dataOrigin = c("mem1", "mem2", "mem3"),
        extra_cols = c(1, 2, 3)
    )

    out <- chromExtract(be, peak_tbl, by = c("msLevel", "dataOrigin"))

    expect_s4_class(out, "ChromBackendMemory")
    expect_true(validObject(out))

    ## chromData should have same number of rows as peak.table
    expect_equal(nrow(chromData(out)), nrow(peak_tbl))

    ## peaksData should be filtered per retention window
    pd_out <- peaksData(out)
    expect_type(pd_out, "list")
    expect_length(pd_out, 3)

    ## Check retention time filtering on first chromatogram
    rt1 <- pdata[[1]]$rtime
    in_range <- rt1 >= peak_tbl$rtMin[1] & rt1 <= peak_tbl$rtMax[1]
    expect_equal(pd_out[[1]]$rtime, rt1[in_range])

    ## intensities align with filtered rt
    expect_equal(length(pd_out[[1]]$rtime), length(pd_out[[1]]$intensity))

    ## Ensure no chromatograms missing
    expect_false(any(vapply(pd_out, nrow, numeric(1)) == 0))

    ## edge case: no match
    bad_tbl <- data.frame(
        rtMin = 1, rtMax = 2,
        msLevel = 2L,
        dataOrigin = "nonexistent"
    )
    expect_warning_or_error <- function(expr) {
        expect_error(expr, regexp = "do not exist|must be present",
                     fixed = FALSE)
    }
    expect_warning_or_error(chromExtract(be, bad_tbl,
                                         by = c("msLevel", "dataOrigin")))

})

test_that("chromExtract handles empty peaksData gracefully", {
    cdata <- data.frame(
        msLevel = c(1L, 1L),
        mz = c(100.1, 110.2),
        dataOrigin = c("x", "y")
    )
    be <- backendInitialize(new("ChromBackendMemory"), chromData = cdata)

    peak_tbl <- data.frame(
        rtMin = c(1, 2),
        rtMax = c(3, 4),
        msLevel = c(1L, 1L),
        dataOrigin = c("x", "y")
    )

    out <- chromExtract(be, peak_tbl, by = c("msLevel", "dataOrigin"))
    expect_s4_class(out, "ChromBackendMemory")
    expect_equal(nrow(chromData(out)), 2L)
    expect_equal(length(peaksData(out)), 2L)
    expect_true(all(vapply(peaksData(out), nrow, numeric(1)) == 0))
})

test_that("imputePeaksData() works correctly for ChromBackendMemory", {
    pdata <- list(
        data.frame(
            rtime = seq(12, 20, by = 0.5),
            intensity = c(123.3, 153.6, NA, NA, NA, 230, NA, 250, 260, 280,
                          300, 320, 340, 360, 380, 400, 420)
        ),
        data.frame(
            rtime = seq(45, 55, by = 0.5),  # length 21
            intensity = c(100, NA, 120, 130, 140, NA, NA, 170, 180, 190, 200,
                          210, 220, 230, 240, 250, 260, 270, 280, 290, 300)
        ),
        data.frame(
            rtime = seq(10, 18, by = 0.5),  # length 17
            intensity = c(NA, 153.6, 2354.3, 243.4, NA, NA, 280, 300, 320,
                          340, 360, 380, 400, 420, 440, NA, NA)
        )
    )

    tmp <- backendInitialize(be_empty, chromData = cdata, peaksData = pdata)

    expect_true(any(is.na(unlist(intensity(tmp)))))

    ## check each method
    methods <- c("linear", "spline")
    for (m in methods) {
        expect_silent({
            be_imp <- imputePeaksData(tmp, method = m, extrapolate = TRUE)
        })

        ## No more NAs
        expect_false(any(is.na(unlist(intensity(be_imp)))),
                     info = paste("NA remains after", m, "imputation"))
        ## Structure unchanged
        expect_equal(length(be_imp), length(tmp))
        expect_equal(names(chromData(be_imp)), names(chromData(be)))
    }

    ## checking Gaussian specifically as it should give a warning.
    expect_warning(be_imp <- imputePeaksData(tmp, method = "gaussian",
                                             extrapolate = TRUE),
                   "could not fill all NAs")
    ## No more NAs
    expect_false(any(is.na(unlist(intensity(be_imp)))),
                 info = paste("NA remains after", m, "imputation"))
    ## Structure unchanged
    expect_equal(length(be_imp), length(tmp))
    expect_equal(names(chromData(be_imp)), names(chromData(be)))

    ## check loess specifically as it should give a warning.
    expect_warning(be_imp <- imputePeaksData(tmp, method = "loess",
                                             extrapolate = TRUE),
                   "Falling back to linear interpolation")
    ## No more NAs
    expect_false(any(is.na(unlist(intensity(be_imp)))),
                 info = paste("NA remains after", m, "imputation"))
    ## Structure unchanged
    expect_equal(length(be_imp), length(tmp))
    expect_equal(names(chromData(be_imp)), names(chromData(be)))

    ## Test extrapolate = FALSE (default): leading/trailing NAs should remain NA
    ## but interior NAs should be interpolated
    pdata_edge <- list(
        data.frame(
            rtime = 1:10,
            intensity = c(NA, NA, 100, NA, 140, 160, 180, NA, NA, NA)
        )
    )
    tmp_edge <- backendInitialize(be_empty, chromData = cdata[1, , drop = FALSE],
                                  peaksData = pdata_edge)

    ## extrapolate = FALSE (default) - leading/trailing NAs remain, interior filled
    be_imp_interp <- imputePeaksData(tmp_edge, method = "linear")
    ints <- intensity(be_imp_interp)[[1]]
    # leading NAs should remain NA
    expect_true(is.na(ints[1]))
    expect_true(is.na(ints[2]))
    # trailing NAs should remain NA
    expect_true(is.na(ints[8]))
    expect_true(is.na(ints[9]))
    expect_true(is.na(ints[10]))
    # interior NA at position 4 should be interpolated (no longer NA)
    expect_false(is.na(ints[4]))
    # all interior values (positions 3-7) should have no NAs
    expect_false(any(is.na(ints[3:7])))

    ## extrapolate = TRUE - no NAs remain
    be_imp_extrap <- imputePeaksData(tmp_edge, method = "linear",
                                     extrapolate = TRUE)
    expect_false(any(is.na(intensity(be_imp_extrap)[[1]])))

    ## Empty backend
    expect_silent({
        be_cd_imp <- imputePeaksData(be_empty, method = "linear")
    })
    expect_equal(length(be_cd_imp), 0L)
})

test_that("ChromBackend methods throw errors", {
    setClass("DummyBackend",
        contains = "ChromBackend"
    )
    dm <- new("DummyBackend")

    expect_error(dm[1], "Not implemented for ")
    expect_error(dm$a, "Not implemented for ")
    expect_error(dm$a <- "a", "Not implemented for ")
    expect_error(backendMerge(dm), "Not implemented for ")
    expect_error(chromData(dm), "Not implemented for ")
    expect_error(chromData(dm) <- data.frame(), "Not implemented for ")
    expect_error(peaksData(dm), "Not implemented for ")
    expect_error(peaksData(dm) <- list(), "Not implemented for ")
    expect_true(!isReadOnly(dm))
    expect_equal(backendParallelFactor(dm), factor())

    expect_true(is(reset(dm), class(dm)))

    # accessor function with default but dependent on chromData() and
    # therefore not implemented
    expect_true(is(backendInitialize(dm), class(dm)))
    expect_error(chromIndex(dm), "Not implemented for ")
    expect_error(chromIndex(dm) <- 1, "Not implemented for ")
    expect_error(collisionEnergy(dm), "Not implemented for ")
    expect_error(collisionEnergy(dm) <- 1, "Not implemented for ")
    expect_error(dataOrigin(dm), "Not implemented for ")
    expect_error(dataOrigin(dm) <- "a", "Not implemented for ")
    expect_error(msLevel(dm), "Not implemented for ")
    expect_error(msLevel(dm) <- 1, "Not implemented for ")
    expect_error(mz(dm), "Not implemented for ")
    expect_error(mz(dm) <- 1, "Not implemented for ")
    expect_error(mzMax(dm), "Not implemented for ")
    expect_error(mzMax(dm) <- 1, "Not implemented for ")
    expect_error(mzMin(dm), "Not implemented for ")
    expect_error(mzMin(dm) <- 1, "Not implemented for ")
    expect_error(precursorMz(dm), "Not implemented for ")
    expect_error(precursorMz(dm) <- 1, "Not implemented for ")
    expect_error(precursorMzMax(dm), "Not implemented for ")
    expect_error(precursorMzMax(dm) <- 1, "Not implemented for ")
    expect_error(precursorMzMin(dm), "Not implemented for ")
    expect_error(precursorMzMin(dm) <- 1, "Not implemented for ")
    expect_error(productMz(dm), "Not implemented for ")
    expect_error(productMz(dm) <- 1, "Not implemented for ")
    expect_error(productMzMax(dm), "Not implemented for ")
    expect_error(productMzMax(dm) <- 1, "Not implemented for ")
    expect_error(productMzMin(dm), "Not implemented for ")
    expect_error(productMzMin(dm) <- 1, "Not implemented for ")
    expect_error(rtime(dm), "Not implemented for ")
    expect_error(rtime(dm) <- c(1, 1, 1), "Not implemented for ")
    expect_error(factorize(dm), "Not implemented for ")
    expect_error(extractByIndex(dm, 1), "not implemented for ")
    expect_error(chromExtract(dm, data.frame(), "dataorigin"),
                 "Not implemented for ")
    expect_error(imputePeaksData(dm), "Not implemented for ")


    expect_identical(chromVariables(dm), names(coreChromVariables()))
    expect_identical(peaksVariables(dm), names(corePeaksVariables()))
    expect_identical(extractByIndex(dm), dm)

    expect_error(dm[[123]], "is supposed to be a character")
    expect_error(dm[["character", "character2"]], "is not supported")
    expect_error(dm[[123]] <- 123, "is supposed to be a character")
    expect_error(dm[["character", "character2"]] <- 123, "is not supported")
    expect_identical(filterPeaksData(dm, ranges = c()), dm)
    expect_identical(filterPeaksData(dm, ranges = c(1, 2), variables = c()), dm)
    expect_error(
        filterPeaksData(dm, ranges = c("a", "3"), variables = c("a")),
        "only support"
    )
    expect_error(
        filterPeaksData(dm, ranges = c(1, 2), variables = c("a")),
        "One or more"
    )
    expect_error(
        filterPeaksData(dm, ranges = c(1, 2), variables = c(33)),
        "needs to be of type"
    )
    expect_error(
        filterPeaksData(dm,
            ranges = c(1, 2),
            variables = c("intensity", "rtime")
        ),
        "needs to be twice"
    )
    expect_false(supportsSetBackend(dm))

})

test_that("Chromatograms, chromData, chromData<- works", {
    expect_equal(chromData(c_full), fillCoreChromVariables(.chromData(c_full)))
    cdata2 <- cdata
    cdata2$mz <- cdata2$mz + 1
    res <- c_full
    chromData(res) <- cdata2
    expect_equal(chromData(res), fillCoreChromVariables(cdata2))
    expect_true(all(chromData(res)$mz != fillCoreChromVariables(cdata)$mz))
})

test_that("Chromatograms, chromIndex, chromIndex<- works", {
    expect_equal(
        chromIndex(c_full),
        rep(NA_integer_, length(c_full))
    )
    res <- c_full
    chromIndex(res) <- c(3L, 3L, 1L)
    expect_equal(chromIndex(res), c(3, 3, 1))
    expect_false(all(is.na(chromIndex(res))))
})

test_that("Chromatograms, collisionEnergy, collisionEnergy<- works", {
    res <- c_full
    chromData(res)$collisionEnergy <- c(35.0, 40.0, 45.0)
    expect_equal(collisionEnergy(res), c(35.0, 40.0, 45.0))
    collisionEnergy(res) <- c(50.0, 55.0, 60.0)
    expect_equal(collisionEnergy(res), c(50.0, 55.0, 60.0))
})

test_that("Chromatograms, dataOrigin, dataOrigin<- works", {
    res <- c_full
    chromData(res)$dataOrigin <- c("file1", "file2", "file3")
    expect_equal(dataOrigin(res), c("file1", "file2", "file3"))
    dataOrigin(res) <- c("source1", "source2", "source3")
    expect_equal(dataOrigin(res), c("source1", "source2", "source3"))
})

test_that("Chromatograms, msLevel, msLevel<- works", {
    res <- c_full
    expect_equal(msLevel(res), c(1L, 1L, 1L))
    msLevel(res) <- c(2L, 3L, 1L)
    expect_equal(msLevel(res), c(2L, 3L, 1L))
})

test_that("Chromatograms, mz, mz<- works", {
    res <- c_full
    expect_equal(mz(res), c(112.2, 123.3, 134.4))
    mz(res) <- c(115.0, 125.0, 135.0)
    expect_equal(mz(res), c(115.0, 125.0, 135.0))
})

test_that("Chromatograms, mzMin, mzMin<- works", {
    res <- c_full
    chromData(res)$mzMin <- c(110.0, 120.0, 130.0)
    expect_equal(mzMin(res), c(110.0, 120.0, 130.0))
    mzMin(res) <- c(111.0, 121.0, 131.0)
    expect_equal(mzMin(res), c(111.0, 121.0, 131.0))
})

test_that("Chromatograms, mzMax, mzMax<- works", {
    res <- c_full
    chromData(res)$mzMax <- c(114.0, 124.0, 134.0)
    expect_equal(mzMax(res), c(114.0, 124.0, 134.0))
    mzMax(res) <- c(116.0, 126.0, 136.0)
    expect_equal(mzMax(res), c(116.0, 126.0, 136.0))
})

test_that("Chromatograms, length works", {
    expect_equal(length(c_full), 3)
    expect_equal(length(c_empty), 0)
})

test_that("Chromatograms, precursorMz, precursorMz<- works", {
    res <- c_full
    chromData(res)$precursorMz <- c(400.0, 500.0, 600.0)
    expect_equal(precursorMz(res), c(400.0, 500.0, 600.0))
    precursorMz(res) <- c(450.0, 550.0, 650.0)
    expect_equal(precursorMz(res), c(450.0, 550.0, 650.0))
})

test_that("Chromatograms, precursorMzMin, precursorMzMin<- works", {
    res <- c_full
    chromData(res)$precursorMzMin <- c(395.0, 495.0, 595.0)
    expect_equal(precursorMzMin(res), c(395.0, 495.0, 595.0))
    precursorMzMin(res) <- c(390.0, 490.0, 590.0)
    expect_equal(precursorMzMin(res), c(390.0, 490.0, 590.0))
})

test_that("Chromatograms, precursorMzMax, precursorMzMax<- works", {
    res <- c_full
    chromData(res)$precursorMzMax <- c(405.0, 505.0, 605.0)
    expect_equal(precursorMzMax(res), c(405.0, 505.0, 605.0))
    precursorMzMax(res) <- c(410.0, 510.0, 610.0)
    expect_equal(precursorMzMax(res), c(410.0, 510.0, 610.0))
})

test_that("Chromatograms, productMz, productMz<- works", {
    res <- c_full
    chromData(res)$productMz <- c(100.0, 150.0, 200.0)
    expect_equal(productMz(res), c(100.0, 150.0, 200.0))
    productMz(res) <- c(110.0, 160.0, 210.0)
    expect_equal(productMz(res), c(110.0, 160.0, 210.0))
})

test_that("Chromatograms, productMzMin, productMzMin<- works", {
    res <- c_full
    chromData(res)$productMzMin <- c(95.0, 145.0, 195.0)
    expect_equal(productMzMin(res), c(95.0, 145.0, 195.0))
    productMzMin(res) <- c(90.0, 140.0, 190.0)
    expect_equal(productMzMin(res), c(90.0, 140.0, 190.0))
})

test_that("Chromatograms, productMzMax, productMzMax<- works", {
    res <- c_full
    chromData(res)$productMzMax <- c(105.0, 155.0, 205.0)
    expect_equal(productMzMax(res), c(105.0, 155.0, 205.0))
    productMzMax(res) <- c(110.0, 160.0, 210.0)
    expect_equal(productMzMax(res), c(110.0, 160.0, 210.0))
})


test_that("filterChromData handles various edge cases", {
    c_full$randomIdx <- c(1, 2, 3)
    expect_identical(
        filterChromData(c_full,
            variables = c("mz"), ranges = numeric(),
            match = "any"
        ),
        c_full
    )
    res <- filterChromData(c_full,
        variables = c("mz"), ranges = c(100, 200),
        match = "all"
    )
    expect_identical(res, c_full)

    res <- filterChromData(c_full,
        variables = c("mz"), ranges = c(500, 600),
        match = "any"
    )
    expect_equal(nrow(chromData(res)), 0)
    expect_error(
        filterChromData(c_full,
            variables = c("mz", "randomIdx"),
            ranges = c(100, 200), match = "any"
        ),
        "be twice the length of the "
    )
    expect_error(
        filterChromData(c_full,
            variables = c("mz"),
            ranges = c("a", "b"), match = "any"
        ),
        "filterChromData only support filtering for numerical"
    )
    expect_error(
        filterChromData(c_full,
            variables = c("nonExistentVar"),
            ranges = c(100, 200), match = "any"
        ),
        " not available"
    )
    res <- filterChromData(c_empty,
        variables = c("mz"),
        ranges = c(100, 200), match = "any"
    )
    expect_equal(nrow(chromData(res)), 0)
    res <- filterChromData(c_full,
        variables = c("mz", "randomIdx"),
        ranges = c(134, 150, 1, 2), match = "any"
    )
    expect_true(nrow(chromData(res)) > 0)
    res <- filterChromData(c_full,
        variables = c("mz", "randomIdx"),
        ranges = c(134, 150, 1, 2), match = "all"
    )
    expect_true(nrow(chromData(res)) <= nrow(chromData(c_full)))
    res <- filterChromData(c_full,
        variables = c("mz", "randomIdx"),
        ranges = c(100, 200, 1, 3), match = "any"
    )
    expect_equal(nrow(chromData(res)), 3)
    res <- filterChromData(c_full,
        variables = c("mz", "randomIdx"),
        ranges = c(500, 600, 10, 20), match = "all"
    )
    expect_equal(nrow(chromData(res)), 0)
    res <- filterChromData(c_full,
        variables = c("mz"), ranges = c(134, 150),
        match = "any", keep = FALSE
    )
    expect_equal(nrow(chromData(res)), 2)

    res <- filterChromData(c_full,
        variables = c("mz"), ranges = c(120, 130),
        match = "any"
    )
    expect_equal(nrow(chromData(res)), 1)
})

#' Tests for peaks variables.
test_that("peaksData", {
    res <- peaksData(be)
    expect_type(res, "list")
    expect_identical(length(res), length(be))
    expect_true(all(vapply(res, is.data.frame, logical(1))))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- lapply(lengths(res), function(z) data.frame(rtime = sort(abs(rnorm(z))),
                                                            intensity = abs(rnorm(z))))
        peaksData(tmp) <- vals
        res <- peaksData(tmp)
        expect_equal(res, vals)
    }

})

test_that("peaksData() output is valid", {
    expect_true(is.null(validPeaksData(peaksData(be))))
})

test_that("peaksVariables", {
    res <- peaksVariables(be)
    expect_type(res, "character")
    expect_true(all(names(corePeaksVariables()) %in% res))

})

test_that("intensity", {
    res <- intensity(be)
    expect_true(is.list(res) || is(res, "list"))
    expect_true(all(vapply(res, is.numeric, logical(1))))
    expect_false(all(vapply(res, is.data.frame, logical(1))))
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- lapply(lengths(res), function(z) abs(rnorm(z)))
        intensity(tmp) <- vals
        res <- intensity(tmp)
        expect_equal(as.list(res), vals)
        ## test that if length is different error message
        expect_error(intensity(tmp) <- list(1:5, 1:6), "same length")
        ## test incorrect length for one element
        if (length(res) >= 2) {
            incorrect_int <- vals
            incorrect_int[[1]] <- incorrect_int[[1]][-1]
            expect_error(intensity(tmp) <- incorrect_int, "Length of ")
        }
    }
    be_empty <- ChromBackendMemory()
    expect_equal(intensity(be_empty), list())
})

test_that("rtime", {
    res <- rtime(be)
    expect_true(is.list(res) || is(res, "list"))
    expect_true(all(vapply(res, is.numeric, logical(1))))
    expect_false(all(vapply(res, is.data.frame, logical(1))))
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- lapply(lengths(res), function(z) sort(abs(rnorm(z))))
        rtime(tmp) <- vals
        res <- rtime(tmp)
        expect_equal(as.list(res), vals)
        expect_error(rtime(tmp) <- list(1:5, 1:6), "same length")
        if (length(res) >= 2) {
            incorrect_rtime <- vals
            incorrect_rtime[[1]] <- incorrect_rtime[[1]][-1]
            expect_error(rtime(be) <- incorrect_rtime,
                         "Length of ")
        }
    }
    be_empty <- ChromBackendMemory()
    expect_equal(rtime(be_empty), list())
})


test_that("$ works", {
    res <- be$rtime
    expect_true(is.list(res) || is(res, "list"))
    expect_equal(length(res), length(be))
})

test_that("[ works", {
    be1 <- be[1]
    expect_equal(length(peaksData(be1)), 1)
    be12 <- be[1:2]
    expect_equal(length(peaksData(be12)), 2)
    be0 <- be[integer()]
    expect_equal(nrow(peaksData(be0)[[1]]), 0)
})

test_that("imputePeaksData works", {
    be1 <- be[1]
    res <- intensity(be)[[1]]
    ## Put NA in interior position (not leading/trailing) to test interpolation
    ## with default extrapolate = FALSE
    if (length(res) >= 3) {
        res[2] <- NA
    } else {
        res[1] <- NA
    }
    intensity(be1) <- list(res)
    expect_true(any(is.na(intensity(be1)[[1]])))
    ## Use extrapolate = TRUE to ensure all NAs are filled regardless of position
    be1 <- imputePeaksData(be1, method = "linear", extrapolate = TRUE)
    expect_false(any(is.na(intensity(be1)[[1]])))
    })

#' To run this unit tests from another package:
#'
#' The unit tests in this suite expect a variable `be` to be defined, which
#' has to represent an **already initialized** backend instance.

## Test chromData
test_that("chromData", {
    res <- chromData(be)
    expect_true(is.data.frame(res))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- res[sample(seq_len(nrow(res))), ]
        chromData(tmp) <- vals
        expect_equal(vals, chromData(tmp))
    }
})

test_that("chromData() output is valid", {
    expect_true(is.null(validChromData(chromData(be))))
})

test_that("chromVariables", {
    res <- chromVariables(be)
    expect_type(res, "character")
    expect_true(all(names(coreChromVariables()) %in% res))
})

test_that("chromIndex", {
    res <- chromIndex(be)
    expect_type(res, "integer")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- sample(1:5, length(res), replace = TRUE)
        chromIndex(tmp) <- vals
        expect_equal(chromIndex(tmp), vals)
    }
})

test_that("collisionEnergy", {
    res <- collisionEnergy(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        collisionEnergy(tmp) <- vals
        expect_equal(unname(collisionEnergy(tmp)), unname(vals))
    }
})

test_that("dataOrigin", {
    res <- dataOrigin(be)
    expect_type(res, "character")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- sample(c("file", "database", "network"), length(be), replace = TRUE)
        dataOrigin(tmp) <- vals
        expect_equal(dataOrigin(tmp), vals)
    }
})

test_that("msLevel", {
    res <- msLevel(be)
    expect_type(res, "integer")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- sample(1:5, length(res), replace = TRUE)
        msLevel(tmp) <- vals
        expect_equal(msLevel(tmp), vals)
    }
})

test_that("mz", {
    res <- mz(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        mz(tmp) <- vals
        expect_equal(unname(mz(tmp)), unname(vals))
    }
})

test_that("mzMax", {
    res <- mzMax(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        mzMax(tmp) <- vals
        expect_equal(unname(mzMax(tmp)), unname(vals))
    }
})

test_that("mzMin", {
    res <- mzMin(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        mzMin(tmp) <- vals
        expect_equal(unname(mzMin(tmp)), unname(vals))
    }
})

test_that("precursorMz", {
    res <- precursorMz(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        precursorMz(tmp) <- vals
        expect_equal(unname(precursorMz(tmp)), unname(vals))
    }
})

test_that("precursorMzMin", {
    res <- precursorMzMin(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        precursorMzMin(tmp) <- vals
        expect_equal(unname(precursorMzMin(tmp)), unname(vals))
    }
})

test_that("precursorMzMax", {
    res <- precursorMzMax(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        precursorMzMax(tmp) <- vals
        expect_equal(unname(precursorMzMax(tmp)), unname(vals))
    }
})

test_that("productMz", {
    res <- productMz(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        productMz(tmp) <- vals
        expect_equal(unname(productMz(tmp)), unname(vals))
    }
})

test_that("productMzMin", {
    res <- productMzMin(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        productMzMin(tmp) <- vals
        expect_equal(unname(productMzMin(tmp)), unname(vals))
    }
})

test_that("productMzMax", {
    res <- productMzMax(be)
    expect_type(res, "double")
    expect_identical(length(res), length(be))
    if (!isReadOnly(be)) {
        tmp <- be
        vals <- abs(rnorm(length(be)))
        productMzMax(tmp) <- vals
        expect_equal(unname(productMzMax(tmp)), unname(vals))
    }
})

## Other
test_that("isEmpty", {
    res <- isEmpty(be)
    expect_type(res, "logical")
    expect_identical(length(res), length(be))
    expect_identical(lengths(be) == 0, res)
})

test_that("lengths", {
    res <- lengths(be)
    expect_type(res, "integer")
    expect_identical(length(res), length(be))
    expect_identical(res, lengths(rtime(be)))
})

test_that("$ works", {
    res <- be$msLevel
    expect_true(is.integer(res))
    expect_equal(length(res), length(be))

    ## Error if spectra variable not available
    expect_error(be$doesnt_exist)
})

test_that("[ works", {
    be$randomIdx <- seq_len(length(be))
    be1 <- be[1]
    expect_equal(nrow(chromData(be1)), 1)
    expect_equal(chromData(be1)$randomIdx, 1)
    be12 <- be[1:2]
    expect_equal(nrow(chromData(be12)), 2)
    expect_equal(chromData(be12)$randomIdx, c(1, 2))

    # Handles empty integer
    be0 <- be[integer()]
    expect_equal(nrow(chromData(be0)), 0)
    expect_equal(be0, be_empty)
})

test_that("[[ works", {
    expect_error(be[["doesnt_exist"]], "The requested variable")
    expect_error(be[["msLevel", "no"]], "is not supported")
    expect_equal(be[["msLevel"]], msLevel(be))
    if (!isReadOnly(be)) {
        be2 <- be
        be2[["msLevel"]] <- c(1L,2L,3L)
        expect_false(all(be2[["msLevel"]] == msLevel(be)))
    }
})

test_that("fillCoreChromVariables works", {
    x <- data.frame(a = numeric(), b = character())
    res <- fillCoreChromVariables(x)
    expect_true(is.data.frame(res))
    expect_true(all(c("a", "b") %in% colnames(res)))
    expect_true(all(names(.CORE_CHROM_VARIABLES) %in%
        c(colnames(res), "intensity", "rtime")))
    expect_true(nrow(res) == 0)

    x <- data.frame(msLevel = c(1L, 2L, 1L), other_col = "a")
    res <- fillCoreChromVariables(x)
    expect_true(is.data.frame(res))
    expect_true(all(c("msLevel", "other_col") %in% colnames(res)))
    expect_true(all(names(.CORE_CHROM_VARIABLES) %in%
        c(colnames(res), "intensity", "rtime")))
    expect_true(nrow(res) == 3L)

    cv <- .CORE_CHROM_VARIABLES[!names(.CORE_CHROM_VARIABLES) %in%
        c("intensity", "rtime", "msLevel")]
    for (i in seq_along(cv)) {
        expect_true(is(res[, names(cv)[i]], cv[i]))
        expect_true(all(is.na(res[, names(cv)[i]])))
    }
    expect_equal(res$msLevel, c(1L, 2L, 1L))

    x <- fillCoreChromVariables(x)
    expect_equal(x, fillCoreChromVariables(x))
})

test_that("core variables functions works", {
    expect_equal(coreChromVariables(), .CORE_CHROM_VARIABLES)
    expect_equal(corePeaksVariables(), .CORE_PEAKS_VARIABLES)
})

test_that("validChromData works", {
    x <- data.frame()
    res <- validChromData(x)
    expect_true(length(res) == 0)

    x <- data.frame(msLevel = c(1L, 2L), other_col = "a")
    res <- validChromData(x)
    expect_true(length(res) == 0)

    x$mz <- "a"
    x$collisionEnergy <- "30ev"
    expect_error(validChromData(x, error = TRUE), "wrong data type")
    res <- validChromData(x, error = FALSE)
    expect_true(is.character(res))
    expect_true(length(res) == 2L)
})

test_that("validPeaksData works", {
    x <- matrix()
    expect_error(validPeaksData(x), "list")
    x <- data.frame()
    expect_equal(validPeaksData(x), NULL)
    x <- list(data.frame(rtime = numeric(), intensity = numeric()))
    res <- validPeaksData(x)
    expect_true(is.null(res))
    # checks order
    x <- list(data.frame(intensity = numeric(), rtime = numeric()))
    expect_error(validPeaksData(x), "Columns should be in the order")

    x <- list(
        data.frame(rtime = c(10.0, 12.0), intensity = c(200, 150)),
        data.frame(rtime = c(30.1, 31.2), intensity = c(110, 90), other_col = c("test", "test"))
    )
    expect_error(
        validPeaksData(x),
        "same order"
    )
})

#' Here are the helper functions used in the package.
#' Please add a description of the function and the methods in which it is
#' used.

#' @note
#' Used for:
#' - `backendMerge()`
#'
#' @author Johannes Rainer
#' @importFrom MsCoreUtils vapply1c rbindFill
#' @noRd
.df_combine <- function(objects) {
    if (length(objects) == 1) {
        return(objects[[1]])
    }
    if (!all(vapply1c(objects, class) == class(objects[[1]]))) {
        stop("Can only merge backends of the same type: ", class(objects[[1]]))
    }
    res <- objects[[1]]
    pv <- names(.peaksData(res)[[1]])
    for (i in 2:length(objects)) {
        res@chromData <- rbindFill(.chromData(res), .chromData(objects[[i]]))
        pv2 <- peaksVariables(objects[[i]])
        if (length(pv) == length(pv2) && all(pv == pv2)) {
            res@peaksData <- c(.peaksData(res), .peaksData(objects[[i]]))
        } else {
            stop(
                "Provided objects have different sets of peak variables. ",
                "Combining such objects is currently not supported."
            )
        }
    }
    res
}

#' Fast rbind for lists of data.frames using `data.table::rbindlist()`.
#' Always returns a plain `data.frame`.
#' @importFrom data.table rbindlist
#' @noRd
.fast_rbind <- function(lst) {
  if (length(lst) == 0L) return(data.frame())
  as.data.frame(rbindlist(lst, use.names = TRUE, fill = TRUE))
}

#' Helper function to check the order and data types of columns
#'
#' @note:
#' used in:
#' - `validPeaksData()`
#' @noRd
.check_column_order_and_types <- function(df, expected_cols, expected_types) {
    if (!identical(colnames(df)[seq_len(2)], expected_cols)) {
        return(paste0("Columns should be in the order 'rtime', 'intensity'."))
    }
    invalid_cols <- vapply(expected_cols, function(col) {
        !is(df[[col]], expected_types[[col]])
    }, logical(1))
    if (any(invalid_cols)) {
        invalid_col_names <- expected_cols[invalid_cols]
        return(paste0(
            "The peaksData variable(s) ", paste(invalid_col_names,
                collapse = ", "
            ),
            " have the wrong data type."
        ))
    }
    return(NULL)
}

#' Helper function to check the properties of the 'rtime' column.
#'
#' @note:
#' used in:
#' - `validPeaksData()`
#' @noRd
.check_rtime <- function(df) {
    if (nrow(df) == 0) {
        return(NULL)
    }
    if (any(is.na(df$rtime))) {
        return("'rtime' column contains NA values.")
    }

    if (!all(diff(df$rtime) > 0)) {
        return("'rtime' column is not strictly increasing.") ## does it need to strictly increase ?
    }

    return(NULL)
}

#' Function to apply the processing queue to the backend, return a peaksData.
#' Used in:
#' - `peaksData(Chromatograms())`
#' - `applyProcessing()`
#'
#' It takes a backend and a preprocessingQueue and applies it. It returns
#' then the backend. This function might need to be  refined later in case the
#' backend is `readOnly == TRUE`.
#'
#' @importFrom BiocParallel bplapply SerialParam
#' @noRd
.run_process_queue <- function(object, queue, f = factor(),
    BPPARAM = SerialParam()) {
    BPPARAM <- backendBpparam(object, BPPARAM)
    if (!length(f) || length(levels(f)) == 1) {
        for (i in seq_along(queue)) {
            object <- do.call(queue[[i]]@FUN, c(object, queue[[i]]@ARGS))
        }
        return(object)
    }
    if (!is(f, "factor")) stop("f must be a factor")
    if (length(f) != length(object)) {
        stop(
            "length 'f' has to be equal to the length of 'object' (",
            length(object), ")"
        )
    }
    processed_data <- bplapply(split(object, f), function(x) {
        for (i in seq_along(queue)) {
            x <- do.call(queue[[i]]@FUN, c(x, queue[[i]]@ARGS))
        }
        x
    }, BPPARAM = BPPARAM)
    backendMerge(processed_data)
}

#' Function to validate each peaksData entry
#'
#' @note:
#' used in:
#' - `validPeaksData()`
#' @noRd
.validate_entry <- function(df, i, expected_cols, expected_types) {
    msgs <- NULL
    if (!is.data.frame(df)) {
        msgs <- c(msgs, paste0(
            "Entry ", i, ": all 'peaksData' ",
            "entries should ",
            "be of class 'data.frame'"
        ))
    } else {
        msgs <- c(
            msgs, .check_column_order_and_types(
                df, expected_cols,
                expected_types
            ),
            .check_rtime(df)
        )
    }
    return(msgs)
}

#' Function to validate the processingQueue slot of a `Chromatograms` object
#'
#' Used in:
#' - `validObject(Chromatograms())`
#' @importFrom MsCoreUtils vapply1l
#' @noRd
.valid_processing_queue <- function(x) {
    if (length(x) && !all(vapply1l(x, inherits, "ProcessingStep"))) {
        stop("'processingQueue' should only contain ProcessingStep objects.")
    }
    NULL
}

#' function to loop through  query column and check if within corresponding
#' ranges. Return an index of the corresponding matches.
#' Used in:
#' - `filterPeaksData()`: looped through the list of data.frame
#' - `filterChromData()`
#' @importFrom MsCoreUtils between
#' @noRd
.filter_ranges <- function(query, ranges, match) {
    nc <- ncol(query)
    nr <- nrow(query)
    if (length(ranges) != 2 * nc) {
        stop(
            "Length of 'ranges' needs to be twice the length of the ",
            "parameter 'query'"
        )
    }

    # Compute within_ranges for each column of the query
    within_ranges <- vapply(seq_len(nc), function(i) {
        pairs <- c(ranges[2 * i - 1], ranges[2 * i])
        between(query[[i]], pairs)
    }, logical(nrow(query)))

    if (match == "all") {
        if (nr == 1) {
            return(as.integer(all(within_ranges)))
        }
        return(which(rowSums(within_ranges) == nc))
    }
    if (nr == 1) {
        return(as.integer(any(within_ranges)))
    }
    return(which(rowSums(within_ranges) > 0))
}

#' Used in:
#' - `filterPeaksData()`
#' @noRd
.logging <- function(x, ...) {
    c(x, paste0(..., " [", date(), "]"))
}

#' Used In:
#' - `ChromBackendMzR()`
#' @noRd
.check_mzR_package <- function() {
    if (!requireNamespace("mzR", quietly = TRUE)) {
        stop(
            "The use of 'ChromBackendMzR' requires package 'mzR'. ",
            "Install it using 'BiocManager::install(\"mzR\")'"
        )
    }
}

#' Function to create chromData form mzml file
#' Used In:
#' - `backendInitialize()` for `ChromBackendMzR` class
#' Helper function to format chromatographic data from mzR files.
#' @noRd
.mzR_format_chromData <- function(file) {
    .check_mzR_package()
    msd <- mzR::openMSfile(file)
    on.exit(mzR::close(msd))
    tmp <- mzR::chromatogramHeader(msd)
    colnames(tmp)[colnames(tmp) ==
        "chromatogramIndex"] <- "chromIndex"
    colnames(tmp)[colnames(tmp) ==
        "precursorCollisionEnergy"] <- "collisionEnergy"
    colnames(tmp)[colnames(tmp) ==
        "productIsolationWindowTargetMZ"] <- "productMz"
    colnames(tmp)[colnames(tmp) ==
        "precursorIsolationWindowTargetMZ"] <- "precursorMz"
    tmp$dataOrigin <- file
    tmp
}

#' Used In:
#' - `peaksData()` for `ChromBackendMzR` class
#' @noRd
.get_chrom_data <- function(fl, idx) {
    .check_mzR_package()
    msd <- mzR::openMSfile(fl)
    on.exit(mzR::close(msd))
    mzR::chromatogram(msd, idx, drop = FALSE)
}

#' Helper function to plot a single chromatogram.
#' @note:
#' Used in:
#' - `plotChromatograms()`
#' - `plotChromatogramsOverlay()`
#'
#' @importFrom graphics plot.new plot.window plot.xy axis box title par
#' @importFrom grDevices dev.hold dev.flush xy.coords n2mfrow
#' @noRd
.plot_single_chromatogram <- function(x, xlab = "rtime (s)",
    ylab = "intensity",
    type = "l", xlim = numeric(),
    ylim = numeric(),
    main = paste("m/z", round(mz(x), 1)),
    col = "#00000080", add = FALSE,
    axes = TRUE, frame.plot = axes,
    orientation = 1, ...) {
    v <- peaksData(x)[[1L]]
    rts <- v$rtime
    raw_ints <- v[, "intensity"]
    ints <- orientation * raw_ints
    if (!length(xlim)) {
        xlim <- range(rts, na.rm = TRUE)
    }
    if (!length(ylim)) {
        ylim <- range(orientation * c(0, max(abs(ints), na.rm = TRUE)))
    }
    if (any(is.infinite(xlim))) {
        xlim <- c(0, 0)
    }
    if (any(is.infinite(ylim))) {
        ylim <- c(0, 0)
    }
    if (!add) {
        dev.hold()
        on.exit(dev.flush())
        plot.new()
        plot.window(xlim = xlim, ylim = ylim)
    }
    if (!add) {
        if (axes) {
            axis(side = 1, ...)
            axis(side = 2, ...)
        }
        if (frame.plot) {
            box(...)
        }
        title(main = main, xlab = xlab, ylab = ylab, ...)
    }
    plot.xy(xy.coords(rts, ints), type = type, col = col, ...)
}
#' Strip NA intensities from paired mz/intensity vectors.
#'
#' Returns a list with cleaned `mz`, `int`, and `n`, or `NULL` if no
#' non-NA values remain.
#'
#' Used in:
#' - `.build_intensity_matrix()`
#' - `.compute_chrom_intensities()`
#' @noRd
.strip_na_peaks <- function(mz_j, int_j) {
  if (!anyNA(int_j))
    return(list(mz = mz_j, int = int_j, n = length(mz_j)))
  keep <- !is.na(int_j)
  int_j <- int_j[keep]
  mz_j <- mz_j[keep]
  n <- length(int_j)
  if (n == 0L) NULL else list(mz = mz_j, int = int_j, n = n)
}

#' Compute m/z range boundary indices via binary search.
#'
#' Wraps `findInterval()` to return the lo (first included) and hi (last
#' included) positions for each m/z range in `mzMins`/`mzMaxs`.
#'
#' Used in:
#' - `.build_intensity_matrix()`
#' - `.compute_chrom_intensities()`
#' @noRd
.mz_boundaries <- function(mzMins, mzMaxs, mz_j) {
  list(
    lo = findInterval(mzMins, mz_j, left.open = TRUE,
                      checkSorted = FALSE, checkNA = FALSE) + 1L,
    hi = findInterval(mzMaxs, mz_j,
                      checkSorted = FALSE, checkNA = FALSE)
  )
}

#' Find spectra indices within a retention time window.
#'
#' Used in:
#' - `.process_peaks_data()`
#' @noRd
.rt_keep <- function(valid_rt, rt_min, rt_max, n_valid) {
  if (is.finite(rt_min) || is.finite(rt_max))
    which(valid_rt >= rt_min & valid_rt <= rt_max)
  else
    seq_len(n_valid)
}

#' Build a single chromatogram result entry.
#'
#' Constructs a data.frame (or vector when `drop = TRUE`) directly without
#' calling `as.data.frame()`, which avoids its per-call overhead (~10x
#' faster for repeated construction).
#'
#' Used in:
#' - `.process_peaks_data()`
#' @noRd
.make_chrom_entry <- function(rtime, intensity, do_rt, do_int, columns, drop) {
  df <- list()
  if (do_rt) df$rtime <- rtime
  if (do_int) df$intensity <- intensity
  class(df) <- "data.frame"
  attr(df, "row.names") <- .set_row_names(
    if (do_rt) length(rtime) else length(intensity)
  )
  if (drop && length(columns) == 1L) df[[columns]] else df
}

#' Pre-extract and pre-filter spectra data for chromatogram computation.
#'
#' Extracts mz, intensity, and rtime from a Spectra object as plain R lists,
#' removes duplicated retention times, and pre-filters spectra to the global
#' retention time range. Also classifies the request (shared rt, TIC/BPC,
#' cumsum-eligible) to guide downstream dispatch.
#'
#' Used in:
#' - `.process_peaks_data()`
#'
#' @param cd `data.frame` with columns rtMin, rtMax, mzMin, mzMax.
#' @param s `Spectra` object.
#' @param fun Summary function (e.g., `sumi`, `maxi`).
#' @return A named `list` with prepared vectors and classification flags.
#' @importFrom Spectra rtime peaksData filterRanges filterRt
#' @importFrom MsCoreUtils sumi maxi reduce
#' @noRd
.prepare_spectra_input <- function(cd, s, fun) {
  global_rt_min <- min(cd$rtMin)
  global_rt_max <- max(cd$rtMax)
  rtr <- do.call(rbind, reduce(cd$rtMin, cd$rtMax, .check = FALSE))
  n_ranges <- ncol(rtr)
  if (n_ranges == 1L) {
    rt_range <- as.vector(rtr)
    if (rt_range[1L] < rt_range[2L])
      s <- filterRt(s, rt = rt_range)
  } else {
    s <- filterRanges(s, as.vector(rtr), match = "any",
                      spectraVariables = rep("rtime", n_ranges))
  }
  rt <- rtime(s)
  valid_idx <- which(!duplicated(rt) & rt >= global_rt_min & rt <= global_rt_max)
  p <- peaksData(s[valid_idx], return.type = "list")
  mz_list <- lapply(p, `[`, , 1L)
  int_list <- lapply(p, `[`, , 2L)
  rt_mins <- cd$rtMin
  rt_maxs <- cd$rtMax
  mzMins <- cd$mzMin
  mzMaxs <- cd$mzMax
  list(
    valid_rt   = rt[valid_idx],
    valid_mz   = mz_list,
    valid_int  = int_list,
    n_valid    = length(valid_idx),
    rt_mins    = rt_mins,
    rt_maxs    = rt_maxs,
    mzMins     = mzMins,
    mzMaxs     = mzMaxs,
    n_cd       = nrow(cd),
    same_rt    = all(rt_mins == rt_mins[1L]) && all(rt_maxs == rt_maxs[1L]),
    all_tic    = all(is.infinite(mzMins) & mzMins < 0) &&
                 all(is.infinite(mzMaxs) & mzMaxs > 0),
    use_cumsum = identical(fun, sumi)
  )
}

#' Compute intensities for one spectrum across one or more m/z ranges.
#'
#' Core per-spectrum logic shared by `.build_intensity_matrix()` (matrix
#' path) and `.compute_chrom_intensities()` (vector path).  Dispatches
#' through three branches: TIC/BPC (no m/z filtering), cumsum (for
#' `sumi`), and generic (e.g. `maxi`).
#'
#' Used in:
#' - `.build_intensity_matrix()`
#' - `.compute_chrom_intensities()`
#'
#' @param mz_j,int_j Numeric vectors of m/z and intensity for one spectrum.
#' @param mzMins,mzMaxs Numeric vectors of m/z bounds (length = n ranges).
#' @param fun Summary function.
#' @param all_tic Logical: all mz ranges infinite (TIC/BPC case)?
#' @param use_cumsum Logical: use cumsum optimisation (for `sumi`)?
#' @return Numeric vector of length `length(mzMins)`.
#' @noRd
.spectrum_intensities <- function(mz_j, int_j, mzMins, mzMaxs,
                                  fun, all_tic, use_cumsum) {
  n_cd <- length(mzMins)
  if (all_tic) return(rep(fun(int_j), n_cd))
  if (use_cumsum) {
    pk <- .strip_na_peaks(mz_j, int_j)
    if (is.null(pk)) return(rep(NA_real_, n_cd))
    cs <- c(0, cumsum(pk$int))
    b <- .mz_boundaries(mzMins, mzMaxs, pk$mz)
    res <- rep(NA_real_, n_cd)
    w <- which(b$lo <= b$hi & b$lo >= 1L & b$hi >= 1L & b$lo <= pk$n)
    if (length(w)) res[w] <- cs[b$hi[w] + 1L] - cs[b$lo[w]]
    return(res)
  }
  n_mz <- length(mz_j)
  if (n_mz == 0L) return(rep(NA_real_, n_cd))
  b <- .mz_boundaries(mzMins, mzMaxs, mz_j)
  vapply(seq_len(n_cd), function(i) {
    if (b$lo[i] <= b$hi[i] && b$lo[i] >= 1L && b$hi[i] <= n_mz)
      fun(int_j[b$lo[i]:b$hi[i]])
    else NA_real_
  }, numeric(1))
}

#' Build intensity matrix for the shared-retention-time path.
#'
#' When all chromatograms share the same retention time window, computes
#' intensities using a spectrum-major loop via `.spectrum_intensities()`.
#'
#' Used in:
#' - `.process_peaks_data()`
#'
#' @param valid_mz,valid_int Lists of m/z and intensity vectors.
#' @param kept Integer indices into `valid_*` for spectra in the rt window.
#' @param mzMins,mzMaxs Numeric vectors of m/z bounds per chromatogram.
#' @param n_cd Number of chromatograms.
#' @param fun Summary function.
#' @param all_tic Logical: all mz ranges infinite (TIC/BPC case)?
#' @param use_cumsum Logical: use cumsum optimisation (for `sumi`)?
#' @return A `matrix` of dimensions `length(kept)` x `n_cd`.
#' @noRd
.build_intensity_matrix <- function(valid_mz, valid_int, kept, mzMins,
                                    mzMaxs, n_cd, fun, all_tic, use_cumsum) {
  n_kept <- length(kept)
  int_mat <- matrix(NA_real_, nrow = n_kept, ncol = n_cd)
  for (j in seq_len(n_kept))
    int_mat[j, ] <- .spectrum_intensities(
      valid_mz[[kept[j]]], valid_int[[kept[j]]],
      mzMins, mzMaxs, fun, all_tic, use_cumsum)
  int_mat
}

#' Compute intensity vector for a single chromatogram (different-rt path).
#'
#' Thin wrapper around `.spectrum_intensities()` that returns a vector
#' (one value per spectrum) for a single m/z range.
#'
#' Used in:
#' - `.process_peaks_data()`
#'
#' @param valid_mz,valid_int Lists of m/z and intensity vectors.
#' @param kept Integer indices into `valid_*` for spectra in this chrom's
#'   rt window.
#' @param mz_lo,mz_hi Scalar m/z bounds for this chromatogram.
#' @param fun Summary function.
#' @param use_cumsum Logical: use cumsum optimisation?
#' @return Numeric vector of length `length(kept)`.
#' @noRd
.compute_chrom_intensities <- function(valid_mz, valid_int, kept, mz_lo,
                                       mz_hi, fun, use_cumsum) {
  all_tic <- is.infinite(mz_lo) && mz_lo < 0 && is.infinite(mz_hi)
  vapply(kept, function(j)
    .spectrum_intensities(valid_mz[[j]], valid_int[[j]],
                          mz_lo, mz_hi, fun, all_tic, use_cumsum),
    numeric(1), USE.NAMES = FALSE)
}

#' Compute peaks data from Spectra for ChromBackendSpectra.
#'
#' Main orchestrator that converts raw spectra into chromatographic peaks data
#' by aggregating intensities within defined m/z and retention time ranges.
#' Dispatches to `.build_intensity_matrix()` (shared-rt path) or
#' `.compute_chrom_intensities()` (per-chromatogram path).
#'
#' Used In:
#' - `peaksData` for `ChromBackendSpectra` class.
#' @importFrom Spectra peaksData
#' @importFrom MsCoreUtils maxi
#' @noRd
.process_peaks_data <- function(cd, s, columns, fun, drop) {
  do_rt <- "rtime" %in% columns
  do_int <- "intensity" %in% columns
  prep <- .prepare_spectra_input(cd, s, fun)
  n_cd <- prep$n_cd
  valid_rt <- prep$valid_rt
  valid_mz <- prep$valid_mz
  valid_int <- prep$valid_int
  n_valid <- prep$n_valid
  if (prep$same_rt) {
    ## Optimized path: all chromatograms share the same rt range
    kept <- .rt_keep(valid_rt, prep$rt_mins[1L], prep$rt_maxs[1L], n_valid)
    n_kept <- length(kept)
    rtime_out <- valid_rt[kept]
    int_mat <- if (do_int && n_kept > 0L)
      .build_intensity_matrix(valid_mz, valid_int, kept, prep$mzMins,
                              prep$mzMaxs, n_cd, fun, prep$all_tic,
                              prep$use_cumsum)
    res <- lapply(seq_len(n_cd), function(i)
      .make_chrom_entry(
        rtime_out,
        if (do_int && n_kept > 0L) int_mat[, i] else numeric(0L),
        do_rt, do_int, columns, drop
      ))
  } else {
    ## Fallback: per-chromatogram rt ranges differ
    res <- vector("list", n_cd)
    for (i in seq_len(n_cd)) {
      kept <- .rt_keep(valid_rt, prep$rt_mins[i], prep$rt_maxs[i], n_valid)
      n_kept <- length(kept)
      rtime_out <- valid_rt[kept]
      intensities <- if (do_int && n_kept > 0L)
        .compute_chrom_intensities(valid_mz, valid_int, kept,
                                   prep$mzMins[i], prep$mzMaxs[i],
                                   fun, prep$use_cumsum)
      else numeric(n_kept)
      res[[i]] <- .make_chrom_entry(
        rtime_out, intensities, do_rt, do_int, columns, drop
      )
    }
  }
  res
}

#' Used in:
#' - `backendInitialize()` for `ChrombackendSpectra`
#' @noRd
.spectra_format_chromData <- function(sps) {
    res <- data.frame(
        msLevel = unique(sps$msLevel),
        rtMin = min(sps$rtime, na.rm = TRUE),
        rtMax = max(sps$rtime, na.rm = TRUE),
        mzMin = -Inf,
        mzMax = Inf,
        mz = Inf,
        dataOrigin = unique(sps$dataOrigin),
        chromSpectraIndex = unique(sps$chromSpectraIndex)
    )
    ## Add optional columns if present
    if ("polarity" %in% spectraVariables(sps)) {
        res$polarity <- sps$polarity[1]
    }
    if ("scanWindowLowerLimit" %in% spectraVariables(sps)) {
        res$scanWindowLowerLimit <- sps$scanWindowLowerLimit[1]
    }
    if ("scanWindowUpperLimit" %in% spectraVariables(sps)) {
        res$scanWindowUpperLimit <- sps$scanWindowUpperLimit[1]
    }
    res
}

#' Used in:
#' - `factorize()` for `ChrombackendSpectra`
#' @noRd
.ensure_rt_mz_columns <- function(chrom_data, spectra, spectra_f) {
    if (!all(c("mzMin", "mzMax") %in% colnames(chrom_data))) {
        if ("mzMin" %in% colnames(chrom_data) ||
            "mzMax" %in% colnames(chrom_data)) {
            stop("Both 'mzMin' and 'mzMax' must be present if one",
                 " is provided.")
        } else {
            chrom_data$mzMin <- -Inf
            chrom_data$mzMax <- Inf
        }
    }
    if (!all(c("rtMin", "rtMax") %in% colnames(chrom_data))) {
        if ("rtMin" %in% colnames(chrom_data) || "rtMax" %in%
            colnames(chrom_data)) {
            stop("Both 'rtMin' and 'rtMax' must be present if one",
                 " is provided.")
        } else {
            levs <- levels(spectra_f)
            if (is.null(levs)) {
                levs <- unique(as.character(spectra_f))
            }
            rt_mat <- vapply(levs, function(lvl) {
                range(spectra$rtime[spectra_f == lvl], na.rm = TRUE)
            }, numeric(2))
            chrom_idx <- as.character(chrom_data$chromSpectraIndex)
            chrom_data$rtMin <- rt_mat[1, chrom_idx]
            chrom_data$rtMax <- rt_mat[2, chrom_idx]
        }
    }
    chrom_data
}

#' Used in:
#' - `chromExtract()`.
#' @noRd
.validate_chromExtract_input <- function(object,
                                         peak.table,
                                         by,
                                         required_cols = c("rtMin", "rtMax",
                                                           by)) {
    cd <- .chromData(object)
    if (!all(required_cols %in% names(peak.table))) {
        stop("`peak.table` must contain columns: ", paste(required_cols,
                                                          collapse = ", "), ".")
    }

    if (anyNA(peak.table$rtMin) || anyNA(peak.table$rtMax)) {
        stop("Columns 'rtMin' and 'rtMax' in `peak.table` cannot ",
             "contain NA values.")
    }
    if (!all(by %in% names(cd))) {
        stop("All 'by' columns must be present in chromData(object).")
    }
    if (nrow(cd) != nrow(unique(cd[, by, drop = FALSE]))) {
        stop("Combinations of 'by' columns must uniquely identify rows ",
             "in chromData.")
    }
}


#' Used in:
#' - `chromExract()`
#' @noRd
.match_chromdata_peaktable <- function(object, peak.table, by) {
    cd <- .chromData(object)
    chrom_keys <- interaction(cd[, by, drop = FALSE], drop = TRUE)
    peak_keys  <- interaction(peak.table[, by, drop = FALSE], drop = TRUE)

    # ensure all peak.table keys exist in chromData
    missing_keys <- setdiff(levels(peak_keys), levels(chrom_keys))
    if (length(missing_keys)) {
        stop("Some combinations in `peak.table` do not exist in chromData: ",
             paste(missing_keys, collapse = ", "))
    }

    ## Subset chromdata and only keep the row of interest.
    keep_idx <- chrom_keys %in% peak_keys
    object <- object[keep_idx]
    chrom_keys <- droplevels(chrom_keys[keep_idx])

    # align factor levels (so splitting matches between cd and peak.table)
    shared_levels <- intersect(levels(peak_keys), levels(chrom_keys))
    chrom_keys <- factor(as.character(chrom_keys), levels = shared_levels)
    peak_keys  <- factor(as.character(peak_keys),  levels = shared_levels)

    list(object = object, chrom_keys = chrom_keys, peak_keys = peak_keys)
}

#' Used in:
#' - `chromExtract()`
#' @noRd
.check_overl_columns <- function(object, peak.table, required_cols) {
    overl_cols <- names(peak.table) %in% chromVariables(object)
    extra_cols <- setdiff(names(peak.table)[overl_cols], required_cols)
    if (length(extra_cols)) {
        warning( "The following columns in `peak.table` already exist in ",
                 "`chromData` and will be replaced in the output: ",
                 paste(extra_cols, collapse = ", ")
        )
    }
    overl_cols
}


#' Used in:
#' - `imputePeaksData()`
#' @importFrom stats approx filter loess spline dnorm sd predict
#' @noRd
.impute <- function(x, method, window = 2, span = 0.25, sd = 1,
                    extrapolate = FALSE) {
    if (all(is.na(x))) return(x)

    na_idx <- which(is.na(x))
    if (length(na_idx) == 0) return(x)
    not_na_idx <- which(!is.na(x))
    x_out <- seq_along(x)
    # rule = 2 extrapolates, rule = 1 returns NA outside range
    approx_rule <- if (extrapolate) 2L else 1L
    x[na_idx] <- switch(method,
        linear = approx(not_na_idx, x[not_na_idx], xout = na_idx,
                        rule = approx_rule)$y,
        spline = {
            vals <- spline(not_na_idx, x[not_na_idx],
                           xout = na_idx, method = "natural")$y
            if (!extrapolate) {
                # Set values outside data range to NA
                outside <- na_idx < min(not_na_idx) | na_idx > max(not_na_idx)
                vals[outside] <- NA
            }
            vals
        },
        gaussian = {
            # Create symmetric Gaussian kernel
            kernel_range <- -window:window
            w <- dnorm(kernel_range, mean = 0, sd = sd)
            w <- w / sum(w)

            # Fill missing with linear approx to allow smoothing
            x_filled <- x
            x_filled[is.na(x_filled)] <- approx(not_na_idx, x[not_na_idx],
                                                xout = which(is.na(x_filled)),
                                                rule = approx_rule)$y
            smoothed <- filter(x_filled, filter = w, sides = 2,
                               circular = FALSE)
            smoothed[na_idx]
        },
        loess = {
            fit <- loess(x[not_na_idx] ~ not_na_idx, span = span)
            vals <- predict(fit, newdata = na_idx)
            if (!extrapolate) {
                outside <- na_idx < min(not_na_idx) | na_idx > max(not_na_idx)
                vals[outside] <- NA
            }
            vals
        }
    )
    # Fallback for any remaining NAs (only if extrapolate = TRUE)
    na_remaining <- is.na(x)
    if (extrapolate && any(na_remaining)) {
        warning("Method chosen could not fill all NAs. ",
                "Falling back to linear interpolation ",
                "for these positions.")
        x[na_remaining] <- approx(not_na_idx, x[not_na_idx],
                                         xout = which(na_remaining),
                                         rule = 2)$y
    }
    x
}


## Used in:
## - BackendInitialize, ChromBackendSpectra method
#' @noRd
.map_spectra_vars <- function(object, spectraVariables) {
    # Check variable validity for mapping from the Spectra object stored in the internal @spectra slot
    spectra <- .spectra(object)
    cd <- .chromData(object)
    if (!all(spectraVariables %in% spectraVariables(spectra)))
        stop("All 'spectraVariables' must exist in the Spectra object.")
    if (any(spectraVariables %in% colnames(cd))) {
        existing <- intersect(spectraVariables, colnames(cd))
        non_replaceable <- vapply(existing, function(v) !all(is.na(cd[[v]])),
                                  logical(1))
        if (any(non_replaceable)) {
            stop("None of the 'spectraVariables' must already exist in the ",
                 "chromData data.frame.")
        }
    }
    idx <- spectra$chromSpectraIndex
    spd <- spectraData(spectra, columns = spectraVariables)

    ## Aggregate and simplify singletons
    aggregated <- as.data.frame(
        lapply(spectraVariables, function(var) {
            res <- tapply(spd[[var]], idx, unique, simplify = FALSE)
            if (all(lengths(res) == 1L)) {
                res <- unlist(res, use.names = TRUE)
            }
            res
        }),
        stringsAsFactors = FALSE
    )
    names(aggregated) <- spectraVariables

    ## match order and combine
    aggregated <- aggregated[as.character(cd$chromSpectraIndex), , drop = FALSE]
    cd <- cbind(cd, aggregated)
    rownames(cd) <- NULL
    object@chromData <- cd
    object
}


#' Compute pairwise similarity between two chromatograms.
#'
#' Interpolates both chromatograms onto the union of their retention time
#' points within the overlapping range, then computes the correlation (or a
#' custom similarity via `FUN`).
#'
#' Used in:
#' - `compareChromatograms()`
#'
#' @param x,y `data.frame` with columns `rtime` and `intensity`.
#' @param MAPFUN function to align two chromatograms' retention times.
#'        Defaults to `matchRtime`.
#' @param FUN function to compute similarity from aligned intensities.
#'        Defaults to `cor`.
#' @param mapfun_args `list` of additional arguments passed to `MAPFUN`.
#' @param fun_args `list` of additional arguments passed to `FUN`.
#' @param minPeaks minimum number of overlapping RT points required to compute
#'        the similarity; pairs with fewer points return `NA` (score) and the
#'        actual count (n_peaks). Defaults to `2L`.
#' @return `numeric(2)`: similarity value and number of overlapping RT points.
#' @noRd
.compare_chrom_pair <- function(x, y, MAPFUN = matchRtime, FUN = cor,
                                mapfun_args = list(), fun_args = list(),
                                minPeaks = 2L) {
    aligned <- do.call(MAPFUN, c(list(x, y), mapfun_args))
    n <- length(aligned$x)
    if (n < minPeaks) return(c(NA_real_, n))
    c(do.call(FUN, c(list(aligned$x, aligned$y), fun_args)), n)
}

#' Compute a pairwise similarity array between two lists of peaks data.frames.
#'
#' Compares each chromatogram in `pd_x` with each chromatogram in `pd_y`.
#' When `pd_y` is not provided (default), `pd_x` is compared against itself
#' and the result is a symmetric n x n x 2 array.
#'
#' Used in:
#' - `compareChromatograms()`
#'
#' @param pd_x `list` of `data.frame` each with columns `rtime` and
#'        `intensity`.
#' @param pd_y `list` of `data.frame` each with columns `rtime` and
#'        `intensity`. Defaults to `pd_x` (self-comparison).
#' @param MAPFUN function to align retention times.
#' @param FUN function to compute similarity.
#' @param minPeaks minimum overlapping RT points to compute similarity.
#' @param labelsColumn optional `character` vector of row/column names. Only
#'        meaningful for self-comparison (when `pd_y` is not supplied).
#' @param ... passed to `.compare_chrom_pair()`.
#' @return A numeric array with dimensions `length(pd_x)` x `length(pd_y)` x 2.
#'         Layer 1 contains similarity scores; layer 2 the number of
#'         overlapping retention-time points used for each comparison.
#' @noRd

#' @importFrom BiocParallel bplapply SerialParam

.compare_chromatograms <- function(pd_x, pd_y = pd_x,
                                   MAPFUN = matchRtime, FUN = cor,
                                   minPeaks = 4L, labelsColumn = NULL,
                                   BPPARAM = SerialParam(),
                                   self = FALSE, ...) {
    nx <- length(pd_x)
    ny <- length(pd_y)
    if (nx == 0L || ny == 0L)
        return(array(numeric(0), dim = c(nx, ny, 2L),
                     dimnames = list(NULL, NULL, c("score", "n_peaks"))))

    dots <- list(...)
    mapfun_args <- dots[names(dots) %in% names(formals(MAPFUN))]
    fun_args <- dots[names(dots) %in% names(formals(FUN))]

    .lapply <- if (inherits(BPPARAM, "SerialParam")) lapply else
        function(X, FUN) bplapply(X, FUN, BPPARAM = BPPARAM)

    scores <- matrix(NA_real_, nx, ny)
    counts <- matrix(0L, nx, ny)

    if (self) {
        diag(scores) <- 1
        diag(counts) <- vapply(pd_x, nrow, integer(1L))
        if (nx > 1L) {
            row_vals <- .lapply(seq_len(nx - 1L), function(i) {
                vapply(seq.int(i + 1L, nx), function(j)
                    .compare_chrom_pair(pd_x[[i]], pd_x[[j]], MAPFUN, FUN,
                                        mapfun_args, fun_args, minPeaks),
                    numeric(2L))
            })
            for (i in seq_len(nx - 1L)) {
                js <- seq.int(i + 1L, nx)
                scores[i, js] <- row_vals[[i]][1L, ]
                scores[js, i] <- row_vals[[i]][1L, ]
                counts[i, js] <- row_vals[[i]][2L, ]
                counts[js, i] <- row_vals[[i]][2L, ]
            }
        }
    } else {
        row_vals <- .lapply(seq_len(nx), function(i) {
            vapply(seq_len(ny), function(j)
                .compare_chrom_pair(pd_x[[i]], pd_y[[j]], MAPFUN, FUN,
                                    mapfun_args, fun_args, minPeaks),
                numeric(2L))
        })
        for (i in seq_len(nx)) {
            scores[i, ] <- row_vals[[i]][1L, ]
            counts[i, ] <- row_vals[[i]][2L, ]
        }
    }

    arr <- array(NA_real_, dim = c(nx, ny, 2L),
                 dimnames = list(NULL, NULL, c("score", "n_peaks")))
    arr[, , 1L] <- scores
    arr[, , 2L] <- counts
    if (!is.null(labelsColumn)) {
        dimnames(arr)[[1L]] <- labelsColumn
        dimnames(arr)[[2L]] <- labelsColumn
    }
    arr
}

#' Resolve and validate the labels vector from a chromData column.
#'
#' Used in:
#' - `compareChromatograms()`
#'
#' @param object A `Chromatograms` object.
#' @param labelsColumn `character(1)` column name in `chromData()`, or `NULL`.
#' @return A character vector of labels, or `NULL`.
#' @noRd
.resolve_labels <- function(object, labelsColumn) {
    if (is.null(labelsColumn)) return(NULL)
    if (!is.character(labelsColumn) || length(labelsColumn) != 1L)
        stop("'labelsColumn' must be a single character string")
    labs <- chromData(object)[[labelsColumn]]
    if (is.null(labs))
        stop("Column '", labelsColumn, "' not found in chromData")
    if (anyDuplicated(labs))
        stop("Column '", labelsColumn, "' contains duplicated values")
    labs
}


#' Below are internal accessor functions, used ubiquitously in the package.
#' These directly access the internal slots (e.g., `@chromData`, `@peaksData`,
#' `@spectra`, etc.).
#' These are NOT to be used by general users.
#' @noRd
.backend <- function(object) {
    object@backend
}
.peaksData <- function(object) {
    if (is(object, "Chromatograms")) {
        return(object@backend@peaksData)
    }
    if (is(object, "ChromBackend")) {
        return(object@peaksData)
    }
    stop("'object' must be of class 'Chromatograms' or 'ChromBackend'.")
}
.chromData <- function(object) {
    if (is(object, "Chromatograms")) {
        return(object@backend@chromData)
    }
    if (is(object, "ChromBackend")) {
        return(object@chromData)
    }
    stop("'object' must be of class 'Chromatograms' or 'ChromBackend'.")
}
.inMemory <- function(object) {
    object@inMemory
}
.processing <- function(object) {
    object@processing
}
.processingQueue <- function(object) {
    object@processingQueue
}
.spectra <- function(object) {
    object@spectra
}

#' Compute peak boundaries for a single chromatogram.
#'
#' Finds the retention time boundaries of the tallest peak in a chromatogram
#' using `MsCoreUtils::valleys()` to locate the flanking valleys. When the
#' valley intensities exceed a baseline-relative threshold, falls back to a
#' threshold-based boundary search.
#'
#' Used in:
#' - `peakBoundary()`
#'
#' @param rtime numeric vector of retention times.
#' @param intensity numeric vector of intensities.
#' @param threshold numeric(1), fraction of the peak height above baseline
#'        used as a fallback cut-off (default 0.1).
#' @param baselineThreshold numeric(1), fraction of the peak height above
#'        baseline used to validate valley positions (default 0.1).
#' @param baselineQuantile numeric(1), quantile of the intensity distribution
#'        used as the baseline estimate (default 0.1).
#' @return A numeric vector of length 2 (left boundary, right boundary
#'   retention times), or `c(NA_real_, NA_real_)` when no valid peak
#'   is found.
#' @importFrom MsCoreUtils valleys
#' @importFrom stats quantile
#' @noRd
.peak_boundary_one <- function(rtime, intensity, threshold = 0.1,
                               baselineThreshold = 0.1,
                               baselineQuantile = 0.1) {
    n <- length(intensity)
    na_res <- c(NA_real_, NA_real_)
    if (n < 3L || all(is.na(intensity)))
        return(na_res)
    max_int <- max(intensity, na.rm = TRUE)
    if (max_int == 0)
        return(na_res)
    max_idx <- which.max(intensity)
    baseline_int <- quantile(intensity, probs = baselineQuantile,
                             na.rm = TRUE)
    peak_height <- max_int - baseline_int
    baseline_thresh <- baseline_int + peak_height * baselineThreshold
    v <- valleys(intensity, max_idx)
    left_idx  <- if ("left" %in% colnames(v)) v[1L, "left"] else 1L
    right_idx <- if ("right" %in% colnames(v)) v[1L, "right"] else n
    left_ok <- !is.na(intensity[left_idx]) &&
        intensity[left_idx] <= baseline_thresh &&
        !(left_idx > 1L && is.na(intensity[left_idx - 1L]))
    right_ok <- !is.na(intensity[right_idx]) &&
        intensity[right_idx] <= baseline_thresh &&
        !(right_idx < n && is.na(intensity[right_idx + 1L]))
    if (!left_ok || !right_ok) {
        thresh_val <- baseline_int + peak_height * threshold
        left_cand  <- which(intensity[seq_len(max_idx)] <= thresh_val)
        right_cand <- which(intensity[max_idx:n] <= thresh_val)
        left_idx  <- if (length(left_cand)) max(left_cand) else 1L
        right_idx <- if (length(right_cand))
            max_idx + min(right_cand) - 1L else n
    }
    c(unname(rtime[left_idx]), unname(rtime[right_idx]))
}

test_that("Chromatograms works", {
  ## empty object
  expect_true(is(.backend(c_empty), "ChromBackendMemory"))
  expect_true(is(c_empty, "Chromatograms"))
  expect_true(processingChunkSize(c_empty) == Inf)
  expect_true(c_empty@version == "0.1")
  expect_identical(.processingQueue(c_empty), list())

  ## object with backend
  expect_true(is(.backend(c_full), "ChromBackendMemory"))
  expect_true(is(c_full, "Chromatograms"))
  expect_true(processingChunkSize(c_full) == Inf)
  expect_true(c_full@version == "0.1")
  expect_identical(.processingQueue(c_full), list())

  expect_equal(processingChunkSize(c_full), Inf)
  c_chunk <- c_full
  processingChunkSize(c_chunk) <- 2
  expect_equal(processingChunkSize(c_chunk), 2)
  expect_equal(levels(processingChunkFactor(c_chunk)), c("1", "2"))

  ## method with Spectra works
  c_sp <- Chromatograms(s[1:2])
  expect_true(is(.backend(c_sp), "ChromBackendSpectra"))
  expect_true(is(c_sp, "Chromatograms"))
  expect_true(processingChunkSize(c_sp) == Inf)
  expect_true(c_sp@version == "0.1")
  expect_identical(.processingQueue(c_sp), list())
})

test_that("Chromatograms constructor from Spectra works with all parameters", {
  ## Basic construction with defaults
  chr <- Chromatograms(s)
  expect_s4_class(chr, "Chromatograms")
  expect_s4_class(.backend(chr), "ChromBackendSpectra")
  expect_equal(length(chr), 3L)

  ## With summarize.method = "sum" (default)
  chr_sum <- Chromatograms(s, summarize.method = "sum")
  expect_s4_class(chr_sum, "Chromatograms")
  expect_identical(.backend(chr_sum)@summaryFun, sumi)

  ## With summarize.method = "max"
  chr_max <- Chromatograms(s, summarize.method = "max")
  expect_s4_class(chr_max, "Chromatograms")
  expect_identical(.backend(chr_max)@summaryFun, maxi)

  ## With empty chromData (should create default)
  chr_empty_cd <- Chromatograms(s, chromData = data.frame())
  expect_s4_class(chr_empty_cd, "Chromatograms")
  expect_true(nrow(chromData(chr_empty_cd)) > 0)
  expect_true(all(
    names(coreChromVariables()) %in%
      chromVariables(chr_empty_cd)
  ))

  ## With custom chromData
  custom_cd <- data.frame(
    msLevel = 1L,
    dataOrigin = unique(s$dataOrigin),
    customCol = "test"
  )
  chr_custom <- Chromatograms(s, chromData = custom_cd)
  expect_s4_class(chr_custom, "Chromatograms")
  expect_true("customCol" %in% colnames(chromData(chr_custom)))
  expect_equal(
    chromData(chr_custom)$customCol,
    rep("test", length(unique(s$dataOrigin)))
  )

  ## With custom factorize.by
  chr_factby <- Chromatograms(s, factorize.by = "dataOrigin")
  expect_s4_class(chr_factby, "Chromatograms")
  expect_true(all(
    chromData(chr_factby)$dataOrigin == chromData(chr_factby)$chromSpectraIndex
  ))

  ## With spectraVariables
  expect_error(
    Chromatograms(s, spectraVariables = "polarity"),
    "must already exist in the chromData"
  )

  chr_specvars <- Chromatograms(s, spectraVariables = c("precursorMz"))
  expect_s4_class(chr_specvars, "Chromatograms")
  if ("precursorMz" %in% Spectra::spectraVariables(s)) {
    expect_true("precursorMz" %in% colnames(chromData(chr_specvars)))
  }

  ## spectraVariables should replace all-NA columns in provided chromData
  sp_small <- Spectra::Spectra(S4Vectors::DataFrame(
    rtime = c(1, 2),
    msLevel = c(1L, 1L),
    dataOrigin = c("A", "A"),
    polarity = c(1L, 1L)
  ))
  cd_na <- data.frame(msLevel = 1L, dataOrigin = "A", polarity = NA_integer_)
  chr_specvars_replace <- Chromatograms(
    sp_small,
    chromData = cd_na,
    spectraVariables = c("polarity")
  )
  expect_s4_class(chr_specvars_replace, "Chromatograms")
  expect_identical(chromData(chr_specvars_replace)$polarity, 1L)
})

test_that("Chromatograms constructor from ChromBackend works", {
  ## From ChromBackendMemory - already initialized
  chr_mem <- Chromatograms(be)
  expect_s4_class(chr_mem, "Chromatograms")
  expect_s4_class(.backend(chr_mem), "ChromBackendMemory")
  expect_equal(length(chr_mem), length(be))

  ## From empty ChromBackendMemory with chromData and peaksData parameters
  cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    mz = c(112.2, 123.3, 134.4),
    dataOrigin = c("mem1", "mem1", "mem1")
  )
  pdata <- list(
    data.frame(
      rtime = c(2.1, 2.5, 3.0, 3.4, 3.9),
      intensity = c(100, 250, 400, 300, 150)
    ),
    data.frame(rtime = numeric(), intensity = numeric()),
    data.frame(
      rtime = c(5.1, 5.8, 6.3, 6.9, 7.5),
      intensity = c(80, 500, 1200, 600, 120)
    )
  )
  chr <- Chromatograms(
    ChromBackendMemory(),
    chromData = cdata,
    peaksData = pdata
  )
  expect_s4_class(chr, "Chromatograms")
  expect_s4_class(.backend(chr), "ChromBackendMemory")
  expect_equal(length(chr), 3L)
  expect_identical(chromData(chr)$mz, cdata$mz)
  expect_identical(peaksData(chr), pdata)

  ## From ChromBackendMzR
  chr_mzr <- Chromatograms(be_mzr)
  expect_s4_class(chr_mzr, "Chromatograms")
  expect_s4_class(.backend(chr_mzr), "ChromBackendMzR")
  expect_equal(length(chr_mzr), length(be_mzr))

  ## From ChromBackendSpectra
  chr_spec <- Chromatograms(be_sp)
  expect_s4_class(chr_spec, "Chromatograms")
  expect_s4_class(.backend(chr_spec), "ChromBackendSpectra")
  expect_equal(length(chr_spec), length(be_sp))

  ## With processingQueue
  pq <- list(ProcessingStep(
    "smooth",
    list(method = "SavitzkyGolay", halfWindowSize = 2L)
  ))
  chr_pq <- Chromatograms(be, processingQueue = pq)
  expect_equal(length(.processingQueue(chr_pq)), 1)
})

test_that("Chromatograms constructor handles edge cases", {
  ## Empty Spectra
  empty_s <- Spectra()
  chr_empty <- Chromatograms(empty_s)
  expect_s4_class(chr_empty, "Chromatograms")
  expect_equal(length(chr_empty), 0)

  ## Missing object (creates empty ChromBackendMemory)
  chr_missing <- Chromatograms()
  expect_s4_class(chr_missing, "Chromatograms")
  expect_s4_class(.backend(chr_missing), "ChromBackendMemory")
  expect_equal(length(chr_missing), 0)
})

test_that("show, Chromatograms - ChromBackendMemory works", {
  expect_output(show(c_full), "ChromBackendMemory")
  res <- c_full
  res@processing <- c("a", "b", "c", "d")
  expect_output(show(res), "1 more processings")
  res@processingQueue <- list("a", "b", "c", "d")
  expect_output(show(res), "4 processing step")
})

test_that("show, Chromatograms - ChromBackendMzR works", {
  expect_output(show(c_mzr), "ChromBackendMzR")
  res <- c_mzr
  res@processing <- c("a", "b", "c", "d")
  expect_output(show(res), "1 more processings")
  res@processingQueue <- list("a", "b", "c", "d")
  expect_output(show(res), "4 processing step")
})


test_that("setBackend works correctly", {
  c_mzr_new <- setBackend(c_mzr, backend = ChromBackendMemory())
  expect_s4_class(.backend(c_mzr_new), "ChromBackendMemory")
  expect_identical(chromData(c_mzr_new), chromData(c_mzr))
  expect_identical(peaksData(c_mzr_new), peaksData(c_mzr))
  expect_identical(peaksData(c_mzr_new), peaksData(c_mzr))

  processingChunkSize(c_mzr) <- 100
  f <- processingChunkFactor(c_mzr)
  expect_true(length(levels(f)) > 1)
  c_mzr_new <- setBackend(c_mzr, backend = ChromBackendMemory(), f = f)
  expect_s4_class(.backend(c_mzr_new), "ChromBackendMemory")
  expect_identical(chromData(c_mzr_new), chromData(c_mzr))
  expect_identical(peaksData(c_mzr_new), peaksData(c_mzr))
  expect_identical(.peaksData(c_mzr_new), peaksData(c_mzr))

  expect_error(
    setBackend(c_mzr, backend = ChromBackendMzR()),
    "does not support"
  )

  c_sp_new <- setBackend(c_sp, backend = ChromBackendMemory())
  expect_true(!all(c("rtMin", "rtMax") %in% colnames(chromData(c_sp_new))))

  ## setBackend should empty the processing queue so steps are not
  ## double-applied. peaksData on the Chromatograms object should equal
  ## peaksData directly from its backend (queue is empty).
  c_queued <- filterPeaksData(c_mzr,
    variables = "rtime", ranges = c(12.5, 45.5)
  )
  expect_length(.processingQueue(c_queued), 1)
  pd_before <- peaksData(c_queued)
  c_queued_mem <- setBackend(c_queued, backend = ChromBackendMemory())
  expect_length(.processingQueue(c_queued_mem), 0)
  expect_identical(peaksData(c_queued_mem), peaksData(.backend(c_queued_mem)))
  expect_equal(peaksData(c_queued_mem), pd_before)
  expect_true(any(grepl("Switch backend", c_queued_mem@processing)))
})

test_that("setBackend preserves peaksData order with duplicate chromSpectraIndex", {
  ## This test reproduces a bug where setBackend would scramble peaksData order
  ## when multiple chromatograms share the same chromSpectraIndex (e.g., EICs
  ## with different mz windows from the same file). The bug occurred because
  ## peaksData() was returning data in factor level order instead of row order.
  sp <- Spectra(DataFrame(
    mz = NumericList(
      c(100, 101, 102), c(100, 101, 102), c(100, 101, 102),
      c(100, 101, 102), c(100, 101, 102), c(100, 101, 102),
      compress = FALSE
    ),
    intensity = NumericList(
      c(10, 20, 30), c(15, 25, 35), c(20, 30, 40),
      c(100, 200, 300), c(150, 250, 350), c(200, 300, 400),
      compress = FALSE
    ),
    rtime = c(1, 2, 3, 4, 5, 6),
    msLevel = rep(1L, 6),
    dataOrigin = c("A", "A", "A", "B", "B", "B")
  ))

  ## Create multiple EICs from same spectra (sharing chromSpectraIndex)
  custom_chromData <- data.frame(
    msLevel = rep(1L, 4),
    mzMin = c(100, 100, 100, 100),
    mzMax = c(101, 102, 101, 102),
    rtMin = c(1, 1, 4, 4),
    rtMax = c(3, 3, 6, 6),
    dataOrigin = c("A", "A", "B", "B")
  )

  chr <- Chromatograms(sp, chromData = custom_chromData)

  ## Get peaksData before setBackend
  pd_before <- peaksData(chr)

  ## Convert to memory backend
  chr_mem <- setBackend(chr, ChromBackendMemory())

  ## Get peaksData after setBackend
  pd_after <- peaksData(chr_mem)

  ## peaksData should be identical before and after setBackend
  expect_identical(length(pd_before), length(pd_after))

  ## Each element should have the same intensity values
  for (i in seq_along(pd_before)) {
    expect_identical(pd_before[[i]]$intensity, pd_after[[i]]$intensity)
    expect_identical(pd_before[[i]]$rtime, pd_after[[i]]$rtime)
  }

  ## Verify subsetting after setBackend works correctly
  ## This tests the scenario where subsetting after setBackend returns wrong data
  chr_sub <- chr_mem[2]  # Get second chromatogram (1_A with mzMax=102)
  pd_sub <- peaksData(chr_sub)
  expect_identical(pd_sub[[1]]$intensity, pd_before[[2]]$intensity)

  chr_sub4 <- chr_mem[4]  # Get fourth chromatogram (1_B with mzMax=102)
  pd_sub4 <- peaksData(chr_sub4)
  expect_identical(pd_sub4[[1]]$intensity, pd_before[[4]]$intensity)
})

test_that("$ works correctly", {
  expect_identical(msLevel(c_full), c_full$msLevel)
  expect_identical(chromIndex(c_mzr), c_mzr$chromIndex)
  expect_identical(intensity(c_full), c_full$intensity)
  expect_identical(intensity(c_mzr), c_mzr$intensity)
  tmp <- c_full
  tmp$msLevel <- c(2L, 2L, 3L)
  expect_identical(msLevel(tmp), c(2L, 2L, 3L))
  tmp$intensity <- lapply(tmp$intensity, function(x) x + 10)
  expect_false(identical(intensity(tmp), intensity(c_full)))
})

test_that("[ works correctly", {
  c_sub <- c_full[1:2]
  expect_true(is(c_sub, "Chromatograms"))
  expect_equal(nrow(chromData(c_sub)), 2)
  expect_equal(length(peaksData(c_sub)), 2)
  expect_error(c_full[1:2, 1], "by columns is not")

  c_sub <- c_full[1]
  expect_equal(c_sub, c_sub[])
})

test_that("[[ works properly", {
  expect_error(c_full[[1]], "character")
  expect_error(c_full[["test", 1]], "not supported")
  expect_error(c_full[["test"]], "No variable")
  expect_equal(c_full[["msLevel"]], msLevel(c_full))
  expect_equal(c_full[["msLevel"]], .backend(c_full)[["msLevel"]])

  ## replace
  expect_error(c_full[[1]] <- 1, "character defining the chromatogram")
  expect_error(c_full[["msLevel", 4]] <- 1, "not supported")
  expect_error(c_full[["test"]] <- 1, "No variable")
  repet <- c_full
  repet[["msLevel"]] <- rep(2L, length(repet))
  expect_false(identical(msLevel(repet), msLevel(c_full)))
})

test_that("factorize() works", {
  tmp <- c_sp
  tmp$msLevel <- c(1L, 2L, 3L)
  idx_before <- chromSpectraIndex(.backend(tmp))
  tmp <- factorize(tmp)
  idx_after <- chromSpectraIndex(.backend(tmp))
  expect_false(identical(idx_before, idx_after))
})

test_that("chromExtract, Chromatograms works correctly", {
  peak.table <- data.frame(
    msLevel = 1L,
    dataOrigin = "mem1",
    rtMin = 12.5,
    rtMax = 14.0
  )
  res <- chromExtract(c_full, peak.table, by = c("msLevel", "dataOrigin"))
  expect_s4_class(res, "Chromatograms")
  expect_true(length(res) == nrow(peak.table))

  pk_nomatch <- transform(peak.table, dataOrigin = "no_such_sample")
  expect_error(
    chromExtract(c_full, pk_nomatch, by = c("msLevel", "dataOrigin")),
    "Some combinations in"
  )
})

test_that("c, Chromatograms works", {
  ## Create two simple Chromatograms objects
  cdata1 <- data.frame(
    msLevel = c(1L, 1L),
    mz = c(112.2, 123.3),
    dataOrigin = c("file1", "file1")
  )
  pdata1 <- list(
    data.frame(rtime = c(1.0, 2.0, 3.0), intensity = c(100, 200, 150)),
    data.frame(rtime = c(1.0, 2.0, 3.0), intensity = c(80, 120, 90))
  )
  chr1 <- Chromatograms(
    ChromBackendMemory(),
    chromData = cdata1,
    peaksData = pdata1
  )

  cdata2 <- data.frame(
    msLevel = c(2L, 2L),
    mz = c(134.4, 145.5),
    dataOrigin = c("file2", "file2")
  )
  pdata2 <- list(
    data.frame(rtime = c(4.0, 5.0, 6.0), intensity = c(300, 400, 350)),
    data.frame(rtime = c(4.0, 5.0, 6.0), intensity = c(200, 250, 180))
  )
  chr2 <- Chromatograms(
    ChromBackendMemory(),
    chromData = cdata2,
    peaksData = pdata2
  )

  ## Test c() method

  res <- c(chr1, chr2)
  expect_s4_class(res, "Chromatograms")
  expect_equal(length(res), length(chr1) + length(chr2))
  expect_equal(msLevel(res), c(msLevel(chr1), msLevel(chr2)))
  expect_equal(chromData(res)$mz, c(cdata1$mz, cdata2$mz))

  ## c() with single object returns same object
  res_single <- c(chr1)
  expect_s4_class(res_single, "Chromatograms")
  expect_equal(length(res_single), length(chr1))

  ## c() with empty object
  res_empty <- c(chr1, Chromatograms())
  expect_s4_class(res_empty, "Chromatograms")
  expect_equal(length(res_empty), length(chr1))

  ## Cannot concatenate Chromatograms with non-empty processing queue
  chr_with_pq <- chr1
  chr_with_pq@processingQueue <- list(ProcessingStep(function(x) x))
  expect_error(c(chr_with_pq, chr2), "non-empty processing queue")

  ## Cannot concatenate different backend types
  expect_error(c(c_full, c_mzr), "same type")
})

test_that("concatenateChromatograms works", {
  ## Create test objects
  cdata1 <- data.frame(
    msLevel = 1L,
    mz = 100.0,
    dataOrigin = "file1"
  )
  pdata1 <- list(
    data.frame(rtime = c(1.0, 2.0), intensity = c(100, 200))
  )
  chr1 <- Chromatograms(
    ChromBackendMemory(),
    chromData = cdata1,
    peaksData = pdata1
  )

  cdata2 <- data.frame(
    msLevel = 2L,
    mz = 200.0,
    dataOrigin = "file2"
  )
  pdata2 <- list(
    data.frame(rtime = c(3.0, 4.0), intensity = c(300, 400))
  )
  chr2 <- Chromatograms(
    ChromBackendMemory(),
    chromData = cdata2,
    peaksData = pdata2
  )

  ## Test with individual Chromatograms arguments
  res <- concatenateChromatograms(chr1, chr2)
  expect_s4_class(res, "Chromatograms")
  expect_equal(length(res), 2L)
  expect_equal(msLevel(res), c(1L, 2L))

  ## Test with a list of Chromatograms
  chr_list <- list(chr1, chr2)
  res_list <- concatenateChromatograms(chr_list)
  expect_s4_class(res_list, "Chromatograms")
  expect_equal(length(res_list), 2L)

  ## Results should be identical
  expect_equal(msLevel(res), msLevel(res_list))
  expect_equal(chromData(res)$mz, chromData(res_list)$mz)

  ## Error with non-Chromatograms objects
  expect_error(concatenateChromatograms(chr1, "not a chromatogram"),
               "Can only concatenate")

  ## Test with empty list
  res_empty <- concatenateChromatograms(list())
  expect_s4_class(res_empty, "Chromatograms")
  expect_equal(length(res_empty), 0L)
})

test_that("filterEmptyChromatograms,Chromatograms works", {
    ## Empty Chromatograms object: returns empty without adding processing log
    res <- filterEmptyChromatograms(c_empty)
    expect_true(is(res, "Chromatograms"))
    expect_equal(length(res), 0)
    expect_equal(length(res@processing), 0)

    ## All chromatograms have peaks data, none should be removed
    res <- filterEmptyChromatograms(c_full)
    expect_equal(length(res), length(c_full))
    expect_true(length(res@processing) == 1)

    ## Create object with some empty chromatograms
    cdata_mix <- data.frame(
        msLevel = c(1L, 1L, 1L, 1L),
        mz = c(100, 200, 300, 400),
        dataOrigin = c("a", "b", "c", "d")
    )
    pdata_mix <- list(
        data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30)),
        data.frame(rtime = numeric(), intensity = numeric()),
        data.frame(rtime = c(4, 5), intensity = c(40, 50)),
        data.frame(rtime = numeric(), intensity = numeric())
    )
    chr_mix <- Chromatograms(
        ChromBackendMemory(),
        chromData = cdata_mix,
        peaksData = pdata_mix
    )
    expect_equal(length(chr_mix), 4)
    res <- filterEmptyChromatograms(chr_mix)
    expect_equal(length(res), 2)
    expect_equal(mz(res), c(100, 300))
    expect_true(all(lengths(res) > 0))
    expect_true(length(res@processing) == 1)

    ## All empty
    pdata_empty <- list(
        data.frame(rtime = numeric(), intensity = numeric()),
        data.frame(rtime = numeric(), intensity = numeric())
    )
    cdata_empty <- data.frame(
        msLevel = c(1L, 1L),
        mz = c(100, 200),
        dataOrigin = c("a", "b")
    )
    chr_all_empty <- Chromatograms(
        ChromBackendMemory(),
        chromData = cdata_empty,
        peaksData = pdata_empty
    )
    res <- filterEmptyChromatograms(chr_all_empty)
    expect_equal(length(res), 0)
})

test_that("split, Chromatograms works", {
  ## Create a Chromatograms with multiple msLevels
  cdata <- data.frame(
    msLevel = c(1L, 1L, 2L, 2L),
    mz = c(100, 110, 200, 210),
    dataOrigin = c("A", "A", "B", "B")
  )
  pdata <- list(
    data.frame(rtime = c(1, 2), intensity = c(10, 20)),
    data.frame(rtime = c(1, 2), intensity = c(15, 25)),
    data.frame(rtime = c(3, 4), intensity = c(30, 40)),
    data.frame(rtime = c(3, 4), intensity = c(35, 45))
  )
  chr <- Chromatograms(
    ChromBackendMemory(),
    chromData = cdata,
    peaksData = pdata
  )

  ## Split by msLevel
  res <- split(chr, f = chr$msLevel)
  expect_true(is.list(res))
  expect_equal(length(res), 2L)
  expect_s4_class(res[[1L]], "Chromatograms")
  expect_s4_class(res[[2L]], "Chromatograms")
  expect_equal(length(res[[1L]]), 2L)
  expect_equal(length(res[[2L]]), 2L)
  expect_true(all(msLevel(res[[1L]]) == 1L))
  expect_true(all(msLevel(res[[2L]]) == 2L))

  ## Split by dataOrigin
  res_origin <- split(chr, f = chr$dataOrigin)
  expect_equal(length(res_origin), 2L)
  expect_equal(length(res_origin[["A"]]), 2L)
  expect_equal(length(res_origin[["B"]]), 2L)

  ## Split with factor that has more levels than data
  f <- factor(c("a", "a", "b", "b"), levels = c("a", "b", "c"))
  res_factor <- split(chr, f = f)
  expect_equal(length(res_factor), 3L)
  expect_equal(length(res_factor[["c"]]), 0L)
})

test_that("peaksData, Chromatograms, ChrombackendMemory works as expected", {
    peaks <- peaksData(c_full)
    backend_peaks <- peaksData(.backend(c_full))
    expect_equal(peaks, backend_peaks)

    rtime_data <- peaksData(c_full, columns = "rtime", drop = TRUE)
    expect_type(rtime_data, "list")
    expect_equal(rtime_data, peaksData(.backend(c_full),
        columns = "rtime",
        drop = TRUE
    ))

    c_empty_queue <- c_full
    c_empty_queue@processingQueue <- list()
    expect_equal(peaksData(c_empty_queue), peaksData(.backend(c_empty_queue)))

    c_queued <- filterPeaksData(c_full,
        variables = c("rtime"),
        ranges = c(12.5, 45.5)
    )
    expect_false(identical(peaksData(c_queued), peaksData(.backend(c_queued))))
    c_queued2 <- filterPeaksData(c_queued,
        variables = c("intensity"),
        ranges = c(100, 1000)
    )
    expect_false(identical(peaksData(c_queued), peaksData(c_queued2)))

    c_queued <- filterPeaksData(c_mzr,
        variables = c("rtime"),
        ranges = c(12.5, 45.5)
    )
    expect_false(identical(peaksData(c_queued),
                           peaksData(.backend(c_queued))))
    c_queued2 <- filterPeaksData(c_queued,
        variables = c("intensity"),
        ranges = c(100, 1000)
    )
    expect_false(identical(peaksData(c_queued), peaksData(c_queued2)))
})

test_that("peaksData, Chromatogram, ChromBackendMzR works as expected", {
    peaks <- peaksData(c_mzr)
    backend_peaks <- peaksData(.backend(c_mzr))
    expect_equal(peaks, backend_peaks)

    rtime_data <- peaksData(c_mzr, columns = "rtime", drop = TRUE)
    expect_type(rtime_data, "list")
    expect_equal(rtime_data, peaksData(.backend(c_mzr),
        columns = "rtime",
        drop = TRUE
    ))

    c_empty_queue <- c_mzr
    c_empty_queue@processingQueue <- list()
    expect_equal(peaksData(c_empty_queue), peaksData(.backend(c_empty_queue)))
})

test_that("peaksData replacement works as expected", {
    new_peaks <- peaksData(c_full)
    new_peaks[[1]][4, "rtime"] <- 999.9

    peaksData(c_full) <- new_peaks
    updated_peaks <- peaksData(c_full)
    expect_equal(new_peaks, updated_peaks)

    expect_error(
        peaksData(c_mzr) <- list(data.frame()),
        "replace peaks data"
    )
})


test_that("peaksVariables works as expected", {
    vars <- peaksVariables(c_full)
    backend_vars <- peaksVariables(.backend(c_full))
    expect_equal(vars, backend_vars)
    expect_true("rtime" %in% vars)
    expect_true("intensity" %in% vars)

    vars <- peaksVariables(c_mzr)
    backend_vars <- peaksVariables(.backend(c_mzr))
    expect_equal(vars, backend_vars)
    expect_true("rtime" %in% vars)
    expect_true("intensity" %in% vars)
})

test_that("rtime accessor and replacement work as expected", {
    rtime_data <- rtime(c_full)
    backend_rtime <- peaksData(.backend(c_full), columns = "rtime",
                               drop = TRUE)
    expect_equal(rtime_data, backend_rtime)

    rtime_data <- rtime(c_mzr)
    backend_rtime <- peaksData(.backend(c_mzr), columns = "rtime", drop = TRUE)
    expect_equal(rtime_data, backend_rtime)

    new_rtime <- rtime(c_full)
    new_rtime[[1]][4] <- 888.8
    rtime(c_full) <- new_rtime
    expect_equal(rtime(c_full), new_rtime)
    expect_error(
        rtime(c_mzr)[[1]] <- rep(1, length(rtime(c_mzr)[[1]])),
        "Cannot replace peaks data in a read-only backend"
    )
})

test_that("intensity accessor and replacement work as expected", {
    intensity_data <- intensity(c_full)
    backend_intensity <- peaksData(.backend(c_full),
        columns = "intensity",
        drop = TRUE
    )
    expect_equal(intensity_data, backend_intensity)

    intensity_data <- intensity(c_mzr)
    backend_intensity <- peaksData(.backend(c_mzr),
        columns = "intensity",
        drop = TRUE
    )
    expect_equal(intensity_data, backend_intensity)

    new_intensity <- intensity(c_full)
    new_intensity[[1]][1] <- 777.7
    intensity(c_full) <- new_intensity
    expect_equal(intensity(c_full), new_intensity)
    expect_error(
        intensity(c_mzr)[[1]] <- rep(1, length(intensity(c_mzr)[[1]])),
        "Cannot replace peaks data in a read-only backend"
    )
})

test_that("filterPeaksData queues the correct processing step", {
    c_filtered <- filterPeaksData(c_full,
        variables = c("rtime"),
        ranges = c(12.5, 45.5)
    )
    queue <- .processingQueue(c_filtered)
    expect_length(queue, 1)
    expect_equal(queue[[1]]@FUN, filterPeaksData)
    expect_equal(queue[[1]]@ARGS, list(
        variables = c("rtime"),
        ranges = c(12.5, 45.5),
        match = c("any", "all"),
        keep = TRUE
    ))

    expect_match(.processing(c_filtered), "Filter: remove peaks")
})

test_that("Chromatograms, imputePeaksData works", {
    pdata <- list(
        data.frame(
            rtime = seq(12, 20, by = 0.5),
            intensity = c(123.3, 153.6, NA, 200, 210, 230, NA, 250, 260, 280,
                          300, 320, 340, 360, 380, 400, 420)
        ),
        data.frame(
            rtime = seq(45, 55, by = 0.5),  # length 21
            intensity = c(100, NA, 120, 130, 140, NA, 160, 170, 180, 190, 200,
                          210, 220, 230, 240, 250, 260, 270, 280, 290, 300)
        ),
        data.frame(
            rtime = seq(10, 18, by = 0.5),  # length 17
            intensity = c(NA, 153.6, 2354.3, 243.4, 260, NA, 280, 300, 320,
                          340, 360, 380, 400, 420, 440, 460, 480)
        )
    )

    tmp <- backendInitialize(be_empty, chromData = cdata, peaksData = pdata)
    c_tmp <- Chromatograms(tmp)

    for (meth in c("linear", "spline")) {
        c_imp <- imputePeaksData(c_tmp, method = meth, span = 0.3,
                                 sd = 1, window = 2, extrapolate = TRUE)
        expect_s4_class(c_imp, "Chromatograms")

        expect_equal(length(peaksData(c_imp)), length(peaksData(c_tmp)))
        expect_true(all(!is.na(unlist(intensity(c_imp)))))
        expect_true(any(grepl(meth, .processing(c_imp))))
    }

    for (meth in c("gaussian", "loess")) {
        c_imp <- imputePeaksData(c_tmp, method = meth, span = 0.3,
                                 sd = 1, window = 2, extrapolate = TRUE)
        expect_warning(c_imp <- applyProcessing(c_imp,
                                                BPPARAM = SerialParam()),
                       "Falling back to linear interpolation")
        expect_s4_class(c_imp, "Chromatograms")
        expect_equal(length(peaksData(c_imp)), length(peaksData(c_tmp)))
        expect_true(all(!is.na(unlist(intensity(c_imp)))))
        expect_true(any(grepl(meth, .processing(c_imp))))
    }

    ## Test extrapolate = FALSE (default)
    ## Interior NAs should be interpolated, leading/trailing NAs remain
    pdata_edge <- list(
        data.frame(
            rtime = 1:10,
            intensity = c(NA, NA, 100, NA, 140, 160, 180, NA, NA, NA)
        )
    )
    tmp_edge <- backendInitialize(be_empty, chromData = cdata[1, , drop = FALSE],
                                  peaksData = pdata_edge)
    c_edge <- Chromatograms(tmp_edge)

    ## extrapolate = FALSE (default) - leading/trailing NAs remain, interior filled
    c_imp_interp <- imputePeaksData(c_edge, method = "linear")
    ints <- intensity(c_imp_interp)[[1]]
    expect_true(is.na(ints[1]))
    expect_true(is.na(ints[2]))
    expect_true(is.na(ints[8]))
    expect_true(is.na(ints[9]))
    expect_true(is.na(ints[10]))
    # interior NA at position 4 should be interpolated
    expect_false(is.na(ints[4]))
    # all interior values (positions 3-7) should have no NAs
    expect_false(any(is.na(ints[3:7])))

    ## extrapolate = TRUE - no NAs remain
    c_imp_extrap <- imputePeaksData(c_edge, method = "linear", extrapolate = TRUE)
    expect_false(any(is.na(unlist(intensity(c_imp_extrap)))))

    ## Check processing log mentions extrapolation when enabled
    expect_true(grepl("with extrapolation", .processing(c_imp_extrap)))
})

test_that("peakBoundary returns per-chromatogram matrix.", {
    cdata <- data.frame(
        msLevel = c(1L, 1L),
        mz = c(100.0, 200.0),
        dataOrigin = c("mem1", "mem1")
    )
    pdata <- list(
        data.frame(rtime = c(2.1, 2.5, 3.0, 3.4, 3.9),
                   intensity = c(100, 250, 400, 300, 150)),
        data.frame(rtime = c(1, 2, 3, 4, 5, 6, 7),
                   intensity = c(0, 10, 50, 100, 50, 10, 0))
    )
    chr <- Chromatograms(ChromBackendMemory(), chromData = cdata,
                         peaksData = pdata)
    tmp <- peakBoundary(chr)
    expect_true(is.matrix(tmp))
    expect_equal(nrow(tmp), length(chr))
    expect_equal(colnames(tmp), c("left_boundary", "right_boundary"))
    expect_true(all(tmp[, "left_boundary"] <= tmp[, "right_boundary"],
                    na.rm = TRUE))
})

test_that("peakBoundary returns correct boundaries for clean symmetric peak.", {
    cdata_pb <- data.frame(msLevel = 1L, mz = 100, dataOrigin = "s1")
    pdata_pb <- list(data.frame(
        rtime = c(1, 2, 3, 4, 5, 6, 7),
        intensity = c(0, 10, 50, 100, 50, 10, 0)
    ))
    chr_pb <- Chromatograms(ChromBackendMemory(), chromData = cdata_pb,
                            peaksData = pdata_pb)
    tmp <- peakBoundary(chr_pb)
    expect_equal(unname(tmp[1, "left_boundary"]), 1)
    expect_equal(unname(tmp[1, "right_boundary"]), 7)
})

test_that("peakBoundary handles empty chromatogram.", {
    cdata_e <- data.frame(msLevel = 1L, mz = 100, dataOrigin = "s1")
    pdata_e <- list(data.frame(rtime = numeric(0), intensity = numeric(0)))
    chr_empty <- Chromatograms(ChromBackendMemory(), chromData = cdata_e,
                               peaksData = pdata_e)
    tmp <- peakBoundary(chr_empty)
    expect_true(all(is.na(tmp)))

    ## Also n < 3 returns NA
    pdata_2 <- list(data.frame(rtime = c(1, 2), intensity = c(10, 20)))
    chr_2 <- Chromatograms(ChromBackendMemory(), chromData = cdata_e,
                           peaksData = pdata_2)
    tmp2 <- peakBoundary(chr_2)
    expect_true(all(is.na(tmp2)))
})

test_that("peakBoundary handles all-NA intensities.", {
    cdata_na <- data.frame(msLevel = 1L, mz = 100, dataOrigin = "s1")
    pdata_na <- list(data.frame(
        rtime = c(1, 2, 3, 4, 5),
        intensity = c(NA_real_, NA_real_, NA_real_, NA_real_, NA_real_)
    ))
    chr_allna <- Chromatograms(ChromBackendMemory(), chromData = cdata_na,
                               peaksData = pdata_na)
    tmp <- peakBoundary(chr_allna)
    expect_true(all(is.na(tmp)))
})

test_that("peakBoundary handles all-zero intensities.", {
    cdata_z <- data.frame(msLevel = c(1L, 1L), mz = c(100, 200),
                          dataOrigin = c("s1", "s1"))
    pdata_z <- list(
        data.frame(rtime = c(1, 2, 3, 4, 5),
                   intensity = c(0, 0, 0, 0, 0)),
        data.frame(rtime = c(1, 2, 3, 4, 5, 6, 7),
                   intensity = c(0, 0, 0, 0, 0, 0, 0))
    )
    chr_zero <- Chromatograms(ChromBackendMemory(), chromData = cdata_z,
                              peaksData = pdata_z)
    tmp <- peakBoundary(chr_zero)
    expect_true(all(is.na(tmp)))
})

test_that("peakBoundary threshold parameter works.", {
    cdata_t <- data.frame(msLevel = 1L, mz = 100, dataOrigin = "s1")
    pdata_t <- list(data.frame(
        rtime = c(1, 2, 3, 4, 5, 6, 7),
        intensity = c(5, 20, 60, 100, 60, 20, 5)
    ))
    chr_t <- Chromatograms(ChromBackendMemory(), chromData = cdata_t,
                           peaksData = pdata_t)
    tmp <- peakBoundary(chr_t, threshold = 0.1)
    expect_true(is.matrix(tmp))
    expect_false(any(is.na(tmp)))
    tmp_05 <- peakBoundary(chr_t, threshold = 0.05)
    expect_true(is.matrix(tmp_05))
    expect_false(any(is.na(tmp_05)))
})

test_that("peakBoundary correctly isolates tallest peak in multi-peak data.", {
    cdata_m <- data.frame(msLevel = 1L, mz = 100, dataOrigin = "s1")
    pdata_m <- list(data.frame(
        rtime = 1:9,
        intensity = c(0, 5, 10, 5, 0, 3, 20, 3, 0)
    ))
    chr_m <- Chromatograms(ChromBackendMemory(), chromData = cdata_m,
                           peaksData = pdata_m)
    tmp <- peakBoundary(chr_m)
    ## Should return boundaries for the tallest peak (rtime 5-9 region)
    expect_equal(unname(tmp[1, "left_boundary"]), 5)
    expect_equal(unname(tmp[1, "right_boundary"]), 9)
})

test_that("peakBoundary handles NA-adjacent boundaries.", {
    cdata_na_adj <- data.frame(msLevel = 1L, mz = 100, dataOrigin = "s1")
    pdata_na_adj <- list(data.frame(
        rtime = c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
        intensity = c(NA, NA, 50, 100, 500, 100, 50, NA, NA, 5)
    ))
    chr_na_adj <- Chromatograms(ChromBackendMemory(), chromData = cdata_na_adj,
                                peaksData = pdata_na_adj)
    expect_no_error(peakBoundary(chr_na_adj))
    tmp <- peakBoundary(chr_na_adj)
    expect_true(is.matrix(tmp))
    expect_false(any(is.na(tmp)))
})

test_that("peakBoundary validates threshold parameter.", {
    cdata_v <- data.frame(msLevel = 1L, mz = 100, dataOrigin = "s1")
    pdata_v <- list(data.frame(
        rtime = c(1, 2, 3, 4, 5),
        intensity = c(10, 50, 100, 50, 10)
    ))
    chr_v <- Chromatograms(ChromBackendMemory(), chromData = cdata_v,
                           peaksData = pdata_v)
    expect_error(peakBoundary(chr_v, threshold = "a"),
                 "single non-missing numeric")
    expect_error(peakBoundary(chr_v, threshold = c(0.1, 0.2)),
                 "single non-missing numeric")
    expect_error(peakBoundary(chr_v, threshold = NA),
                 "single non-missing numeric")
    expect_error(peakBoundary(chr_v, threshold = -0.1),
                 "must be >= 0 and < 1")
    expect_error(peakBoundary(chr_v, threshold = 1),
                 "must be >= 0 and < 1")
    expect_error(peakBoundary(chr_v, threshold = 2),
                 "must be >= 0 and < 1")
    ## baselineThreshold validation
    expect_error(peakBoundary(chr_v, baselineThreshold = "a"),
                 "baselineThreshold.*single non-missing numeric")
    expect_error(peakBoundary(chr_v, baselineThreshold = -0.1),
                 "baselineThreshold.*must be >= 0 and < 1")
    ## baselineQuantile validation
    expect_error(peakBoundary(chr_v, baselineQuantile = "a"),
                 "baselineQuantile.*single non-missing numeric")
    expect_error(peakBoundary(chr_v, baselineQuantile = 1.5),
                 "baselineQuantile.*must be >= 0 and <= 1")
})

## ---- compareChromatograms tests ----

test_that("compareChromatograms default returns correct array structure.", {
    res <- compareChromatograms(c_full)
    expect_true(is.array(res))
    expect_equal(dim(res), c(3L, 3L, 2L))
    expect_equal(dimnames(res)[[3L]], c("score", "n_peaks"))
    ## Scores layer: symmetric, diagonal = 1
    expect_equal(diag(res[, , 1L]), c(1, 1, 1))
    expect_equal(res[1, 2, 1L], res[2, 1, 1L])
    expect_equal(res[1, 3, 1L], res[3, 1, 1L])
    expect_equal(res[2, 3, 1L], res[3, 2, 1L])
    expect_equal(unname(res[1, 3, 1L]), 1)
    expect_true(is.na(res[1, 2, 1L]))
    expect_true(is.na(res[2, 3, 1L]))
    ## Count layer: diagonal > 0, non-overlapping = 0
    expect_true(all(diag(res[, , 2L]) > 0))
    expect_equal(unname(res[1, 2, 2L]), 0)
    expect_equal(unname(res[2, 3, 2L]), 0)
    expect_null(rownames(res))
    expect_null(colnames(res))
})

test_that("compareChromatograms default, empty object returns 0x0x2 array.", {
    res <- compareChromatograms(c_empty)
    expect_true(is.array(res))
    expect_equal(dim(res), c(0L, 0L, 2L))
})

test_that("compareChromatograms default, single chromatogram returns 1x1x2.", {
    res <- compareChromatograms(c_full[1])
    expect_true(is.array(res))
    expect_equal(dim(res), c(1L, 1L, 2L))
    expect_equal(unname(res[1, 1, 1L]), 1)
    expect_true(res[1, 1, 2L] > 0)
})

test_that("compareChromatograms method = 'spearman' and 'kendall' work.", {
    res_s <- compareChromatograms(c_full, method = "spearman")
    expect_true(is.array(res_s))
    expect_equal(dim(res_s), c(3L, 3L, 2L))
    expect_equal(diag(res_s[, , 1L]), c(1, 1, 1))
    ## Identical chromatograms → 1 regardless of method
    expect_equal(unname(res_s[1, 3, 1L]), 1)
    ## Non-overlapping → NA
    expect_true(is.na(res_s[1, 2, 1L]))

    res_k <- compareChromatograms(c_full, method = "kendall")
    expect_true(is.array(res_k))
    expect_equal(diag(res_k[, , 1L]), c(1, 1, 1))
    expect_equal(unname(res_k[1, 3, 1L]), 1)
})

test_that("compareChromatograms with custom FUN produces expected values.", {
    cosine_sim <- function(x, y, ...) {
        sum(x * y, na.rm = TRUE) /
            (sqrt(sum(x^2, na.rm = TRUE)) * sqrt(sum(y^2, na.rm = TRUE)))
    }
    res <- compareChromatograms(c_full, FUN = cosine_sim)
    expect_true(is.array(res))
    expect_equal(dim(res), c(3L, 3L, 2L))
    expect_equal(diag(res[, , 1L]), c(1, 1, 1))
    expect_equal(unname(res[1, 3, 1L]), 1)
})

test_that("compareChromatograms returns NA for non-overlapping chromatograms.", {
    cdata_no <- data.frame(
        msLevel = c(1L, 1L),
        mz = c(100, 200),
        dataOrigin = c("s1", "s2")
    )
    pdata_no <- list(
        data.frame(rtime = c(1, 2, 3, 4), intensity = c(10, 50, 100, 50)),
        data.frame(rtime = c(10, 11, 12, 13), intensity = c(20, 80, 200, 80))
    )
    chr_no <- Chromatograms(ChromBackendMemory(), chromData = cdata_no,
                            peaksData = pdata_no)
    res <- compareChromatograms(chr_no)
    expect_true(is.na(res[1, 2, 1L]))
    expect_true(is.na(res[2, 1, 1L]))
    expect_equal(diag(res[, , 1L]), c(1, 1))
})

test_that("compareChromatograms returns NA when overlap < minPeaks.", {
    cdata_short <- data.frame(
        msLevel = c(1L, 1L),
        mz = c(100, 200),
        dataOrigin = c("s1", "s2")
    )
    pdata_short <- list(
        data.frame(rtime = 1, intensity = 50),
        data.frame(rtime = c(1, 2, 3), intensity = c(10, 50, 100))
    )
    chr_short <- Chromatograms(ChromBackendMemory(), chromData = cdata_short,
                               peaksData = pdata_short)
    res <- compareChromatograms(chr_short)
    expect_true(is.na(res[1, 2, 1L]))
    expect_equal(unname(res[1, 1, 1L]), 1)
    expect_equal(unname(res[2, 2, 1L]), 1)
})

test_that("compareChromatograms computes correct Pearson value for known data.", {
    ## Two chromatograms with overlapping RT and known correlation
    cdata_known <- data.frame(
        msLevel = c(1L, 1L),
        mz = c(100, 200),
        dataOrigin = c("s1", "s2")
    )
    pdata_known <- list(
        data.frame(rtime = c(1, 2, 3, 4, 5),
                   intensity = c(10, 20, 30, 20, 10)),
        data.frame(rtime = c(1, 2, 3, 4, 5),
                   intensity = c(5, 15, 25, 15, 5))
    )
    chr_known <- Chromatograms(ChromBackendMemory(), chromData = cdata_known,
                               peaksData = pdata_known)
    res <- compareChromatograms(chr_known)
    ## Same RT grid → no interpolation needed; these are perfectly correlated
    expect_equal(unname(res[1, 2, 1L]), 1, tolerance = 1e-10)
    expect_equal(unname(res[1, 2, 2L]), 5)  ## 5 overlapping RT points
})

test_that("compareChromatograms minPeaks filters low-overlap pairs.", {
    ## chr_known has 5 overlapping RT points
    cdata_known <- data.frame(
        msLevel = c(1L, 1L),
        mz = c(100, 200),
        dataOrigin = c("s1", "s2")
    )
    pdata_known <- list(
        data.frame(rtime = c(1, 2, 3, 4, 5),
                   intensity = c(10, 20, 30, 20, 10)),
        data.frame(rtime = c(1, 2, 3, 4, 5),
                   intensity = c(5, 15, 25, 15, 5))
    )
    chr_known <- Chromatograms(ChromBackendMemory(), chromData = cdata_known,
                               peaksData = pdata_known)
    ## Default minPeaks = 4: 5 points ≥ 4 → score computed
    res <- compareChromatograms(chr_known)
    expect_false(is.na(res[1, 2, 1L]))
    ## minPeaks = 6: 5 points < 6 → score is NA, but count is still 5
    res_strict <- compareChromatograms(chr_known, minPeaks = 6L)
    expect_true(is.na(res_strict[1, 2, 1L]))
    expect_equal(unname(res_strict[1, 2, 2L]), 5)
    ## minPeaks = 2: always compute when ≥ 2 points overlap
    res_loose <- compareChromatograms(chr_known, minPeaks = 2L)
    expect_false(is.na(res_loose[1, 2, 1L]))
})

test_that("compareChromatograms labelsColumn parameter sets row/col names.", {
    ## 'mz' column has unique values in c_full
    res <- compareChromatograms(c_full, labelsColumn = "mz")
    expect_equal(dimnames(res)[[1L]], as.character(chromData(c_full)[["mz"]]))
    expect_equal(dimnames(res)[[2L]], as.character(chromData(c_full)[["mz"]]))
    ## NULL labelsColumn → no names on first two dims
    res_no <- compareChromatograms(c_full, labelsColumn = NULL)
    expect_null(dimnames(res_no)[[1L]])
    expect_null(dimnames(res_no)[[2L]])
    ## Non-existent column → error
    expect_error(compareChromatograms(c_full, labelsColumn = "nonexistent"),
                 "not found")
    ## Non-unique column → error
    c_dup <- c_full
    c_dup$msLevel <- rep(1L, length(c_dup))
    expect_error(compareChromatograms(c_dup, labelsColumn = "msLevel"),
                 "duplicated")
    ## Non-character labelsColumn → error
    expect_error(compareChromatograms(c_full, labelsColumn = 42),
                 "single character string")
})

test_that("compareChromatograms works with ChromBackendMzR.", {
    res <- compareChromatograms(c_mzr)
    expect_true(is.array(res))
    expect_equal(dim(res)[1L], length(c_mzr))
    expect_equal(dim(res)[2L], length(c_mzr))
    expect_equal(dim(res)[3L], 2L)
    expect_equal(diag(res[, , 1L]), rep(1, length(c_mzr)))
    expect_true(is.numeric(res))
})

test_that("compareChromatograms per-group via split() works.", {
    ## The 'by' parameter is no longer supported; users split beforehand.
    c_grp <- c_full
    c_grp$group <- c("A", "B", "A")
    grp_list <- split(c_grp, c_grp$group)
    res <- lapply(grp_list, compareChromatograms)
    expect_equal(sort(names(res)), c("A", "B"))
    ## Group "A" has 2 identical chromatograms → off-diagonal score = 1
    expect_equal(dim(res[["A"]]), c(2L, 2L, 2L))
    expect_equal(unname(res[["A"]][1, 2, 1L]), 1)
    ## Group "B" has 1 chromatogram → 1x1x2 array
    expect_equal(dim(res[["B"]]), c(1L, 1L, 2L))
    expect_equal(unname(res[["B"]][1, 1, 1L]), 1)
})

test_that("compareChromatograms minPeaks works for x,y comparison.", {
    ## chr_known has 5 overlapping RT points (defined above in minPeaks test)
    cdata_known <- data.frame(
        msLevel = c(1L, 1L), mz = c(100, 200), dataOrigin = c("s1", "s2")
    )
    pdata_known <- list(
        data.frame(rtime = c(1, 2, 3, 4, 5), intensity = c(10, 20, 30, 20, 10)),
        data.frame(rtime = c(1, 2, 3, 4, 5), intensity = c(5, 15, 25, 15, 5))
    )
    chr_known <- Chromatograms(ChromBackendMemory(), chromData = cdata_known,
                               peaksData = pdata_known)
    ## strict threshold blocks score but count is still recorded
    res <- compareChromatograms(chr_known[1], chr_known[2], minPeaks = 6L)
    expect_true(is.na(res[1, 1, 1L]))
    expect_equal(unname(res[1, 1, 2L]), 5)
})

## ---- compareChromatograms x, y tests ----

test_that("compareChromatograms(x, y) returns n x m x 2 array.", {
    ## c_full has 3 chromatograms; compare first 2 vs last 1
    res <- compareChromatograms(c_full[1:2], c_full[3])
    expect_true(is.array(res))
    expect_equal(dim(res), c(2L, 1L, 2L))
    ## Chromatograms 1 and 3 are identical
    expect_equal(unname(res[1, 1, 1L]), 1)
    ## Chromatograms 2 and 3 have non-overlapping RT
    expect_true(is.na(res[2, 1, 1L]))
    ## Overlap count for identical pair = number of RT points
    expect_true(res[1, 1, 2L] > 0)
    expect_equal(unname(res[2, 1, 2L]), 0)
})

test_that("compareChromatograms(x, y) with empty objects.", {
    res <- compareChromatograms(c_empty, c_full)
    expect_true(is.array(res))
    expect_equal(dim(res)[1L], 0L)
    res2 <- compareChromatograms(c_full, c_empty)
    expect_true(is.array(res2))
    expect_equal(dim(res2)[2L], 0L)
})

test_that("compareChromatograms(x, y) with custom FUN.", {
    cosine_sim <- function(x, y, ...) {
        sum(x * y, na.rm = TRUE) /
            (sqrt(sum(x^2, na.rm = TRUE)) * sqrt(sum(y^2, na.rm = TRUE)))
    }
    res <- compareChromatograms(c_full[1], c_full[3], FUN = cosine_sim)
    expect_equal(unname(res[1, 1, 1L]), 1)
})

## ---- matchRtime tests ----

test_that("matchRtime returns aligned intensities for overlapping RT", {
    pd_a <- data.frame(rtime = c(1, 2, 3, 4), intensity = c(10, 20, 30, 40))
    pd_b <- data.frame(rtime = c(2, 3, 4, 5), intensity = c(20, 30, 40, 50))
    res <- matchRtime(pd_a, pd_b)
    expect_true(is.list(res))
    expect_equal(length(res$x), length(res$y))
    expect_true(length(res$x) >= 2L)
    ## Overlapping RT is 2-4; both have identical intensities at those points
    expect_equal(res$x, res$y)
})

test_that("matchRtime returns empty vectors for non-overlapping RT", {
    pd_a <- data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
    pd_b <- data.frame(rtime = c(10, 11, 12), intensity = c(40, 50, 60))
    res <- matchRtime(pd_a, pd_b)
    expect_equal(res$x, numeric())
    expect_equal(res$y, numeric())
})

test_that("matchRtime returns empty vectors when fewer than 2 points", {
    pd_short <- data.frame(rtime = 1, intensity = 50)
    pd_ok <- data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
    res <- matchRtime(pd_short, pd_ok)
    expect_equal(res$x, numeric())
    expect_equal(res$y, numeric())
})

test_that("matchRtime interpolates onto union of RT grids", {
    pd_a <- data.frame(rtime = c(1, 3), intensity = c(10, 30))
    pd_b <- data.frame(rtime = c(1, 2, 3), intensity = c(100, 200, 300))
    res <- matchRtime(pd_a, pd_b, tolerance = 0.5)
    expect_equal(res$x, c(10, 20, 30))
    expect_equal(res$y, c(100, 200, 300))
})

test_that("matchRtime x larger than y - y interpolates at x unmatched points", {
    pd_a <- data.frame(rtime = c(1, 2, 3, 4, 5), intensity = c(10, 20, 30, 40, 50))
    pd_b <- data.frame(rtime = c(1, 3, 5),        intensity = c(100, 300, 500))
    res <- matchRtime(pd_a, pd_b, tolerance = 0.5)
    expect_equal(res$x, c(10, 20, 30, 40, 50))
    expect_equal(res$y, c(100, 200, 300, 400, 500))
})

test_that("matchRtime both sides interpolate at their respective unmatched points", {
    pd_a <- data.frame(rtime = c(1, 3, 5),    intensity = c(10, 30, 50))
    pd_b <- data.frame(rtime = c(1, 2, 4, 5), intensity = c(10, 20, 40, 50))
    res <- matchRtime(pd_a, pd_b, tolerance = 0.5)
    expect_equal(res$x, c(10, 20, 30, 40, 50))
    expect_equal(res$y, c(10, 20, 30, 40, 50))
})

## ---- matchRtime tolerance tests ----

test_that("matchRtime tolerance = Inf gives same result as no tolerance", {
    pd_a <- data.frame(rtime = c(1, 2, 3, 4), intensity = c(10, 20, 30, 40))
    pd_b <- data.frame(rtime = c(1, 2, 3, 4), intensity = c(10, 20, 30, 40))
    expect_equal(matchRtime(pd_a, pd_b), matchRtime(pd_a, pd_b, tolerance = Inf))
})

test_that("matchRtime tolerance keeps interior points and approximates missing values", {
    ## pd_a has points at 1, 2, 3; pd_b at 1, 3 (no point at rt=2).
    ## With tolerance = 0.5, rt=2 from pd_a has no match in pd_b (distance = 1),
    ## but it lies inside the matched range [1, 3] so it is retained.
    ## pd_b's intensity at rt=2 is approximated by linear interpolation → 200.
    pd_a <- data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
    pd_b <- data.frame(rtime = c(1, 3),    intensity = c(100, 300))
    res <- matchRtime(pd_a, pd_b, tolerance = 0.5)
    expect_equal(res$x, c(10, 20, 30))
    expect_equal(res$y, c(100, 200, 300))
})

test_that("matchRtime tolerance = 1 keeps points exactly at the boundary", {
    ## pd_a has rt=2; nearest pd_b point is rt=1 or rt=3 (distance = 1).
    ## tolerance = 1 uses <=, so distance == tolerance is kept.
    pd_a <- data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
    pd_b <- data.frame(rtime = c(1, 3),    intensity = c(100, 300))
    res <- matchRtime(pd_a, pd_b, tolerance = 1)
    expect_equal(res$x, c(10, 20, 30))
})

test_that("matchRtime tolerance returns empty when no points survive", {
    pd_a <- data.frame(rtime = c(1, 10), intensity = c(10, 100))
    pd_b <- data.frame(rtime = c(5, 10), intensity = c(50, 100))
    ## rt=1 from pd_a: nearest pd_b is rt=5, distance=4 > tolerance=2 → not matched
    ## rt=5 from pd_b: nearest pd_a is rt=1, distance=4 > tolerance=2 → not matched
    ## rt=10 in both: distance=0 → matched in each
    ## Only one matched point across both → rt_min == rt_max = 10 → empty
    res_strict <- matchRtime(pd_a, pd_b, tolerance = 2)
    expect_equal(res_strict$x, numeric())
    expect_equal(res_strict$y, numeric())
})

test_that("matchRtime tolerance affects the matched range", {
    pd_a <- data.frame(rtime = c(0.5, 1, 2, 3), intensity = c(5, 10, 20, 30))
    pd_b <- data.frame(rtime = c(0, 1, 2, 3),   intensity = c(0, 10, 20, 30))
    res_inf <- matchRtime(pd_a, pd_b, tolerance = Inf)
    res_tol <- matchRtime(pd_a, pd_b, tolerance = 0.4)
    expect_true(length(res_inf$x) > length(res_tol$x))
})

test_that("compareChromatograms with custom MAPFUN.", {
    ## A MAPFUN that returns raw intensities (assumes identical RT)
    identity_map <- function(x, y, ...) {
        list(x = x$intensity, y = y$intensity)
    }
    ## Use chromatograms 1 and 3 which are identical (same number of points)
    res <- compareChromatograms(c_full[c(1, 3)], MAPFUN = identity_map)
    expect_true(is.array(res))
    expect_equal(dim(res), c(2L, 2L, 2L))
    expect_equal(diag(res[, , 1L]), c(1, 1))
    expect_equal(unname(res[1, 2, 1L]), 1)
})

test_that("compareChromatograms routes tolerance to MAPFUN, not FUN.", {
    ## tolerance is a matchRtime argument; passing it to compareChromatograms
    ## must not error even though FUN = cor does not accept tolerance.
    expect_no_error(compareChromatograms(c_full, tolerance = 1))
    expect_no_error(compareChromatograms(c_full[1], c_full[2], tolerance = 1))
})

test_that("compareChromatograms routes MAPFUN and FUN args independently.", {
    ## tolerance goes to MAPFUN (matchRtime), method goes to FUN (cor).
    ## Both should be accepted without error.
    expect_no_error(
        compareChromatograms(c_full, tolerance = 1, method = "spearman")
    )
})

test_that("ChromBackendSpectra works", {
    expect_true(isReadOnly(be_sp))
    expect_false(be_sp@inMemory)
    expect_false(identical(peaksData(be_sp),.peaksData(be_sp)))
    expect_true(identical(length(peaksData(be_sp)), length((.peaksData(be_sp)))))
    expect_true(identical(
        chromData(be_sp),
        fillCoreChromVariables(be_sp@chromData)
    ))
    expect_false(supportsSetBackend(be_sp))

    ## works ofr single column factorisation too
    expect_silent(be_tmp <- factorize(be_sp, factorize.by = "dataOrigin"))
    expect_true(is(be_tmp, "ChromBackendSpectra"))
    expect_true(all(.chromData(be_tmp)$dataOrigin ==
                    .chromData(be_tmp)$chromSpectraIndex))
})

test_that("backendInitialize works", {
    expect_equal(
        backendInitialize(ChromBackendSpectra(), spectra = Spectra()),
        ChromBackendSpectra()
    )
    expect_error(
        backendInitialize(ChromBackendSpectra(), spectra = numeric()),
        "must be a 'Spectra'"
    )
    expect_error(backendInitialize(ChromBackendSpectra(),
        spectra = s,
        factorize.by = c("nope", "nope2")
    ), "variables must exist in")
    expect_error(
        backendInitialize(ChromBackendSpectra(),
            summarize.method = "nope"
        ),
        "'arg' should be one of "
    )
    expect_error(
        backendInitialize(ChromBackendSpectra(),
            spectra = s,
            chromData = matrix()
        ),
        "must be a 'data.frame'"
    )
    expect_error(
        backendInitialize(ChromBackendSpectra(),
            spectra = s,
            chromData = data.frame(
                a = c(1, 2),
                b = c("yes", "no")
            )
        ),
        "All 'factorize.by' variables"
    )
    expect_error(
        backendInitialize(ChromBackendSpectra(),
            summarize.method = "mean"
        ),
        "should be one of"
    )
    df <- data.frame(msLevel = 1L, mz = 100,
                     dataOrigin = dataOrigin(s)[1])
    bd_tmp <- backendInitialize(ChromBackendSpectra(), spectra = s,
                                chromData = df,
                                factorize.by = c("msLevel", "dataOrigin")
                              )
    expect_identical(bd_tmp@chromData[, colnames(df)], df)
    expect_true(all(c("rtMin", "rtMax", "mzMin", "mzMax", "chromSpectraIndex") %in%
                colnames(bd_tmp@chromData)))
})

test_that("show method for ChromBackendSpectra works correctly", {
    expect_output(show(be_sp), "1651 spectra")
    expect_output(show(be_sp), "3 chromatograms")
    expect_output(show(be_sp), "chromIndex")
    expect_output(show(be_sp), "msLevel")
    expect_output(show(be_sp), "mz")
    tmp <- be_sp
    tmp@inMemory <- TRUE
    expect_output(show(tmp), "Peaks data is cached in memory")
})

test_that("replacement method works", {
    tmp <- be_sp
    peaksData(tmp)[[1]] <- peaksData(be_sp)[[1]] + 1
    expect_false(identical(peaksData(tmp)[[1]], peaksData(be_sp)[[1]]))
    expect_true(identical(peaksData(tmp)[[1]], peaksData(be_sp)[[1]] + 1))
    expect_true(tmp@inMemory)
    expect_false(be_sp@inMemory)

    cd <- chromData(tmp)
    cd$mz <- 1
    expect_true(!identical(chromData(tmp), cd))
    chromData(tmp) <- cd
    expect_equal(chromData(tmp), cd)
})

test_that("error message work", {
    expect_error(
        peaksData(be_sp, columns = "notacolumn"),
        "requested peaks variables"
    )
})

test_that("factorize() works", {
    expect_error(factorize(be_sp, factorize.by = "nope"),
                 "variables must be in the Spectra")
    expect_error(factorize(be_sp, factorize.by = "chromIndex"),
                 "variables must be in the")
    tmp <- be_sp
    tmp$msLevel <- c(1L, 2L, 3L)
    idx_before <- chromSpectraIndex(tmp)
    tmp <- factorize(tmp)
    idx_after <- chromSpectraIndex(tmp)
    expect_false(identical(idx_before, idx_after))
    expect_identical(levels(chromSpectraIndex(tmp)),
                     levels(tmp@spectra$chromSpectraIndex))
    tmp@spectra$extra_col <- seq_len(length(tmp@spectra))
    expect_error(factorize(tmp, factorize.by = c("msLevel", "extra_col")),
                 "must be in chromData")
})

test_that("factorize() fills rt columns with numeric vectors", {
    sp <- Spectra(DataFrame(
        mz = replicate(4, c(1, 2), simplify = FALSE),
        intensity = replicate(4, c(10, 20), simplify = FALSE),
        rtime = c(20, 10, 5, 15),
        msLevel = rep(1L, 4),
        dataOrigin = c("A", "A", "B", "B")
    ))

    cb <- ChromBackendSpectra()
    cb@spectra <- sp
    cb@chromData <- data.frame(
        msLevel = 1L,
        dataOrigin = c("B", "A")
    )
    ## Set spectraSortIndex manually since this unsorted data
    sort_idx <- order(sp$dataOrigin, sp$rtime)
    if (!identical(sort_idx, seq_along(sp))) {
        cb@spectraSortIndex <- sort_idx
    }

    cb <- factorize(cb, factorize.by = "dataOrigin")

    expect_type(cb@chromData$rtMin, "double")
    expect_type(cb@chromData$rtMax, "double")
    expect_false(is.list(cb@chromData$rtMin))
    expect_equal(cb@chromData$rtMin, c(5, 10))
    expect_equal(cb@chromData$rtMax, c(15, 20))
})

test_that("spectraSortIndex is empty for pre-sorted data", {
    ## Create spectra that are already sorted by dataOrigin and rtime
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(1, 2, 3, 4, 5),
        msLevel = rep(1L, 5),
        dataOrigin = rep("A", 5)
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    ## For pre-sorted data, spectraSortIndex should be empty
    expect_identical(length(cb@spectraSortIndex), 0L)
    expect_identical(cb@spectraSortIndex, integer())
})

test_that("spectraSortIndex is set for unsorted data", {
    ## Create spectra that are NOT sorted by rtime
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(5, 1, 3, 2, 4),
        msLevel = rep(1L, 5),
        dataOrigin = rep("A", 5)
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    ## For unsorted data, spectraSortIndex should be set
    expect_true(length(cb@spectraSortIndex) > 0)
    expected_sort <- order(sp$dataOrigin, sp$rtime)
    expect_identical(cb@spectraSortIndex, expected_sort)

    ## Verify the sort index produces sorted rtimes
    sorted_rtimes <- sp$rtime[cb@spectraSortIndex]
    expect_identical(sorted_rtimes, c(1, 2, 3, 4, 5))
})

test_that("spectraSortIndex is empty for pre-sorted data with multiple dataOrigins", {
    ## Create spectra sorted by dataOrigin then rtime
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(1, 2, 3, 4, 5, 6),
        msLevel = rep(1L, 6),
        dataOrigin = c("A", "A", "A", "B", "B", "B")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    ## For pre-sorted data, spectraSortIndex should be empty
    expect_identical(length(cb@spectraSortIndex), 0L)
})

test_that("spectraSortIndex is set for unsorted data with multiple dataOrigins", {
    ## Create spectra NOT sorted by dataOrigin and rtime
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(3, 1, 2, 6, 4, 5),
        msLevel = rep(1L, 6),
        dataOrigin = c("B", "A", "A", "B", "A", "B")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    ## For unsorted data, spectraSortIndex should be set
    expect_true(length(cb@spectraSortIndex) > 0)
    expected_sort <- order(sp$dataOrigin, sp$rtime)
    expect_identical(cb@spectraSortIndex, expected_sort)

    ## Verify sorting is correct
    sorted_do <- sp$dataOrigin[cb@spectraSortIndex]
    sorted_rt <- sp$rtime[cb@spectraSortIndex]
    expect_identical(sorted_do, c("A", "A", "A", "B", "B", "B"))
    expect_identical(sorted_rt, c(1, 2, 4, 3, 5, 6))
})

test_that("factorize() clears spectraSortIndex when data is sorted", {
    ## Create unsorted spectra
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(4, 1, 3, 2),
        msLevel = rep(1L, 4),
        dataOrigin = rep("A", 4)
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    ## Initially unsorted, so spectraSortIndex should be set
    expect_true(length(cb@spectraSortIndex) > 0)

    ## Now manually sort the spectra and refactorize
    sorted_indices <- order(sp$dataOrigin, sp$rtime)
    cb@spectra <- sp[sorted_indices]
    cb@spectra$rtime <- sp$rtime[sorted_indices]
    cb <- factorize(cb)

    ## After refactorize with sorted data, spectraSortIndex should be empty
    expect_identical(length(cb@spectraSortIndex), 0L)
})

test_that("factorize() recalculates spectraSortIndex for unsorted data", {
    ## Create spectra and manually set it to unsorted state
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(1, 2, 3, 4),
        msLevel = rep(1L, 4),
        dataOrigin = rep("A", 4)
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    ## Initially sorted, so spectraSortIndex should be empty
    expect_identical(length(cb@spectraSortIndex), 0L)

    ## Modify msLevel to create different factorization
    cb@spectra$msLevel <- c(1L, 2L, 1L, 2L)
    cb <- factorize(cb)

    ## spectraSortIndex should still be empty since data is still sorted by rtime
    expect_identical(length(cb@spectraSortIndex), 0L)
})

test_that("peaksData works correctly with empty spectraSortIndex", {
    ## Create pre-sorted spectra
    sp <- Spectra(DataFrame(
        mz = NumericList(c(100, 101), c(100, 101), c(100, 101), c(100, 101), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(1, 2, 3, 4),
        msLevel = rep(1L, 4),
        dataOrigin = rep("A", 4)
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    ## spectraSortIndex should be empty
    expect_identical(length(cb@spectraSortIndex), 0L)

    ## peaksData should still work correctly
    pd <- peaksData(cb)
    expect_true(is.list(pd))
    expect_identical(length(pd), 1L)
    expect_true(is.data.frame(pd[[1]]))
})

test_that("peaksData works correctly with populated spectraSortIndex", {
    ## Create unsorted spectra
    sp <- Spectra(DataFrame(
        mz = NumericList(c(100, 101), c(100, 101), c(100, 101), c(100, 101), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(3, 1, 4, 2),
        msLevel = rep(1L, 4),
        dataOrigin = c("B", "A", "B", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    ## spectraSortIndex should be set
    expect_true(length(cb@spectraSortIndex) > 0)

    ## peaksData should still work correctly
    pd <- peaksData(cb)
    expect_true(is.list(pd))
    expect_identical(length(pd), 2L)
    expect_true(all(sapply(pd, is.data.frame)))
})

test_that("spectraSortIndex is set and used for sorting", {
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(3, 1, 4, 2),
        msLevel = rep(1L, 4),
        dataOrigin = c("B", "A", "B", "A")
    ))

    cb <- ChromBackendSpectra()
    cb@spectra <- sp
    csi <- interaction(
        as.list(spectraData(sp)[, c("msLevel", "dataOrigin"), drop = FALSE]),
        drop = TRUE, sep = "_"
    )
    cb@spectra$chromSpectraIndex <- csi
    cb@chromData <- fillCoreChromVariables(data.frame(
        msLevel = c(1L, 1L),
        dataOrigin = c("A", "B"),
        chromSpectraIndex = c("1_A", "1_B")
    ))
    ## Manually set spectraSortIndex since this is unsorted data
    sort_idx <- order(sp$dataOrigin, sp$rtime)
    if (!identical(sort_idx, seq_along(sp))) {
        cb@spectraSortIndex <- sort_idx
    }

    expect_true(length(cb@spectraSortIndex) > 0)
    expect_identical(cb@spectraSortIndex, order(sp$dataOrigin, sp$rtime))
})

test_that("[ maintains spectra and spectraSortIndex", {
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(3, 1, 4, 2),
        msLevel = rep(1L, 4),
        dataOrigin = c("B", "A", "B", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Verify initial state
    expect_identical(length(cb), 2L)  # 2 unique combinations of msLevel and dataOrigin
    expect_identical(length(cb@spectra), 4L)  # 4 spectra
    # spectraSortIndex should be set since data is unsorted
    expect_true(length(cb@spectraSortIndex) > 0)

    # Verify spectraSortIndex is correctly set
    expected_sort <- order(sp$dataOrigin, sp$rtime)
    expect_identical(cb@spectraSortIndex, expected_sort)

    # Get chromSpectraIndex before subsetting
    all_chrom_idx <- chromSpectraIndex(cb)
    expect_identical(length(all_chrom_idx), 2L)

    # Subset to keep only "1_A" chromatograms
    keep <- all_chrom_idx == "1_A"
    cb_sub <- cb[keep]

    # Verify subsetting worked correctly
    expect_s4_class(cb_sub, "ChromBackendSpectra")
    expect_identical(length(cb_sub), 1L)  # Only 1 chromatogram kept

    # Verify chromData was subsetted
    sub_chrom_idx <- chromSpectraIndex(cb_sub)
    expect_identical(length(sub_chrom_idx), 1L)
    expect_identical(as.character(sub_chrom_idx), "1_A")

    # Verify spectra were subsetted to only include spectra for "1_A"
    expect_identical(length(cb_sub@spectra), 2L)  # Only 2 spectra belong to "1_A"
    expect_true(all(cb_sub@spectra$dataOrigin == "A"))

    # Verify spectraSortIndex is valid (may be empty if subsetted data is sorted)
    if (length(cb_sub@spectraSortIndex) > 0) {
        expect_true(all(cb_sub@spectraSortIndex <= length(cb_sub@spectra)))
        # Verify the sort index still provides correct ordering
        sorted_rtime <- cb_sub@spectra$rtime[cb_sub@spectraSortIndex]
        expect_true(all(diff(sorted_rtime) >= 0))
    } else {
        # If spectraSortIndex is empty, spectra should already be sorted
        sorted_rtime <- cb_sub@spectra$rtime
        expect_true(all(diff(sorted_rtime) >= 0))
    }

    # Verify chromSpectraIndex is properly factorized
    expect_true(is.factor(cb_sub@spectra$chromSpectraIndex))
    expect_identical(levels(cb_sub@spectra$chromSpectraIndex), "1_A")
})

test_that("factorize() handles empty chromData correctly", {
    # Test the scenario from the vignette: creating Chromatograms from Spectra
    # without providing explicit chromData
    sp <- Spectra(DataFrame(
        mz = NumericList(
            c(100, 101), c(100, 101), c(100, 101), c(100, 101), c(100, 101),
            compress = FALSE
        ),
        intensity = NumericList(
            c(10, 20), c(15, 25), c(30, 5), c(12, 18), c(40, 2),
            compress = FALSE
        ),
        rtime = c(100, 110, 120, 130, 140),
        msLevel = rep(1L, 5),
        dataOrigin = rep("example", 5)
    ))

    # Create Chromatograms with empty chromData
    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp, chromData = data.frame())

    # Should create chromData from spectra
    expect_identical(nrow(cb@chromData), 1L)  # One chromatogram: 1_example
    expect_identical(length(cb@spectra), 5L)  # All 5 spectra retained

    # Verify chromSpectraIndex was created correctly
    expect_true(is.factor(cb@spectra$chromSpectraIndex))
    expect_identical(as.character(unique(cb@spectra$chromSpectraIndex)), "1_example")

    # Verify spectraSortIndex is valid (should be empty since data is pre-sorted)
    expect_identical(length(cb@spectraSortIndex), 0L)
    # For pre-sorted data, rtimes should already be in order
    expect_identical(cb@spectra$rtime, c(100, 110, 120, 130, 140))

    # Verify chromData has correct rt range
    expect_equal(cb@chromData$rtMin, 100)
    expect_equal(cb@chromData$rtMax, 140)
})

test_that("factorize() recalculates spectraSortIndex correctly - in-memory backend", {
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2),
                                   compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20),
                                          compress = FALSE),
        rtime = c(3, 1, 4, 2, 5),
        msLevel = c(1L, 1L, 2L, 2L, 1L),
        dataOrigin = rep("A", 5)
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Verify initial sort order is by dataOrigin, rtime
    original_sort_idx <- cb@spectraSortIndex
    sorted_rtimes <- sp$rtime[original_sort_idx]
    expect_true(all(diff(sorted_rtimes) >= 0))

    # Modify msLevel and refactorize
    cb@spectra$msLevel <- c(2L, 1L, 1L, 2L, 1L)
    cb <- factorize(cb)

    # After factorize, spectraSortIndex should be recalculated
    # and should still produce correctly sorted rtimes
    new_sort_idx <- cb@spectraSortIndex
    sorted_rtimes_after <- cb@spectra$rtime[new_sort_idx]
    expect_true(all(diff(sorted_rtimes_after) >= 0))

    # Verify chromData reflects the new factorization
    expect_identical(nrow(cb@chromData), 2L)  # Now 2 groups: 1_A and 2_A
})

test_that("factorize() works correctly with on-disk spectra backend", {
    # Use the pre-loaded on-disk spectra from test setup (be_sp)
    cb_test <- be_sp

    # Verify initial state
    expect_true(length(cb_test@spectra) > 0)
    expect_true(length(cb_test@spectraSortIndex) > 0)

    # Verify sort index produces sorted rtimes within groups
    sorted_rtimes <- cb_test@spectra$rtime[cb_test@spectraSortIndex]
    sorted_dataOrigin <- cb_test@spectra$dataOrigin[cb_test@spectraSortIndex]
    # Check rtimes are sorted within each dataOrigin group
    for (do in unique(sorted_dataOrigin)) {
        rtimes_in_group <- sorted_rtimes[sorted_dataOrigin == do]
        expect_true(all(diff(rtimes_in_group) >= 0))
    }

    # Factorize with different grouping
    cb_single <- factorize(cb_test, factorize.by = "dataOrigin")

    # Should have one row per unique dataOrigin
    expect_true(nrow(cb_single@chromData) >= 1)

    # spectraSortIndex should still be valid
    expect_true(all(cb_single@spectraSortIndex <= length(cb_single@spectra)))
    sorted_rtimes_new <- cb_single@spectra$rtime[cb_single@spectraSortIndex]
    sorted_dataOrigin_new <- cb_single@spectra$dataOrigin[cb_single@spectraSortIndex]
    # Check rtimes are sorted within each dataOrigin group
    for (do in unique(sorted_dataOrigin_new)) {
        rtimes_in_group <- sorted_rtimes_new[sorted_dataOrigin_new == do]
        expect_true(all(diff(rtimes_in_group) >= 0))
    }
})

test_that("factorize() maintains consistency between chromData and spectra", {
    sp <- Spectra(DataFrame(
        mz = replicate(6, c(1, 2), simplify = FALSE),
        intensity = replicate(6, c(10, 20), simplify = FALSE),
        rtime = c(1, 2, 3, 4, 5, 6),
        msLevel = c(1L, 1L, 2L, 2L, 1L, 2L),
        dataOrigin = c("A", "A", "A", "B", "B", "B")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Get the chromSpectraIndex values
    chrom_idx <- chromSpectraIndex(cb)
    spectra_idx <- cb@spectra$chromSpectraIndex

    # Both should have the same set of levels
    expect_identical(sort(levels(chrom_idx)), sort(levels(spectra_idx)))

    # All chromSpectraIndex values in spectra should be in chromData levels
    unique_spectra_idx <- unique(as.character(spectra_idx))
    expect_true(all(unique_spectra_idx %in% levels(spectra_idx)))

    # Verify consistency after subsetting
    keep_idx <- chrom_idx == levels(chrom_idx)[1]
    cb_sub <- cb[keep_idx]

    chrom_idx_sub <- chromSpectraIndex(cb_sub)
    spectra_idx_sub <- cb_sub@spectra$chromSpectraIndex

    # After subsetting, both should have same single level
    expect_identical(nlevels(chrom_idx_sub), 1L)
    expect_identical(nlevels(spectra_idx_sub), 1L)
    expect_identical(levels(chrom_idx_sub), levels(spectra_idx_sub))
})

test_that("peaksData generation respects spectraSortIndex - in-memory", {
    sp <- Spectra(DataFrame(
        mz = replicate(4, c(100, 101), simplify = FALSE),
        intensity = replicate(4, c(10, 20), simplify = FALSE),
        rtime = c(3, 1, 4, 2),
        msLevel = rep(1L, 4),
        dataOrigin = c("B", "A", "B", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Get peaksData
    pd <- peaksData(cb)

    # Should return a list with one entry per chromatogram
    expect_identical(length(pd), 2L)

    # Each entry should be a data.frame with rtime and intensity
    for (i in seq_along(pd)) {
        expect_true(is.data.frame(pd[[i]]))
        expect_true("rtime" %in% colnames(pd[[i]]))
        expect_true("intensity" %in% colnames(pd[[i]]))
    }
})

test_that("subsetting and peaksData consistency - in-memory", {
    sp <- Spectra(DataFrame(
        mz = replicate(4, c(100, 101), simplify = FALSE),
        intensity = replicate(4, c(10, 20), simplify = FALSE),
        rtime = c(3, 1, 4, 2),
        msLevel = rep(1L, 4),
        dataOrigin = c("B", "A", "B", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Subset to one chromatogram
    keep <- chromSpectraIndex(cb) == "1_A"
    cb_sub <- cb[keep]

    # Get peaksData from subsetted backend
    pd_sub <- peaksData(cb_sub)

    # Should have one entry
    expect_identical(length(pd_sub), 1L)

    # The data should be valid
    expect_true(is.data.frame(pd_sub[[1]]))
    expect_true(nrow(pd_sub[[1]]) > 0)
})

test_that("chromSpectraIndex works", {
    expect_error(
        chromSpectraIndex(1),
        "object must be a"
    )
    expect_equal(be_sp@chromData$chromSpectraIndex, chromSpectraIndex(be_sp))
    tmp <- be_sp
    tmp@chromData$chromSpectraIndex <- seq_len(nrow(tmp@chromData))
    expect_true(is.factor(chromSpectraIndex(tmp)))
})

test_that("backendParallelFactor works", {
    expect_identical(backendParallelFactor(be_sp), factor())
})

test_that("chromExtract works for ChromBackendSpectra", {
    tmp <- be_sp
    do <- dataOrigin(tmp)[1:2]
    peak_tbl <- data.frame(
        rtMin = c(1, 100),
        rtMax = c(56, 600),
        mzMin = c(100, 140),
        mzMax = c(130, 160),
        msLevel = c(1L, 1L),
        dataOrigin = do,
        extra_cols = c("peak1", "peak2")
    )

    out <- chromExtract(tmp, peak_tbl, by = c("msLevel", "dataOrigin"))

    # Output should be a ChromBackendSpectra
    expect_s4_class(out, "ChromBackendSpectra")

    # chromData should match rows in peak.table
    expect_equal(nrow(.chromData(out)), nrow(peak_tbl))

    # spectra should be preserved
    expect_true(!is.null(out@spectra))
    expect_true(!nrow(.peaksData(out)[[1]]) > 0)

    # Ensure replacement of mz columns occurred
    expect_true(all(c("mzMin", "mzMax", "extra_cols") %in% names(.chromData(out))))
    expect_equal(length(.peaksData(out)), nrow(peak_tbl))
})

test_that("[ subsetting with empty spectraSortIndex works", {
    ## Create pre-sorted spectra (empty spectraSortIndex)
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(1, 2, 3, 4),
        msLevel = c(1L, 1L, 2L, 2L),
        dataOrigin = c("A", "A", "B", "B")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Verify initial state: pre-sorted data should have empty spectraSortIndex
    expect_identical(length(cb@spectraSortIndex), 0L)

    # Subset to keep only first chromatogram (msLevel=1, dataOrigin=A)
    keep <- chromSpectraIndex(cb) == levels(chromSpectraIndex(cb))[1]
    cb_sub <- cb[keep]

    # Verify subset is correct
    expect_identical(length(cb_sub), 1L)
    expect_identical(length(cb_sub@spectra), 2L)
    expect_true(all(cb_sub@spectra$msLevel == 1L))

    # Verify spectraSortIndex is still empty (pre-sorted data stays pre-sorted)
    expect_identical(length(cb_sub@spectraSortIndex), 0L)
})

test_that("[ subsetting with populated spectraSortIndex works", {
    ## Create unsorted spectra (populated spectraSortIndex)
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(3, 1, 4, 2, 5, 6),  # Unsorted
        msLevel = c(1L, 1L, 2L, 2L, 1L, 2L),
        dataOrigin = c("B", "A", "B", "A", "A", "B")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Verify initial state: unsorted data should have populated spectraSortIndex
    expect_true(length(cb@spectraSortIndex) > 0)
    original_sort_idx <- cb@spectraSortIndex

    # Subset to keep only "1_A" chromatogram (msLevel=1, dataOrigin=A)
    keep <- chromSpectraIndex(cb) == "1_A"
    cb_sub <- cb[keep]

    # Verify subset is correct
    expect_identical(length(cb_sub), 1L)
    expect_identical(length(cb_sub@spectra), 2L)

    # Verify spectra in subset have correct values
    expect_true(all(cb_sub@spectra$msLevel == 1L))
    expect_true(all(cb_sub@spectra$dataOrigin == "A"))

    # Verify spectraSortIndex is remapped correctly
    ## Original positions that map to "1_A" are positions 2 and 5
    ## After subsetting, they should be renumbered to 1, 2
    expect_true(all(cb_sub@spectraSortIndex <= length(cb_sub@spectra)))

    # Verify the sort index produces correct ordering
    if (length(cb_sub@spectraSortIndex) > 0) {
        sorted_rtimes <- cb_sub@spectra$rtime[cb_sub@spectraSortIndex]
        expect_true(all(diff(sorted_rtimes) >= 0))
    }
})

test_that("[ subsetting remaps spectraSortIndex correctly", {
    ## Test remapping with a specific example
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(5, 1, 3, 2, 4),  # Unsorted
        msLevel = rep(1L, 5),
        dataOrigin = c("A", "A", "B", "B", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Get the original spectraSortIndex
    original_sort_idx <- cb@spectraSortIndex
    expect_true(length(original_sort_idx) > 0)

    # Subset to keep only dataOrigin="A" (indices 1, 2, 5)
    keep <- chromSpectraIndex(cb) == "1_A"
    cb_sub <- cb[keep]

    # Verify we kept the right spectra
    expect_identical(length(cb_sub@spectra), 3L)
    kept_positions <- c(1, 2, 5)  # These are the original positions we kept
    expect_true(all(cb_sub@spectra$dataOrigin == "A"))

    # Verify spectraSortIndex is valid
    expect_true(all(cb_sub@spectraSortIndex > 0))
    expect_true(all(cb_sub@spectraSortIndex <= length(cb_sub@spectra)))

    # Verify sort order is preserved
    sorted_rtimes <- cb_sub@spectra$rtime[cb_sub@spectraSortIndex]
    expect_true(all(diff(sorted_rtimes) >= 0))
})

test_that("[ subsetting with multiple indices works", {
    ## Test subsetting with multiple chromatograms
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(2, 1, 4, 3),  # Unsorted
        msLevel = c(1L, 1L, 2L, 2L),
        dataOrigin = c("A", "A", "A", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Subset to keep both chromatograms
    keep <- c(TRUE, TRUE)  # Keep both msLevel=1 and msLevel=2
    cb_sub <- cb[keep]

    # Verify all spectra are retained
    expect_identical(length(cb_sub@spectra), 4L)
    expect_identical(length(cb_sub), 2L)

    # Verify spectraSortIndex is still valid
    if (length(cb_sub@spectraSortIndex) > 0) {
        expect_true(all(cb_sub@spectraSortIndex <= length(cb_sub@spectra)))
        sorted_do <- cb_sub@spectra$dataOrigin[cb_sub@spectraSortIndex]
        sorted_rt <- cb_sub@spectra$rtime[cb_sub@spectraSortIndex]
        # Verify sorting is correct
        expect_true(all(diff(sorted_rt) >= 0))
    }
})

test_that("[ subsetting with single index works", {
    ## Test subsetting to keep only one chromatogram
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(3, 1, 4, 2),  # Unsorted
        msLevel = c(1L, 1L, 2L, 2L),
        dataOrigin = c("A", "A", "A", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Subset to keep only first chromatogram
    keep <- c(TRUE, FALSE)
    cb_sub <- cb[keep]

    # Verify subset is correct
    expect_identical(length(cb_sub), 1L)
    expect_identical(length(cb_sub@spectra), 2L)
    expect_true(all(cb_sub@spectra$msLevel == 1L))
})

test_that("factorize() with single factorize.by variable works", {
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(1, 2, 3, 4, 5, 6),
        msLevel = c(1L, 1L, 2L, 2L, 1L, 2L),
        dataOrigin = c("A", "A", "A", "A", "B", "B")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)  # Default: factorize.by = c("msLevel", "dataOrigin")

    # Verify initial state creates 4 groups (1_A, 1_B, 2_A, 2_B)
    expect_identical(length(cb), 4L)

    # Refactorize by only dataOrigin
    cb_fact <- factorize(cb, factorize.by = "dataOrigin")

    # After refactorization, chromSpectraIndex changes to only have dataOrigin groups
    expect_true(is.factor(cb_fact@spectra$chromSpectraIndex))
    expect_identical(nlevels(cb_fact@spectra$chromSpectraIndex), 2L)  # Now only 2 levels: A and B

    # Verify spectra are correctly assigned to groups
    expect_true(all(cb_fact@spectra$chromSpectraIndex[cb_fact@spectra$dataOrigin == "A"] == "A"))
    expect_true(all(cb_fact@spectra$chromSpectraIndex[cb_fact@spectra$dataOrigin == "B"] == "B"))
})

test_that("factorize() with multiple factorize.by variables works", {
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(1, 2, 3, 4, 5, 6),
        msLevel = c(1L, 1L, 2L, 2L, 1L, 2L),
        dataOrigin = c("A", "A", "A", "A", "B", "B")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Factorize by msLevel and dataOrigin (default)
    cb_fact <- factorize(cb, factorize.by = c("msLevel", "dataOrigin"))

    # Should create 4 chromatograms
    expect_identical(length(cb_fact), 4L)
    expect_identical(nrow(cb_fact@chromData), 4L)

    # Verify chromData levels
    levels_str <- sort(as.character(unique(cb_fact@spectra$chromSpectraIndex)))
    expected_levels <- c("1_A", "1_B", "2_A", "2_B")
    expect_identical(levels_str, sort(expected_levels))
})

test_that("factorize() recalculates spectraSortIndex for unsorted data", {
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(5, 1, 3, 2, 6, 4),  # Unsorted
        msLevel = c(1L, 1L, 2L, 2L, 1L, 2L),
        dataOrigin = c("A", "A", "A", "A", "B", "B")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Verify initial spectraSortIndex is set (unsorted data)
    expect_true(length(cb@spectraSortIndex) > 0)
    initial_sort_idx <- cb@spectraSortIndex

    # Modify rtime and refactorize
    cb@spectra$rtime <- c(1, 2, 3, 4, 5, 6)  # Make it sorted
    cb <- factorize(cb)

    # After refactorize with sorted data, spectraSortIndex should be empty
    expect_identical(length(cb@spectraSortIndex), 0L)
})

test_that("factorize() updates chromSpectraIndex correctly", {
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(10, 20), c(10, 20), c(10, 20), compress = FALSE),
        rtime = c(1, 2, 3, 4),
        msLevel = c(1L, 1L, 2L, 2L),
        dataOrigin = rep("A", 4)
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Verify chromSpectraIndex is a factor
    expect_true(is.factor(cb@spectra$chromSpectraIndex))

    # Refactorize
    cb <- factorize(cb)

    # Verify chromSpectraIndex is still a factor
    expect_true(is.factor(cb@spectra$chromSpectraIndex))

    # Verify all spectra have valid chromSpectraIndex
    expect_true(all(!is.na(cb@spectra$chromSpectraIndex)))

    # Verify levels match chromData
    expect_identical(sort(levels(cb@spectra$chromSpectraIndex)),
                     sort(as.character(cb@chromData$chromSpectraIndex)))
})

test_that("factorize() preserves spectra data integrity", {
    sp <- Spectra(DataFrame(
        mz = NumericList(c(100, 101), c(100, 101), c(100, 101), c(100, 101), compress = FALSE),
        intensity = NumericList(c(10, 20), c(15, 25), c(30, 5), c(12, 18), compress = FALSE),
        rtime = c(1, 2, 3, 4),
        msLevel = c(1L, 1L, 2L, 2L),
        dataOrigin = rep("A", 4)
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    original_mz <- cb@spectra$mz
    original_intensity <- cb@spectra$intensity

    # Refactorize
    cb <- factorize(cb)

    # Verify spectra data is preserved
    expect_identical(cb@spectra$mz, original_mz)
    expect_identical(cb@spectra$intensity, original_intensity)
})

test_that("[ subsetting with reordering works", {
    ## Test that reordering chromatograms works correctly
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(15, 25), c(30, 5), c(12, 18), c(40, 2), c(50, 10), compress = FALSE),
        rtime = c(1, 2, 3, 4, 5, 6),
        msLevel = c(1L, 1L, 2L, 1L, 2L, 2L),
        dataOrigin = c("A", "A", "A", "B", "B", "B")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Get original chromSpectraIndex levels
    original_idx <- chromSpectraIndex(cb)
    expect_identical(length(original_idx), 4L)  # 4 chromatograms: 1_A, 1_B, 2_A, 2_B

    # Reorder: reverse order
    cb_reordered <- cb[c(4, 3, 2, 1)]

    # Verify length is correct
    expect_identical(length(cb_reordered), 4L)

    # Verify order is reversed
    reordered_idx <- chromSpectraIndex(cb_reordered)
    expect_identical(as.character(reordered_idx), as.character(original_idx[c(4, 3, 2, 1)]))

    # Verify chromData is reordered
    expect_identical(nrow(cb_reordered@chromData), 4L)

    # Verify spectra are still correctly associated
    expect_true(all(cb_reordered@spectra$chromSpectraIndex %in% reordered_idx))

    # Verify spectraSortIndex is still valid
    if (length(cb_reordered@spectraSortIndex) > 0) {
        expect_true(all(cb_reordered@spectraSortIndex <= length(cb_reordered@spectra)))
    }
})

test_that("[ subsetting with duplication works", {
    ## Test that duplicating chromatograms works correctly
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(15, 25), c(30, 5), c(12, 18), compress = FALSE),
        rtime = c(1, 2, 3, 4),
        msLevel = c(1L, 1L, 2L, 2L),
        dataOrigin = c("A", "A", "A", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Get original chromSpectraIndex levels
    original_idx <- chromSpectraIndex(cb)
    expect_identical(length(original_idx), 2L)  # 2 chromatograms: 1_A, 2_A

    # Duplicate: keep first chromatogram twice, then second
    cb_dup <- cb[c(1, 1, 2)]

    # Verify length includes duplicates
    expect_identical(length(cb_dup), 3L)

    # Verify chromData includes duplicates
    expect_identical(nrow(cb_dup@chromData), 3L)
    dup_idx <- chromSpectraIndex(cb_dup)
    expect_identical(as.character(dup_idx), as.character(original_idx[c(1, 1, 2)]))

    # Verify spectra: keeps all unique spectra from subsetted chromatograms
    ## Both chrom 1 and 2 are kept, so all 4 original spectra are kept
    expect_identical(length(cb_dup@spectra), 4L)

    # Verify all spectra belong to the right chromatograms
    expect_true(all(cb_dup@spectra$chromSpectraIndex %in% dup_idx))

    # Verify spectraSortIndex is still valid
    if (length(cb_dup@spectraSortIndex) > 0) {
        expect_true(all(cb_dup@spectraSortIndex <= length(cb_dup@spectra)))
    }

    # Verify peaksData works with duplicated chromatograms
    pd <- peaksData(cb_dup)
    expect_identical(length(pd), 3L)
})

test_that("[ subsetting with reordering and duplication works", {
    ## Test combined reordering and duplication
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(15, 25), c(30, 5), c(12, 18), compress = FALSE),
        rtime = c(1, 2, 3, 4),
        msLevel = c(1L, 1L, 2L, 2L),
        dataOrigin = c("A", "A", "A", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    original_idx <- chromSpectraIndex(cb)
    expect_identical(length(original_idx), 2L)  # 2 chromatograms

    # Reorder and duplicate: c(2, 1, 1)
    cb_mixed <- cb[c(2, 1, 1)]

    # Verify length is correct
    expect_identical(length(cb_mixed), 3L)

    # Verify order matches requested subset
    mixed_idx <- chromSpectraIndex(cb_mixed)
    expect_identical(as.character(mixed_idx), as.character(original_idx[c(2, 1, 1)]))

    # Verify chromData has correct structure
    expect_identical(nrow(cb_mixed@chromData), 3L)

    # Verify spectra are correctly associated
    expect_true(all(cb_mixed@spectra$chromSpectraIndex %in% mixed_idx))

    # Verify number of spectra: keeps unique spectra from both chromatograms
    ## Both chrom 1 and 2 are in the subset, so all 4 spectra are kept
    expect_identical(length(cb_mixed@spectra), 4L)

    # Verify spectraSortIndex is valid
    if (length(cb_mixed@spectraSortIndex) > 0) {
        expect_true(all(cb_mixed@spectraSortIndex > 0))
        expect_true(all(cb_mixed@spectraSortIndex <= length(cb_mixed@spectra)))

        # Verify sort index still provides correct ordering
        sorted_rtimes <- cb_mixed@spectra$rtime[cb_mixed@spectraSortIndex]
        sorted_do <- cb_mixed@spectra$dataOrigin[cb_mixed@spectraSortIndex]

        # Check rtimes are sorted within each dataOrigin group
        for (do in unique(sorted_do)) {
            rtimes_in_group <- sorted_rtimes[sorted_do == do]
            expect_true(all(diff(rtimes_in_group) >= 0))
        }
    }
})

test_that("[ subsetting with unsorted data and duplication works", {
    ## Test duplication with unsorted spectra (non-empty spectraSortIndex)
    sp <- Spectra(DataFrame(
        mz = NumericList(c(1, 2), c(1, 2), c(1, 2), c(1, 2), compress = FALSE),
        intensity = NumericList(c(10, 20), c(15, 25), c(30, 5), c(12, 18), compress = FALSE),
        rtime = c(3, 1, 4, 2),  # Unsorted
        msLevel = c(1L, 1L, 2L, 2L),
        dataOrigin = c("B", "A", "B", "A")
    ))

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp)

    # Verify initial state has populated spectraSortIndex
    expect_true(length(cb@spectraSortIndex) > 0)

    original_idx <- chromSpectraIndex(cb)
    expect_identical(length(original_idx), 4L)  # 4 chromatograms: 1_A, 1_B, 2_A, 2_B

    # Duplicate and reorder: keep chromatogram 2, then 1 twice
    cb_dup <- cb[c(2, 1, 1)]

    # Verify basic structure
    expect_identical(length(cb_dup), 3L)
    expect_identical(nrow(cb_dup@chromData), 3L)

    # Verify order
    dup_idx <- chromSpectraIndex(cb_dup)
    expect_identical(as.character(dup_idx), as.character(original_idx[c(2, 1, 1)]))

    # Verify spectra count: keeps unique spectra from chromatograms 1 and 2
    ## Chrom 1_A has 1 spectrum, chrom 1_B has 1 spectrum → 2 unique spectra
    expect_identical(length(cb_dup@spectra), 2L)

    # Verify spectraSortIndex is valid after duplication
    if (length(cb_dup@spectraSortIndex) > 0) {
        expect_true(all(cb_dup@spectraSortIndex > 0))
        expect_true(all(cb_dup@spectraSortIndex <= length(cb_dup@spectra)))

        # Verify sorting is maintained
        sorted_rtimes <- cb_dup@spectra$rtime[cb_dup@spectraSortIndex]
        sorted_do <- cb_dup@spectra$dataOrigin[cb_dup@spectraSortIndex]

        # Within each dataOrigin, rtimes should be sorted
        for (do in unique(sorted_do)) {
            rtimes_in_group <- sorted_rtimes[sorted_do == do]
            expect_true(all(diff(rtimes_in_group) >= 0))
        }
    }

    # Verify peaksData works with duplicated chromatograms
    pd <- peaksData(cb_dup)
    expect_identical(length(pd), 3L)
    expect_true(all(sapply(pd, is.data.frame)))
})

test_that("peaksData returns data in correct order when chromSpectraIndex has duplicates", {
    ## Test case that reproduces the setBackend bug where peaksData was returned
    ## in factor level order instead of chromData row order.
    ## When multiple chromatograms share the same chromSpectraIndex (e.g., multiple
    ## EICs from the same file), split() groups them together by level. The fix
    ## ensures peaksData is returned in the original chromData row order.
    sp <- Spectra(DataFrame(
        mz = NumericList(
            c(100, 101, 102), c(100, 101, 102), c(100, 101, 102),
            c(100, 101, 102), c(100, 101, 102), c(100, 101, 102),
            compress = FALSE
        ),
        intensity = NumericList(
            c(10, 20, 30), c(15, 25, 35), c(20, 30, 40),
            c(100, 200, 300), c(150, 250, 350), c(200, 300, 400),
            compress = FALSE
        ),
        rtime = c(1, 2, 3, 4, 5, 6),
        msLevel = rep(1L, 6),
        dataOrigin = c("A", "A", "A", "B", "B", "B")
    ))

    ## Create custom chromData with multiple EICs sharing same chromSpectraIndex
    ## This simulates extracting different m/z windows from the same spectra
    custom_chromData <- data.frame(
        msLevel = rep(1L, 4),
        mzMin = c(100, 100, 100, 100),
        mzMax = c(101, 102, 101, 102),
        rtMin = c(1, 1, 4, 4),
        rtMax = c(3, 3, 6, 6),
        dataOrigin = c("A", "A", "B", "B")
    )

    cb <- ChromBackendSpectra()
    cb <- backendInitialize(cb, spectra = sp, chromData = custom_chromData)

    ## chromSpectraIndex should have duplicates: 1_A, 1_A, 1_B, 1_B
    chrom_idx <- chromSpectraIndex(cb)
    expect_identical(length(chrom_idx), 4L)
    expect_identical(as.character(chrom_idx), c("1_A", "1_A", "1_B", "1_B"))

    ## Get peaksData
    pd <- peaksData(cb)

    ## Verify peaksData has same length as chromData rows
    expect_identical(length(pd), 4L)

    ## Verify peaksData is returned in chromData row order, NOT factor level order
    ## Row 1: 1_A with mzMax=101 (should have intensities from mz=100,101)
    ## Row 2: 1_A with mzMax=102 (should have intensities from mz=100,101,102)
    ## Row 3: 1_B with mzMax=101 (should have intensities from mz=100,101)
    ## Row 4: 1_B with mzMax=102 (should have intensities from mz=100,101,102)

    ## Check that the intensity patterns match the expected order
    ## The key test: row 1 and row 2 should have different intensity sums
    ## because they have different mz windows
    int_sums <- sapply(pd, function(x) sum(x$intensity, na.rm = TRUE))

    ## Row 1 (1_A, mz 100-101): should have lower sum than Row 2 (1_A, mz 100-102)
    expect_true(int_sums[1] < int_sums[2])
    ## Row 3 (1_B, mz 100-101): should have lower sum than Row 4 (1_B, mz 100-102)
    expect_true(int_sums[3] < int_sums[4])

    ## Also verify the pattern matches across files
    ## Row 1 and Row 3 should have similar intensity ratios (both use mz 100-101)
    ## Row 2 and Row 4 should have similar intensity ratios (both use mz 100-102)
})

test_that("addProcessing adds processing steps correctly", {
    c_queued <- addProcessing(c_full, filterPeaksData,
        variables = c("rtime"), ranges = c(12.5, 45.5)
    )
    queue <- .processingQueue(c_queued)
    expect_length(queue, 1)
    expect_equal(queue[[1]]@FUN, filterPeaksData)
    expect_equal(queue[[1]]@ARGS, list(
        variables = c("rtime"),
        ranges = c(12.5, 45.5)
    ))
    c_queued <- addProcessing(c_queued, filterPeaksData,
        variables = c("intensity"), ranges = c(100, 200)
    )
    queue <- .processingQueue(c_queued)
    expect_length(queue, 2)

    ## Missing FUNs
    c_queued <- addProcessing(c_mzr,
        variables = c("rtime"), ranges = c(12.5, 45.5)
    )
    expect_equal(c_queued, c_mzr)
})

test_that("processingChunkFactor is handled correctly in peaksData", {
    f <- processingChunkFactor(c_full, chunkSize = 2)
    factorized_peaks <- peaksData(c_full, f = f)
    non_factorized_peaks <- peaksData(c_full, f = f())
    expect_equal(factorized_peaks, non_factorized_peaks)

    f <- processingChunkFactor(c_mzr, chunkSize = 100)
    factorized_peaks <- peaksData(c_mzr, f = f)
    non_factorized_peaks <- peaksData(c_mzr, f = f())
    expect_equal(factorized_peaks, non_factorized_peaks)
})

test_that("applyProcessing applies all queued processing steps", {
    c_queued <- filterPeaksData(c_full,
        variables = c("rtime"),
        ranges = c(12.5, 45.5)
    )
    c_queued <- filterPeaksData(c_queued,
        variables = c("intensity"),
        ranges = c(100, 200), keep = FALSE
    )
    c_applied <- applyProcessing(c_queued)
    expect_length(.processingQueue(c_applied), 0)
    expect_equal(peaksData(c_applied), peaksData(.backend(c_applied)))

    ## empty queue
    c_queued@processingQueue <- list()
    c_applied <- applyProcessing(c_queued)
    expect_equal(c_applied, c_queued)

    c_queued <- filterPeaksData(c_mzr,
        variables = c("rtime"),
        ranges = c(12.5, 45.5)
    )
    expect_error(applyProcessing(c_queued), "Cannot apply")
})

test_that("ChromBackendMzR works", {
    expect_true(isReadOnly(be_mzr))
    expect_false(be_mzr@inMemory)
    expect_false(identical(peaksData(be_mzr),.peaksData(be_mzr)))
    expect_true(identical(length(peaksData(be_mzr)), length(.peaksData(be_mzr))))
    expect_true(identical(
        chromData(be_mzr),
        fillCoreChromVariables(.chromData(be_mzr))
    ))
    expect_false(supportsSetBackend(be_mzr))
})

test_that("backendInitialize works", {
    expect_equal(
        backendInitialize(ChromBackendMzR(), files = character()),
        ChromBackendMzR()
    )
    expect_error(
        backendInitialize(ChromBackendMzR(), files = numeric(2)),
        "must be"
    )
})

test_that("show method for ChromBackendMzR works correctly", {
    expect_output(show(be_mzr), "file")
    expect_output(show(be_mzr), "138 chromatograms")
    expect_output(show(be_mzr), "chromIndex")
    expect_output(show(be_mzr), "msLevel")
    expect_output(show(be_mzr), "mz")
    tmp <- be_mzr
    tmp@inMemory <- TRUE
    expect_output(show(tmp), "Peaks data is cached in memory")
})

test_that("replacement method works", {
    tmp <- be_mzr
    peaksData(tmp)[[1]] <- peaksData(be_mzr)[[1]] + 1
    expect_false(identical(peaksData(tmp)[[1]], peaksData(be_mzr)[[1]]))
    expect_true(identical(peaksData(tmp)[[1]], peaksData(be_mzr)[[1]] + 1))
    expect_true(.inMemory(tmp))
    expect_false(.inMemory(be_mzr))

    cd <- chromData(tmp)
    cd$mz <- 1
    expect_true(!identical(chromData(tmp), cd))
    chromData(tmp) <- cd
    expect_equal(chromData(tmp), cd)
})

test_that("backendParallelFactor works", {
    expect_equal(levels(backendParallelFactor(be_mzr)), unique(dataOrigin(be_mzr)))
})

test_that("error message work", {
    expect_error(
        peaksData(be_mzr, columns = "notacolumn"),
        "requested peaks variables"
    )
})

test_that("chromExtract works correctly for ChromBackendMzR", {
    cd <- chromData(be_mzr)
    first_cd <- cd[100:102,]

    ## We'll use known retention time ranges and identifiers from chromData
    peak_tbl <- data.frame(
        rtMin = c(13.7, 16.2, 18 ),
        rtMax = c(18.1,19,20.6 ),
        dataOrigin = first_cd$dataOrigin,
        chromIndex = c(100, 101, 102)
    )
    out <- chromExtract(be_mzr, peak_tbl, by = c("dataOrigin", "chromIndex"))

    expect_s4_class(out, "ChromBackendMzR")
    expect_true(validObject(out))
    expect_equal(nrow(chromData(out)), nrow(peak_tbl))

    ## Chromatographic data should have been restricted to the same identifiers
    out_cd <- chromData(out)
    expect_true(all(out_cd$dataOrigin %in% unique(peak_tbl$dataOrigin)))
    expect_true(all(out_cd$chromIndex %in% unique(peak_tbl$chromIndex)))

    ## rtMin/rtMax should match peak.table values
    expect_equal(out_cd$rtMin, peak_tbl$rtMin)
    expect_equal(out_cd$rtMax, peak_tbl$rtMax)

    bad_tbl <- data.frame(
        rtMin = 1, rtMax = 2,
        msLevel = 9L, dataOrigin = "nonexistent_file.mzML"
    )
    expect_error(
        chromExtract(be, bad_tbl, by = c("dataOrigin", "msLevel")),
        regexp = "do not exist|must be present"
    )
    pd <- peaksData(out) # should work
    expect_equal(length(pd), nrow(peak_tbl))
    expect_true(all(sapply(pd, nrow) > 0))

    expect_equal(length(.peaksData(out)), nrow(peak_tbl)) ## i think this fails

})

test_that(".df_combine works as expected", {
  combined_backend <- .df_combine(list(be, be_cd))
  expect_equal(nrow(combined_backend@chromData), nrow(cdata) + nrow(cdata))
  expect_equal(
    length(.peaksData(combined_backend)),
    length(.peaksData(be)) + length(.peaksData(be_cd))
  )

  expect_equal(.df_combine(list(be)), be)
  incompatible_data <- list(
    data.frame(
      rtime = c(10.0, 12.0),
      intensity = c(200, 150),
      other_col = c("test", "test")
    ),
    data.frame(
      rtime = c(30.1, 31.2),
      intensity = c(110, 90),
      other_col = c("test", "test")
    )
  )
  be_incompatible <- backendInitialize(
    be_empty,
    chromData = cdata,
    peaksData = incompatible_data
  )

  expect_error(
    .df_combine(c(be, be_incompatible)),
    "Provided objects have different sets of peak variables."
  )

  setClass("DummyBackend", contains = "ChromBackend")
  dm <- new("DummyBackend")

  expect_error(
    .df_combine(c(be, dm)),
    "merge backends of the same type"
  )
})


test_that(".filter_ranges helper function works correctly", {
  query <- data.frame(
    mz = c(112.2, 123.3, 134.4),
    chromIndex = c(1L, 2L, 3L)
  )
  ranges <- numeric()
  expect_error(.filter_ranges(query, ranges, match = "any"), "needs to be")
  ranges <- c(1, 2, 3)
  expect_error(
    .filter_ranges(query, ranges, match = "any"),
    "Length of 'ranges'"
  )

  # All ranges match (full data retained)
  ranges <- c(100, 200, 1, 3)
  res <- .filter_ranges(query, ranges, match = "all")
  expect_equal(res, c(1, 2, 3))
  res <- .filter_ranges(query, ranges, match = "any")
  expect_equal(res, c(1, 2, 3))

  # No matches (empty result)
  ranges <- c(500, 600, 4, 5)
  res <- .filter_ranges(query, ranges, match = "any")
  expect_equal(res, integer(0))

  # Partially overlapping ranges
  ranges <- c(120, 130, 4, 5)
  res <- .filter_ranges(query, ranges, match = "any")
  expect_equal(res, 2)
  res <- .filter_ranges(query, ranges, match = "all")
  expect_equal(res, integer(0))

  # Single row edge case
  query_single <- data.frame(mz = 125.0, chromIndex = 2L)
  ranges <- c(120, 130, 1, 3)
  res <- .filter_ranges(query_single, ranges, match = "all")
  expect_equal(res, 1)
  res <- .filter_ranges(query_single, ranges, match = "any")
  expect_equal(res, 1)

  # Edge case: empty query
  query_empty <- data.frame(mz = numeric(), chromIndex = integer())
  ranges <- c(100, 200, 1, 3)
  res <- .filter_ranges(query_empty, ranges, match = "all")
  expect_equal(res, integer(0))
  res <- .filter_ranges(query_empty, ranges, match = "any")
  expect_equal(res, integer(0))
})

test_that(".check_column_order_and_types works", {
  df_valid <- data.frame(
    rtime = c(1, 2, 3),
    intensity = c(10, 20, 30)
  )
  df_invalid_order <- data.frame(
    intensity = c(10, 20, 30),
    rtime = c(1, 2, 3)
  )
  df_invalid_type <- data.frame(
    rtime = c("a", "b", "c"),
    intensity = c(10, 20, 30)
  )
  df_empty <- data.frame(rtime = numeric(), intensity = numeric())

  expect_null(.check_column_order_and_types(
    df_valid,
    names(.CORE_PEAKS_VARIABLES),
    .CORE_PEAKS_VARIABLES
  ))
  expect_equal(
    .check_column_order_and_types(
      df_invalid_order,
      names(.CORE_PEAKS_VARIABLES),
      .CORE_PEAKS_VARIABLES
    ),
    "Columns should be in the order 'rtime', 'intensity'."
  )
  expect_equal(
    .check_column_order_and_types(
      df_invalid_type,
      names(.CORE_PEAKS_VARIABLES),
      .CORE_PEAKS_VARIABLES
    ),
    "The peaksData variable(s) rtime have the wrong data type."
  )
  expect_null(.check_column_order_and_types(
    df_empty,
    names(.CORE_PEAKS_VARIABLES),
    .CORE_PEAKS_VARIABLES
  ))
})


test_that(".check_rtime works", {
  df_valid <- data.frame(
    rtime = c(1, 2, 3),
    intensity = c(10, 20, 30)
  )
  df_invalid_na <- data.frame(
    rtime = c(1, NA, 3),
    intensity = c(10, 20, 30)
  )
  df_invalid_increasing <- data.frame(
    rtime = c(3, 2, 1),
    intensity = c(10, 20, 30)
  )
  df_empty <- data.frame(rtime = numeric(), intensity = numeric())

  expect_null(.check_rtime(df_valid))
  expect_equal(
    .check_rtime(df_invalid_na),
    "'rtime' column contains NA values."
  )
  expect_equal(
    .check_rtime(df_invalid_increasing),
    "'rtime' column is not strictly increasing."
  )
  expect_null(.check_rtime(df_empty))
})

test_that(".validate_entry works", {
  df_valid <- data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
  df_empty <- data.frame(rtime = numeric(), intensity = numeric())

  expect_null(.validate_entry(
    df_valid,
    1,
    names(.CORE_PEAKS_VARIABLES),
    .CORE_PEAKS_VARIABLES
  ))
  expect_null(.validate_entry(
    df_empty,
    3,
    names(.CORE_PEAKS_VARIABLES),
    .CORE_PEAKS_VARIABLES
  ))
  expect_equal(
    .validate_entry(
      c(1, 2, 3),
      1,
      names(.CORE_PEAKS_VARIABLES),
      .CORE_PEAKS_VARIABLES
    ),
    "Entry 1: all 'peaksData' entries should be of class 'data.frame'"
  )
})

test_that(".run_process_queue ChromBackendMemory work", {
  result <- .run_process_queue(
    c_empty@backend,
    f = processingChunkFactor(c_empty),
    queue = c_empty@processingQueue
  )
  expect_equal(result, c_empty@backend)

  result <- .run_process_queue(
    c_full@backend,
    f = processingChunkFactor(c_full),
    queue = c_full@processingQueue
  )
  expect_equal(result, c_full@backend)

  c_queued <- filterPeaksData(
    c_full,
    variables = c("rtime"),
    ranges = c(12.5, 45.5)
  )
  c_queued <- filterPeaksData(
    c_queued,
    variables = c("intensity"),
    ranges = c(100, 200)
  )

  # this test for f = factor() and queue >1
  result <- .run_process_queue(
    c_queued@backend,
    f = processingChunkFactor(c_queued),
    queue = c_queued@processingQueue
  )
  expect_true(inherits(result, "ChromBackend"))
  peaks_result <- peaksData(result)
  expect_equal(length(peaks_result), length(peaksData(c_full@backend)))
  expect_false(identical(peaks_result, peaksData(c_full@backend)))

  f <- factor(c(1, 1, 2))
  c_queued <- filterPeaksData(
    c_full,
    variables = c("rtime"),
    ranges = c(12.5, 45.5)
  )
  c_queued <- filterPeaksData(
    c_queued,
    variables = c("intensity"),
    ranges = c(100, 200)
  )

  result <- .run_process_queue(
    c_queued@backend,
    f = f,
    queue = c_queued@processingQueue
  )

  expect_true(inherits(result, "ChromBackend"))
  expect_equal(length(peaksData(result)), length(peaksData(c_full@backend)))

  split_data <- split(c_full@backend, f)
  expect_equal(length(split_data), length(levels(f)))

  f <- factor(c(1, 2))
  tmp <- c_full
  tmp@processingQueue <- list(1)
  expect_error(
    .run_process_queue(tmp@backend, f = f, queue = tmp@processingQueue),
    "length 'f' has to be equal to the length of 'object'"
  )

  f <- c(1, 2, 3)
  expect_error(
    .run_process_queue(tmp@backend, f = f, queue = tmp@processingQueue),
    "f must be a factor"
  )
})

test_that(".run_processing_queue, ChromBackendMzr work", {
  ## without factor and queue == 1
  c_queued <- filterPeaksData(
    c_mzr,
    variables = c("rtime"),
    ranges = c(12.5, 25.5),
    keep = FALSE
  )

  result1 <- .run_process_queue(
    c_queued@backend,
    f = processingChunkFactor(c_queued),
    queue = c_queued@processingQueue
  )

  expect_true(result1@inMemory)
  expect_false(c_queued@backend@inMemory)

  expect_true(inherits(result1, "ChromBackendMzR"))
  expect_false(identical(
    lengths(rtime(result1)),
    lengths(rtime(c_mzr@backend))
  ))
  expect_equal(length(peaksData(result1)), length(peaksData(c_mzr@backend)))

  ## with levels(factor) > 1 and queue == 1
  processingChunkSize(c_queued) <- 100
  f <- processingChunkFactor(c_queued) # > 1
  result2 <- .run_process_queue(
    c_queued@backend,
    f = f,
    queue = c_queued@processingQueue
  )
  expect_true(inherits(result2, "ChromBackendMzR"))
  expect_equal(length(peaksData(result2)), length(peaksData(c_mzr@backend)))
  expect_false(identical(
    lengths(rtime(result2)),
    lengths(rtime(c_mzr@backend))
  ))
  expect_true(result2@inMemory)
  expect_false(c_queued@backend@inMemory)
  expect_identical(result1, result2)

  ## without factor and queue > 1
  c_queued <- filterPeaksData(
    c_queued,
    variables = c("intensity"),
    ranges = c(45, 50)
  )
  result3 <- .run_process_queue(
    c_queued@backend,
    f = factor(),
    queue = c_queued@processingQueue
  )

  expect_true(inherits(result3, "ChromBackendMzR"))
  expect_equal(length(peaksData(result3)), length(peaksData(c_mzr@backend)))
  expect_false(identical(
    lengths(rtime(result3)),
    lengths(rtime(c_mzr@backend))
  ))
  expect_true(result3@inMemory)
  expect_false(c_queued@backend@inMemory)

  ## with factor and queue > 1
  f <- processingChunkFactor(c_queued)
  result4 <- .run_process_queue(
    c_queued@backend,
    f = f,
    queue = c_queued@processingQueue
  )
  expect_true(inherits(result4, "ChromBackendMzR"))
  expect_equal(
    length(peaksData(result4)),
    length(peaksData(c_mzr@backend))
  )
  expect_false(identical(
    lengths(rtime(result4)),
    lengths(rtime(c_mzr@backend))
  ))
  expect_true(result4@inMemory)
  expect_false(c_queued@backend@inMemory)
  expect_identical(result3, result4)
})

test_that(".valid_processing_queue works correctly", {
  expect_null(.valid_processing_queue(list()))
  valid_queue <- list(new("ProcessingStep"))
  expect_null(.valid_processing_queue(valid_queue))

  invalid_queue <- list("not_a_processing_step")
  expect_error(
    .valid_processing_queue(invalid_queue),
    "'processingQueue' should only contain ProcessingStep objects."
  )
})

test_that("ensure_rt_mz_columns correctly handles mz and rt columns", {
  spectra <- s
  spectra_f <- factor(
    do.call(
      paste,
      c(
        as.list(Spectra::spectraData(s)[, c("msLevel", "dataOrigin")]),
        sep = "_"
      )
    )
  )
  levs <- levels(spectra_f)
  chrom_data <- data.frame(msLevel = c(1, 2, 3), chromSpectraIndex = levs[1:3])
  chrom_data <- .ensure_rt_mz_columns(chrom_data, spectra, spectra_f)
  expect_equal(chrom_data$mzMin, c(-Inf, -Inf, -Inf))
  expect_equal(chrom_data$mzMax, c(Inf, Inf, Inf))

  chrom_data <- data.frame(mzMin = c(100), chromSpectraIndex = levs[1])
  expect_error(
    .ensure_rt_mz_columns(chrom_data, spectra, spectra_f),
    "must be present if one is provided."
  )

  chrom_data <- data.frame(mzMax = c(200), chromSpectraIndex = levs[1])
  expect_error(
    .ensure_rt_mz_columns(chrom_data, spectra, spectra_f),
    "must be present if one is provided."
  )
  chrom_data <- data.frame(msLevel = c(1, 2, 3), chromSpectraIndex = levs[1:3])
  chrom_data <- .ensure_rt_mz_columns(chrom_data, spectra, spectra_f)
  s_plit <- split(spectra, spectra_f)
  expect_equal(chrom_data$rtMin[[1]], min(s_plit[[1]]$rtime, na.rm = TRUE))
  expect_equal(chrom_data$rtMax[[1]], max(s_plit[[1]]$rtime, na.rm = TRUE))

  chrom_data <- data.frame(rtMin = c(10), chromSpectraIndex = levs[1])
  expect_error(
    .ensure_rt_mz_columns(chrom_data, spectra, spectra_f),
    " must be present if one is provided."
  )
  chrom_data <- data.frame(rtMax = c(50), chromSpectraIndex = levs[1])
  expect_error(
    .ensure_rt_mz_columns(chrom_data, spectra, spectra_f),
    "must be present if one is provided."
  )

  chrom_data <- data.frame(
    mzMin = c(100),
    mzMax = c(200),
    rtMin = c(10),
    rtMax = c(50),
    chromSpectraIndex = levs[1]
  )
  chrom_data <- .ensure_rt_mz_columns(chrom_data, spectra, spectra_f)
  expect_equal(chrom_data$mzMin, 100)
  expect_equal(chrom_data$mzMax, 200)
  expect_equal(chrom_data$rtMin, 10)
  expect_equal(chrom_data$rtMax, 50)
})

test_that(".validate_chromExtract_input works correctly", {
  cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    dataOrigin = c("A", "B", "C")
  )
  be <- backendInitialize(new("ChromBackendMemory"), chromData = cdata)

  peak_tbl <- data.frame(
    rtMin = c(1, 2, 3),
    rtMax = c(5, 6, 7),
    msLevel = c(1L, 1L, 1L),
    dataOrigin = c("A", "B", "C")
  )

  # should pass
  expect_silent(
    .validate_chromExtract_input(be, peak_tbl, by = c("msLevel", "dataOrigin"))
  )

  # missing required column
  bad_tbl <- peak_tbl[, !names(peak_tbl) %in% "rtMax", drop = FALSE]
  expect_error(
    .validate_chromExtract_input(be, bad_tbl, by = c("msLevel", "dataOrigin")),
    "must contain columns"
  )

  # NA in rtMin
  bad_tbl2 <- peak_tbl
  bad_tbl2$rtMin[1] <- NA
  expect_error(
    .validate_chromExtract_input(be, bad_tbl2, by = c("msLevel", "dataOrigin")),
    "cannot contain NA"
  )

  # missing 'by' columns in chromData
  bad_be <- backendInitialize(
    new("ChromBackendMemory"),
    chromData = cdata[, "msLevel", drop = FALSE]
  )
  expect_error(
    .validate_chromExtract_input(
      bad_be,
      peak_tbl,
      by = c("msLevel", "dataOrigin")
    ),
    "must be present"
  )

  ## unique
  expect_error(
    .validate_chromExtract_input(bad_be, peak_tbl, by = "msLevel"),
    "must uniquely identify rows"
  )
})

test_that(".match_chromdata_peaktable aligns correctly", {
  tmp_cdata <- data.frame(
    msLevel = c(1L, 1L, 2L),
    dataOrigin = c("A", "B", "A")
  )
  tmp <- backendInitialize(new("ChromBackendMemory"), chromData = tmp_cdata)

  peak_tbl <- data.frame(
    msLevel = c(1L, 2L),
    dataOrigin = c("A", "A"),
    rtMin = c(1, 2),
    rtMax = c(5, 6)
  )

  matched <- .match_chromdata_peaktable(
    tmp,
    peak_tbl,
    by = c("msLevel", "dataOrigin")
  )

  # Expect a subset of object
  expect_s4_class(matched$object, "ChromBackendMemory")
  expect_equal(length(matched$chrom_keys), nrow(.chromData(matched$object)))

  # Check factor levels alignment
  expect_true(all(levels(matched$peak_keys) %in% levels(matched$chrom_keys)))

  # missing key should error
  bad_tbl <- data.frame(
    msLevel = 3L,
    dataOrigin = "Z",
    rtMin = 1,
    rtMax = 2
  )
  expect_error(
    .match_chromdata_peaktable(tmp, bad_tbl, by = c("msLevel", "dataOrigin")),
    "do not exist"
  )
})

test_that(".check_overl_columns warns correctly", {
  tmp_cdata <- data.frame(
    msLevel = 1L,
    dataOrigin = "X",
    mz = 100,
    extracol = "info"
  )
  tmp <- backendInitialize(new("ChromBackendMemory"), chromData = tmp_cdata)

  peak_tbl <- data.frame(
    rtMin = 1,
    rtMax = 2,
    mzMin = 99,
    mzMax = 101,
    msLevel = 1L,
    dataOrigin = "X",
    mz = 123,
    extracol = "test"
  )

  req_cols <- c("rtMin", "rtMax", "mzMin", "mzMax", "msLevel", "dataOrigin")

  expect_warning(
    overl <- .check_overl_columns(tmp, peak_tbl, req_cols),
    "already exist"
  )

  # overlapping should include "mz"
  expect_true(all(c("mz", "extracol") %in% names(peak_tbl)[overl]))
})

test_that(".impute() works correctly and without warnings", {
  # Base signal with gaps
  x <- c(1:5, NA, 7:10, NA, 12:15, rep(NA, 2), 18:20)

  ## linear
  expect_silent({
    res_lin <- .impute(x, method = "linear")
  })
  expect_false(anyNA(res_lin))
  expect_true(all(diff(res_lin) > 0)) # still increasing

  ## spline
  expect_silent({
    res_spl <- .impute(x, method = "spline")
  })
  expect_false(anyNA(res_spl))
  expect_equal(length(res_spl), length(x))

  ## Gaussian
  expect_silent(
    res_gauss <- .impute(x, method = "gaussian", window = 2, sd = 1)
  )
  expect_false(anyNA(res_gauss))
  expect_equal(length(res_gauss), length(x))

  ## loess
  expect_warning({
    res_loess <- .impute(x, method = "loess", span = 0.3)
    "could not fill all NAs"
  })
  expect_false(anyNA(res_loess))
  expect_equal(length(res_loess), length(x))

  ## Consecutive NAs
  x_na <- c(1, 2, NA, NA, 5, 6, 7, 8, NA, NA, 11, 12)
  expect_silent({
    res_consec <- .impute(x_na, method = "linear")
  })
  expect_false(anyNA(res_consec))
  expect_equal(length(res_consec), length(x_na))

  ## No NA
  x_nomiss <- 1:10
  expect_silent({
    res_nomiss <- .impute(x_nomiss, method = "spline")
  })
  expect_identical(res_nomiss, x_nomiss)

  ## all Nas returns NA
  x_allna <- rep(NA_real_, 8)
  expect_silent({
    res_allna <- .impute(x_allna, method = "gaussian")
  })
  expect_true(all(is.na(res_allna)))
})

test_that(".map_spectra_vars() correctly maps spectra variables", {
  vars <- c("scanIndex", "mtbls_id")

  mapped <- .map_spectra_vars(be_sp, vars)

  expect_s4_class(mapped, "ChromBackendSpectra")
  cd <- chromData(mapped)

  expect_true(all(vars %in% names(cd)))

  expect_true(all(vapply(
    cd$mtbls_id,
    function(x) length(unique(x)) == 1,
    logical(1)
  )))

  expect_true(all(vapply(cd$scanIndex, function(x) length(x) >= 1, logical(1))))

  expect_identical(
    rownames(cd),
    rownames(chromData(be_sp))
  )
  expect_error(
    .map_spectra_vars(be_sp, c("scanIndex", "fake_var")),
    "must exist in the Spectra"
  )

  cd_names <- names(chromData(be_sp))
  fake_var <- cd_names[1L] # pick a real chromData column
  expect_error(
    .map_spectra_vars(be_sp, fake_var),
    "must already exist in the chromData"
  )
})

test_that(".prepare_spectra_input extracts and classifies correctly", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200, 300), c(100, 200, 300), c(100, 200, 300),
                     compress = FALSE),
    intensity = NumericList(c(10, 20, 30), c(15, 25, 35), c(20, 30, 40),
                            compress = FALSE),
    rtime = c(1, 2, 3),
    msLevel = rep(1L, 3),
    dataOrigin = rep("A", 3)
  ))

  ## EIC case (finite mz range)
  cd <- data.frame(rtMin = 1, rtMax = 3, mzMin = 100, mzMax = 200)
  prep <- .prepare_spectra_input(cd, sp, sumi)
  expect_equal(prep$n_cd, 1L)
  expect_equal(prep$n_valid, 3L)
  expect_true(prep$same_rt)
  expect_false(prep$all_tic)
  expect_true(prep$use_cumsum)
  expect_equal(prep$valid_rt, c(1, 2, 3))
  expect_length(prep$valid_mz, 3L)
  expect_length(prep$valid_int, 3L)
  ## Each mz/int entry should be a plain numeric vector
  expect_type(prep$valid_mz[[1]], "double")
  expect_type(prep$valid_int[[1]], "double")

  ## TIC case (infinite mz ranges)
  cd_tic <- data.frame(rtMin = 1, rtMax = 3, mzMin = -Inf, mzMax = Inf)
  prep_tic <- .prepare_spectra_input(cd_tic, sp, sumi)
  expect_true(prep_tic$all_tic)

  ## maxi function → use_cumsum = FALSE
  prep_max <- .prepare_spectra_input(cd, sp, maxi)
  expect_false(prep_max$use_cumsum)

  ## Multiple chromatograms with same rt
  cd_multi <- data.frame(
    rtMin = c(1, 1), rtMax = c(3, 3),
    mzMin = c(100, 200), mzMax = c(200, 300)
  )
  prep_multi <- .prepare_spectra_input(cd_multi, sp, sumi)
  expect_equal(prep_multi$n_cd, 2L)
  expect_true(prep_multi$same_rt)
  expect_false(prep_multi$all_tic)

  ## Different rt ranges → same_rt = FALSE
  cd_diff <- data.frame(
    rtMin = c(1, 2), rtMax = c(2, 3),
    mzMin = c(100, 100), mzMax = c(200, 200)
  )
  prep_diff <- .prepare_spectra_input(cd_diff, sp, sumi)
  expect_false(prep_diff$same_rt)

  ## Narrow rt filter reduces n_valid
  cd_narrow <- data.frame(rtMin = 2, rtMax = 2, mzMin = 100, mzMax = 200)
  prep_narrow <- .prepare_spectra_input(cd_narrow, sp, sumi)
  expect_equal(prep_narrow$n_valid, 1L)
  expect_equal(prep_narrow$valid_rt, 2)
})

test_that(".prepare_spectra_input handles duplicated retention times", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200), c(100, 200), c(100, 200), c(100, 200),
                     compress = FALSE),
    intensity = NumericList(c(10, 20), c(15, 25), c(30, 40), c(35, 45),
                            compress = FALSE),
    rtime = c(1, 1, 2, 3),
    msLevel = rep(1L, 4),
    dataOrigin = rep("A", 4)
  ))
  cd <- data.frame(rtMin = 1, rtMax = 3, mzMin = -Inf, mzMax = Inf)
  prep <- .prepare_spectra_input(cd, sp, sumi)
  ## Only unique rtimes should be kept
  expect_equal(prep$n_valid, 3L)
  expect_equal(prep$valid_rt, c(1, 2, 3))
})


test_that(".spectrum_intensities TIC case returns replicated fun(int_j)", {
  mz_j <- c(100, 200, 300)
  int_j <- c(10, 20, 30)
  ## sumi: sum = 60, replicated for 3 chromatograms
  res <- .spectrum_intensities(mz_j, int_j,
    mzMins = c(-Inf, -Inf, -Inf), mzMaxs = c(Inf, Inf, Inf),
    fun = sumi, all_tic = TRUE, use_cumsum = TRUE
  )
  expect_equal(res, rep(60, 3))
  ## maxi: max = 30
  res_max <- .spectrum_intensities(mz_j, int_j,
    mzMins = c(-Inf, -Inf), mzMaxs = c(Inf, Inf),
    fun = maxi, all_tic = TRUE, use_cumsum = FALSE
  )
  expect_equal(res_max, rep(30, 2))
})

test_that(".spectrum_intensities cumsum EIC case", {
  mz_j <- c(100, 200, 300)
  int_j <- c(10, 20, 30)
  ## Two ranges: mz 100-200 → 10+20=30, mz 200-300 → 20+30=50
  res <- .spectrum_intensities(mz_j, int_j,
    mzMins = c(100, 200), mzMaxs = c(200, 300),
    fun = sumi, all_tic = FALSE, use_cumsum = TRUE
  )
  expect_equal(res, c(30, 50))
})

test_that(".spectrum_intensities generic (maxi) case", {
  mz_j <- c(100, 200, 300)
  int_j <- c(10, 20, 30)
  ## mz 100-200 → max(10,20)=20, mz 200-300 → max(20,30)=30
  res <- .spectrum_intensities(mz_j, int_j,
    mzMins = c(100, 200), mzMaxs = c(200, 300),
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  expect_equal(res, c(20, 30))
})

test_that(".spectrum_intensities empty mz returns NA", {
  ## cumsum path
  res_cs <- .spectrum_intensities(numeric(0), numeric(0),
    mzMins = 100, mzMaxs = 200,
    fun = sumi, all_tic = FALSE, use_cumsum = TRUE
  )
  expect_true(is.na(res_cs))
  ## generic path
  res_gen <- .spectrum_intensities(numeric(0), numeric(0),
    mzMins = 100, mzMaxs = 200,
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  expect_true(is.na(res_gen))
})

test_that(".spectrum_intensities no mz in range returns NA", {
  mz_j <- c(500, 600)
  int_j <- c(10, 20)
  res_cs <- .spectrum_intensities(mz_j, int_j,
    mzMins = 100, mzMaxs = 200,
    fun = sumi, all_tic = FALSE, use_cumsum = TRUE
  )
  expect_true(is.na(res_cs))
  res_gen <- .spectrum_intensities(mz_j, int_j,
    mzMins = 100, mzMaxs = 200,
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  expect_true(is.na(res_gen))
})

test_that(".spectrum_intensities TIC with NAs uses fun directly", {
  mz_j <- c(100, 200, 300)
  int_j <- c(10, NA, 30)
  ## sumi (na.rm): 10 + 30 = 40 (MsCoreUtils::sumi removes NAs)
  res <- .spectrum_intensities(mz_j, int_j,
    mzMins = -Inf, mzMaxs = Inf,
    fun = sumi, all_tic = TRUE, use_cumsum = TRUE
  )
  expect_equal(res, sumi(int_j))
  ## maxi (na.rm): max = 30
  res_max <- .spectrum_intensities(mz_j, int_j,
    mzMins = -Inf, mzMaxs = Inf,
    fun = maxi, all_tic = TRUE, use_cumsum = FALSE
  )
  expect_equal(res_max, maxi(int_j))
})

test_that(".spectrum_intensities cumsum with NAs strips NA peaks", {
  mz_j <- c(100, 200, 300, 400)
  int_j <- c(10, NA, 30, NA)
  ## After stripping NAs: mz = c(100, 300), int = c(10, 30)
  ## mz 100-300: 10+30=40, mz 300-400: 30
  res <- .spectrum_intensities(mz_j, int_j,
    mzMins = c(100, 300), mzMaxs = c(300, 400),
    fun = sumi, all_tic = FALSE, use_cumsum = TRUE
  )
  expect_equal(res, c(40, 30))
})

test_that(".spectrum_intensities generic (maxi) with NAs", {
  mz_j <- c(100, 200, 300, 400)
  int_j <- c(10, NA, 30, NA)
  ## mz 100-300: maxi(10, NA, 30) = 30
  res <- .spectrum_intensities(mz_j, int_j,
    mzMins = c(100, 300), mzMaxs = c(300, 400),
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  expect_equal(res[1], 30) # max(10, NA, 30) via maxi = 30
  expect_equal(res[2], 30) # max(30, NA) via maxi = 30
})

test_that(".spectrum_intensities all-NA spectrum returns NA", {
  mz_j <- c(100, 200, 300)
  int_j <- c(NA_real_, NA_real_, NA_real_)
  ## cumsum: .strip_na_peaks returns NULL → NA
  res_cs <- .spectrum_intensities(mz_j, int_j,
    mzMins = c(100, 200), mzMaxs = c(200, 300),
    fun = sumi, all_tic = FALSE, use_cumsum = TRUE
  )
  expect_true(all(is.na(res_cs)))
  ## generic: maxi(NA, NA, NA) → NA
  res_gen <- .spectrum_intensities(mz_j, int_j,
    mzMins = c(100, 200), mzMaxs = c(200, 300),
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  expect_true(all(is.na(res_gen)))
})


test_that(".build_intensity_matrix TIC/BPC fast path", {
  mz_list <- list(c(100, 200, 300), c(100, 200, 300))
  int_list <- list(c(10, 20, 30), c(15, 25, 35))
  int_mat <- .build_intensity_matrix(
    mz_list, int_list, kept = 1:2,
    mzMins = c(-Inf, -Inf), mzMaxs = c(Inf, Inf), n_cd = 2L,
    fun = sumi, all_tic = TRUE, use_cumsum = TRUE
  )
  expect_equal(nrow(int_mat), 2L)
  expect_equal(ncol(int_mat), 2L)
  ## TIC: sum of all intensities per spectrum, same for all chroms
  expect_equal(int_mat[1, 1], 60)  # 10+20+30
  expect_equal(int_mat[1, 2], 60)
  expect_equal(int_mat[2, 1], 75)  # 15+25+35
  expect_equal(int_mat[2, 2], 75)
})

test_that(".build_intensity_matrix cumsum EIC case", {
  mz_list <- list(c(100, 200, 300), c(100, 200, 300))
  int_list <- list(c(10, 20, 30), c(15, 25, 35))
  ## 2 chromatograms: mz 100-200 and mz 200-300
  int_mat <- .build_intensity_matrix(
    mz_list, int_list, kept = 1:2,
    mzMins = c(100, 200), mzMaxs = c(200, 300), n_cd = 2L,
    fun = sumi, all_tic = FALSE, use_cumsum = TRUE
  )
  expect_equal(nrow(int_mat), 2L)
  expect_equal(ncol(int_mat), 2L)
  ## Chrom 1 (mz 100-200): spec1 = 10+20=30, spec2 = 15+25=40
  expect_equal(int_mat[1, 1], 30)
  expect_equal(int_mat[2, 1], 40)
  ## Chrom 2 (mz 200-300): spec1 = 20+30=50, spec2 = 25+35=60
  expect_equal(int_mat[1, 2], 50)
  expect_equal(int_mat[2, 2], 60)
})

test_that(".build_intensity_matrix generic function (maxi)", {
  mz_list <- list(c(100, 200, 300))
  int_list <- list(c(10, 20, 30))
  int_mat <- .build_intensity_matrix(
    mz_list, int_list, kept = 1L,
    mzMins = c(100, 200), mzMaxs = c(200, 300), n_cd = 2L,
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  expect_equal(int_mat[1, 1], 20)  # max(10, 20)
  expect_equal(int_mat[1, 2], 30)  # max(20, 30)
})

test_that(".build_intensity_matrix handles empty mz and no match", {
  mz_list <- list(numeric(0), c(100, 200), c(500, 600))
  int_list <- list(numeric(0), c(10, 20), c(50, 60))
  int_mat <- .build_intensity_matrix(
    mz_list, int_list, kept = 1:3,
    mzMins = 100, mzMaxs = 200, n_cd = 1L,
    fun = sumi, all_tic = FALSE, use_cumsum = TRUE
  )
  expect_true(is.na(int_mat[1, 1]))   # empty mz → NA
  expect_equal(int_mat[2, 1], 30)     # 10+20
  expect_true(is.na(int_mat[3, 1]))   # no mz in range → NA
})


test_that(".build_intensity_matrix BPC multi-range (maxi)", {
  ## Multiple chromatograms with different mz ranges
  mz_list <- list(c(100, 150, 200, 250, 300),
                  c(100, 150, 200, 250, 300))
  int_list <- list(c(10, 50, 20, 80, 30),
                   c(15, 60, 25, 70, 35))
  int_mat <- .build_intensity_matrix(
    mz_list, int_list, kept = 1:2,
    mzMins = c(100, 200), mzMaxs = c(200, 300), n_cd = 2L,
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  ## Chrom 1 (mz 100-200): spec1 max(10,50,20)=50, spec2 max(15,60,25)=60
  expect_equal(int_mat[1, 1], 50)
  expect_equal(int_mat[2, 1], 60)
  ## Chrom 2 (mz 200-300): spec1 max(20,80,30)=80, spec2 max(25,70,35)=70
  expect_equal(int_mat[1, 2], 80)
  expect_equal(int_mat[2, 2], 70)
})

test_that(".build_intensity_matrix BPC single-element range", {
  mz_list <- list(c(100, 200, 300))
  int_list <- list(c(10, 20, 30))
  int_mat <- .build_intensity_matrix(
    mz_list, int_list, kept = 1L,
    mzMins = c(200), mzMaxs = c(200), n_cd = 1L,
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  expect_equal(int_mat[1, 1], 20)
})

test_that(".build_intensity_matrix BPC handles NA intensities", {
  mz_list <- list(c(100, 200, 300, 400))
  int_list <- list(c(10, NA, 30, NA))
  ## Sparse table path with NAs
  int_mat <- .build_intensity_matrix(
    mz_list, int_list, kept = 1L,
    mzMins = c(100, 300), mzMaxs = c(300, 400), n_cd = 2L,
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  ## mz 100-300: non-NA peaks are 10, 30 → max = 30

  expect_equal(int_mat[1, 1], 30)
  ## mz 300-400: non-NA peaks are 30 → max = 30
  expect_equal(int_mat[1, 2], 30)
})

test_that(".build_intensity_matrix cumsum handles NA intensities", {
  mz_list <- list(c(100, 200, 300, 400))
  int_list <- list(c(10, NA, 30, NA))
  int_mat <- .build_intensity_matrix(
    mz_list, int_list, kept = 1L,
    mzMins = c(100, 300), mzMaxs = c(300, 400), n_cd = 2L,
    fun = sumi, all_tic = FALSE, use_cumsum = TRUE
  )
  ## mz 100-300: non-NA are 10, 30 → sum = 40
  expect_equal(int_mat[1, 1], 40)
  ## mz 300-400: non-NA is 30 → sum = 30
  expect_equal(int_mat[1, 2], 30)
})

test_that(".build_intensity_matrix all-NA spectrum returns NA", {
  mz_list <- list(c(100, 200, 300))
  int_list <- list(c(NA_real_, NA_real_, NA_real_))
  ## cumsum path
  res_sum <- .build_intensity_matrix(
    mz_list, int_list, kept = 1L,
    mzMins = 100, mzMaxs = 300, n_cd = 1L,
    fun = sumi, all_tic = FALSE, use_cumsum = TRUE
  )
  expect_true(is.na(res_sum[1, 1]))
  ## BPC path
  res_max <- .build_intensity_matrix(
    mz_list, int_list, kept = 1L,
    mzMins = 100, mzMaxs = 300, n_cd = 1L,
    fun = maxi, all_tic = FALSE, use_cumsum = FALSE
  )
  expect_true(is.na(res_max[1, 1]))
})

test_that(".compute_chrom_intensities TIC case", {
  mz_list <- list(c(100, 200), c(100, 200, 300))
  int_list <- list(c(10, 20), c(15, 25, 35))
  res <- .compute_chrom_intensities(
    mz_list, int_list, kept = 1:2,
    mz_lo = -Inf, mz_hi = Inf, fun = sumi, use_cumsum = TRUE
  )
  expect_equal(res, c(30, 75))
})

test_that(".compute_chrom_intensities cumsum case", {
  mz_list <- list(c(100, 200, 300), c(100, 200, 300))
  int_list <- list(c(10, 20, 30), c(15, 25, 35))
  res <- .compute_chrom_intensities(
    mz_list, int_list, kept = 1:2,
    mz_lo = 100, mz_hi = 200, fun = sumi, use_cumsum = TRUE
  )
  expect_equal(res, c(30, 40))
})

test_that(".compute_chrom_intensities generic function (maxi)", {
  mz_list <- list(c(100, 200, 300))
  int_list <- list(c(10, 20, 30))
  res <- .compute_chrom_intensities(
    mz_list, int_list, kept = 1L,
    mz_lo = 100, mz_hi = 300, fun = maxi, use_cumsum = FALSE
  )
  expect_equal(res, 30)
})

test_that(".compute_chrom_intensities handles empty mz and no match", {
  mz_list <- list(numeric(0), c(500, 600))
  int_list <- list(numeric(0), c(10, 20))
  res <- .compute_chrom_intensities(
    mz_list, int_list, kept = 1:2,
    mz_lo = 100, mz_hi = 200, fun = sumi, use_cumsum = TRUE
  )
  expect_true(is.na(res[1]))
  expect_true(is.na(res[2]))
})

test_that(".compute_chrom_intensities TIC with NAs", {
  mz_list <- list(c(100, 200), c(100, 200, 300))
  int_list <- list(c(10, NA), c(NA, 25, 35))
  res <- .compute_chrom_intensities(
    mz_list, int_list, kept = 1:2,
    mz_lo = -Inf, mz_hi = Inf, fun = sumi, use_cumsum = TRUE
  )
  ## sumi removes NAs: spec1=10, spec2=60
  expect_equal(res, c(sumi(c(10, NA)), sumi(c(NA, 25, 35))))
})

test_that(".compute_chrom_intensities cumsum with NAs", {
  mz_list <- list(c(100, 200, 300), c(100, 200, 300))
  int_list <- list(c(10, NA, 30), c(NA, 25, 35))
  res <- .compute_chrom_intensities(
    mz_list, int_list, kept = 1:2,
    mz_lo = 100, mz_hi = 200, fun = sumi, use_cumsum = TRUE
  )
  ## After stripping NAs: spec1 mz=c(100,300) int=c(10,30) → mz 100-200 = 10
  ## spec2 mz=c(200,300) int=c(25,35) → mz 100-200: only mz=200 → 25
  expect_equal(res[1], 10)
  expect_equal(res[2], 25)
})

test_that(".compute_chrom_intensities generic (maxi) with NAs", {
  mz_list <- list(c(100, 200, 300))
  int_list <- list(c(10, NA, 30))
  res <- .compute_chrom_intensities(
    mz_list, int_list, kept = 1L,
    mz_lo = 100, mz_hi = 300, fun = maxi, use_cumsum = FALSE
  )
  ## maxi(10, NA, 30) = 30  (maxi ignores NAs)
  expect_equal(res, 30)
})

test_that(".compute_chrom_intensities all-NA spectrum returns NA", {
  mz_list <- list(c(100, 200, 300))
  int_list <- list(c(NA_real_, NA_real_, NA_real_))
  ## cumsum: .strip_na_peaks → NULL → NA
  res_sum <- .compute_chrom_intensities(
    mz_list, int_list, kept = 1L,
    mz_lo = 100, mz_hi = 300, fun = sumi, use_cumsum = TRUE
  )
  expect_true(is.na(res_sum))
  ## generic path
  res_max <- .compute_chrom_intensities(
    mz_list, int_list, kept = 1L,
    mz_lo = 100, mz_hi = 300, fun = maxi, use_cumsum = FALSE
  )
  expect_true(is.na(res_max))
})


test_that(".process_peaks_data same-rt TIC case", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200, 300), c(100, 200, 300), compress = FALSE),
    intensity = NumericList(c(10, 20, 30), c(15, 25, 35), compress = FALSE),
    rtime = c(1, 2),
    msLevel = rep(1L, 2),
    dataOrigin = rep("A", 2)
  ))
  cd <- data.frame(
    rtMin = c(1, 1), rtMax = c(2, 2),
    mzMin = c(-Inf, -Inf), mzMax = c(Inf, Inf)
  )
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), sumi, FALSE)
  expect_length(res, 2L)
  expect_true(all(vapply(res, is.data.frame, logical(1))))
  expect_equal(res[[1]]$rtime, c(1, 2))
  expect_equal(res[[1]]$intensity, c(60, 75))
  ## Both TIC with same rt → identical
  expect_equal(res[[1]], res[[2]])
})

test_that(".process_peaks_data same-rt EIC (mz filtering)", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200, 300), c(100, 200, 300), compress = FALSE),
    intensity = NumericList(c(10, 20, 30), c(15, 25, 35), compress = FALSE),
    rtime = c(1, 2),
    msLevel = rep(1L, 2),
    dataOrigin = rep("A", 2)
  ))
  cd <- data.frame(
    rtMin = c(1, 1), rtMax = c(2, 2),
    mzMin = c(100, 200), mzMax = c(200, 300)
  )
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), sumi, FALSE)
  expect_equal(res[[1]]$intensity, c(30, 40))
  expect_equal(res[[2]]$intensity, c(50, 60))
})

test_that(".process_peaks_data different-rt case", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200), c(100, 200), c(100, 200), compress = FALSE),
    intensity = NumericList(c(10, 20), c(15, 25), c(30, 40), compress = FALSE),
    rtime = c(1, 2, 3),
    msLevel = rep(1L, 3),
    dataOrigin = rep("A", 3)
  ))
  cd <- data.frame(
    rtMin = c(1, 2), rtMax = c(2, 3),
    mzMin = c(-Inf, -Inf), mzMax = c(Inf, Inf)
  )
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), sumi, FALSE)
  expect_equal(res[[1]]$rtime, c(1, 2))
  expect_equal(res[[1]]$intensity, c(30, 40))
  expect_equal(res[[2]]$rtime, c(2, 3))
  expect_equal(res[[2]]$intensity, c(40, 70))
})

test_that(".process_peaks_data with maxi function", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200, 300), c(100, 200, 300), compress = FALSE),
    intensity = NumericList(c(10, 20, 30), c(15, 25, 35), compress = FALSE),
    rtime = c(1, 2),
    msLevel = rep(1L, 2),
    dataOrigin = rep("A", 2)
  ))
  cd <- data.frame(rtMin = 1, rtMax = 2, mzMin = 100, mzMax = 300)
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), maxi, FALSE)
  expect_equal(res[[1]]$intensity, c(30, 35))
})

test_that(".process_peaks_data drop=TRUE returns vectors", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200), compress = FALSE),
    intensity = NumericList(c(10, 20), compress = FALSE),
    rtime = 1,
    msLevel = 1L,
    dataOrigin = "A"
  ))
  cd <- data.frame(rtMin = 1, rtMax = 1, mzMin = -Inf, mzMax = Inf)
  res <- .process_peaks_data(cd, sp, "intensity", sumi, drop = TRUE)
  expect_type(res[[1]], "double")
  expect_equal(res[[1]], 30)
})

test_that(".process_peaks_data handles empty rt range", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200), compress = FALSE),
    intensity = NumericList(c(10, 20), compress = FALSE),
    rtime = 1,
    msLevel = 1L,
    dataOrigin = "A"
  ))
  cd <- data.frame(rtMin = 5, rtMax = 10, mzMin = -Inf, mzMax = Inf)
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), sumi, FALSE)
  expect_equal(nrow(res[[1]]), 0L)
})

test_that(".process_peaks_data rtime-only request", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200), c(100, 200), compress = FALSE),
    intensity = NumericList(c(10, 20), c(15, 25), compress = FALSE),
    rtime = c(1, 2),
    msLevel = rep(1L, 2),
    dataOrigin = rep("A", 2)
  ))
  cd <- data.frame(rtMin = 1, rtMax = 2, mzMin = 100, mzMax = 200)
  res <- .process_peaks_data(cd, sp, "rtime", sumi, drop = TRUE)
  expect_equal(res[[1]], c(1, 2))
})

test_that(".process_peaks_data returns correct results via ChromBackendSpectra", {
  ## Integration test using the actual fixture backend
  pd <- peaksData(be_sp)
  expect_true(is.list(pd))
  expect_true(all(vapply(pd, is.data.frame, logical(1))))
  expect_true(all(vapply(pd, function(d) all(c("rtime", "intensity") %in% names(d)),
                         logical(1))))
  ## rtime should be monotonically non-decreasing in each chromatogram
  for (i in seq_along(pd)) {
    if (nrow(pd[[i]]) > 1)
      expect_true(all(diff(pd[[i]]$rtime) >= 0))
  }
})

test_that(".process_peaks_data different-rt EIC with cumsum", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 150, 200), c(100, 150, 200), c(100, 150, 200),
                     compress = FALSE),
    intensity = NumericList(c(10, 5, 20), c(15, 8, 25), c(30, 12, 40),
                            compress = FALSE),
    rtime = c(1, 2, 3),
    msLevel = rep(1L, 3),
    dataOrigin = rep("A", 3)
  ))
  cd <- data.frame(
    rtMin = c(1, 2), rtMax = c(2, 3),
    mzMin = c(100, 100), mzMax = c(150, 200)
  )
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), sumi, FALSE)
  ## Chrom 1: rt 1-2, mz 100-150 → spec1: 10+5=15, spec2: 15+8=23
  expect_equal(res[[1]]$intensity, c(15, 23))
  ## Chrom 2: rt 2-3, mz 100-200 → spec2: 15+8+25=48, spec3: 30+12+40=82
  expect_equal(res[[2]]$intensity, c(48, 82))
})

test_that(".process_peaks_data different-rt EIC with maxi", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200, 300), c(100, 200, 300), compress = FALSE),
    intensity = NumericList(c(10, 20, 30), c(15, 25, 35), compress = FALSE),
    rtime = c(1, 2),
    msLevel = rep(1L, 2),
    dataOrigin = rep("A", 2)
  ))
  cd <- data.frame(
    rtMin = c(1, 2), rtMax = c(1, 2),
    mzMin = c(100, 100), mzMax = c(200, 300)
  )
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), maxi, FALSE)
  ## Chrom 1: rt 1 only, mz 100-200 → max(10, 20) = 20
  expect_equal(res[[1]]$intensity, 20)
  ## Chrom 2: rt 2 only, mz 100-300 → max(15, 25, 35) = 35
  expect_equal(res[[2]]$intensity, 35)
})

test_that(".process_peaks_data same-rt TIC with NAs in intensities", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200, 300), c(100, 200, 300), compress = FALSE),
    intensity = NumericList(c(10, NA, 30), c(NA, 25, 35), compress = FALSE),
    rtime = c(1, 2),
    msLevel = rep(1L, 2),
    dataOrigin = rep("A", 2)
  ))
  cd <- data.frame(
    rtMin = c(1, 1), rtMax = c(2, 2),
    mzMin = c(-Inf, -Inf), mzMax = c(Inf, Inf)
  )
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), sumi, FALSE)
  ## TIC: sumi ignores NAs → spec1: 10+30=40, spec2: 25+35=60
  expect_equal(res[[1]]$intensity, c(sumi(c(10, NA, 30)), sumi(c(NA, 25, 35))))
  expect_equal(res[[1]], res[[2]])
})

test_that(".process_peaks_data same-rt EIC cumsum with NAs", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200, 300), c(100, 200, 300), compress = FALSE),
    intensity = NumericList(c(10, NA, 30), c(NA, 25, 35), compress = FALSE),
    rtime = c(1, 2),
    msLevel = rep(1L, 2),
    dataOrigin = rep("A", 2)
  ))
  cd <- data.frame(
    rtMin = c(1, 1), rtMax = c(2, 2),
    mzMin = c(100, 200), mzMax = c(200, 300)
  )
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), sumi, FALSE)
  ## cumsum path strips NAs first, then sums within range
  ## spec1: non-NA mz=c(100,300) int=c(10,30) → mz 100-200: 10, mz 200-300: 30
  ## spec2: non-NA mz=c(200,300) int=c(25,35) → mz 100-200: 25, mz 200-300: 60
  expect_equal(res[[1]]$intensity, c(10, 25))
  expect_equal(res[[2]]$intensity, c(30, 60))
})

test_that(".process_peaks_data different-rt with NAs in intensities", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200), c(100, 200), c(100, 200), compress = FALSE),
    intensity = NumericList(c(10, NA), c(NA, 25), c(30, 40), compress = FALSE),
    rtime = c(1, 2, 3),
    msLevel = rep(1L, 3),
    dataOrigin = rep("A", 3)
  ))
  cd <- data.frame(
    rtMin = c(1, 2), rtMax = c(2, 3),
    mzMin = c(-Inf, -Inf), mzMax = c(Inf, Inf)
  )
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), sumi, FALSE)
  ## TIC: sumi ignores NAs → spec1: 10, spec2: 25, spec3: 70
  expect_equal(res[[1]]$intensity, c(sumi(c(10, NA)), sumi(c(NA, 25))))
  expect_equal(res[[2]]$intensity, c(sumi(c(NA, 25)), sumi(c(30, 40))))
})

test_that(".process_peaks_data same-rt all-NA spectrum", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200), c(100, 200), compress = FALSE),
    intensity = NumericList(c(NA_real_, NA_real_), c(10, 20), compress = FALSE),
    rtime = c(1, 2),
    msLevel = rep(1L, 2),
    dataOrigin = rep("A", 2)
  ))
  cd <- data.frame(rtMin = 1, rtMax = 2, mzMin = 100, mzMax = 200)
  ## cumsum: spec1 all-NA → NA, spec2 → 30
  res <- .process_peaks_data(cd, sp, c("rtime", "intensity"), sumi, FALSE)
  expect_true(is.na(res[[1]]$intensity[1]))
  expect_equal(res[[1]]$intensity[2], 30)
  ## maxi: spec1 all-NA → NA, spec2 max = 20
  res_max <- .process_peaks_data(cd, sp, c("rtime", "intensity"), maxi, FALSE)
  expect_true(is.na(res_max[[1]]$intensity[1]))
  expect_equal(res_max[[1]]$intensity[2], 20)
})


test_that("intensity/rtime bypass peaksData when queue is empty", {
  ## Direct backend dispatch when no processing queue
  int_direct <- intensity(c_full)
  int_via_be <- peaksData(.backend(c_full), columns = "intensity", drop = TRUE)
  expect_identical(int_direct, int_via_be)

  rt_direct <- rtime(c_full)
  rt_via_be <- peaksData(.backend(c_full), columns = "rtime", drop = TRUE)
  expect_identical(rt_direct, rt_via_be)

  ## Same test with ChromBackendMzR
  int_mzr <- intensity(c_mzr)
  rt_mzr <- rtime(c_mzr)
  expect_identical(int_mzr,
    peaksData(.backend(c_mzr), columns = "intensity", drop = TRUE))
  expect_identical(rt_mzr,
    peaksData(.backend(c_mzr), columns = "rtime", drop = TRUE))

  ## With processing queue, should go through full peaksData dispatch
  c_queued <- filterPeaksData(c_full, variables = "rtime",
                              ranges = c(12.5, 45.5))
  int_queued <- intensity(c_queued)
  expect_false(identical(int_queued, int_direct))
})

test_that("lengths works for ChromBackendMemory", {
  l <- lengths(be)
  expect_equal(l, vapply(peaksData(be), nrow, integer(1L)))
  expect_type(l, "integer")
  ## Empty backend still has one chromatogram, with zero data points
  expect_equal(lengths(be_empty), 0L)
})

test_that("lengths works for Chromatograms", {
  ## No queue
  l_full <- lengths(c_full)
  expect_equal(l_full, vapply(peaksData(c_full), nrow, integer(1L)))

  l_mzr <- lengths(c_mzr)
  expect_equal(l_mzr, vapply(peaksData(c_mzr), nrow, integer(1L)))

  l_empty <- lengths(c_empty)
  expect_equal(l_empty, 0L)

  ## With queue
  c_queued <- filterPeaksData(c_full, variables = "rtime",
                              ranges = c(12.5, 14.0))
  l_queued <- lengths(c_queued)
  expect_equal(l_queued, vapply(peaksData(c_queued), nrow, integer(1L)))
})

test_that("peaksData fast [[ path for single column", {
  ## drop=TRUE, single column should use fast path
  int_fast <- peaksData(be, columns = "intensity", drop = TRUE)
  int_ref <- lapply(peaksData(be), `[[`, "intensity")
  expect_equal(int_fast, int_ref)

  rt_fast <- peaksData(be, columns = "rtime", drop = TRUE)
  rt_ref <- lapply(peaksData(be), `[[`, "rtime")
  expect_equal(rt_fast, rt_ref)
})

test_that("setBackend clears processing queue", {
  c_queued <- filterPeaksData(c_mzr, variables = "rtime",
                              ranges = c(12.5, 45.5))
  expect_length(.processingQueue(c_queued), 1L)
  pd_before <- peaksData(c_queued)

  c_mem <- setBackend(c_queued, ChromBackendMemory())
  expect_length(.processingQueue(c_mem), 0L)
  expect_identical(peaksData(c_mem), peaksData(.backend(c_mem)))
  expect_equal(peaksData(c_mem), pd_before)
})


test_that(".logging appends a timestamped entry", {
  res <- .logging(character(), "test message")
  expect_length(res, 1L)
  expect_match(res, "^test message \\[")

  res2 <- .logging(res, "second")
  expect_length(res2, 2L)
  expect_match(res2[2], "^second \\[")
})

test_that(".strip_na_peaks no NAs returns unchanged", {
  mz <- c(100, 200, 300)
  int <- c(10, 20, 30)
  res <- .strip_na_peaks(mz, int)
  expect_equal(res$mz, mz)
  expect_equal(res$int, int)
  expect_equal(res$n, 3L)
})

test_that(".strip_na_peaks partial NAs strips correctly", {
  mz <- c(100, 200, 300, 400)
  int <- c(10, NA, 30, NA)
  res <- .strip_na_peaks(mz, int)
  expect_equal(res$mz, c(100, 300))
  expect_equal(res$int, c(10, 30))
  expect_equal(res$n, 2L)
})

test_that(".strip_na_peaks all NAs returns NULL", {
  mz <- c(100, 200)
  int <- c(NA_real_, NA_real_)
  expect_null(.strip_na_peaks(mz, int))
})

test_that(".strip_na_peaks empty vectors returns unchanged", {
  res <- .strip_na_peaks(numeric(0), numeric(0))
  expect_equal(res$mz, numeric(0))
  expect_equal(res$int, numeric(0))
  expect_equal(res$n, 0L)
})

## ── .mz_boundaries ──────────────────────────────────────────────────────────

test_that(".mz_boundaries returns correct lo/hi indices", {
  mz <- c(100, 200, 300, 400, 500)

  ## Single range enclosing all
  b <- .mz_boundaries(100, 500, mz)
  expect_equal(b$lo, 1L)
  expect_equal(b$hi, 5L)

  ## Range in the middle
  b <- .mz_boundaries(200, 400, mz)
  expect_equal(b$lo, 2L)
  expect_equal(b$hi, 4L)

  ## Range excluding all (below)
  b <- .mz_boundaries(10, 50, mz)
  expect_true(b$lo > b$hi) # empty range

  ## Multiple ranges at once
  b <- .mz_boundaries(c(100, 300), c(200, 500), mz)
  expect_equal(b$lo, c(1L, 3L))
  expect_equal(b$hi, c(2L, 5L))
})

test_that(".rt_keep finite bounds returns correct indices", {
  rt <- c(1, 2, 3, 4, 5)
  res <- .rt_keep(rt, 2, 4, 5L)
  expect_equal(res, 2:4)
})

test_that(".rt_keep infinite bounds returns all", {
  rt <- c(1, 2, 3)
  res <- .rt_keep(rt, -Inf, Inf, 3L)
  expect_equal(res, 1:3)
})

test_that(".rt_keep narrow window returns subset", {
  rt <- c(1.0, 2.0, 3.0, 4.0)
  res <- .rt_keep(rt, 2.5, 3.5, 4L)
  expect_equal(res, 3L)
})

test_that(".rt_keep no match returns empty", {
  rt <- c(1, 2, 3)
  res <- .rt_keep(rt, 10, 20, 3L)
  expect_equal(res, integer(0))
})

## ── .make_chrom_entry ────────────────────────────────────────────────────────

test_that(".make_chrom_entry returns data.frame with both columns", {
  res <- .make_chrom_entry(
    rtime = c(1, 2, 3), intensity = c(10, 20, 30),
    do_rt = TRUE, do_int = TRUE, columns = c("rtime", "intensity"),
    drop = FALSE
  )
  expect_s3_class(res, "data.frame")
  expect_equal(res$rtime, c(1, 2, 3))
  expect_equal(res$intensity, c(10, 20, 30))
  expect_equal(nrow(res), 3L)
})

test_that(".make_chrom_entry rtime-only", {
  res <- .make_chrom_entry(
    rtime = c(1, 2), intensity = numeric(0),
    do_rt = TRUE, do_int = FALSE, columns = "rtime", drop = FALSE
  )
  expect_s3_class(res, "data.frame")
  expect_equal(names(res), "rtime")
})

test_that(".make_chrom_entry drop=TRUE single column returns vector", {
  res <- .make_chrom_entry(
    rtime = numeric(0), intensity = c(10, 20, 30),
    do_rt = FALSE, do_int = TRUE, columns = "intensity", drop = TRUE
  )
  expect_type(res, "double")
  expect_equal(res, c(10, 20, 30))
})

test_that(".make_chrom_entry drop=TRUE two columns returns data.frame", {
  res <- .make_chrom_entry(
    rtime = c(1, 2), intensity = c(10, 20),
    do_rt = TRUE, do_int = TRUE, columns = c("rtime", "intensity"),
    drop = TRUE
  )
  expect_s3_class(res, "data.frame")
})

## ── .spectra_format_chromData ────────────────────────────────────────────────

test_that(".spectra_format_chromData returns correct structure", {
  sp <- Spectra(DataFrame(
    mz = NumericList(c(100, 200), c(100, 200), compress = FALSE),
    intensity = NumericList(c(10, 20), c(15, 25), compress = FALSE),
    rtime = c(1.5, 3.0),
    msLevel = rep(1L, 2),
    dataOrigin = rep("fileA", 2),
    chromSpectraIndex = rep("grp1", 2)
  ))
  res <- .spectra_format_chromData(sp)
  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 1L)
  expect_equal(res$msLevel, 1L)
  expect_equal(res$rtMin, 1.5)
  expect_equal(res$rtMax, 3.0)
  expect_equal(res$mzMin, -Inf)
  expect_equal(res$mzMax, Inf)
  expect_equal(res$dataOrigin, "fileA")
  expect_equal(res$chromSpectraIndex, "grp1")
})

test_that(".peaksData errors for wrong object type", {
  expect_error(.peaksData("not_a_backend"),
               "must be of class 'Chromatograms' or 'ChromBackend'")
})

test_that(".chromData errors for wrong object type", {
  expect_error(.chromData(42),
               "must be of class 'Chromatograms' or 'ChromBackend'")
})

## ── .compare_chrom_pair ─────────────────────────────────────────────────────

test_that(".compare_chrom_pair returns c(score, n) for identical chromatograms", {
    pd <- data.frame(rtime = c(1, 2, 3, 4), intensity = c(10, 50, 100, 50))
    res <- .compare_chrom_pair(pd, pd)
    expect_equal(length(res), 2L)
    expect_equal(res[[1L]], 1)
    expect_equal(res[[2L]], 4L)  ## all 4 RT points overlap
})

test_that(".compare_chrom_pair returns c(NA, 0) for non-overlapping RT", {
    pd_a <- data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
    pd_b <- data.frame(rtime = c(10, 11, 12), intensity = c(40, 50, 60))
    res <- .compare_chrom_pair(pd_a, pd_b)
    expect_true(is.na(res[[1L]]))
    expect_equal(res[[2L]], 0L)
})

test_that(".compare_chrom_pair returns c(NA, 0) when fewer than 2 points", {
    pd_short <- data.frame(rtime = 1, intensity = 50)
    pd_ok <- data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
    res1 <- .compare_chrom_pair(pd_short, pd_ok)
    expect_true(is.na(res1[[1L]]))
    expect_equal(res1[[2L]], 0L)
    res2 <- .compare_chrom_pair(pd_ok, pd_short)
    expect_true(is.na(res2[[1L]]))
    expect_equal(res2[[2L]], 0L)
})

test_that(".compare_chrom_pair uses custom FUN when provided", {
    pd_a <- data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
    pd_b <- data.frame(rtime = c(1, 2, 3), intensity = c(5, 15, 25))
    cosine <- function(x, y, ...) {
        sum(x * y) / (sqrt(sum(x^2)) * sqrt(sum(y^2)))
    }
    res <- .compare_chrom_pair(pd_a, pd_b, FUN = cosine, minPeaks = 3L)
    expected <- cosine(pd_a$intensity, pd_b$intensity)
    expect_equal(res[[1L]], expected, tolerance = 1e-10)
    expect_equal(res[[2L]], 3L)
})

test_that(".compare_chrom_pair interpolates onto common RT grid", {
    ## Different RT grids, but overlapping range (RT 2-4 → 3 common points)
    pd_a <- data.frame(rtime = c(1, 2, 3, 4), intensity = c(10, 20, 30, 40))
    pd_b <- data.frame(rtime = c(2, 3, 4, 5), intensity = c(20, 30, 40, 50))
    res <- .compare_chrom_pair(pd_a, pd_b, minPeaks = 3L)
    expect_true(is.numeric(res))
    expect_true(!is.na(res[[1L]]))
    ## Overlap is RT 2-4, perfectly correlated linear data → correlation = 1
    expect_equal(res[[1L]], 1, tolerance = 1e-10)
    expect_equal(res[[2L]], 3L)
})

test_that(".compare_chrom_pair respects method argument", {
    pd_a <- data.frame(rtime = c(1, 2, 3, 4, 5),
                       intensity = c(10, 20, 30, 20, 10))
    pd_b <- data.frame(rtime = c(1, 2, 3, 4, 5),
                       intensity = c(5, 15, 25, 15, 5))
    res_p <- .compare_chrom_pair(pd_a, pd_b, fun_args = list(method = "pearson"))
    res_s <- .compare_chrom_pair(pd_a, pd_b, fun_args = list(method = "spearman"))
    expect_true(is.numeric(res_p))
    expect_true(is.numeric(res_s))
    ## Both should be 1 for perfectly linearly related data
    expect_equal(res_p[[1L]], 1, tolerance = 1e-10)
    expect_equal(res_s[[1L]], 1, tolerance = 1e-10)
})

## ── .compare_chromatograms ──────────────────────────────────────────────────

test_that(".compare_chromatograms self-comparison returns symmetric array", {
    pd <- list(
        data.frame(rtime = c(1, 2, 3, 4), intensity = c(10, 50, 100, 50)),
        data.frame(rtime = c(1, 2, 3, 4), intensity = c(5, 25, 50, 25)),
        data.frame(rtime = c(10, 11, 12), intensity = c(40, 50, 60))
    )
    res <- .compare_chromatograms(pd, minPeaks = 3L)
    expect_true(is.array(res))
    expect_equal(dim(res), c(3L, 3L, 2L))
    expect_equal(dimnames(res)[[3L]], c("score", "n_peaks"))
    expect_equal(diag(res[, , 1L]), c(1, 1, 1))
    expect_equal(res[1, 2, 1L], res[2, 1, 1L])
    expect_equal(res[1, 3, 1L], res[3, 1, 1L])
    ## Chromatograms 1 & 2 overlap and are correlated
    expect_true(!is.na(res[1, 2, 1L]))
    expect_true(res[1, 2, 2L] > 0)
    ## Chromatograms 1/2 vs 3 don't overlap
    expect_true(is.na(res[1, 3, 1L]))
    expect_equal(unname(res[1, 3, 2L]), 0)
})

test_that(".compare_chromatograms cross-comparison returns n x m x 2 array", {
    pd_x <- list(
        data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30)),
        data.frame(rtime = c(1, 2, 3), intensity = c(5, 15, 25))
    )
    pd_y <- list(
        data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
    )
    res <- .compare_chromatograms(pd_x, pd_y, minPeaks = 3L)
    expect_true(is.array(res))
    expect_equal(dim(res), c(2L, 1L, 2L))
    ## First pair is identical
    expect_equal(unname(res[1, 1, 1L]), 1, tolerance = 1e-10)
    expect_equal(unname(res[1, 1, 2L]), 3)
    ## Second pair is linearly related → correlation = 1
    expect_equal(unname(res[2, 1, 1L]), 1, tolerance = 1e-10)
})

test_that(".compare_chromatograms empty lists return 0-dim array", {
    expect_equal(dim(.compare_chromatograms(list())), c(0L, 0L, 2L))
    pd <- list(data.frame(rtime = 1:3, intensity = c(10, 20, 30)))
    res <- .compare_chromatograms(list(), pd)
    expect_equal(dim(res)[1L], 0L)
    res2 <- .compare_chromatograms(pd, list())
    expect_equal(dim(res2)[2L], 0L)
})

test_that(".compare_chromatograms applies labels for self-comparison", {
    pd <- list(
        data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30)),
        data.frame(rtime = c(1, 2, 3), intensity = c(5, 15, 25))
    )
    res <- .compare_chromatograms(pd, labelsColumn = c("a", "b"))
    expect_equal(dimnames(res)[[1L]], c("a", "b"))
    expect_equal(dimnames(res)[[2L]], c("a", "b"))
})

test_that(".compare_chromatograms with custom FUN", {
    cosine <- function(x, y, ...) {
        sum(x * y) / (sqrt(sum(x^2)) * sqrt(sum(y^2)))
    }
    pd <- list(
        data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30)),
        data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30))
    )
    res <- .compare_chromatograms(pd, FUN = cosine, minPeaks = 3L)
    expect_equal(unname(res[1, 2, 1L]), 1, tolerance = 1e-10)
    expect_equal(unname(res[2, 1, 1L]), 1, tolerance = 1e-10)
})

test_that(".compare_chromatograms single element returns 1x1x2 with score = 1", {
    pd <- list(data.frame(rtime = c(1, 2, 3), intensity = c(10, 20, 30)))
    res <- .compare_chromatograms(pd, minPeaks = 3L)
    expect_equal(dim(res), c(1L, 1L, 2L))
    expect_equal(unname(res[1, 1, 1L]), 1)
    expect_equal(unname(res[1, 1, 2L]), 3)
})

## ── .resolve_labels ─────────────────────────────────────────────────────────

test_that(".resolve_labels returns NULL when labels is NULL", {
    expect_null(.resolve_labels(c_full, NULL))
})

test_that(".resolve_labels returns values from a valid column", {
    res <- .resolve_labels(c_full, "mz")
    expect_equal(res, chromData(c_full)[["mz"]])
})

test_that(".resolve_labels errors on non-existent column", {
    expect_error(.resolve_labels(c_full, "nonexistent"), "not found")
})

test_that(".resolve_labels errors on duplicated column values", {
    c_dup <- c_full
    c_dup$msLevel <- rep(1L, length(c_dup))
    expect_error(.resolve_labels(c_dup, "msLevel"), "duplicated")
})

test_that(".resolve_labels errors on non-character or length != 1", {
    expect_error(.resolve_labels(c_full, 42), "single character string")
    expect_error(.resolve_labels(c_full, c("mz", "msLevel")),
                 "single character string")
})

#' @include helpers.R
#' @include hidden_aliases.R
#' @include ChromBackend.R
NULL

#' @title Improved in-memory Chromatographic data backend
#'
#' @name ChromBackendMemory
#'
#'
#' @description
#' `ChromBackendMemory`: This backend stores chromatographic data directly
#' in memory, making it ideal for small datasets or testing. It can be
#' initialized with a `data.frame` of chromatographic data via the `chromData`
#' parameter and a `list` of `data.frame` entries for peaks data using the
#' `peaksData` parameter. These data can be accessed with the `chromData()` and
#' `peaksData()` functions.
#'
#' @param chromData For `backendInitialize()` of a `ChromBackendMemory`
#'        backend, a `data.frame` with the chromatographic data. If not
#'        provided (or if empty), a default `data.frame` with the core
#'        chromatographicvariables will be created.
#'
#' @param object  A `ChromBackendMemory` object.
#'
#' @param peaksData For `backendInitialize()` of a `ChromBackendMemory`
#'        backend, a `list` of `data.frame` with the peaks data. If not
#'        provided (or if empty), a default `list` of empty `data.frame` with
#'        the core peaks variables will be created. The length of the list
#'        should match the number of chromatograms in the `chromData`
#'        parameter.
#'
#' @param ... Additional parameters to be passed.
#'
#' @author Philippine Louail
#'
#' @exportClass ChromBackendMemory
#'
#' @return Refer to the individual function description for information on the
#'         return value.
#'
#' @examples
#'
#' ## Method 1: Initialize backend directly
#' cdata <- data.frame(
#'     msLevel = c(1L, 1L, 1L),
#'     mz = c(112.2, 123.3, 134.4),
#'     dataOrigin = c("mem1", "mem2", "mem3")
#' )
#'
#' pdata <- list(
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     ),
#'     data.frame(
#'         rtime = c(45.1, 46.2),
#'         intensity = c(100, 80.1)
#'     ),
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     )
#' )
#'
#' cbm <- ChromBackendMemory()
#' cbm <- backendInitialize(cbm, chromData = cdata, peaksData = pdata)
#' cbm
#'
#' ## Method 2: Use Chromatograms constructor (recommended)
#' chr <- Chromatograms(ChromBackendMemory(), chromData = cdata, peaksData = pdata)
#' chr
#'
NULL

#' @noRd
setClass("ChromBackendMemory",
    contains = "ChromBackend",
    slots = c(
        chromData = "data.frame",
        peaksData = "list"
    ),
    prototype = prototype(
        chromData = fillCoreChromVariables(data.frame()),
        peaksData = list(.EMPTY_PEAKS_DATA),
        readonly = FALSE,
        version = "0.1"
    )
)

#' @rdname ChromBackendMemory
#'
#' @importFrom methods new
#' @export ChromBackendMemory
ChromBackendMemory <- function() {
    new("ChromBackendMemory")
}

#' @rdname ChromBackendMemory
setMethod(
    "backendInitialize", "ChromBackendMemory",
    function(object,
    chromData = fillCoreChromVariables(data.frame()),
    peaksData = list(.EMPTY_PEAKS_DATA), ...) {
        if (!is(chromData, "data.frame")) {
            stop(
                "'chromData' needs to be a 'data.frame' with the general",
                " chromatogram variables"
            )
        }
        n_cd <- nrow(chromData)
        if (n_cd) {
            if (is.null(chromData$dataOrigin)) {
                chromData$dataOrigin <- NA_character_
            }
            validChromData(chromData)
            chromData <- chromData[, !vapply(chromData, function(x) {
                all(is.na(x))
            }, logical(1)), drop = FALSE]
        } else {
            chromData <- fillCoreChromVariables(data.frame())
        }
        object@chromData <- chromData
        if (length(peaksData) > 1 || length(peaksData) == n_cd) {
            validPeaksData(peaksData)
        } else {
            peaksData <- rep(list(.EMPTY_PEAKS_DATA), n_cd)
        }
        object@peaksData <- peaksData
        object
    }
)

#' @rdname hidden_aliases
setMethod("backendMerge", "ChromBackendMemory", function(object, ...) {
    object <- unname(c(object, ...))
    not_empty <- lengths(object) > 0
    if (any(not_empty)) {
        res <- .df_combine(object[not_empty])
    } else {
        res <- object[[1L]]
    }
    validChromData(.chromData(res))
    validPeaksData(.peaksData(res))
    res
})

#' @rdname hidden_aliases
#' @description This method returns the chromatographic data stored in the
#' backend. If not specified otherwise it will return all defined columns in
#' the chromData slot as well as adding the `coreChromVariables` missing with
#' NA values.
setMethod(
    "chromData", "ChromBackendMemory",
    function(object, columns = chromVariables(object), drop = FALSE) {
        if (!all(columns %in% chromVariables(object))) {
            stop(
                "Some of the requested Chromatogram variables are not ",
                "available"
            )
        }
        res <- fillCoreChromVariables(.chromData(object))
        res <- res[, columns, drop = drop]
        res
    }
)

#' @rdname hidden_aliases
setReplaceMethod("chromData", "ChromBackendMemory", function(object, value) {
    if (!inherits(value, "data.frame")) {
        stop("'value' is expected to be a 'data.frame'")
    }
    if (length(object) && length(object) != nrow(value)) {
        stop("'value' has to be a 'data.frame' with ", length(object), " rows")
    }
    validChromData(value)
    object@chromData <- value
    object
})

#' @rdname hidden_aliases
setMethod("chromVariables", "ChromBackendMemory", function(object) {
    union(names(.chromData(object)), names(coreChromVariables()))
})

#' @rdname hidden_aliases
setMethod(
    "peaksData", "ChromBackendMemory",
    function(object, columns = peaksVariables(object),
    drop = FALSE, ...) {
        pv <- peaksVariables(object)
        if (!all(columns %in% pv)) {
            stop(
                "Some of the requested peaks variables are not",
                " available"
            )
        }
        if (identical(as.vector(pv), as.vector(columns))) {
            return(.peaksData(object))
        }
        if (drop && length(columns) == 1L) {
            return(lapply(.peaksData(object), function(z) z[[columns]]))
        }
        lapply(.peaksData(object), function(x) x[, columns, drop = drop])
    }
)

#' @rdname hidden_aliases
setReplaceMethod("peaksData", "ChromBackendMemory", function(object, value) {
    if (!is.list(value)) {
        stop("'value' is expected to be a list")
    }
    if (length(object) && length(object) != length(value)) {
        stop("'value' has to be a list with ", length(object), " elements")
    }
    validPeaksData(value)
    object@peaksData <- value
    object
})

#' @rdname hidden_aliases
setMethod("peaksVariables", "ChromBackendMemory", function(object) {
    union(names(.peaksData(object)[[1]]), names(corePeaksVariables()))
})

#' @rdname hidden_aliases
#' @importMethodsFrom ProtGenerics backendParallelFactor
setMethod("backendParallelFactor", "ChromBackendMemory",
          function(object, ...) {
    do <- dataOrigin(object)
    if (length(unique(do[!is.na(do)])) > 1L)
        factor(do, levels = unique(do))
    else factor()
})

#' @rdname hidden_aliases
#' @export
setMethod("isReadOnly", "ChromBackendMemory", function(object) FALSE)

#' @rdname hidden_aliases
setMethod("lengths", "ChromBackendMemory", function(x) {
    vapply(.peaksData(x), nrow, integer(1L))
})

#' Optimized intensity accessor for ChromBackendMemory.
#' Bypasses peaksData() dispatch for direct slot access.
#' @rdname hidden_aliases
setMethod("intensity", "ChromBackendMemory", function(object) {
    if (!length(object)) return(list())
    lapply(.peaksData(object), function(z) z$intensity)
})

#' Optimized rtime accessor for ChromBackendMemory.
#' Bypasses peaksData() dispatch for direct slot access.
#' @rdname hidden_aliases
setMethod("rtime", "ChromBackendMemory", function(object) {
    if (!length(object)) return(list())
    lapply(.peaksData(object), function(z) z$rtime)
})

#' @importFrom utils capture.output head
#' @rdname hidden_aliases
setMethod("show", "ChromBackendMemory", function(object) {
    cpd <- .chromData(object)
    cat(class(object), "with", nrow(cpd), "chromatograms\n")
    if (nrow(cpd)) {
        cpd <- fillCoreChromVariables(cpd)
        cpd <- cpd[, c("chromIndex", "msLevel", "mz"), drop = FALSE]
        txt <- capture.output(print(head(cpd)))
        cat(txt, sep = "\n")
        cp_cols <- .chromData(object)[
            ,
            !(colnames(.chromData(object)) %in%
                colnames(cpd))
        ]
        cat("...", length(cp_cols), "more  chromatogram variables/columns\n")
        cat("...", ncol(.peaksData(object)[[1]]), "peaksData variables\n")
    }
})

#' @rdname hidden_aliases
#' @export
setMethod(
    "supportsSetBackend", "ChromBackendMemory",
    function(object, ...) TRUE
)

#' @importFrom MsCoreUtils i2index
#' @importMethodsFrom S4Vectors [ [<-
#' @rdname hidden_aliases
setMethod("[", "ChromBackendMemory", function(x, i, j, ..., drop = FALSE) {
    if (!length(i)) {
        return(ChromBackendMemory())
    }
    i <- i2index(i, length = length(x))
    x@chromData <- .chromData(x)[i, ]
    x@peaksData <- .peaksData(x)[i]
    x
})

#' @rdname hidden_aliases
setMethod("$", "ChromBackendMemory", function(x, name) {
    if (name %in% union(chromVariables(x), names(coreChromVariables()))) {
        res <- chromData(x, columns = name, drop = TRUE)
    } else if (name %in% peaksVariables(x)) {
        res <- peaksData(x, columns = name, drop = TRUE)
    } else {
        stop("The requested variable '", name, "' is not available")
    }
    res
})

#' @rdname hidden_aliases
setReplaceMethod("$", "ChromBackendMemory", function(x, name, value) {
    if (length(x) && length(value) != length(x)) {
        stop(
            "length of 'value' needs to match the number of chromatograms ",
            "in object."
        )
    }
    if (name %in% peaksVariables(x)) {
        if (!is.list(value)) {
            stop("The value for peaksData should be a list")
        }
        for (i in seq_along(value)) peaksData(x)[[i]][[name]] <- value[[i]]
    } else {
        chromData(x)[, name] <- value
    }
    x
})

#' @rdname hidden_aliases
setMethod("chromExtract", "ChromBackendMemory", function(object, peak.table, by) {
    required_cols <- c("rtMin", "rtMax", by)
    .validate_chromExtract_input(
        object = object,
        peak.table = peak.table,
        by = by, required_cols = required_cols
    )

    matched <- .match_chromdata_peaktable(
        object = object,
        peak.table = peak.table,
        by = by
    )
    object <- matched$object
    chrom_keys <- matched$chrom_keys
    peak_keys  <- matched$peak_keys
    obj_sp <- split(object, chrom_keys) ##  UT need to check that
    pk_split <- split(peak.table, peak_keys)

    overl_cols <- .check_overl_columns(peak.table = peak.table,
                                        object = object,
                                        required_cols = required_cols)
    new_data <- mapply(function(obj, pks) { ## could switch to bpmapply ?
        d <- .chromData(obj)
        d <- suppressWarnings(cbind(d, pks[!overl_cols]))
        d[, names(peak.table)[overl_cols]] <- pks[, overl_cols]
        p <- vector("list", nrow(d))
        for (z in seq(nrow(d))) {
            rt <- rtime(obj)[[1]]
            inrt <- rt >= d$rtMin[z] & rt <= d$rtMax[z]
            p[[z]] <- peaksData(obj)[[1]][inrt, , drop = FALSE]
        }
        list(cd = d, pd = p)
    }, obj = obj_sp,
    pks = pk_split, SIMPLIFY = FALSE)

    new_cdata <- .fast_rbind(lapply(new_data, function(z) z$cd))
    rownames(new_cdata) <- NULL
    new_pdata <- unlist(lapply(new_data, function(z) z$pd), recursive = FALSE)
    backendInitialize(new("ChromBackendMemory"),
                      chromData = new_cdata, peaksData = new_pdata)

})

#' @include ChromBackend.R ChromBackendMemory.R
NULL

#' @title The Chromatograms class to manage and access chromatographic data
#'
#' @name Chromatograms
#'
#' @aliases Chromatograms-class Chromatograms
#' @aliases [,Chromatograms-method
#' @aliases [<-,Chromatograms-method
#' @aliases [[,Chromatograms-method
#' @aliases [[<-,Chromatograms-method
#' @aliases chromExtract
#' @aliases filterEmptyChromatograms
#'
#' @description
#' The `Chromatograms` class encapsules chromatographic data and related
#' metadata. The chromatographic data is represented by a *backend* extending
#' the virtual [ChromBackend] class which provides the raw data to the
#' `Chromatograms` object. Different backends and their properties are
#' described in the [ChromBackend] class documentation.
#'
#' **Available Backends**: The package provides several backends:
#' - `ChromBackendMemory`: Stores data in memory (default, ideal for small datasets).
#' - `ChromBackendMzR`: Reads peaks data from raw MS files on demand.
#' - `ChromBackendSpectra`: Generates chromatographic data from a `Spectra` object.
#'   This backend supports both in-memory and file-backed `Spectra` objects, using
#'   an internal `spectraSortIndex` to avoid physically reordering the spectra.
#'
#' @section Creation of objects:
#'
#' `Chromatograms` objects can be created using the `Chromatograms()`
#' construction function. Either by providing a `ChromBackend` object or by
#' providing a `Spectra` object. The `Spectra` object will be used to generate
#' a `Chromatograms` object with a backend of class [`ChromBackendSpectra`].
#'
#' @section Data stored in a `Chromatograms` object:
#'
#' The `Chromatograms` object is a container for chromatographic data, which
#' includes peaks data (*retention time* and related intensity values, also
#' referred to as *peaks data variables* in the context of `Chromatograms`) and
#' metadata of individual chromatogram (so called *chromatograms variables*).
#' While a core set of chromatograms variables (the
#' `coreChromatogramsVariables()`) and peaks data variables (the
#' `corePeaksVariables()`) are guaranteed to be provided by a `Chromatograms`,
#' it is possible to add arbitrary variables to a `Chromatograms` object.
#'
#' The `Chromatograms` object is designed to contain chromatographic data of a
#' (large) set of chromatograms. The data is organized *linearly* and can be
#' thought of a list of chromatograms, i.e. each element in the `Chromatograms`
#' is one chromatogram.
#'
#' The *chromatograms variables* information in the `Chromatograms` object can
#' be accessed using the `chromData()` function. Specific chromatograms
#' variables can be accessed by either precising the `"columns"` parameter in
#' `chromData()` or using `$`. `@chromData` can be accessed, replaced but
#' also filtered/subsetted. Refer to the [chromData] documentation for more
#' details.
#'
#' The *peaks data variables* information in the `Chromatograms` object can be
#' accessed using the `peaksData()` function. Specific peaks variables can be
#' accessed by either precising the `"columns"` parameter in `peaksData()` or
#' using `$`. `@peaksData` can be accessed, replaced but also
#' filtered/subsetted. Refer to the [peaksData] documentation for more details.
#'
#' @section Processing of `Chromatograms` objects:
#'
#' Functions that process the chromatograms data in some ways can be applied to
#' the object either directly or by using the `processingQueue` mechanism. The
#' `@processingQueue` is a list of processing steps that are stored within the
#' object and only applied when needed. This was created so that the data can be
#' processed in a single step and is very useful for larger datasets. This is
#' even more true as this processing queue will call function that can be
#' applied on the data in a chunk-wise manner. This allows for parallel
#' processing of the data and reduces the memory demand. To read more about the
#' `processingQueue`, and how to parallelize your processes, see the
#' [processingQueue] documentation.
#'
#' @section Subsetting and accessing data:
#' The `Chromatograms` class supports subsetting by chromatogram (i.e. rows) using
#' the `[` operator. The `[` operator does not support subsetting by columns.
#' Specific chromatograms or peaks variables can be accessed using the `[[`
#' operator or the `$` operator. The `[[` operator can also be used to
#' replace specific chromatograms or peaks variables.
#'
#' @section Changing the backend:
#' The `setBackend()` function can be used to change the backend of a
#' `Chromatograms` object. This can be useful to switch to a backend that
#' better suits the needs of the user, for example switching to a memory-based
#' backend for smaller datasets or to a file-based backend for larger datasets.
#' The `setBackend()` function supports parallelization of the backend
#' conversion using the `BPPARAM` parameter. Note that any queued processing
#' steps are applied during the backend switch (since `peaksData()` is called
#' to transfer the data) and the processing queue is emptied afterwards.
#'
#' @section Filtering chromatograms:
#'
#' - `filterEmptyChromatograms()`: removes empty chromatograms (i.e.
#'   chromatograms without peaks) from the object. Returns the filtered
#'   `Chromatograms` object (with chromatograms in their original order).
#'
#' @section Extracting chromatograms based on a peak table:
#'
#' The `chromExtract()` function allows users to extract specific regions of
#' interest from a `Chromatograms` object based on a user-provided peak table.
#' Each row in the `peak.table` defines a region to extract, using minimum and
#' maximum retention time (and m/z in the case of `chromBackendSpectra`)
#' boundaries, and identifiers that uniquely match chromatograms in the object.
#'
#' The resulting **new** `Chromatograms` object contains only chromatograms
#' overlapping the specified regions, with updated metadata reflecting the
#' extracted boundaries.
#'
#' This function is most commonly used to subset chromatographic data around
#' detected peaks or predefined time/mass ranges, for example to reprocess,
#' visualize, or quantify extracted chromatograms corresponding to known
#' features. It's important to notes that filtering by m/z is only supported
#' when using a `ChromBackendSpectra` backend. if the `mzMin` and `mzMax`
#' columns are provided when using other backends, they will be ignored.
#'
#'
#' @param BPPARAM Parallel setup configuration. See [BiocParallel::bpparam()]
#'        for more information.
#'
#' @param backend [ChromBackend] object providing the raw data for the
#'        `Chromatograms` object.
#'
#' @param by A `character` vector naming one or more columns that uniquely
#'        identify chromatograms in both `peak.table` and
#'        `chromData(object)`. The combination of these columns must be unique
#'        within `chromData(object)`. Typically includes `"dataOrigin"`,
#'        `"msLevel"`, or both.
#'
#' @param chromData For `Chromatograms()` build from a `Spectra` object backend,
#'        a `data.frame` with the chromatographic data. If not provided
#'        (or if empty), a default `data.frame` with the core chromatographic
#'        variables will be created.
#'
#' @param drop For `[`: `logical(1)` default to `FALSE`.
#'
#' @param f `factor` defining the grouping to split the `Chromatograms` object.
#'
#' @param factorize.by A `character` vector with the names of the variables in
#'        the `Spectra` object and the `chromData` slot that should be used
#'        to factorize the `Spectra` object data to generate the
#'        chromatographic data.
#'
#' @param i For `[`: `integer`, `logical` or `character` to subset the object.
#'
#' @param j For `[` and `[[`: ignored.
#'
#' @param name A `character` string specifying the name of the variable to
#'        access.
#'
#' @param object A [Chromatograms] object.
#'
#' @param peak.table For `chromExtract()` A `data frame` containing the
#'        following minimum columns:
#'          - rtMin: Minimum retention time for each peak. Cannot be NA.
#'          - rtMax: Maximum retention time for each peak. Cannot be NA.
#'          - mzMin: Minimum m/z value for each peak.
#'          - mzMax: Maximum m/z value for each peak.
#'        Additionally, the `peak.table` must include columns that uniquely
#'        identify chromatograms in the `object`. Common choices are
#'        "msLevel" and/or "dataOrigin". These columns must also be present
#'        in the `chromData` of the `object`. Any extra columns in
#'        `peak.table` will be added to the `chromData` of the newly created
#'        object.
#'
#' @param processingQueue [list] a list of processing steps (i.e. functions) to
#'        be applied to the chromatographic data. The processing steps are
#'        applied in the order they are listed in the `processingQueue`.
#'
#' @param summarize.method For `Chromatograms` created with a `Spectra` object:
#'        A `character` vector with the name of the function to be used to
#'        summaries the spectra data intensity. The available methods are "sum"
#'        and "max". The default is "sum".
#'
#' @param spectraVariables A `character` vector specifying which variables
#'        from the `Spectra` object should be added to the chromData. These
#'        will be mapped using the `chromSpectraIndex` variable.
#'
#' @param value The value to replace the variable with.
#'
#' @param x A [Chromatograms] object.
#'
#' @param ... Additional arguments.
#'
#' @md
#'
#' @exportClass Chromatograms
#'
#' @return Refer to the individual function description for information on
#' the return value.
#'
#' @seealso [chromData] for a general description of the chromatographic
#'          metadata available in the object, as well as how to access, replace
#'          and subset them.
#'          [peaksData] for a general description of the chromatographic peaks
#'          data available in the object, as well as how to access, replace and
#'          subset them.
#'          [processingQueue] for more information on the queuing of
#'          processings and parallelization for larger dataset.
#'
#' @examples
#'
#' ## Create a Chromatograms object with ChromBackendMemory
#' cdata <- data.frame(
#'     msLevel = c(1L, 1L, 1L),
#'     mz = c(112.2, 123.3, 134.4),
#'     dataOrigin = c("mem1", "mem2", "mem3")
#' )
#' pdata <- list(
#'     data.frame(rtime = c(2.1, 2.5, 3.0, 3.4, 3.9),
#'                intensity = c(100, 250, 400, 300, 150)),
#'     data.frame(rtime = c(3.5, 4.0, 4.5),
#'                intensity = c(80, 120, 90)),
#'     data.frame(rtime = c(5.1, 5.8, 6.3, 6.9, 7.5),
#'                intensity = c(80, 500, 1200, 600, 120))
#' )
#' chr <- Chromatograms(ChromBackendMemory(), chromData = cdata, peaksData = pdata)
#' chr
#'
#' ## Create a Chromatograms object from a Spectra object
#' library(MsBackendMetaboLights)
#' library(Spectra)
#'
#' be <- backendInitialize(MsBackendMetaboLights(),
#'     mtblsId = "MTBLS39",
#'     filePattern = c("63B.cdf")
#' )
#' s <- Spectra(be)
#' s <- setBackend(s, MsBackendMemory())
#' chr <- Chromatograms(s)
#'
#' ## Subset
#' chr[1:2]
#'
#' ## Access a specific variable
#' chr[["msLevel"]]
#' chr$msLevel
#'
#' ## Replace data of a specific variable
#' chr$msLevel <- c(2L, 2L, 2L)
#'
#' ## Re-factorize the data
#' chr <- factorize(chr)
#'
#' ## Change the backend to memory
#' chr <- setBackend(chr, ChromBackendMemory())
#'
NULL

setClassUnion("ChromBackendOrMissing", c("ChromBackend", "missing"))

#' The `Chromatograms` class
#'
#' The `Chromatograms` class is a container for chromatographic data.
#'
#' @slot backend [ChromBackend] the *backend* object providing the raw data for
#'       the `Chromatograms` object.
#'
#' @slot processingQueue `list` a list of processing steps to be applied to the
#'       `Chromatograms`. Each element in the list is a function that
#'       processes the data. The processing steps are applied in
#'       the order they are listed in the `processingQueue`.
#'
#' @slot processing `character` a character vector with the names of the
#'       processing steps that have been applied to the `Chromatograms` object.
#'       This is mainly used in the "show" method to display the processing
#'       steps that have been applied to the `Chromatograms` object.
#'
#' @slot processingChunkSize `numeric(1)` the number of chromatograms to be
#'       processed in a single chunk. This is useful for processing large
#'       data sets in smaller chunks to avoid memory issues.
#'
#' @slot version `character(1)` the version of the `Chromatograms` object.
#'
#' @noRd
setClass("Chromatograms",
    slots = c(
        backend = "ChromBackend",
        processingQueue = "list",
        processing = "character",
        processingChunkSize = "numeric",
        version = "character"
    ),
    prototype = prototype(
        version = "0.1",
        processingChunkSize = Inf,
        processingQueue = list(),
        processing = character()
    )
)

setValidity("Chromatograms", function(object) {
    msg <- character()
    if (!is(.backend(object), "ChromBackend")) {
        msg <- ("backend must be a ChromBackend object")
    }
    if (!is.numeric(processingChunkSize(object)) ||
        length(processingChunkSize(object)) != 1) {
        msg <- c(msg, "processingChunkSize must be a numeric value")
    }
    msg <- c(msg, .valid_processing_queue(.processingQueue(object)))
    if (length(msg)) {
        msg
    } else {
        TRUE
    }
})

#' @rdname Chromatograms
#' @export
setMethod(
    "Chromatograms", "ChromBackendOrMissing",
    function(object = ChromBackendMemory(), processingQueue = list(), ...) {
        if (missing(object))
            object <- ChromBackendMemory()
        ## Extract backend-specific parameters from ... and initialize backend
        dots <- list(...)
        if (length(dots) > 0 && length(object) == 0) {
            ## Backend is empty, initialize it with provided parameters
            object <- do.call(backendInitialize, c(list(object), dots))
            dots <- list()
        }
        do.call(new, c(list("Chromatograms", backend = object,
                             processingQueue = processingQueue), dots))
    }
)

#' @rdname Chromatograms
#' @importFrom methods new
#' @export
setMethod(
    "Chromatograms", "Spectra",
    function(object, summarize.method = c("sum", "max"),
             chromData = data.frame(),
             factorize.by = c("msLevel", "dataOrigin"),
             spectraVariables = character(), ...) {
        bd <- backendInitialize(ChromBackendSpectra(),
            spectra = object, factorize.by = factorize.by,
            chromData = chromData,
            summarize.method = summarize.method,
            spectraVariables = spectraVariables, ...)
        new("Chromatograms", backend = bd, processingQueue = list(), ...)
    }
)


#' @rdname hidden_aliases
#'
#' @param object A [Chromatograms] object.
#' @importMethodsFrom methods show
#' @importFrom utils capture.output
#'
#' @exportMethod show
setMethod(
    "show", "Chromatograms",
    function(object) {
        cat("Chromatographic data (", class(object)[1L], ") with ",
            length(.backend(object)), " chromatograms in a ",
            class(.backend(object)), " backend:\n", sep = "")
        if (length(.backend(object))) {
            txt <- capture.output(show(.backend(object)))
            cat(txt[-1], sep = "\n")
        }
        if (length(.processingQueue(object)))
            cat("Lazy evaluation queue:",
                length(.processingQueue(object)),
                "processing step(s)\n")
        lp <- length(.processing(object))
        if (lp) {
            lps <- .processing(object)
            if (lp > 3) {
                lps <- lps[seq_len(3)]
                lps <- c(lps, paste0("...", lp - 3,
                    " more processings. ",
                    "Use 'processingLog' to list all."))
            }
            cat("Processing:\n", paste(lps, collapse = "\n "), "\n")
        }
    }
)

#' @rdname Chromatograms
#'
#' @note
#' This needs to be discussed, if we want for example to be able to set a
#' a backend to `ChromBackendMzR` we need to implement backendInitialize()
#' better. = Support peaksData and chromData as arguments AND have a way to
#' write .mzml files (which we do not have for chromatographic data).
#'
#' @importMethodsFrom ProtGenerics setBackend
#'
#' @exportMethod setBackend
setMethod(
    "setBackend", c("Chromatograms", "ChromBackend"),
    function(object, backend, f = processingChunkFactor(object),
             BPPARAM = SerialParam(), ...) {
        backend_class <- class(.backend(object))
        BPPARAM <- backendBpparam(.backend(object), BPPARAM)
        BPPARAM <- backendBpparam(backend, BPPARAM)
        if (!supportsSetBackend(backend))
            stop(class(backend), " does not support 'setBackend'")
        if (!length(f) || length(levels(f)) == 1 || !length(object)) {
            bd_new <- backendInitialize(backend,
                peaksData = peaksData(object),
                chromData = chromData(object))
        } else {
            queue <- .processingQueue(object)
            bd_new <- bplapply(
                split(.backend(object), f = f),
                function(z, queue, ...) {
                    if (length(queue))
                        z <- .run_process_queue(z, queue)
                    backendInitialize(backend,
                        peaksData = peaksData(z),
                        chromData = chromData(z),
                        BPPARAM = SerialParam())
                },
                queue = queue, ...,
                BPPARAM = BPPARAM)
            bd_new <- backendMerge(bd_new)
        }
        if (any(colnames(chromData(bd_new)) %in% c("rtMin", "rtMax")))
            chromData(bd_new) <- chromData(bd_new)[,
                !colnames(chromData(bd_new)) %in% c("rtMin", "rtMax")]
        object@backend <- bd_new
        object@processingQueue <- list()
        object@processing <- .logging(
            object@processing, "Switch backend from ",
            backend_class, " to ", class(.backend(object)))
        object
    }
)

#' @rdname Chromatograms
#' @export
setMethod("$", signature = "Chromatograms", function(x, name) {
    .backend(x)[[name]]
})

#' @rdname Chromatograms
#' @export
setReplaceMethod("$", signature = "Chromatograms", function(x, name, value) {
    x@backend[[name]] <- value
    x
})

#' @rdname Chromatograms
#' @importFrom methods slot<-
#' @importFrom MsCoreUtils i2index
#' @export
setMethod("[", "Chromatograms", function(x, i, j, ..., drop = FALSE) {
    if (!missing(j))
        stop("Subsetting 'Chromatograms' by columns is not (yet) supported")
    if (missing(i))
        return(x)
    slot(x, "backend", check = FALSE) <- extractByIndex(
        .backend(x), i2index(i, length(x)))
    x
})

#' @rdname Chromatograms
#' @export
setMethod("[[", "Chromatograms", function(x, i, j, ...) {
    if (!is.character(i))
        stop("'i' is supposed to be a character defining the chromatogram or ",
             "peak variable to access.")
    if (!missing(j))
        stop("'j' is not supported.")
    if (!(i %in% peaksVariables(x)) && !(i %in% chromVariables(x)))
        stop("No variable '", i, "' available")
    else
        do.call("[[", list(.backend(x), i))
})

#' @rdname Chromatograms
#'
#' @export
setReplaceMethod("[[", "Chromatograms", function(x, i, j, ..., value) {
    if (!is.character(i))
        stop("'i' is supposed to be a character defining the chromatogram ",
             "or peak variable to replace or create.")
    if (!(i %in% peaksVariables(x)) && !(i %in% chromVariables(x)))
        stop("No variable '", i, "' available")
    if (!missing(j))
        stop("'j' is not supported.")
    x@backend <- do.call("[[<-", list(.backend(x), i = i, value = value))
    x
})

#' @rdname Chromatograms
#' @export
setMethod(
    "factorize", "Chromatograms",
    function(object, factorize.by = c("msLevel", "dataOrigin"), ...) {
        object@backend <- factorize(.backend(object), ...)
        object
    }
)

#' @rdname Chromatograms
#' @export
setMethod(
    "chromExtract", "Chromatograms",
    function(object, peak.table, by, ...) {
        new_bd <- chromExtract(.backend(object), peak.table, by, ...)
        return(Chromatograms(new_bd))
    }
)

#' @rdname Chromatograms
#' @export
setMethod(
    "filterEmptyChromatograms", "Chromatograms",
    function(object, ...) {
        if (!length(object)) return(object)
        object@backend <- extractByIndex(
            .backend(object), which(as.logical(lengths(object))))
        object@processing <- .logging(
            object@processing, "Filter: removed empty chromatograms.")
        object
    }
)


#' @title Merging, combining and splitting Chromatograms
#'
#' @name concatenateChromatograms
#'
#' @aliases c,Chromatograms-method
#' @aliases concatenateChromatograms
#' @aliases split,Chromatograms,ANY-method
#'
#' @description
#'
#' Various functions are available to combine or split data from one or more
#' `Chromatograms` objects. These are:
#'
#' - `c()` and `concatenateChromatograms()`: combines several `Chromatograms`
#'   objects into a single object. The resulting `Chromatograms` contains all
#'   data from all individual `Chromatograms`, i.e. the union of all their
#'   chromatograms variables. Concatenation will fail if the processing queue
#'   of any of the `Chromatograms` objects is not empty or if different backends
#'   are used for the `Chromatograms` objects. In such cases it is suggested to
#'   first change the backends of all `Chromatograms` to the same type of
#'   backend (using the [setBackend()] function) and to eventually (if needed)
#'   apply the processing queue using the [applyProcessing()] function.
#'
#' - `split()`: splits the `Chromatograms` object based on a provided grouping
#'   factor returning a `list` of `Chromatograms` objects.
#'
#' @param x A `Chromatograms` object.
#'
#' @param f `factor` defining the grouping to split the `Chromatograms` object.
#'
#' @param drop For `split()`: not considered.
#'
#' @param ... For `c()` and `concatenateChromatograms()`: `Chromatograms`
#'        objects or a `list` of `Chromatograms` objects.
#'
#' @return
#'
#' - `c()` and `concatenateChromatograms()`: a single `Chromatograms` object
#'   containing the data from all input objects.
#'
#' - `split()`: a `list` of `Chromatograms` objects.
#'
#' @examples
#'
#' ## Create two Chromatograms objects
#' cdata1 <- data.frame(
#'     msLevel = c(1L, 1L),
#'     mz = c(112.2, 123.3),
#'     dataOrigin = c("file1", "file1")
#' )
#' pdata1 <- list(
#'     data.frame(rtime = c(1.0, 2.0, 3.0), intensity = c(100, 200, 150)),
#'     data.frame(rtime = c(1.0, 2.0, 3.0), intensity = c(80, 120, 90))
#' )
#' chr1 <- Chromatograms(
#'     ChromBackendMemory(),
#'     chromData = cdata1,
#'     peaksData = pdata1
#' )
#'
#' cdata2 <- data.frame(
#'     msLevel = c(2L, 2L),
#'     mz = c(134.4, 145.5),
#'     dataOrigin = c("file2", "file2")
#' )
#' pdata2 <- list(
#'     data.frame(rtime = c(4.0, 5.0, 6.0), intensity = c(300, 400, 350)),
#'     data.frame(rtime = c(4.0, 5.0, 6.0), intensity = c(200, 250, 180))
#' )
#' chr2 <- Chromatograms(
#'     ChromBackendMemory(),
#'     chromData = cdata2,
#'     peaksData = pdata2
#' )
#'
#' ## Combine using c()
#' chr_combined <- c(chr1, chr2)
#' chr_combined
#'
#' ## Combine using concatenateChromatograms
#' chr_combined2 <- concatenateChromatograms(chr1, chr2)
#'
#' ## Combine a list of Chromatograms
#' chr_list <- list(chr1, chr2)
#' chr_combined3 <- concatenateChromatograms(chr_list)
#'
#' ## Split by msLevel
#' chr_split <- split(chr_combined, f = chr_combined$msLevel)
#' chr_split
#'
NULL

#' Internal function to concatenate a list of `Chromatograms` objects.
#'
#' @param x `list` of `Chromatograms` objects.
#'
#' @return `Chromatograms`.
#'
#' @author Philippine Louail, Johannes Rainer
#'
#' @noRd
.concatenate_chromatograms <- function(x) {
    cls <- vapply(x, class, character(1))
    if (any(cls != "Chromatograms"))
        stop("Can only concatenate 'Chromatograms' objects")
    x <- x[lengths(x) > 0]
    if (length(x) == 0) return(Chromatograms())
    if (length(x) == 1) return(x[[1L]])
    pqs <- lapply(x, function(z) z@processingQueue)
    if (any(lengths(pqs)))
        stop("Can not concatenate 'Chromatograms' objects with non-empty ",
             "processing queue. Consider calling 'applyProcessing' before.")
    procs <- unique(unlist(lapply(x, function(z) z@processing)))
    object <- new("Chromatograms",
        backend = backendMerge(lapply(x, function(z) z@backend)),
        processing = c(procs,
            paste0("Merged ", length(x),
                   " Chromatograms into one [", date(), "]")))
    validObject(object)
    object
}

#' @export
#'
#' @rdname concatenateChromatograms
concatenateChromatograms <- function(x, ...) {
    if (is.list(x))
        .concatenate_chromatograms(unlist(unname(c(x, list(...)))))
    else
        .concatenate_chromatograms(unlist(unname(list(x, ...))))
}

#' @rdname concatenateChromatograms
#'
#' @exportMethod c
setMethod("c", "Chromatograms", function(x, ...) {
    .concatenate_chromatograms(unname(list(unname(x), ...)))
})

#' @rdname concatenateChromatograms
setMethod("split", "Chromatograms", function(x, f, drop = FALSE, ...) {
    bcknds <- split(x@backend, f, ...)
    lapply(bcknds, function(b) {
        slot(x, "backend", check = FALSE) <- b
        x
    })
})

## ----style, echo = FALSE, results = 'asis', message=FALSE---------------------
BiocStyle::markdown()

## ----echo = FALSE, message = FALSE--------------------------------------------
library(Chromatograms)
library(BiocStyle)
register(SerialParam())

## -----------------------------------------------------------------------------
# A data.frame with chromatogram variables.
cdata <- data.frame(
    msLevel = c(1L, 1L),
    mz = c(112.2, 123.3),
    chromIndex = c(1L, 2L)
)

# Retention time and intensity values for each chromatogram.
pdata <- list(
    data.frame(
        rtime = c(11, 12.4, 12.8, 13.2, 14.6, 15.1, 16.5),
        intensity = c(50.5, 123.3, 153.6, 2354.3, 243.4, 123.4, 83.2)
    ),
    data.frame(
        rtime = c(45.1, 46.2, 53, 54.2, 55.3, 56.4, 57.5),
        intensity = c(100, 180.1, 300.45, 1400, 1200.3, 300.2, 150.1)
    )
)

# Create and initialize the backend
be <- backendInitialize(ChromBackendMemory(),
    chromData = cdata, peaksData = pdata
)

# Create Chromatograms object
chr <- Chromatograms(be)
chr

## -----------------------------------------------------------------------------
MRM_file <- system.file("proteomics", "MRM-standmix-5.mzML.gz",
    package = "msdata"
)

be <- backendInitialize(ChromBackendMzR(),
    files = MRM_file,
    BPPARAM = SerialParam()
)

chr_mzr <- Chromatograms(be)

## -----------------------------------------------------------------------------
length(chr)
length(chr_mzr)

## -----------------------------------------------------------------------------
peaksData(chr)

## -----------------------------------------------------------------------------
peaksData(chr, columns = c("rtime"), drop = TRUE)

chr$rtime

## -----------------------------------------------------------------------------
peaksData(chr) <- list(
    data.frame(
        rtime = c(1, 2, 3, 4, 5, 6, 7),
        intensity = c(1, 2, 3, 4, 5, 6, 7)
    ),
    data.frame(
        rtime = c(1, 2, 3, 4, 5, 6, 7),
        intensity = c(1, 2, 3, 4, 5, 6, 7)
    )
)

## -----------------------------------------------------------------------------
chr$rtime <- list(
    c(8, 9, 10, 11, 12, 13, 14),
    c(8, 9, 10, 11, 12, 13, 14)
)

## -----------------------------------------------------------------------------
chr_filt <- filterPeaksData(chr, variables = "rtime", ranges = c(12, 15))

length(chr_filt)

length(rtime(chr_filt))

## -----------------------------------------------------------------------------
chromData(chr)

## -----------------------------------------------------------------------------
chromData(chr, columns = c("msLevel"))

chr$chromIndex

## -----------------------------------------------------------------------------
chr$msLevel <- c(2L, 2L)

chromData(chr)

## -----------------------------------------------------------------------------
chr$extra <- c("extra1", "extra2")
chromData(chr)

## -----------------------------------------------------------------------------
chr_filt <- filterChromData(chr,
    variables = "chromIndex", ranges = c(1, 2),
    keep = TRUE
)

length(chr_filt)
length(chr)

## ----define-function----------------------------------------------------------
## Define a function that takes the backend as an input, divides the intensity
## by parameter y and returns it. Note that ... is required in
## the function's definition.
divide_intensities <- function(x, y, ...) {
    intensity(x) <- lapply(intensity(x), `/`, y)
    x
}

## Add the function to the procesing queue
chr_2 <- addProcessing(chr, divide_intensities, y = 2)
chr_2

## ----custom-processing--------------------------------------------------------
intensity(chr_2)
intensity(chr)

## ----applyProcessing----------------------------------------------------------
length(chr_2@processingQueue)
chr_2 <- applyProcessing(chr_2)

length(chr_2@processingQueue)
chr_2

## -----------------------------------------------------------------------------
processingChunkFactor(chr)

## -----------------------------------------------------------------------------
processingChunkFactor(chr_mzr) |>
  head()

## -----------------------------------------------------------------------------
processingChunkSize(chr_mzr) <- 10

processingChunkFactor(chr_mzr) |> table()

## -----------------------------------------------------------------------------
print(object.size(chr_mzr), units = "Mb")
chr_mzr <- setBackend(chr_mzr, ChromBackendMemory(), BPPARAM = SerialParam())

chr_mzr

chr_mzr@backend@peaksData[[1]] |> head() # data is now in memory

## -----------------------------------------------------------------------------
print(object.size(chr_mzr), units = "Mb")

## ----message=FALSE------------------------------------------------------------
library(Spectra)
library(IRanges)
sp <- Spectra(
  DataFrame(
    rtime = c(100, 110, 120, 130, 140),
    msLevel = c(1L, 1L, 1L, 1L, 1L),
    dataOrigin = rep("example", 5L),
    mz = NumericList(
      c(100, 101), c(100, 101), c(100, 101), c(100, 101), c(100, 101),
      compress = FALSE
    ),
    intensity = NumericList(
      c(10, 20), c(15, 25), c(30, 5), c(12, 18), c(40, 2),
      compress = FALSE
    )
  ),
  source = MsBackendDataFrame()
)

chr_s <- Chromatograms(sp)

## -----------------------------------------------------------------------------
chr_s

## -----------------------------------------------------------------------------
## Create custom metadata for EIC extraction
custom_cd <- data.frame(
    msLevel = c(1L, 1L),
    dataOrigin = rep(dataOrigin(sp)[1], 2),
    mzMin = c(100, 200),
    mzMax = c(100.5, 200.5)
)

chr_custom <- Chromatograms(sp, chromData = custom_cd)
chr_custom

## -----------------------------------------------------------------------------
## Work on a copy so downstream examples are not affected
chr_s_tmp <- chr_s

## Modify metadata
chr_s_tmp$msLevel <- rep(2L, length(chr_s_tmp))

## Re-factorize to update the groupings
chr_s_tmp <- factorize(chr_s_tmp)

chromData(chr_s_tmp)

## -----------------------------------------------------------------------------
chromData(chr_s)$rtmin <- 125
chromData(chr_s)$rtmax <- 180
chromData(chr_s)$mzmin <- 100
chromData(chr_s)$mzmax <- 100.5

## -----------------------------------------------------------------------------
library(RColorBrewer)
col3 <- brewer.pal(3, "Dark2")

plotChromatograms(chr_s, col = col3)

## -----------------------------------------------------------------------------
plotChromatogramsOverlay(chr_s, col = col3)

## -----------------------------------------------------------------------------
## Define peaks of interest with retention time windows
peak_table <- data.frame(
    rtMin = c(8, 11),
    rtMax = c(10, 13),
    msLevel = c(2L, 2L),
    chromIndex = c(1L, 2L)
)

## Extract those regions
chr_extracted <- chromExtract(chr, peak_table,
                              by = c("msLevel", "chromIndex"))

chr_extracted

## -----------------------------------------------------------------------------
chromData(chr_extracted)

## -----------------------------------------------------------------------------
## Define peak table with both retention time and m/z windows
peak_table_mz <- data.frame(
    rtMin = c(125, 125),
    rtMax = c(180, 180),
    mzMin = c(100, 140),
    mzMax = c(100.5, 140.5),
    msLevel = c(1L, 1L),
    dataOrigin = rep(dataOrigin(chr_s)[1], 2),
    featureID = c("feature_1", "feature_2")
)

## Extract EICs for these features
chr_eics <- chromExtract(chr_s, peak_table_mz,
                        by = c("msLevel", "dataOrigin"))

chr_eics

## -----------------------------------------------------------------------------
chromData(chr_eics)

## -----------------------------------------------------------------------------
## A small Spectra with some missing intensities at m/z 100
sp_gaps <- Spectra(
    DataFrame(
        rtime      = c(100, 110, 120, 130, 140, 150, 160, 170, 180),
        msLevel    = rep(1L, 9),
        dataOrigin = rep("impute_demo", 9),
        mz = NumericList(100, 100, 100, 100, 100, 100, 100, 100, 100,
                         compress = FALSE),
        intensity = NumericList(50, NA, 120, 200, NA, NA, 80, 30, 10,
                               compress = FALSE)
    ),
    source = MsBackendDataFrame()
)

## Derive a Chromatograms and extract the EIC for m/z ≈ 100
chr_gaps <- Chromatograms(sp_gaps)
eic_table <- data.frame(
    rtMin = 100, rtMax = 180,
    mzMin = 99.5, mzMax = 100.5,
    msLevel = 1L,
    dataOrigin = "impute_demo"
)

chr_eic <- chromExtract(chr_gaps, eic_table,
                        by = c("msLevel", "dataOrigin"))
chr_eic

## ----fig.height=10, fig.width=8-----------------------------------------------
## Create copies for comparison
chr_linear <- imputePeaksData(chr_eic, method = "linear")
chr_spline <- imputePeaksData(chr_eic, method = "spline")
chr_gaussian <- imputePeaksData(chr_eic, method = "gaussian",
                                window = 2, sd = 1)
chr_loess <- imputePeaksData(chr_eic, method = "loess", span = 0.75)

## Plot all methods for comparison
par(mfrow = c(3, 2), mar = c(4, 4, 2, 1))

## Original data
plotChromatograms(chr_eic, main = "Original EIC")

## Linear interpolation
plotChromatograms(chr_linear, main = "Linear Imputation")

## Spline interpolation
plotChromatograms(chr_spline, main = "Spline Imputation")

## Gaussian smoothing
plotChromatograms(chr_gaussian, main = "Gaussian Smoothing (window=2, sd=1)")

## LOESS smoothing
plotChromatograms(chr_loess, main = "LOESS Smoothing (span=0.75)")

## -----------------------------------------------------------------------------
## For on-disk backends, add imputation to the lazy queue
chr_mzr_imputed <- imputePeaksData(
  chr_mzr,
  method = "gaussian",
  window = 5,
  sd = 2
)

chr_mzr_imputed

## -----------------------------------------------------------------------------
## This reads from disk and applies imputation in one step
peak_data <- peaksData(chr_mzr_imputed[1])

## -----------------------------------------------------------------------------
length(chr_mzr_imputed@processingQueue)

## -----------------------------------------------------------------------------
## For in-memory backends, you can persist the imputation
chr_in_memory <- setBackend(chr_mzr_imputed, ChromBackendMemory())
chr_in_memory <- applyProcessing(chr_in_memory)

# Now imputation is permanently applied
length(chr_in_memory@processingQueue)

## ----compare-chromatograms-mrm------------------------------------------------
## Pick 8 MRM chromatograms, skipping the first (a TIC with no m/z info)
chr_sub <- chr_mzr[2:9]

cor_arr <- compareChromatograms(chr_sub)
cor_arr[, , "score"]   ## similarity scores
cor_arr[, , "n_peaks"] ## number of overlapping RT points per pair

## Use a chromData column as row/column labels
cor_arr_labeled <- compareChromatograms(chr_sub, labelsColumn = "chromIndex")
cor_arr_labeled[, , "score"]

## ----compare-chromatograms-heatmap, fig.width=6, fig.height=5-----------------
library(pheatmap)
## Label rows/columns with precursor → product m/z transitions
mz_labels <- paste0(round(chromData(chr_sub)$precursorMz, 1), " → ",
                     round(chromData(chr_sub)$productMz, 1))
score_mat <- cor_arr[, , "score"]
rownames(score_mat) <- colnames(score_mat) <- mz_labels
pheatmap(score_mat, main = "Pairwise Pearson correlation",
         color = hcl.colors(30, palette = "RdYlBu", rev = TRUE))

## ----compare-chromatograms-custom---------------------------------------------
cosine <- function(x, y) {
    sum(x * y) / (sqrt(sum(x^2)) * sqrt(sum(y^2)))
}

cos_arr <- compareChromatograms(chr_sub, FUN = cosine)
cos_arr[, , "score"]

## ----compare-chromatograms-minpeaks-------------------------------------------
## Require at least 10 overlapping RT points to compute a score
cor_strict <- compareChromatograms(chr_sub, minPeaks = 10L)
cor_strict[, , "score"]   ## NAs for pairs with < 10 common RT points
cor_strict[, , "n_peaks"] ## actual overlap counts are always recorded

## ----compare-chromatograms-xy-------------------------------------------------
## Compare the first 4 chromatograms against the last 4
res <- compareChromatograms(chr_sub[1:4], chr_sub[5:8])
res[, , "score"]

## ----compare-chromatograms-groups, eval=FALSE---------------------------------
# grp_list <- split(chr_sub, chromData(chr_sub)$dataOrigin)
# lapply(grp_list, compareChromatograms)

## ----si-----------------------------------------------------------------------
sessionInfo()

#' @include helpers.R
#' @include hidden_aliases.R
#' @include ChromBackend.R
NULL

#' @title Chromatographic Data Backend for Spectra Objects
#'
#' @name ChromBackendSpectra
#'
#' @description
#' The `ChromBackendSpectra` class extends `ChromBackendMemory`, inheriting
#' all its slots and methods while providing additional functionality for
#' summarizing chromatographic data from [Spectra::Spectra()] objects.
#'
#' It can be initialized with a `Spectra` object, which is stored in the
#' `spectra` slot of the backend. Users can also provide a `data.frame`
#' containing chromatographic metadata, stored in `@chromData`. This metadata
#' filters the `Spectra` object and generates `peaksData`. If `chromData` is
#' not provided, a default `data.frame` is created from the `Spectra` data.
#' An "rtMin", "rtMax", "mzMin", and "mzMax" column will be created by
#' condensing the `Spectra` data corresponding to each unique combination of
#' the `factorize.by` variables.
#'
#' By "factorization" we mean the process of grouping spectra
#' into chromatograms based on specified variables. For example, using
#' `factorize.by = c("msLevel", "dataOrigin")` means that all MS1 spectra from
#' file "A" form one chromatogram, all MS2 spectra from file "A" form another,
#' and so on. Each unique combination of the factorization variables creates
#' a separate chromatogram. This is essential for organizing spectral data into
#' meaningful chromatographic traces that can be visualized and analyzed.
#'
#' The *dataOrigin* core chromatogram variable should reflect the *dataOrigin*
#' of the `Spectra` object. The `factorize.by` parameter defines the variables
#' for grouping `Spectra` data into chromatographic data. The default is
#' `c("msLevel", "dataOrigin")`, which will define separate chromatograms for
#' each combination of `msLevel` and `dataOrigin`. These variables must be in
#' both the `spectraData()` of the `Spectra` and `chromData` (if provided).
#'
#' The `summarize.method` parameter defines how spectral data intensity is
#' summarized:
#' - **"sum"**: Sums intensity to create a Total Ion Chromatogram (TIC).
#' - **"max"**: Takes max intensity for a Base Peak Chromatogram (BPC).
#'
#' If `chromData` or its factorization columns are modified, the `factorize()`
#' method must be called to update `chromSpectraIndex`.
#'
#' @details
#' No `peaksData` is stored until the user calls a function that generates it
#' (e.g., `rtime()`, `peaksData()`, `intensity()`). The `@peaksData` slot
#' replacement is unsupported — modifications are temporary to optimize memory.
#' The `@inMemory` slot indicates this with `TRUE`.
#'
#' **Spectra Sort Index**: The `ChromBackendSpectra` backend maintains a
#' `spectraSortIndex` slot that stores a sort order for the internal `Spectra`
#' object based on `dataOrigin` and `rtime`. To optimize performance, the sort
#' index is only computed and stored when the spectra are unsorted; if already
#' sorted (which is typical for most real-world data), `spectraSortIndex` remains
#' empty (`integer()`). This avoids unnecessary subsetting operations. The sort
#' index is automatically recalculated whenever the `factorize()` method is called,
#' ensuring it remains valid and consistent. This approach avoids the need to
#' physically reorder disk-backed `Spectra` objects, which would require loading
#' all data into memory.
#'
#' **Factorize and Subsetting**: The `factorize()` method updates the
#' `chromSpectraIndex` in both `chromData` and the `@spectra` to reflect
#' the current grouping, and recalculates `spectraSortIndex` to maintain the
#' correct sort order. The `[` subsetting operator properly handles subsetting
#' of both `@chromData`, `@peaksData`, and `@spectra`, while updating the
#' `spectraSortIndex` to reference valid positions in the subsetted data.
#'
#' `ChromBackendSpectra` should reuse `ChromBackendMemory` methods whenever
#' possible to keep implementations simple.
#'
#' @param chromData A `data.frame` with chromatographic data for use in
#'        `backendInitialize()`. If missing, a default is generated. Columns
#'        like `rtMin`, `rtMax`, `mzMin`, and `mzMax` must be provided and not
#'        contain `NA` values. Use `-Inf/Inf` for unspecified values. The
#'        `"dataOrigin"` column must match the `Spectra` object's
#'        `"dataOrigin"`.
#'
#' @param factorize.by A `character` vector of `spectraVariables` for grouping
#'        `Spectra` data into chromatographic data (i.e., creating separate
#'        chromatograms for each unique combination of these variables).
#'        Default: `c("msLevel", "dataOrigin")`, which creates one chromatogram
#'        per MS level per data file.
#'        If `chromData` is provided, it **must** also contain these columns.
#'
#' @param object A `ChromBackendSpectra` object.
#'
#' @param spectra A `Spectra` object.
#'
#' @param spectraVariables A `character` vector specifying which variables
#'        from the `Spectra` object should be added to the chromData. These
#'        will be mapped using the `chromSpectraIndex` variable.
#'
#' @param summarize.method A `character` string specifying intensity summary:
#'        `"sum"` (default) or `"max"`.
#'
#' @param ... Additional parameters.
#'
#' @author Philippine Louail, Johannes Rainer.
#'
#' @exportClass ChromBackendSpectra
#'
#' @return Refer to the individual function description for information on the
#'         return value.
#'
#' @importClassesFrom Spectra Spectra
#' @importFrom Spectra Spectra spectraVariables spectraData concatenateSpectra
#'
#' @examples
#' library(Spectra)
#' library(MsBackendMetaboLights)
#'
#' ## Get Spectra data from MetaboLights
#' be <- backendInitialize(MsBackendMetaboLights(),
#'     mtblsId = "MTBLS39",
#'     filePattern = c("63B.cdf")
#' )
#' s <- Spectra(be)
#'
#' s <- setBackend(s, MsBackendMemory())
#'
#' ## Initialize ChromBackendSpectra
#' be_empty <- new("ChromBackendSpectra")
#' be <- backendInitialize(be_empty, s)
#'
#' ## replace the msLevel data
#' msLevel(be) <- c(1L, 2L, 3L)
#'
#' ## re-factorize the data
#' be <- factorize(be)
#'
#' ## Create BPC : we summarize the intensity present in the Spectra object
#' ## by the maximum value, thus creating a Base Peak Chromatogram.
#' be <- backendInitialize(be_empty, s, summarize.method = "max")
#'
#' ## Can now see the details of this bpc by looking at the chromData of our
#' ## object
#' chromData(be)
#'
#' ## Another possibilities is to create eics from the Spectra object.
#' ## Here we create an EIC with a specific m/z and retention time window.
#' df <- data.frame(mzMin = 100.01, mzMax = 100.02 , rtMin = 50, rtMax = 100)
#' be <- backendInitialize(be_empty, s, summarize.method = "sum")
#' chromData(be) <- cbind(chromData(be), df)
#'
#' ## now when we call the peaksData function, we will get the intensity
#' ## of the spectra object that are in the m/z and retention time window
#' ## defined in the chromData.
#' peaksData(be)
#'
NULL

#' @noRd
ChromBackendSpectra <- setClass(
    "ChromBackendSpectra",
    contains = "ChromBackendMemory",
    slots = c(
        inMemory = "logical",
        spectra = "Spectra",
        summaryFun = "function",
        spectraSortIndex = "integer"
    ),
    prototype = prototype(
        chromData = fillCoreChromVariables(data.frame()),
        peaksData = list(.EMPTY_PEAKS_DATA),
        readonly = TRUE,
        spectra = Spectra(),
        version = "0.1",
        inMemory = FALSE,
        summaryFun = sumi,
        spectraSortIndex = integer()
    )
)

#' @rdname ChromBackendSpectra
#' @importFrom methods new
#' @export ChromBackendSpectra
ChromBackendSpectra <- function() {
    new("ChromBackendSpectra")
}

#' @rdname ChromBackendSpectra
#' @importFrom methods callNextMethod
#' @importFrom MsCoreUtils rbindFill sumi maxi
setMethod("backendInitialize", "ChromBackendSpectra",
          function(object, spectra = Spectra(),
                   factorize.by = c("msLevel" , "dataOrigin"),
                   summarize.method = c("sum", "max"),
                   chromData = fillCoreChromVariables(),
                   spectraVariables = character(),
                   ...) {
              summarize.method <- match.arg(summarize.method)
              object@summaryFun <- if (summarize.method == "sum") sumi else maxi
              if (!is(spectra, "Spectra"))
                  stop("'spectra' must be a 'Spectra' object.")
              if (!length(spectra)) return(object)
              if (!all(factorize.by %in% spectraVariables(spectra)))
                  stop("All 'factorize.by' variables must exist in 'spectra'.")
              if (!is.data.frame(chromData))
                  stop("'chromData' must be a 'data.frame'.")
              if(!nrow(chromData))
                  chromData <- fillCoreChromVariables(data.frame())
              else  validChromData(chromData)
              if (!all(factorize.by %in% colnames(chromData)))
                  stop("All 'factorize.by' variables must exist ",
                       "in 'chromData'. If no chromData was provided, ",
                       "it needs to be part of the `coreChromVariables()` ",
                       "available.")
              ## Spectra object are not expected to be ordered by rtime,
              ## so we store a sort index instead of concatenating.
              ## This allows us to keep disk-backed backends intact.
              ## Only store sort index if data is actually unsorted (optimization).
              sort_idx <- order(
                  spectra$dataOrigin,
                  spectra$rtime
              )
              if (!identical(sort_idx, seq_along(spectra))) {
                  object@spectraSortIndex <- sort_idx
              }
              object@chromData <- chromData
              object@spectra <- spectra
              object <- factorize(object, factorize.by = factorize.by)
              ## map additional spectraVariables if any
              if (length(spectraVariables)) {
                  object <- .map_spectra_vars(object,
                                              spectraVariables = spectraVariables)
              }
              callNextMethod(object, chromData = .chromData(object))
              }
          )

#' @rdname hidden_aliases
#' @importFrom methods callNextMethod
setMethod("show", "ChromBackendSpectra", function(object) {
    callNextMethod()
    cat("\nThe Spectra object contains", length(object@spectra), "spectra\n")
    if (.inMemory(object)) cat("\nPeaks data is cached in memory\n")
})

#' @rdname ChromBackendSpectra
#' @export
chromSpectraIndex <- function(object) {
    if (!is(object, "ChromBackendSpectra"))
        stop("The object must be a 'ChromBackendSpectra' object.")
    cd <- chromData(object, columns = "chromSpectraIndex", drop = TRUE)
    if (!is.factor(cd))
        cd <- factor(cd)
    cd <- droplevels(cd)
    cd
}

#' @rdname hidden_aliases
setMethod("factorize", "ChromBackendSpectra",
          function(object, factorize.by = c("msLevel", "dataOrigin"),...) {
            if (!all(factorize.by %in%
                     spectraVariables(.spectra(object))))
                  stop("All 'factorize.by' variables must be in the ",
                       "Spectra object.")
            spectra_f <- interaction(as.list(
               spectraData(.spectra(object))[,
                                            factorize.by, drop = FALSE]),
               drop = TRUE, sep = "_")
            cd <- .chromData(object)

            if (nrow(cd)) {
                ## chromData exists: validate and align spectra to it
                if (!all(factorize.by %in% chromVariables(object)))
                    stop("All 'factorize.by' variables must be in chromData.")
                cd$chromSpectraIndex <- interaction(cd[, factorize.by,
                                                        drop = FALSE],
                                                     drop = TRUE, sep = "_")
                object@spectra$chromSpectraIndex <- factor(as.character(spectra_f),
                                                           levels = levels(cd$chromSpectraIndex))
                ## Apply sort index for processing if needed
                if (length(object@spectraSortIndex)) {
                    sorted_spectra <- .spectra(object)[object@spectraSortIndex]
                    sorted_spectra_f <- spectra_f[object@spectraSortIndex]
                } else {
                    sorted_spectra <- .spectra(object)
                    sorted_spectra_f <- spectra_f
                }
                object@chromData <- .ensure_rt_mz_columns(cd,
                                                          sorted_spectra,
                                                          sorted_spectra_f)
            } else {
                ## chromData is empty: create it from spectra
                object@spectra$chromSpectraIndex <- spectra_f
                full_sp <- do.call(rbindFill,
                                   lapply(split(.spectra(object), spectra_f),
                                          .spectra_format_chromData))
                rownames(full_sp) <- NULL
                object@chromData <- full_sp
            }
            ## Recalculate sort index: only store if data is unsorted (optimization)
            sort_idx <- order(
                object@spectra$dataOrigin,
                object@spectra$rtime
            )
            if (!identical(sort_idx, seq_along(object@spectra))) {
                object@spectraSortIndex <- sort_idx
            } else {
                object@spectraSortIndex <- integer()
            }
            object
          })

#' @rdname hidden_aliases
#' @importMethodsFrom ProtGenerics backendParallelFactor
setMethod(
    "backendParallelFactor", "ChromBackendSpectra",
    function(object, ...) {
        factor()
    }
)

#' @rdname hidden_aliases
#' @export
setMethod("isReadOnly", "ChromBackendSpectra", function(object) TRUE)

#' @rdname hidden_aliases
setMethod(
    "peaksData", "ChromBackendSpectra",
    function(object, columns = peaksVariables(object),
    drop = FALSE, ...) {
        if (.inMemory(object) || !length(object))
            return(callNextMethod())
        pv <- peaksVariables(object)
        if (!all(columns %in% pv))
            stop("Some of the requested peaks variables are not",
                 " available: ",
                 paste(setdiff(columns, pv), collapse = ", "),
                 ". Use 'peaksVariables()' to list available variables.")
        ## Ensure chromSpectraIndex only contains relevant levels needed
        valid_f <- chromSpectraIndex(object)
        ## Apply the sort index to spectra for processing (only if unsorted)
        if (length(object@spectraSortIndex)) {
            sorted_spectra <- .spectra(object)[object@spectraSortIndex]
        } else {
            sorted_spectra <- .spectra(object)
        }
        current_vals <- as.character(sorted_spectra$chromSpectraIndex)
        if (!setequal(unique(current_vals), levels(valid_f))) {
            sorted_spectra$chromSpectraIndex <- factor(
                current_vals,
                levels = levels(valid_f)
            )
        }
        ## Track original row indices to restore order after split/mapply
        ## split() groups by factor levels, which may reorder the data
        orig_idx <- seq_along(valid_f)
        split_idx <- split(orig_idx, valid_f)
        ## Process peaks data
        pd <- mapply(.process_peaks_data,
            cd = split(chromData(object), valid_f),
            s = split(
                sorted_spectra,
                sorted_spectra$chromSpectraIndex
            ),
            MoreArgs = list(
                columns = columns,
                fun = object@summaryFun,
                drop = drop
            ),
            SIMPLIFY = FALSE
        )
        pd_flat <- unlist(pd, use.names = FALSE, recursive = FALSE)
        ## Restore original row order: split_idx tells us which original rows
        ## are in each group, so unlist(split_idx) gives the order after split
        reorder_idx <- order(unlist(split_idx, use.names = FALSE))
        pd_flat[reorder_idx]
    }
)

#' @rdname hidden_aliases
setReplaceMethod("peaksData", "ChromBackendSpectra", function(object, value) {
    message(
        "The `peaksData` slot will be modified but the changes will not",
        " affect the Spectra object."
    )
    object <- callNextMethod()
    object@inMemory <- TRUE
    object
})


#' @rdname hidden_aliases
#' @export
setMethod(
    "supportsSetBackend", "ChromBackendSpectra",
    function(object, ...) FALSE
)

#' @rdname hidden_aliases
setMethod("intensity", "ChromBackendSpectra", function(object) {
    if (.inMemory(object)) return(lapply(.peaksData(object), function(z) z$intensity))
    if (!length(object)) return(list())
    peaksData(object, columns = "intensity", drop = TRUE)
})

#' @rdname hidden_aliases
setMethod("rtime", "ChromBackendSpectra", function(object) {
    if (.inMemory(object)) return(lapply(.peaksData(object), function(z) z$rtime))
    if (!length(object)) return(list())
    peaksData(object, columns = "rtime", drop = TRUE)
})

#' @rdname hidden_aliases
setMethod("lengths", "ChromBackendSpectra", function(x) {
    if (.inMemory(x)) return(vapply(.peaksData(x), nrow, integer(1L)))
    lengths(intensity(x))
})

#' @rdname hidden_aliases
#' @importMethodsFrom S4Vectors [ [[
#' @importFrom MsCoreUtils i2index
#' @importFrom stats setNames
#' @export
setMethod("[", "ChromBackendSpectra", function(x, i, j, ...) {
    if (!length(i))
        return(ChromBackendSpectra())

    i <- i2index(i, length = length(x))
    kept_indices <- chromSpectraIndex(x)[i]
    x@chromData <- .chromData(x)[i, , drop = FALSE]
    x@peaksData <- .peaksData(x)[i]
    spectra_keep <- x@spectra$chromSpectraIndex %in% kept_indices
    x@spectra <- x@spectra[spectra_keep]

    ## Update spectraSortIndex: remap old positions to new positions
    if (length(x@spectraSortIndex)) {
        old_positions_kept <- which(spectra_keep)
        ## Create mapping from old position to new position
        ## e.g., if we kept positions c(2, 5, 7), they become c(1, 2, 3)
        position_mapping <- setNames(seq_along(old_positions_kept),
                                     old_positions_kept)
        ## Keep only sort indices that reference kept positions
        kept_sort_positions <- x@spectraSortIndex %in% old_positions_kept
        x@spectraSortIndex <- as.integer(
            position_mapping[as.character(x@spectraSortIndex[kept_sort_positions])]
        )
    }

    x@chromData$chromSpectraIndex <- droplevels(x@chromData$chromSpectraIndex)
    x@spectra$chromSpectraIndex <- droplevels(x@spectra$chromSpectraIndex)
    x
})

#' @rdname hidden_aliases
setMethod("chromExtract", "ChromBackendSpectra",
          function(object, peak.table, by, ...) {
              required_cols <- c("rtMin", "rtMax", "mzMin", "mzMax", by)
              .validate_chromExtract_input(
                  object = object,
                  peak.table = peak.table,
                  by = by, required_cols = required_cols
              )

              matched <- .match_chromdata_peaktable(
                  object = object,
                  peak.table = peak.table,
                  by = by
              )
              cd <- .chromData(matched$object)
              chrom_keys <- matched$chrom_keys
              peak_keys  <- matched$peak_keys
              cd_split <- split(cd, chrom_keys) ##  UT need to check that
              pk_split <- split(peak.table, peak_keys)

              ## Check overlapping columns
              overl_cols <- .check_overl_columns(peak.table = peak.table,
                                                 object = object,
                                                 required_cols = required_cols)

              ## Merge peak.table into chromData safely.
              new_cdata <- mapply(function(cd_row, pks) { ## could switch to bpmapply ?
                  d <- suppressWarnings(cbind(cd_row, pks[!overl_cols]))
                  d[, names(peak.table)[overl_cols]] <- pks[, overl_cols]
                  d
              }, cd_row = cd_split,
              pks = pk_split, SIMPLIFY = FALSE)

              new_cdata <- .fast_rbind(new_cdata)
              rownames(new_cdata) <- NULL

              object@chromData <- new_cdata
              object@peaksData <- rep(list(.EMPTY_PEAKS_DATA),
                                          nrow(new_cdata))
              object
          })

#' @include helpers.R
#' @include hidden_aliases.R
#' @include ChromBackend.R
NULL

#' @title Chromatographic Data Backend for Reading mzML Files
#'
#' @name ChromBackendMzR
#'
#' @description
#' The `ChromBackendMzR` inherits all slots and methods from the base
#' `ChromBackendMemory` backend, providing additional functionality for reading
#' chromatographic data from mzML files.
#'
#' Unlike the `ChromBackendMemory` backend, the `ChromBackendMzR` backend
#' should have the *dataOrigin* chromatographic variables populated with the
#' file path of the mzML file from which the chromatographic data was read.
#'
#' Note that the `ChromBackendMzR` backend is read-only and does not support
#' direct modification of chromatographic data. However, it does support
#' `peaksData` slot replacement, which will modify the `@peaksData` slot but not
#' the local mzML files. This is indicated by the "inMemory" slot being set to
#' TRUE.
#'
#' Implementing functionalities with the `ChromBackendMzR` backend should be
#' simplified as much as possible and reuse the methods already implemented for
#' `ChromBackendMemory` when possible.
#'
#' @param BPPARAM Parallel setup configuration. See [BiocParallel::bpparam()]
#'        for more information.
#' @param files A character vector of file paths to mzML files.
#' @param object A `ChromBackendMzR` object.
#' @param ... Additional parameters to be passed.
#'
#' @author Philippine Louail
#'
#' @exportClass ChromBackendMzR
#'
#' @return Refer to the individual function description for information on the
#'         return value.
#'
#' @examples
#' library(mzR)
#' library(msdata)
#'
#' ## Load an mzML file
#' MRM_file <- system.file("proteomics", "MRM-standmix-5.mzML.gz",
#'     package = "msdata"
#' )
#'
#' ## Initialize the ChromBackendMzR object
#' be_empty <- ChromBackendMzR()
#' be <- backendInitialize(be_empty, files = MRM_file, BPPARAM = SerialParam())
#'
NULL

#' @noRd
setClass("ChromBackendMzR",
    contains = "ChromBackendMemory",
    slots = c(inMemory = "logical"),
    prototype = prototype(
        chromData = fillCoreChromVariables(data.frame()),
        peaksData = list(.EMPTY_PEAKS_DATA),
        readonly = FALSE,
        version = "0.1",
        inMemory = FALSE
    )
)

#' @rdname ChromBackendMzR
#' @export ChromBackendMzR
ChromBackendMzR <- function() {
  .check_mzR_package()
  new("ChromBackendMzR")
}

#' @rdname ChromBackendMzR
#' @importFrom methods callNextMethod
#' @importFrom MsCoreUtils rbindFill
#' @export
setMethod(
    "backendInitialize", "ChromBackendMzR",
    function(object, files = character(), BPPARAM = bpparam(), ...) {
        if (!length(files)) return(object)
        if (!is.character(files))
            stop("Parameter 'files' must be a character vector of ",
                 "file paths")
        files <- normalizePath(files, mustWork = FALSE)
        chromData <- do.call(rbindFill, bplapply(files,
            FUN = function(fl) cbind(.mzR_format_chromData(fl)),
            BPPARAM = BPPARAM))
        callNextMethod(object, chromData = chromData, ...)
    }
)

#' @rdname hidden_aliases
#' @export
setMethod("show", "ChromBackendMzR", function(object) {
    callNextMethod()
    fls <- unique(dataOrigin(object))
    if (length(fls)) {
        to <- min(3, length(fls))
        cat("\nfile(s):\n",
            paste(basename(fls[seq_len(to)]), collapse = "\n"),
            "\n", sep = "")
        if (length(fls) > 3) cat(" ...", length(fls) - 3, "more files\n")
    }
    if (object@inMemory) cat("\nPeaks data is cached in memory\n")
})

#' @rdname hidden_aliases
#' @importMethodsFrom ProtGenerics backendParallelFactor
setMethod("backendParallelFactor", "ChromBackendMzR", function(object, ...) {
  factor(dataOrigin(object), levels = unique(dataOrigin(object)))
})

#' @rdname hidden_aliases
#' @export
setMethod("isReadOnly", "ChromBackendMzR", function(object) TRUE)

#' @rdname hidden_aliases
#' @importFrom BiocParallel bplapply
setMethod(
    "peaksData", "ChromBackendMzR",
    function(object, columns = peaksVariables(object),
             drop = FALSE, BPPARAM = SerialParam(), ...) {
        if (.inMemory(object) || !length(object))
            return(callNextMethod())
        pv <- peaksVariables(object)
        if (!all(columns %in% pv))
            stop("Some of the requested peaks variables are not",
                 " available")
        ret <- all(pv %in% columns)
        f <- factor(dataOrigin(object), levels = unique(dataOrigin(object)))
        pd <- bplapply(split(object, f = f), function(ob) {
            chr <- .get_chrom_data(
                fl = .chromData(ob)$dataOrigin[1L], idx = chromIndex(ob))
            if (ret) chr
            else lapply(chr, `[`, , columns, drop = drop)
        }, BPPARAM = BPPARAM)
        unsplit(pd, f = f)
    }
)

#' @rdname hidden_aliases
setReplaceMethod("peaksData", "ChromBackendMzR", function(object, value) {
    message("Please keep in mind the 'ChromBackendMzR' backend is read-only.",
            " The peaksData slot will be modified but the changes will not",
            " affect the local mzML files.")
    object <- callNextMethod()
    object@inMemory <- TRUE
    object
})

#' @rdname hidden_aliases
setReplaceMethod("chromData", "ChromBackendMzR", function(object, value) {
    message("Please keep in mind the 'ChromBackendMzR' backend is read-only.",
            " The chromData slot will be modified but the changes will not",
            " affect the local mzML files.")
    callNextMethod()
})

#' @rdname hidden_aliases
#' @export
setMethod("supportsSetBackend", "ChromBackendMzR", function(object, ...) FALSE)

#' @rdname hidden_aliases
setMethod("intensity", "ChromBackendMzR", function(object) {
  if (.inMemory(object)) return(lapply(.peaksData(object), function(z) z$intensity))
  if (!length(object)) return(list())
  peaksData(object, columns = "intensity", drop = TRUE)
})

#' @rdname hidden_aliases
setMethod("rtime", "ChromBackendMzR", function(object) {
  if (.inMemory(object)) return(lapply(.peaksData(object), function(z) z$rtime))
  if (!length(object)) return(list())
  peaksData(object, columns = "rtime", drop = TRUE)
})

#' @rdname hidden_aliases
setMethod("lengths", "ChromBackendMzR", function(x) {
  if (.inMemory(x)) return(vapply(.peaksData(x), nrow, integer(1L)))
  lengths(intensity(x))
})

#' @rdname hidden_aliases
#' @importMethodsFrom S4Vectors [ [<-
setMethod("[", "ChromBackendMzR", function(x, i, j, ...) {
    if (!length(i)) return(ChromBackendMzR())
    callNextMethod()
})

#' @rdname hidden_aliases
setMethod(
    "chromExtract", "ChromBackendMzR",
    function(object, peak.table, by, ...) {
        required_cols <- c("rtMin", "rtMax", by)
        .validate_chromExtract_input(
            object = object, peak.table = peak.table,
            by = by, required_cols = required_cols)
        matched <- .match_chromdata_peaktable(
            object = object, peak.table = peak.table, by = by)
        cd <- .chromData(matched$object)
        chrom_keys <- matched$chrom_keys
        peak_keys <- matched$peak_keys
        cd_split <- split(cd, chrom_keys)
        pk_split <- split(peak.table, peak_keys)
        overl_cols <- .check_overl_columns(
            peak.table = peak.table, object = object,
            required_cols = required_cols)
        new_cdata <- mapply(function(cd_row, pks) {
            d <- suppressWarnings(cbind(cd_row, pks[!overl_cols]))
            d[, names(peak.table)[overl_cols]] <- pks[, overl_cols]
            d
        }, cd_row = cd_split, pks = pk_split, SIMPLIFY = FALSE)
        new_cdata <- .fast_rbind(new_cdata)
        rownames(new_cdata) <- NULL
        object@chromData <- new_cdata
        object@peaksData <- rep(list(.EMPTY_PEAKS_DATA), nrow(new_cdata))
        return(object)
    }
)

#' @title Chromatographic MS Data Backends
#'
#' @name ChromBackend
#'
#' @aliases ChromBackend-class
#' @aliases ChromBackendMemory-class
#' @aliases ChromBackendMzR-class
#' @aliases ChromBackendSpectra-class
#' @aliases [,ChromBackend-method
#' @aliases [[,ChromBackend-method
#' @aliases [[<-,ChromBackend-method
#' @aliases extractByIndex
#' @aliases factorize
#'
#' @description
#'
#' `ChromBackend` is a virtual class that defines what different backends need
#' to provide to be used by the `Chromatograms` package and classes.
#'
#' The backend should provide access to the chromatographic data which mainly
#' consists of (paired) intensity and retention time values. Additional
#' chromatographic metadata such as MS level and precursor and product m/z
#' should also be provided.
#'
#' Through their implementation different backends can be either optimized for
#' minimal memory requirements or performance. Each backend needs to implement
#' data access methods listed in section *Backend functions:* below.
#'
#' And example implementation and more details and descriptions are provided
#' in the *Creating new `ChromBackend` classes for Chromatograms* vignette.
#'
#' Currently available backends are:
#'
#' - `ChromBackendMemory`: This backend stores chromatographic data directly
#'   in memory, making it ideal for small datasets or testing. It can be
#'   initialized with a `data.frame` of chromatographic data via the
#'   `chromData` parameter and a `list` of `data.frame` entries for peaks data
#'   using the `peaksData` parameter. These data can be accessed with the
#'   `chromData()` and `peaksData()` functions.
#'
#' - `ChromBackendMzR`: The `ChromBackendMzR` inherits all slots and methods
#'   from the base `ChromBackendMemory` backend, providing additional
#'   functionality for reading chromatographic data from mzML files.
#'
#' - `ChromBackendSpectra`: The `ChromBackendSpectra` inherits all slots and
#'   methods from the base `ChromBackendMemory` backend, providing additional
#'   functionality for reading chromatographic data from `Spectra` objects.
#'
#' @section Core chromatogram variables:
#'
#' The *core* chromatogram variables are variables (metadata) that can/should
#' be provided by a backend. For each of these variables a value needs to be
#' returned, if none is defined, a missing value (of the correct data type)
#' should be returned. The names of the chromatogram variables in your current
#' chromatogram object are returned with the `chromVariables()` function.
#'
#' For each core chromatogram variable a dedicated access method exists. In
#' contrast to the peaks data described below, a single value should be
#' returned for each chromatogram.
#'
#' The `coreChromVariables()` function returns the core chromatogram variables
#' along with their expected (defined) data type.
#'
#' The core chromatogram variables (in alphabetical order) are:
#'
#' - `chromIndex`: an `integer` with the index of the chromatogram in the
#'   original source file (e.g. *mzML* file). In backedn with no original
#'   source file, this variable should be set to `NA_integer_`.
#' - `collisionEnergy`: for SRM data, `numeric` with the collision energy of
#'   the precursor.
#' - `dataOrigin`: optional `character` with the origin of a chromatogram.
#' - `msLevel`: `integer` defining the MS level of the data.
#' - `mz`: optional `numeric` with the (target) m/z value for the
#'   chromatographic data.
#' - `mzMin`: optional `numeric` with the lower m/z value of the m/z range in
#'   case the data (e.g. an extracted ion chromatogram EIC) was extracted from
#'   a `Spectra` object.
#' - `mzMax`: optional `numeric` with the upper m/z value of the m/z range.
#' - `precursorMz`: for SRM data, `numeric` with the target m/z of the
#'   precursor (parent).
#' - `precursorMzMin`: for SRM data, optional `numeric` with the lower m/z of
#'   the precursor's isolation window.
#' - `precursorMzMax`: for SRM data, optional `numeric` with the upper m/z of
#'   the precursor's isolation window.
#' - `productMz` for SRM data, `numeric` with the target m/z of the
#'   product ion.
#' - `productMzMin`: for SRM data, optional `numeric` with the lower m/z of
#'   the product's isolation window.
#' - `productMzMax`: for SRM data, optional `numeric` with the upper m/z of
#'   the product's isolation window.
#'
#' @section Core Peaks variables:
#'
#' Similar to the *core* chromatogram variables, *core* peaks variables
#' represent  metadata that should be provided by a backend. Each of these
#' variables should return a value, and if undefined, a missing value (with the
#' appropriate data type) is returned. The number of values for a peaks
#' variable in a single chromatogram can vary, from none to multiple, and may
#' differ between chromatograms.
#'
#' The names of peaks variables in the current chromatogram object can be
#' obtained with the `peaksVariables()` function.
#'
#' Each core peaks variable has a dedicated accessor method.
#'
#'
#' The `corePeaksVariables()` function returns the core peaks variables along
#' with their expected (defined) data type.
#'
#' The core peaks variables, listed in the required order for `peaksData`, are:
#'
#' - `rtime`: A `numeric` vector containing retention time values.
#' - `intensity`: A `numeric` vector containing intensity values.
#'
#' They should be provided for each chromatogram in the backend,
#' **in this order**, No NAs are allowed for the `rtime` values. These
#' characteristics will be checked with the `validPeaksData()` function.
#'
#' @param BPPARAM Parallel setup configuration. See [BiocParallel::bpparam()]
#'        for more information.
#'
#' @param by for `chromExtract() `A `character` vector specifying one or more
#'        column names that are present in both `peak.table` and
#'        `chromData(object)`. These columns uniquely identify chromatograms.
#'        The combination of these columns must be unique in
#'        `chromData(object)`. Can be of length 1 or greater.
#'
#' @param columns For `chromData()` accessor: optional `character` with column
#'        names (chromatogram variables) that should be included in the
#'        returned `data.frame`. By default, all columns are returned.
#'
#' @param drop For `chromData()` and `peaksData()`: `logical(1)` default to
#'        `FALSE`. If `TRUE`, and one column is requested by the user, the
#'        method should return a vector (or list of vector for `peaksData()`)
#'        of the single column requested.
#'
#' @param f `factor` defining the grouping to split `x`. See [split()].
#'
#' @param i For `[`: `integer`, `logical` or `character` to subset the object.
#'
#' @param j For `[` and `[[`: ignored.
#'
#' @param keep For `filterChromData()`: `logical(1)`
#'        defining whether to keep (`keep = TRUE`) or remove (`keep = FALSE`)
#'        the chromatogram data that match the condition.
#'
#' @param match For `filterChromData()` : `character(1) `
#'        defining whether the condition has to match for all provided
#'        `ranges` (`match = "all"`; the default), or for any of them
#'        (`match = "any"`) for chromatogram data to be retained.
#'
#' @param method For `imputePeaksData()`: `character(1)`: Imputation
#'        method ("linear", "spline", "gaussian", "loess")
#'
#' @param name For `$` and `$<-`: the name of the chromatogram variable to
#'        return or set.
#'
#' @param object Object extending `ChromBackend`.
#'
#' @param peak.table For `chromExtract()` A `data frame` containing the
#'        following minimum columns:
#'          - rtMin: Minimum retention time for each peak. Cannot be NA.
#'          - rtMax: Maximum retention time for each peak. Cannot be NA.
#'          - mzMin: Minimum m/z value for each peak.
#'          - mzMax: Maximum m/z value for each peak.
#'        Additionally, the `peak.table` must include columns that uniquely
#'        identify chromatograms in the `object`. Common choices are
#'        "msLevel" and/or "dataOrigin". These columns must also be present
#'        in the `chromData` of the `object`. Any extra columns in
#'        `peak.table` will be added to the `chromData` of the newly created
#'        object.
#'
#' @param ranges For `filterChromData()` : a `numeric`
#'        vector of paired values (upper and lower boundary) that define the
#'        ranges to filter the `object`. These paired values need to be in the
#'        same order as the `variables` parameter (see below).
#'
#' @param extrapolate For `imputePeaksData`: `logical(1)` (default `FALSE`).
#'        If `TRUE`, missing values at the beginning and end of a chromatogram
#'        (outside the range of observed values) will be extrapolated. If
#'        `FALSE`, only interpolation is performed and leading/trailing `NA`
#'        values remain `NA`.
#'
#' @param sd For `imputePeaksData`: `numeric(1)`, for the gaussian method:
#'        Standard deviation for Gaussian kernel
#'        (only used if method == "gaussian")
#'
#' @param span For `imputePeaksData`: `numeric(1)`, for the loess method:
#'        Smoothing parameter (only used if method == "loess")
#'
#' @param value Replacement value for `<-` methods. See individual
#'        method description or expected data type.
#'
#' @param variables For `filterChromData()`: `character` vector with the names
#'        of the chromatogram variables to filter for. The list of available
#'        chromatogram variables can be obtained with `chromVariables()`.
#'
#' @param window For `imputePeaksData`: `integer`, for the gaussian method:
#'        Half-width of Gaussian kernel window (e.g., 2 gives window size 5)
#'
#' @param x Object extending `ChromBackend`.
#'
#' @param ... Additional arguments.
#'
#' @section Mandatory methods:
#'
#' New backend classes **must** extend the base `ChromBackend` class and
#' implement the following mandatory methods:
#'
#' - `backendInitialize()`: initialises the backend. This method is
#'   supposed to be called right after creating an instance of the
#'   backend class and should prepare the backend.
#'   Parameters can be defined freely for each backend, depending on what is
#'   needed to initialize the backend.
#'   This method has to ensure to set the chromatogram variable `dataOrigin`
#'   correctly.
#'
#' - `backendBpparam()`: returns the parallel processing setup supported by
#'   the backend class. This function can be used by any higher
#'   level function to evaluate whether the provided parallel processing
#'   setup (or the default one returned by `bpparam()`) is supported
#'   by the backend. Backends not supporting parallel processing (e.g.
#'   because they contain a connection to a database that can not be
#'   shared across processes) should extend this method to return only
#'   `SerialParam()` and hence disable parallel processing for (most)
#'   methods and functions. See also `backendParallelFactor()` for a
#'   function to provide a preferred splitting of the backend for parallel
#'   processing.
#'
#' - `backendParallelFactor()`: returns a `factor` defining an optimal
#'   (preferred) way how the backend can be split for parallel processing
#'   used for all peak data accessor or data manipulation functions.
#'   The default implementation returns a factor of length 0 (`factor()`)
#'   providing thus no default splitting. `backendParallelFactor()` for
#'   `ChromBackendMzR` on the other hand returns `factor(dataOrigin(object))`
#'   hence suggesting to split the object by data file.
#'
#' - `chromData()`, `chromData<-`: gets or sets general chromatogram metadata
#'   (annotation). `chromData()` returns a `data.frame`, `chromData<-` expects
#'   a `data.frame` with the same number of rows as there are chromatograms in
#'   `object`. Read-only backends might not need to implement the
#'   replacement method `chromData<-` (unless some internal caching mechanism
#'   could be used). `chromData()` should be implemented with the parameter
#'   `drop` set to `FALSE` as default. With `drop = FALSE` the method should
#'   return a `data.frame` even if one column is requested. If `drop = TRUE`
#'   is specified, the output will be a vector of the single column requested.
#'   New backends should be implemented such as if empty, the method returns a
#'   `data.frame` with 0 rows and the columns defined by `chromVariables()`.
#'   By default, the function *should* return at minimum the
#'   coreChromVariables, even if NAs.
#'
#' - `chromExtract()`: return A new `Chrombackend` object containing separated
#'   chromatographic area as individual chromatograms. The chromatographic areas
#'   are defined by the `peak.table` parameter. The new object will contain
#'   chromatograms that match the conditions defined in `peak.table`. If no
#'   chromatograms match the conditions, an empty `ChromBackend` object
#'   should be returned.
#'
#' - `extractByIndex()`: function to subset a backend to selected elements
#'   defined by the provided index. Similar to `[`, this method should allow
#'   extracting (or to subset) the data in any order. In contrast to `[`,
#'   however, `i` is expected to be an `integer` (while `[` should also
#'   support `logical` and eventually `character`). While being apparently
#'   redundant to `[`, this methods avoids package namespace errors/problems
#'   that can result in implementations of `[` being not found by R (which
#'   can happen sometimes in parallel processing using the
#'   [BiocParallel::SnowParam()]). This method is used internally to
#'   extract/subset its backend. Implementation of this method is mandatory.
#'
#' - `peaksData()`: returns a `list` of `data.frame` with the data
#'   (e.g. retention time - intensity pairs) from each chromatogram. The length
#'   of the `list` is equal to the number of chromatograms in `object`. For an
#'   empty chromatogram a `data.frame` with 0 rows and two columns (named
#'   `"rtime"` and `"intensity"`) has to be returned. The optional parameter
#'   `columns`, if supported by the backend allows to define which peak
#'   variables should be returned in each array. As default (minimum) columns
#'   `"rtime"` and `"intensity"` have to be provided. `peaksData()` should be
#'   implemented with the parameter `drop` set to `FALSE` as default.  With
#'   `drop = FALSE` the method should return a `data.frame` even if only one
#'   column is requested. If `drop = TRUE`  is specified, the output will be a
#'   vector of the single column requested.
#'
#' - `peaksData<-` replaces the peak data (retention time and intensity values)
#'   of the backend. This method expects a `list` of two-dimensional arrays
#'   (`data.frame`) with columns representing the peak variables.
#'   All existing peaks data are expected to be replaced with these new values.
#'   The length of the `list` has to match the number of chromatogram of
#'   `object`. Note that only writeable backends need to support this method.
#'
#' - `[`: subset the backend. Only subsetting by element (*row*/`i`) is
#'   allowed. This method should be implemented as to support empty integer.
#'
#' - `$`, `$<-`: access or set/add a single chromatogram variable (column) in
#'   the backend.
#'
#' - `backendMerge()`: merges (combines) `ChromBackend` objects into a single
#'   instance. All objects to be merged have to be of the same type.
#'
#' @section Optional methods with default implementations:
#'
#' Additional methods that might be implemented, but for which default
#' implementations are already present are:
#'
#' - `[[`
#'
#' - `backendParallelFactor()`: returns a `factor` defining an optimal
#'   (preferred) way how the backend can be split for parallel processing
#'   used for all *peak* data accessor or data manipulation functions.
#'   The default implementation returns a factor of length 0 (`factor()`)
#'   providing thus no default splitting.
#'
#' - `chromIndex()`: returns an `integer` vector with the index of the
#'   chromatograms in the original source file.
#'
#' - `chromVariables()`: returns a `character` vector with the
#'   available chromatogram variables (columns, fields or attributes)
#'   available in `object`. Variables listed by this function are expected to
#'   be returned (if requested) by the `chromData()` function.
#'
#' - `collisionEnergy()`, `collisionEnergy<-`: gets or sets the collision
#'   energy for the precursor (for SRM data). `collisionEnergy()` returns a
#'   `numeric` of length equal to the number of chromatograms in `object`.
#'
#' - `dataOrigin()`, `dataOrigin<-`: gets or sets the *data origin* variable.
#'   `dataOrigin()` returns a `character` of length equal to the number of
#'   chromatograms, `dataOrigin<-` expects a `character` of length equal
#'   `length(object)`.
#'
#' - `filterChromData()`: filters any numerical chromatographic data variables
#'   based on the provided numerical `ranges`. The method should return a
#'   `ChromBackend` object with the chromatograms that match the condition.
#'   This function will results in an object with less chromatogram than the
#'   original.
#'
#' - `filterEmptyChromatograms()`: removes empty chromatograms (i.e.
#'   chromatograms without peaks). Implementation of this method is optional
#'   since a default implementation for `ChromBackend` is available.
#'
#' - `intensity()`: gets the intensity values from the chromatograms. Returns
#'   a `list` of `numeric` vectors (intensity values for each
#'   chromatogram). The length of the list is equal to the number of
#'   chromatograms in `object`.
#'
#' - `intensity<-`: replaces the intensity values. `value` has to be a `list`
#'   of length equal to the number of chromatograms and the number of values
#'   within each list element identical to the number of data pairs in each
#'   chromatogram. Note that just writeable backends need to support this
#'   method.
#'
#' - `imputePeaksData()`: Imputes missing intensity values in the
#'   chromatographic peaks data using various methods such as linear
#'   interpolation, spline interpolation, Gaussian kernel smoothing, or LOESS
#'   smoothing. This method modifies the peaks data in place and returns the
#'   same `ChromBackend` object with imputed values.
#'
#' - `isReadOnly()`: returns a `logical(1)` whether the backend is *read
#'   only* or does allow also to write/update data. Defaults to FALSE.
#'
#' - `isEmpty()`: returns a `logical` of length equal to the number of
#'   chromatograms with `TRUE` for chromatograms without any data pairs.
#'
#' - `length()`: returns the number of chromatograms in the object.
#'
#' - `lengths()`: returns the number of data pairs (retention time and
#'   intensity values) per chromatogram.
#'
#' - `msLevel()`: gets the chromatogram's MS level. Returns an `integer`
#'   vector (of length equal to the number of chromatograms) with the MS
#'   level for each chromatogram (or `NA_integer_` if not available).
#'
#' - `mz()`,`mz<-`: gets or sets the m/z value of the chromatograms. `mz()`
#'   returns a `numeric` of length equal to the number of chromatograms in `
#'   object`, `mz<-` expects a `numeric` of length `length(object)`.
#'
#' - `mzMax()`,`mzMax<-`: gets or sets the upper m/z of the mass-to-charge
#'   range from which a chromatogram contains signal (e.g. if the chromatogram
#'   was extracted from MS data in spectra format and a m/z range was
#'   provided). `mzMax()` returns a `numeric` of length equal to the number of
#'   chromatograms in `object`, `mzMax<-` expects a `numeric` of length equal
#'   to the number of chromatograms in `object`.
#'
#' - `mzMin()`,`mzMin<-`: gets or sets the lower m/z of the mass-to-charge
#'   range from which a chromatogram contains signal (e.g. if the chromatogram
#'   was extracted from MS data in spectra format and a m/z range was
#'   provided). `mzMin()` returns a `numeric` of length equal to the number of
#'   chromatograms in `object`, `mzMin<-` expects a `numeric` of length equal
#'   to the number of chromatograms in `object`.
#'
#' - `peaksVariables()`: lists the available data variables for the
#'   chromatograms. Default peak variables are `"rtime"` and `"intensity"`
#'   (which all backends need to support and provide), but some backends
#'   might provide additional variables.
#'   Variables listed by this function are expected to be returned (if
#'   requested) by the `peaksData()` function.
#'
#' - `precursorMz()`,`precursorMz<-`: gets or sets the (target) m/z of the
#'   precursor (for SRM data). `precursorMz()` returns a `numeric` of length
#'   equal to the number of chromatograms in `object`. `precursorMz<-` expects
#'   a `numeric` of length equal to the number of chromatograms.
#'
#' - `precursorMzMin()`,`precursorMzMax()`,`productMzMin()`, `productMzMax()`:
#'   gets the lower and upper margin for the precursor or product isolation
#'   windows. These functions might return the value of `productMz()` if the
#'   respective minimal or maximal m/z values are not defined in `object`.
#'
#' - `productMz()`,`productMz<-`: gets or sets the (target) m/z of the
#'   product (for SRM data). `productMz()` returns a `numeric` of length
#'   equal to the number of chromatograms in `object`. `productMz<-` expects
#'   a `numeric` of length equal to the number of chromatograms.
#'
#' - `rtime()`: gets the retention times from the chromatograms. returns a
#'   `list` of `numeric` vectors (retention times for each
#'   chromatogram). The length of the returned list is equal to the number of
#'   chromatograms in `object`.
#'
#' - `rtime<-`: replaces the retention times. `value` has to be a `list` of
#'   length equal to the number of chromatograms and the
#'   number of values within each list element identical to the number of
#'   data pairs in each chromatogram. Note that just writeable backends support
#'   this method.
#'
#' - `split()`: splits the backend into a `list` of backends (depending on
#'   parameter `f`). The default method for `ChromBackend` uses
#'   [split.default()], thus backends extending `ChromBackend` don't
#'   necessarily need to implement this method.
#'
#' - `supportsSetBackend()`: whether a `ChromBackend` supports the
#'   `Chromatograms` `setBackend()` function. The default function will
#'   take the `peaksData()` and `chromData()` of the user's backend and pass it
#'   to the new backend. If the backend does not support this function, it
#'   should return `FALSE`. Therefore both backend in question should have a
#'   adequate `peaksData()` and `chromData()` method as well as their
#'   respective replacement method.
#'
#' @section Implementation notes:
#'
#' Backends extending `ChromBackend` **must** implement all of its methods
#' (listed above). A guide to create new backend classes is provided as a
#' dedicated vignette. Additional information and an example for a backend
#' implementation is provided in the respective vignette.
#'
#' @author Johannes Rainer, Philippine Louail
#'
#' @md
#'
#' @exportClass ChromBackend
#'
#' @return Refer to the individual function description for information on the
#'         return value.
#'
#' @examples
#'
#' ## Create a simple backend implementation
#' ChromBackendDummy <- setClass("ChromBackendDummy",
#'     contains = "ChromBackend"
#' )
#'
#' ## We will show examples on a `ChromBackendMemory` backend.
#' be <- ChromBackendMemory()
#'
#' ## The `backendInitialize()` method initializes the backend filling it with
#' ## data. This method can take any parameters needed for the backend to
#' ## get loaded with the data.
#' cdata <- data.frame(
#'     msLevel = c(1L, 1L, 1L),
#'     mz = c(112.2, 123.3, 134.4),
#'     dataOrigin = c("mem1", "mem2", "mem3")
#' )
#'
#' pdata <- list(
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     ),
#'     data.frame(
#'         rtime = c(45.1, 46.2),
#'         intensity = c(100, 80.1)
#'     ),
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     )
#' )
#' be <- backendInitialize(be, chromData = cdata, peaksData = pdata)
#'
#' be
#'
#' ## Data can be accessed with the accessor methods
#' msLevel(be)
#'
#' rtime(be)
#'
#' ## Even if no data was provided for all chromatogram variables, its accessor
#' ## methods are supposed to return a value.
#' precursorMz(be)
#'
#' ## The `peaksData()` method is supposed to return data/frames of rtime and
#' ## intensity pairs as a `list`.
#' peaksData(be)
#'
#' ## Use columns to extract specific peaks variables. Below we extract rtime
#' ## and intensity values, but in reversed order to the default.
#' peaksData(be, columns = c("intensity", "rtime"))
#'
#' ## List available chromatographic variables
#' chromVariables(be)
#'
#' ## List available peak variables
#' peaksVariables(be)
#'
#' ## Extract multiple chromatographic variables
#' chromData(be, c("dataOrigin", "mz", "msLevel"))
#'
#' ## Single variables can also be accessed and replaced
#' mz(be)
#' mz(be) <- c(123.4, 134.5, 145.6)
#'
#' be$msLevel
#' be$msLevel <- c(2L, 2L, 2L)
#'
#' be[["rtime"]]
#' be[["rtime"]] <- list(
#'     c(12.4, 12.8, 13.2, 14.6),
#'     c(45.1, 46.2),
#'     c(12.4, 12.8, 13.2, 14.6)
#' )
NULL

setClass("ChromBackend",
         contains = "VIRTUAL",
         slots = c(
             version = "character"
         ),
         prototype = prototype(readonly = FALSE, version = "0.1")
)

#' @importMethodsFrom S4Vectors $ $<-
#'
#' @exportMethod $
#'
#' @rdname ChromBackend
setMethod("$", "ChromBackend", function(x, name) {
    stop("Not implemented for ", class(x), ".")
})

#' @exportMethod $<-
#'
#' @rdname ChromBackend
setReplaceMethod("$", "ChromBackend", function(x, name, value) {
    stop("Not implemented for ", class(x), ".")
})

#' @exportMethod backendMerge
#'
#' @importMethodsFrom ProtGenerics backendMerge
#'
#' @rdname ChromBackend
setMethod("backendMerge", "ChromBackend", function(object, ...) {
    stop("Not implemented for ", class(object), ".")
})

#' @exportMethod chromData
#'
#' @rdname ChromBackend
setMethod(
    "chromData", "ChromBackend",
    function(object, columns = chromVariables(object), drop = FALSE) {
        stop("Not implemented for ", class(object), ".")
    }
)

#' @exportMethod chromData<-
#'
#' @rdname ChromBackend
setReplaceMethod("chromData", "ChromBackend", function(object, value) {
    stop("Not implemented for ", class(object), ".")
})


#' @exportMethod chromExtract
#'
#' @rdname ChromBackend
setMethod("chromExtract", "ChromBackend", function(object,
                                                   peak.table,
                                                   by) {
    stop("Not implemented for ", class(object), ".")
})

#' @exportMethod factorize
#'
#' @rdname ChromBackend
setMethod("factorize", "ChromBackend", function(object, ...) {
    stop("Not implemented for ", class(object), ".")
})

#' @exportMethod peaksData
#'
#' @importMethodsFrom ProtGenerics peaksData
#'
#' @rdname ChromBackend
setMethod(
    "peaksData", "ChromBackend",
    function(object, columns = c("rtime", "intensity"), drop = FALSE, ...) {
        stop("Not implemented for ", class(object), ".")
    }
)

#' @exportMethod peaksData<-
#'
#' @importMethodsFrom ProtGenerics peaksData<-
#'
#' @rdname ChromBackend
setReplaceMethod("peaksData", "ChromBackend", function(object, value) {
    stop("Not implemented for ", class(object), ".")
})

###############################################################################
## Methods with default implementations below.

#' @exportMethod [
#' @rdname ChromBackend
setMethod("[", "ChromBackend", function(x, i, j, ..., drop = FALSE) {
    extractByIndex(x, i2index(i, length = length(x)))
})

#' @rdname ChromBackend
#'
#' @importMethodsFrom S4Vectors [[ [[<-
#'
#' @export
setMethod("[[", "ChromBackend", function(x, i, j, ...) {
    if (!is.character(i)) {
        stop(
            "'i' is supposed to be a character defining the chromatogram ",
            "variable to access."
        )
    }
    if (!missing(j)) {
        stop("'j' is not supported.")
    }
    do.call("$", list(x, i))
})

#' @rdname ChromBackend
#'
#' @export
setReplaceMethod("[[", "ChromBackend", function(x, i, j, ..., value) {
    if (!is.character(i)) {
        stop(
            "'i' is supposed to be a character defining the chromatogram ",
            "variable to replace or create."
        )
    }
    if (!missing(j)) {
        stop("'j' is not supported.")
    }
    do.call("$<-", list(x, i, value))
})

#' @importMethodsFrom ProtGenerics backendBpparam
#'
#' @exportMethod backendBpparam
#'
#' @rdname ChromBackend
#'
#' @export
setMethod("backendBpparam",
          signature = "ChromBackend",
          function(object, BPPARAM = bpparam()) BPPARAM
)

#' @exportMethod backendInitialize
#'
#' @importFrom methods .valueClassTest validObject
#'
#' @importMethodsFrom ProtGenerics backendInitialize
#'
#' @rdname ChromBackend
setMethod("backendInitialize",
          signature = "ChromBackend",
          definition = function(object, ...) {
              validObject(object)
              object
          }
)

#' @rdname ChromBackend
#'
#' @importMethodsFrom ProtGenerics backendParallelFactor
#'
#' @exportMethod backendParallelFactor
setMethod("backendParallelFactor", "ChromBackend", function(object, ...) {
    factor()
})

#' @rdname ChromBackend
#'
#' @importMethodsFrom ProtGenerics backendMerge
setMethod("backendMerge", "list", function(object, ...) {
    backendMerge(object[[1]], object[-1])
})

#' @exportMethod chromIndex
#'
#' @rdname ChromBackend
setMethod(
    "chromIndex", "ChromBackend",
    function(object) chromData(object, columns = "chromIndex", drop = TRUE)
)

#' @exportMethod chromIndex<-
#'
#' @rdname ChromBackend
setReplaceMethod("chromIndex", "ChromBackend", function(object, value) {
    object$chromIndex <- value
    object
})

#' @exportMethod chromVariables
#'
#' @rdname ChromBackend
setMethod("chromVariables", "ChromBackend", function(object) {
    names(coreChromVariables())
})

#' @exportMethod collisionEnergy
#'
#' @importMethodsFrom ProtGenerics collisionEnergy
#'
#' @rdname ChromBackend
setMethod("collisionEnergy", "ChromBackend", function(object) {
    chromData(object, columns = "collisionEnergy", drop = TRUE)
})

#' @exportMethod collisionEnergy<-
#'
#' @importMethodsFrom ProtGenerics collisionEnergy<-
#'
#' @rdname ChromBackend
setReplaceMethod("collisionEnergy", "ChromBackend", function(object, value) {
    object$collisionEnergy <- value
    object
})

#' @exportMethod dataOrigin
#'
#' @importMethodsFrom ProtGenerics dataOrigin
#'
#' @rdname ChromBackend
setMethod("dataOrigin", "ChromBackend", function(object) {
    chromData(object, columns = "dataOrigin", drop = TRUE)
})

#' @exportMethod dataOrigin<-
#'
#' @importMethodsFrom ProtGenerics dataOrigin<-
#'
#' @rdname ChromBackend
setReplaceMethod("dataOrigin", "ChromBackend", function(object, value) {
    object$dataOrigin <- value
    object
})

#' @rdname ChromBackend
#' @importFrom methods existsMethod
#' @importMethodsFrom ProtGenerics extractByIndex
#' @exportMethod extractByIndex
setMethod("extractByIndex", c("ChromBackend", "ANY"), function(object, i) {
    if (existsMethod("[", class(object)[1L])) {
        object[i = i]
    } else {
        stop("'extractByIndex' not implemented for ", class(object)[1L], ".")
    }
})

#' @rdname ChromBackend
setMethod("extractByIndex", c("ChromBackend", "missing"), function(object, i) {
    object
})

#' @exportMethod intensity
#'
#' @importMethodsFrom ProtGenerics intensity
#'
#' @rdname ChromBackend
setMethod("intensity", "ChromBackend", function(object) {
    if (length(object)) {
        peaksData(object, columns = "intensity", drop = TRUE)
    } else {
        list()
    }
})

#' @exportMethod intensity<-
#'
#' @importMethodsFrom ProtGenerics intensity<-
#'
#' @rdname ChromBackend
setReplaceMethod("intensity", "ChromBackend", function(object, value) {
    pd <- peaksData(object)
    if (!is.list(value) || length(pd) != length(value)) {
        stop("'value' should be a list of the same length as 'object'")
    }
    for (i in seq_along(pd)) {
        if (length(value[[i]]) != nrow(pd[[i]])) {
            stop(
                "Length of 'value[[", i, "]]' does not match ",
                "the number of rows in the intensity of chromatogram: ",
                i, "'"
            )
        }
    }
    peaksData(object) <- lapply(seq_along(pd), function(i) {
        pd[[i]]$intensity <- value[[i]]
        return(pd[[i]])
    })
    object
})

#' @exportMethod isEmpty
#'
#' @rdname ChromBackend
#'
#' @importMethodsFrom S4Vectors isEmpty
setMethod("isEmpty", "ChromBackend", function(x) {
    lengths(x) == 0L
})

#' @exportMethod isReadOnly
#'
#' @importMethodsFrom ProtGenerics isReadOnly
#'
#' @rdname ChromBackend
setMethod("isReadOnly", "ChromBackend", function(object) FALSE)

#' @exportMethod length
#'
#' @rdname ChromBackend
setMethod("length", "ChromBackend", function(x) {
    nrow(chromData(x, columns = "dataOrigin"))
})

#' @exportMethod lengths
#'
#' @rdname ChromBackend
setMethod("lengths", "ChromBackend", function(x) {
    lengths(intensity(x))
})

#' @exportMethod msLevel
#'
#' @importMethodsFrom ProtGenerics msLevel
#'
#' @rdname ChromBackend
setMethod("msLevel", "ChromBackend", function(object) {
    chromData(object, columns = "msLevel", drop = TRUE)
})

#' @exportMethod msLevel<-
#'
#' @importMethodsFrom ProtGenerics msLevel<-
#'
#' @rdname ChromBackend
setReplaceMethod("msLevel", "ChromBackend", function(object, value) {
    object$msLevel <- value
    object
})

#' @exportMethod mz
#'
#' @importMethodsFrom ProtGenerics mz
#'
#' @rdname ChromBackend
setMethod("mz", "ChromBackend", function(object) {
    chromData(object, columns = "mz", drop = TRUE)
})

#' @exportMethod mz<-
#'
#' @importMethodsFrom ProtGenerics mz<-
#'
#' @rdname ChromBackend
setReplaceMethod("mz", "ChromBackend", function(object, value) {
    object$mz <- value
    object
})

#' @exportMethod mzMax
#'
#' @rdname ChromBackend
setMethod("mzMax", "ChromBackend", function(object) {
    chromData(object, columns = "mzMax", drop = TRUE)
})

#' @exportMethod mzMax<-
#'
#' @rdname ChromBackend
setReplaceMethod("mzMax", "ChromBackend", function(object, value) {
    object$mzMax <- value
    object
})

#' @exportMethod mzMin
#'
#' @rdname ChromBackend
setMethod("mzMin", "ChromBackend", function(object) {
    chromData(object, columns = "mzMin", drop = TRUE)
})

#' @exportMethod mzMin<-
#'
#' @rdname ChromBackend
setReplaceMethod("mzMin", "ChromBackend", function(object, value) {
    object$mzMin <- value
    object
})

#' @exportMethod peaksData
#'
#' @importMethodsFrom ProtGenerics peaksVariables
#'
#' @rdname ChromBackend
setMethod("peaksVariables", "ChromBackend", function(object) {
    names(corePeaksVariables())
})

#' @exportMethod precursorMz
#'
#' @importMethodsFrom ProtGenerics precursorMz
#'
#' @rdname ChromBackend
setMethod("precursorMz", "ChromBackend", function(object) {
    chromData(object, columns = "precursorMz", drop = TRUE)
})

#' @exportMethod precursorMz<-
#'
#' @importMethodsFrom ProtGenerics precursorMz<-
#'
#' @rdname ChromBackend
setReplaceMethod("precursorMz", "ChromBackend", function(object, value) {
    object$precursorMz <- value
    object
})

#' @exportMethod precursorMzMax
#'
#' @rdname ChromBackend
setMethod("precursorMzMax", "ChromBackend", function(object) {
    chromData(object, columns = "precursorMzMax", drop = TRUE)
})

#' @exportMethod precursorMzMax<-
#'
#' @rdname ChromBackend
setReplaceMethod("precursorMzMax", "ChromBackend", function(object, value) {
    object$precursorMzMax <- value
    object
})

#' @exportMethod precursorMzMin
#'
#' @rdname ChromBackend
setMethod("precursorMzMin", "ChromBackend", function(object) {
    chromData(object, columns = "precursorMzMin", drop = TRUE)
})

#' @exportMethod precursorMzMin<-
#'
#' @rdname ChromBackend
setReplaceMethod("precursorMzMin", "ChromBackend", function(object, value) {
    object$precursorMzMin <- value
    object
})

#' @exportMethod productMz
#'
#' @importMethodsFrom ProtGenerics productMz
#'
#' @rdname ChromBackend
setMethod("productMz", "ChromBackend", function(object) {
    chromData(object, columns = "productMz", drop = TRUE)
})

#' @exportMethod productMz<-
#'
#' @importMethodsFrom ProtGenerics productMz<-
#'
#' @rdname ChromBackend
setReplaceMethod("productMz", "ChromBackend", function(object, value) {
    object$productMz <- value
    object
})

#' @exportMethod productMzMax
#'
#' @rdname ChromBackend
setMethod("productMzMax", "ChromBackend", function(object) {
    chromData(object, columns = "productMzMax", drop = TRUE)
})

#' @exportMethod productMzMax<-
#'
#' @rdname ChromBackend
setReplaceMethod("productMzMax", "ChromBackend", function(object, value) {
    object$productMzMax <- value
    object
})

#' @exportMethod productMzMin
#'
#' @rdname ChromBackend
setMethod("productMzMin", "ChromBackend", function(object) {
    chromData(object, columns = "productMzMin", drop = TRUE)
})

#' @exportMethod productMzMin<-
#'
#' @rdname ChromBackend
setReplaceMethod("productMzMin", "ChromBackend", function(object, value) {
    object$productMzMin <- value
    object
})

#' @exportMethod reset
#'
#' @rdname ChromBackend
setMethod("reset", "ChromBackend", function(object) {
    object
})

#' @exportMethod rtime
#'
#' @importMethodsFrom ProtGenerics rtime
#'
#' @rdname ChromBackend
setMethod("rtime", "ChromBackend", function(object) {
    if (length(object)) {
        peaksData(object, columns = "rtime", drop = TRUE)
    } else {
        list()
    }
})

#' @exportMethod rtime<-
#'
#' @importMethodsFrom ProtGenerics rtime<-
#'
#' @rdname ChromBackend
setReplaceMethod("rtime", "ChromBackend", function(object, value) {
    pd <- peaksData(object)
    if (!is.list(value) || length(pd) != length(value)) {
        stop("'value' should be a list of the same length as 'object'")
    }
    for (i in seq_along(pd)) {
        if (length(value[[i]]) != nrow(pd[[i]])) {
            stop(
                "Length of 'value[[", i, "]]' does not match ",
                "the number of rows in 'the rtime of chromatogram: ", i, "'"
            )
        }
    }
    peaksData(object) <- lapply(seq_along(pd), function(i) {
        pd[[i]]$rtime <- value[[i]]
        return(pd[[i]])
    })
    object
})

#' @exportMethod split
#'
#' @importMethodsFrom S4Vectors split
#'
#' @rdname ChromBackend
setMethod("split", "ChromBackend", function(x, f, drop = FALSE, ...) {
    split.default(x, f, drop = drop, ...)
})

###############################################################################
## Filter functions

#' @exportMethod filterChromData
#'
#' @rdname ChromBackend
#'
#' @importFrom MsCoreUtils between
#'
setMethod(
    "filterChromData", "ChromBackend",
    function(object, variables = character(),
             ranges = numeric(), match = c("any", "all"),
             keep = TRUE) {
        if (!length(variables) || !length(ranges))
            return(object)
        if (!is.numeric(ranges))
            stop("filterChromData only support filtering for numerical ",
                 "'variables'")
        match <- match.arg(match)
        if (is.character(variables)) {
            if (!all(variables %in% chromVariables(object)))
                stop("One or more values passed with parameter ",
                     "'variables' are not available as chromatogram ",
                     "variables in object. Use the 'chromVariables()' ",
                     "function to list possible values." )
        } else
            stop("The 'variables' parameter needs to be of type ",
                 "'character'." )
        if (length(variables) != length(ranges) / 2)
            stop("Length of 'ranges' needs to be twice the length of ",
                 "the parameter 'variables' and define the lower ",
                 "and upper bound for values of each chromatogram ",
                 "variable defined with parameter 'variables'." )
        query <- chromData(object, columns = variables)
        idx <- .filter_ranges(query, ranges, match)
        if (keep) return(object[idx])
        else {
            if (length(idx)) return(object[-idx])
            else return(object)
        }
    }
)

#' @exportMethod filterEmptyChromatograms
#'
#' @rdname ChromBackend
setMethod(
    "filterEmptyChromatograms", "ChromBackend",
    function(object, ...) {
        if (!length(object)) return(object)
        object[as.logical(lengths(object))]
    }
)

#' @exportMethod filterPeaksData
#'
#' @rdname ChromBackend
#'
#' @description
#' Filter the peak data based on the provided ranges for the given variables.
#'
#' @export
setMethod(
    "filterPeaksData", "ChromBackend",
    function(object, variables = character(),
             ranges = numeric(), match = c("any", "all"),
             keep = TRUE) {
        if (!length(ranges) || !length(variables))
            return(object)
        if (!is.numeric(ranges))
            stop( "filterPeaksData only support filtering for ",
                  "numerical peak variables")
        match <- match.arg(match)
        if (is.character(variables)) {
            if (!all(variables %in% peaksVariables(object)))
                stop("One or more values passed with parameter ",
                     "'variables' are not available as peaks ",
                     "variables in object. Use the 'peaksVariables()' ",
                     "function to list possible values.")
        } else
            stop("The 'variables' parameter needs to be of type ",
                 "'character'.")
        if (length(variables) != length(ranges) / 2)
            stop("Length of 'ranges' needs to be twice the length of ",
                 "the parameter 'variables' and define the lower and ",
                 "upper bound for values of each peak variable ",
                 "defined with parameter 'variables'.")
        if (keep) sel_fun <- function(z, idx) z[idx, , drop = FALSE]
        else sel_fun <- function(z, idx) {
            if (!length(idx)) return(z)
            else return(z[-idx, , drop = FALSE])
        }
        peaksData(object) <- lapply(peaksData(object), function(pd) {
            sel_fun(pd, .filter_ranges(
                pd[, variables, drop = FALSE],
                ranges, match
            ))
        })
        object
    }
)

#' @importMethodsFrom ProtGenerics supportsSetBackend
#'
#' @exportMethod supportsSetBackend
#'
#' @rdname ChromBackend
setMethod("supportsSetBackend", "ChromBackend", function(object, ...) FALSE)


#' @exportMethod imputePeaksData
#' @importFrom stats approx filter
#' @rdname ChromBackend
setMethod("imputePeaksData", signature(object = "ChromBackend"),
          function(object,
                   method = c("linear", "spline", "gaussian", "loess"),
                   span = 0.3,
                   sd = 1,
                   window = 2,
                   extrapolate = FALSE,
                   ...) {

              method <- match.arg(method)
              if (!length(object) || all(lengths(object) == 0L))
                  return(object)
              object$intensity <- lapply(object$intensity, .impute,
                           method = method,
                           span = span, window = window, sd = sd,
                           extrapolate = extrapolate)
              object
          })

#' @title Efficiently processing `Chromatograms` objects.
#'
#' @name processingQueue
#'
#' @description
#' The `processingQueue` of a `Chromatograms` object is a list of processing
#' steps (i.e., functions) that are stored within the object and applied only
#' when needed. This design allows data to be processed in a single step,
#' which is particularly useful for larger datasets. The processing queue
#' enables functions to be applied in a chunk-wise manner, facilitating
#' parallel processing and reducing memory demand.
#'
#' Since the peaks data can be quite large, a processing queue is used to
#' ensure efficiency. Generally, the processing queue is applied either
#' temporarily when calling `peaksData()` or permanently when calling
#' `applyProcessing()`. As explained below the processing efficiency can be
#' further improved by enabling chunk-wise processing.
#'
#' @section Apply Processing:
#'
#' The `applyProcessing()` function applies the processing queue to the backend
#' and returns the updated `Chromatograms` object. The processing queue is a
#' list of processing steps applied to the chromatograms data. Each element in
#' the list is a function that processes the chromatograms data. To apply
#' processing to the peaks data, the backend must be set to a non-read-only
#' backend using the `setBackend()` function.
#'
#' @section Parallel and Chunk-wise Processing of `Chromatograms`:
#'
#' Many operations on `Chromatograms` objects, especially those involving the
#' actual peaks data (see [peaksData]), support chunk-wise processing. This
#' involves splitting the `Chromatograms` into smaller parts (chunks) that are
#' processed iteratively. This enables parallel processing by data chunk and
#' reduces memory demand since only the peak data of the currently processed
#' subset is loaded into memory. Chunk-wise processing, which is disabled by
#' default, can be enabled by setting the processing chunk size of a
#' `Chromatograms` object using the `processingChunkSize()` function to a value
#' smaller than the length of the `Chromatograms` object. For example, setting
#' `processingChunkSize(chr) <- 1000` will cause any data manipulation
#' operation on `chr`, such as `filterPeaksData()`, to be performed in parallel
#' for sets of 1000 chromatograms in each iteration.
#'
#' Chunk-wise processing is particularly useful for `Chromatograms` objects
#' using an *on-disk* backend or for very large experiments. For small datasets
#' or `Chromatograms` using an in-memory backend, direct processing might be
#' more efficient. Setting the chunk size to `Inf` will disable chunk-wise
#' processing.
#'
#' Some backends may prefer a specific type of splitting and chunk-wise
#' processing. For example, the `ChromBackendMzR` backend needs to load MS data
#' from the original (mzML) files, so chunk-wise processing on a per-file basis
#' is ideal. The [backendParallelFactor()][ChromBackend] function for
#' `ChromBackend` allows backends to suggest a preferred data chunking by
#' returning a `factor` defining the respective data chunks. The
#' `ChromBackendMzR` returns a `factor` based on the *dataOrigin*
#' chromatograms variable. A `factor` of length 0 is returned if no particular
#' preferred splitting is needed. The suggested chunk definition will be used
#' if no finite `processingChunkSize()` is defined. Setting the
#' `processingChunkSize` overrides `backendParallelFactor`.
#'
#' Functions to configure parallel or chunk-wise processing:
#' - `processingChunkSize()`: Gets or sets the size of the chunks for parallel
#'   or chunk-wise processing of a `Chromatograms` object. With a value of
#'   `Inf` (the default), no chunk-wise processing will be performed.
#' - `processingChunkFactor()`: Returns a `factor` defining the chunks into
#'   which a `Chromatograms` object will be split for chunk-wise (parallel)
#'   processing. A `factor` of length 0 indicates that no chunk-wise processing
#'   will be performed.
#'
#' @note
#' Some backends might not support parallel processing. For these, the
#' `backendBpparam()` function will always return a `SerialParam()` regardless
#' of how parallel processing was defined.
#'
#' @param BPPARAM Parallel setup configuration. See [BiocParallel::bpparam()]
#'        for more information.
#'
#' @param chunkSize `integer(1)` for `processingChunkFactor` defining the chunk
#'        size. The default is the value stored in the `Chromatograms` object's
#'        `processingChunkSize` slot.
#'
#' @param f `factor` defining the grouping to split the `Chromatograms` object.
#'
#' @param FUN For `addProcessing()`, a function to be added to the
#'        `Chromatograms` object's processing queue.
#'
#' @param object A `Chromatograms` object.
#'
#' @param value `integer(1)` defining the chunk size.
#'
#' @param ... Additional arguments passed to the methods.
#'
#' @return `processingChunkSize()` returns the currently defined processing
#'         chunk size (or `Inf` if it is not defined). `processingChunkFactor()`
#'         returns a `factor` defining the chunks into which `object` will be
#'         split for (parallel) chunk-wise processing or a `factor` of length 0
#'         if no splitting is defined.
#'
#' @author Johannes Rainer, Philippine Louail
#'
#' @return Refer to the individual function description for information on the
#'         return value.
#'
#' @importMethodsFrom ProtGenerics processingChunkSize processingChunkSize<-
#'   processingChunkFactor
#'
#' @examples
#' # Create a Chromatograms object
#' cdata <- data.frame(
#'     msLevel = c(1L, 1L, 1L),
#'     mz = c(112.2, 123.3, 134.4),
#'     chromIndex = c(1L, 2L, 3L)
#' )
#'
#' pdata <- list(
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     ),
#'     data.frame(
#'         rtime = c(45.1, 46.2),
#'         intensity = c(100, 80.1)
#'     ),
#'     data.frame(
#'         rtime = c(12.4, 12.8, 13.2, 14.6),
#'         intensity = c(123.3, 153.6, 2354.3, 243.4)
#'     )
#' )
#'
#' be <- backendInitialize(new("ChromBackendMemory"),
#'     chromData = cdata,
#'     peaksData = pdata
#' )
#'
#' chr <- Chromatograms(be)
#'
#' divide_intensities <- function(x, y, ...) {
#'     intensity(x) <- lapply(intensity(x), `/`, y)
#'     x
#' }
#'
#' ## Add the function to the procesing queue
#' chr <- addProcessing(chr, divide_intensities, y = 2)
#' chr
#'
#' # Apply the processing queue
#' chr <- applyProcessing(chr)
#'
NULL

#' @rdname processingQueue
#' @importMethodsFrom ProtGenerics applyProcessing
#' @export
setMethod(
    "applyProcessing", "Chromatograms",
    function(object, f = processingChunkFactor(object),
    BPPARAM = bpparam(), ...) {
        if (isReadOnly(.backend(object))) {
            stop("Cannot apply processing to a read-only backend")
        }
        queue <- .processingQueue(object)
        if (!length(queue)) {
            return(object)
        }
        object@backend <- .run_process_queue(.backend(object),
            queue = queue,
            f = f, BPPARAM, ...
        )
        object@processing <- .logging(
            .processing(object),
            "Applied processing queue with ",
            length(queue),
            " steps"
        )
        object@processingQueue <- list()
        object
    }
)

#' @importFrom ProtGenerics ProcessingStep
#' @importMethodsFrom ProtGenerics addProcessing
#' @importClassesFrom ProtGenerics ProcessingStep
#' @rdname processingQueue
setMethod("addProcessing", "Chromatograms", function(object, FUN, ...) {
    if (missing(FUN)) {
        return(object)
    }
    object@processingQueue <- c(
        .processingQueue(object),
        list(ProcessingStep(FUN, ARGS = list(...)))
    )
    validObject(object)
    object
})

#' @rdname processingQueue
#' @export
setMethod("processingChunkSize",
    signature = "Chromatograms",
    function(object, ...) object@processingChunkSize
)

#' @rdname processingQueue
#' @export
setReplaceMethod("processingChunkSize",
    signature = "Chromatograms",
    function(object, value) {
        object@processingChunkSize <- value
        object
    }
)

#' @rdname processingQueue
#' @importMethodsFrom ProtGenerics backendParallelFactor
#' @export
setMethod("processingChunkFactor",
    signature = "Chromatograms",
    function(object, chunkSize = processingChunkSize(object), ...) {
        lx <- length(object)
        if (is.finite(chunkSize) && chunkSize < lx) {
            fres <- as.factor(rep(seq_len(ceiling(lx / chunkSize)),
                each = chunkSize
            )[seq_len(lx)])
            return(fres)
        } else {
            backendParallelFactor(.backend(object))
        }
    }
)

setGeneric("Chromatograms", function(object, ...) {
    standardGeneric("Chromatograms")
})
setGeneric("chromData", function(object, ...) standardGeneric("chromData"))
setGeneric("chromData<-", function(object, value) {
    standardGeneric("chromData<-")
})
setGeneric("chromExtract", function(object, peak.table, by, ...)
    standardGeneric("chromExtract"))
setGeneric("chromIndex", function(object, ...) standardGeneric("chromIndex"))
setGeneric("chromIndex<-", function(object, value) {
    standardGeneric("chromIndex<-")
})
setGeneric("chromVariables", function(object, ...) {
    standardGeneric("chromVariables")
})
setGeneric("factorize", function(object, ...) standardGeneric("factorize"))
setGeneric("filterChromData", function(object, ...) {
    standardGeneric("filterChromData")
})
setGeneric("filterEmptyChromatograms", function(object, ...) {
    standardGeneric("filterEmptyChromatograms")
})
setGeneric("filterPeaksData", function(object, ...) {
    standardGeneric("filterPeaksData")
})
setGeneric("mzMax", function(object, ...) standardGeneric("mzMax"))
setGeneric("mzMax<-", function(object, value) standardGeneric("mzMax<-"))
setGeneric("mzMin", function(object, ...) standardGeneric("mzMin"))
setGeneric("mzMin<-", function(object, value) standardGeneric("mzMin<-"))
setGeneric("precursorMzMin", function(object, ...) {
    standardGeneric("precursorMzMin")
})
setGeneric("precursorMzMin<-", function(object, value) {
    standardGeneric("precursorMzMin<-")
})
setGeneric("precursorMzMax", function(object, ...) {
    standardGeneric("precursorMzMax")
})
setGeneric("precursorMzMax<-", function(object, value) {
    standardGeneric("precursorMzMax<-")
})
setGeneric("productMzMax", function(object, ...) {
    standardGeneric("productMzMax")
})
setGeneric("productMzMax<-", function(object, value) {
    standardGeneric("productMzMax<-")
})
setGeneric("productMzMin", function(object, ...) {
    standardGeneric("productMzMin")
})
setGeneric("productMzMin<-", function(object, value) {
    standardGeneric("productMzMin<-")
})
setGeneric("imputePeaksData", function(object, ...)
    standardGeneric("imputePeaksData"))
setGeneric("peakBoundary", function(object, threshold = 0.1,
    baselineThreshold = 0.1, baselineQuantile = 0.1, ...)
    standardGeneric("peakBoundary"))
setGeneric("compareChromatograms", function(x, y, ...)
    standardGeneric("compareChromatograms"))
#' @rdname hidden_aliases
setGeneric("reset", function(object, ...) {
    standardGeneric("reset")
}) ## needs to be moved to ProtGenerics

#' @include Chromatograms.R hidden_aliases.R

#' @title Chromatographic Peaks Metadata.
#'
#' @name chromData
#'
#' @aliases chromData
#' @aliases msLevel
#' @aliases chromIndex
#' @aliases collisionEnergy
#' @aliases dataOrigin
#' @aliases mz
#' @aliases mzMin
#' @aliases mzMax
#' @aliases precursorMz
#' @aliases precursorMzMin
#' @aliases precursorMzMax
#' @aliases productMz
#' @aliases productMzMin
#' @aliases productMzMax
#' @aliases filterChromData
#' @aliases chromVariables
#' @aliases chromData<-
#' @aliases msLevel<-
#' @aliases chromIndex<-
#' @aliases collisionEnergy<-
#' @aliases dataOrigin<-
#' @aliases mz<-
#' @aliases mzMin<-
#' @aliases mzMax<-
#' @aliases precursorMz<-
#' @aliases precursorMzMin<-
#' @aliases precursorMzMax<-
#' @aliases productMz<-
#' @aliases productMzMin<-
#' @aliases productMzMax<-
#' @aliases chromVariables<-
#'
#' @description
#'
#' As explained in the [`Chromatograms`] class documentation, the
#' `Chromatograms` object is a container for chromatogram data that includes
#' chromatographic peaks data (*retention time* and related intensity values,
#' also referred to as *peaks data variables* in the context of
#' `Chromatograms`) and metadata of individual chromatograms (so called
#'  *chromatograms variables*).
#'
#' The *chromatograms variables* information can be accessed using the
#' `chromData()` function. it is also possible to access specific
#' chromatograms variables using `$`.
#'
#' `@chromData` can be accessed, replaced but also filtered/subsetted. Refer to
#' the sections below for more details.
#'
#' @param columns A `character` vector of chromatograms variables to extract.
#'
#' @param drop A `logical` indicating whether to drop dimensions when
#'        extracting a single variable.
#'
#' @param keep For `filterChromData()`: `logical(1)`
#'        defining whether to keep (`keep = TRUE`) or remove (`keep = FALSE`)
#'        the chromatogram data that match the condition.
#'
#' @param match For `filterChromData()` : `character(1) `
#'        defining whether the condition has to match for all provided
#'        `ranges` (`match = "all"`; the default), or for any of them
#'        (`match = "any"`) for chromatogram data to be retained.
#'
#' @param ranges For `filterChromData()` : a `numeric`
#'        vector of paired values (upper and lower boundary) that define the
#'        ranges to filter the `object`. These paired values need to be in the
#'        same order as the `variables` parameter (see below).
#'
#' @param object A [Chromatograms] object.
#'
#' @param value replacement value for `<-` methods. See individual
#'        method description or expected data type.
#'
#' @param variables For `filterChromData()`: `character` vector with the names
#'        of the chromatogram variables to filter for. The list of available
#'        chromatogram variables can be obtained with `chromVariables()`.
#'
#' @param x A [Chromatograms] object.
#'
#' @section Chromatograms variables and accessor functions:
#'
#' The following chromatograms variables are guaranteed to be provided by a
#' `Chromatograms` object and to be accessible with either the `chromData()` or
#'  a specific function named after the variables names:
#'
#' - `chromIndex`: an `integer` with the index of the chromatogram in the
#'   original source file (e.g. *mzML* file).
#' - `collisionEnergy`: for SRM data, `numeric` with the collision energy of
#'   the precursor.
#' - `dataOrigin`: optional `character` with the origin of the data.
#' - `msLevel`: `integer` defining the MS level of the data.
#' - `mz`: optional `numeric` with the (target) m/z value for the
#'   chromatographic data.
#' - `mzMin`: optional `numeric` with the lower m/z value of the m/z range in
#'   case the data (e.g. an extracted ion chromatogram EIC) was extracted from
#'   a `Spectra` object.
#' - `mzMax`: optional `numeric` with the upper m/z value of the m/z range.
#' - `precursorMz`: for SRM data, `numeric` with the target m/z of the
#'   precursor (parent).
#' - `precursorMzMin`: for SRM data, optional `numeric` with the lower m/z of
#'   the precursor's isolation window.
#' - `precursorMzMax`: for SRM data, optional `numeric` with the upper m/z of
#'   the precursor's isolation window.
#' - `productMz` for SRM data, `numeric` with the target m/z of the
#'   product ion.
#' - `productMzMin`: for SRM data, optional `numeric` with the lower m/z of
#'   the product's isolation window.
#' - `productMzMax`: for SRM data, optional `numeric` with the upper m/z of
#'   the product's isolation window.
#'
#' @section Filter Chromatograms variables:
#'
#' Functions that filter `Chromatograms` based on chromatograms variables
#' (i.e, `@chromData` ) will remove chromatographic data that do not meet the
#' specified conditions. This means that if a chromatogram is filtered out, its
#' corresponding `@chromData` and `@peaksData` will be removed from the object
#' immediately.
#'
#' The available functions to filter chromatogram data are:
#'
#' - `filterChromData()`: Filters numerical chromatographic data variables
#'   based on the provided numerical `ranges`. The method returns a
#'   `Chromatograms` object containing only the chromatograms that match the
#'   specified conditions. This function results in an object with fewer
#'   chromatograms than the original.
#'
#'
#' @seealso [Chromatograms] for a general description of the `Chromatograms`
#'          object.
#'          [peaksData] for a general description of the chromatographic peaks
#'          data available in the object, as well as how to access, replace and
#'          subset them.
#'          [processingQueue] for more information on the queuing
#'          of processings and parallelization for larger dataset processing.
#' @md
#'
#' @author Philippine Louail
#'
#' @return Refer to the individual function description for information on the
#'  return value.
#'
#' @examples
#'
#' # Create a Chromatograms object
#' cdata <- data.frame(
#'     msLevel = c(1L, 1L, 1L),
#'     mz = c(112.2, 123.3, 134.4),
#'     chromIndex = c(1L, 2L, 3L)
#' )
#'
#' be <- backendInitialize(new("ChromBackendMemory"), chromData = cdata)
#'
#' chr <- Chromatograms(be)
#'
#' # Access chromatograms variables
#' chromData(chr)
#'
#' # Access specific chromatograms variables
#' chromData(chr, columns = "msLevel")
#'
#' msLevel(chr)
#'
#' # Replace chromatograms variables
#' msLevel(chr) <- c(1L, 2L, 2L)
#'
#' # Filter chromatograms variables
#' filterChromData(chr,
#'     variables = "msLevel", ranges = c(1L, 1L),
#'     keep = FALSE
#' )
#'
NULL

#' @rdname chromData
setMethod(
    "chromData", "Chromatograms",
    function(object, columns = chromVariables(object), drop = FALSE) {
        chromData(.backend(object), columns = columns, drop = drop)
    }
)

#' @rdname chromData
setReplaceMethod("chromData", "Chromatograms", function(object, value) {
    chromData(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("chromVariables", "Chromatograms", function(object) {
    chromVariables(.backend(object))
})

#' @rdname chromData
setMethod("chromIndex", "Chromatograms", function(object) {
    chromIndex(.backend(object))
})

#' @rdname chromData
setReplaceMethod("chromIndex", "Chromatograms", function(object, value) {
    chromIndex(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("collisionEnergy", "Chromatograms", function(object) {
    collisionEnergy(.backend(object))
})

#' @rdname chromData
setReplaceMethod("collisionEnergy", "Chromatograms", function(object, value) {
    collisionEnergy(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("dataOrigin", "Chromatograms", function(object) {
    dataOrigin(.backend(object))
})

#' @rdname chromData
setReplaceMethod("dataOrigin", "Chromatograms", function(object, value) {
    dataOrigin(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("msLevel", "Chromatograms", function(object) {
    msLevel(.backend(object))
})

#' @rdname chromData
setReplaceMethod("msLevel", "Chromatograms", function(object, value) {
    msLevel(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("mz", "Chromatograms", function(object) mz(.backend(object)))

#' @rdname chromData
setReplaceMethod("mz", "Chromatograms", function(object, value) {
    mz(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("mzMax", "Chromatograms", function(object) mzMax(.backend(object)))

#' @rdname chromData
setReplaceMethod("mzMax", "Chromatograms", function(object, value) {
    mzMax(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("mzMin", "Chromatograms", function(object) mzMin(.backend(object)))

#' @rdname chromData
setReplaceMethod("mzMin", "Chromatograms", function(object, value) {
    mzMin(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("length", "Chromatograms", function(x) length(x@backend))

#' @rdname chromData
setMethod("precursorMz", "Chromatograms", function(object) {
    precursorMz(.backend(object))
})

#' @rdname chromData
setReplaceMethod("precursorMz", "Chromatograms", function(object, value) {
    precursorMz(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("precursorMzMin", "Chromatograms", function(object) {
    precursorMzMin(.backend(object))
})

#' @rdname chromData
setReplaceMethod("precursorMzMin", "Chromatograms", function(object, value) {
    precursorMzMin(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("precursorMzMax", "Chromatograms", function(object) {
    precursorMzMax(.backend(object))
})

#' @rdname chromData
setReplaceMethod("precursorMzMax", "Chromatograms", function(object, value) {
    precursorMzMax(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("productMz", "Chromatograms", function(object) {
    productMz(.backend(object))
})

#' @rdname chromData
setReplaceMethod("productMz", "Chromatograms", function(object, value) {
    productMz(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("productMzMin", "Chromatograms", function(object) {
    productMzMin(.backend(object))
})

#' @rdname chromData
setReplaceMethod("productMzMin", "Chromatograms", function(object, value) {
    productMzMin(object@backend) <- value
    object
})

#' @rdname chromData
setMethod("productMzMax", "Chromatograms", function(object) {
    productMzMax(.backend(object))
})

#' @rdname chromData
setReplaceMethod("productMzMax", "Chromatograms", function(object, value) {
    productMzMax(object@backend) <- value
    object
})

#' @rdname chromData
setMethod(
    "filterChromData", "Chromatograms",
    function(object,
    variables = character(), ranges = numeric(),
    match = c("any", "all"), keep = TRUE) {
        object@backend <- filterChromData(.backend(object),
            variables = variables,
            ranges = ranges,
            match = match,
            keep = keep)
        object
    }
)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ChromBackend-functions.R, R/ChromBackend.R
\name{coreChromVariables}
\alias{coreChromVariables}
\alias{corePeaksVariables}
\alias{ChromBackend}
\alias{ChromBackend-class}
\alias{ChromBackendMemory-class}
\alias{ChromBackendMzR-class}
\alias{ChromBackendSpectra-class}
\alias{[,ChromBackend-method}
\alias{[[,ChromBackend-method}
\alias{[[<-,ChromBackend-method}
\alias{extractByIndex}
\alias{factorize}
\alias{$,ChromBackend-method}
\alias{$<-,ChromBackend-method}
\alias{backendMerge,ChromBackend-method}
\alias{chromData,ChromBackend-method}
\alias{chromData<-,ChromBackend-method}
\alias{chromExtract,ChromBackend-method}
\alias{factorize,ChromBackend-method}
\alias{peaksData,ChromBackend-method}
\alias{peaksData<-,ChromBackend-method}
\alias{backendBpparam,ChromBackend-method}
\alias{backendInitialize,ChromBackend-method}
\alias{backendParallelFactor,ChromBackend-method}
\alias{backendMerge,list-method}
\alias{chromIndex,ChromBackend-method}
\alias{chromIndex<-,ChromBackend-method}
\alias{chromVariables,ChromBackend-method}
\alias{collisionEnergy,ChromBackend-method}
\alias{collisionEnergy<-,ChromBackend-method}
\alias{dataOrigin,ChromBackend-method}
\alias{dataOrigin<-,ChromBackend-method}
\alias{extractByIndex,ChromBackend,ANY-method}
\alias{extractByIndex,ChromBackend,missing-method}
\alias{intensity,ChromBackend-method}
\alias{intensity<-,ChromBackend-method}
\alias{isEmpty,ChromBackend-method}
\alias{isReadOnly,ChromBackend-method}
\alias{length,ChromBackend-method}
\alias{lengths,ChromBackend-method}
\alias{msLevel,ChromBackend-method}
\alias{msLevel<-,ChromBackend-method}
\alias{mz,ChromBackend-method}
\alias{mz<-,ChromBackend-method}
\alias{mzMax,ChromBackend-method}
\alias{mzMax<-,ChromBackend-method}
\alias{mzMin,ChromBackend-method}
\alias{mzMin<-,ChromBackend-method}
\alias{peaksVariables,ChromBackend-method}
\alias{precursorMz,ChromBackend-method}
\alias{precursorMz<-,ChromBackend-method}
\alias{precursorMzMax,ChromBackend-method}
\alias{precursorMzMax<-,ChromBackend-method}
\alias{precursorMzMin,ChromBackend-method}
\alias{precursorMzMin<-,ChromBackend-method}
\alias{productMz,ChromBackend-method}
\alias{productMz<-,ChromBackend-method}
\alias{productMzMax,ChromBackend-method}
\alias{productMzMax<-,ChromBackend-method}
\alias{productMzMin,ChromBackend-method}
\alias{productMzMin<-,ChromBackend-method}
\alias{reset,ChromBackend-method}
\alias{rtime,ChromBackend-method}
\alias{rtime<-,ChromBackend-method}
\alias{split,ChromBackend,ANY-method}
\alias{filterChromData,ChromBackend-method}
\alias{filterEmptyChromatograms,ChromBackend-method}
\alias{filterPeaksData,ChromBackend-method}
\alias{supportsSetBackend,ChromBackend-method}
\alias{imputePeaksData,ChromBackend-method}
\title{Chromatographic MS Data Backends}
\usage{
coreChromVariables()

corePeaksVariables()

\S4method{$}{ChromBackend}(x, name)

\S4method{$}{ChromBackend}(x, name) <- value

\S4method{backendMerge}{ChromBackend}(object, ...)

\S4method{chromData}{ChromBackend}(object, columns = chromVariables(object), drop = FALSE)

\S4method{chromData}{ChromBackend}(object) <- value

\S4method{chromExtract}{ChromBackend}(object, peak.table, by)

\S4method{factorize}{ChromBackend}(object, ...)

\S4method{peaksData}{ChromBackend}(object, columns = c("rtime", "intensity"), drop = FALSE, ...)

\S4method{peaksData}{ChromBackend}(object) <- value

\S4method{[}{ChromBackend}(x, i, j, ..., drop = FALSE)

\S4method{[[}{ChromBackend}(x, i, j, ...)

\S4method{[[}{ChromBackend}(x, i, j, ...) <- value

\S4method{backendBpparam}{ChromBackend}(object, BPPARAM = bpparam())

\S4method{backendInitialize}{ChromBackend}(object, ...)

\S4method{backendParallelFactor}{ChromBackend}(object, ...)

\S4method{backendMerge}{list}(object, ...)

\S4method{chromIndex}{ChromBackend}(object)

\S4method{chromIndex}{ChromBackend}(object) <- value

\S4method{chromVariables}{ChromBackend}(object)

\S4method{collisionEnergy}{ChromBackend}(object)

\S4method{collisionEnergy}{ChromBackend}(object) <- value

\S4method{dataOrigin}{ChromBackend}(object)

\S4method{dataOrigin}{ChromBackend}(object) <- value

\S4method{extractByIndex}{ChromBackend,ANY}(object, i)

\S4method{extractByIndex}{ChromBackend,missing}(object, i)

\S4method{intensity}{ChromBackend}(object)

\S4method{intensity}{ChromBackend}(object) <- value

\S4method{isEmpty}{ChromBackend}(x)

\S4method{isReadOnly}{ChromBackend}(object)

\S4method{length}{ChromBackend}(x)

\S4method{lengths}{ChromBackend}(x)

\S4method{msLevel}{ChromBackend}(object)

\S4method{msLevel}{ChromBackend}(object) <- value

\S4method{mz}{ChromBackend}(object)

\S4method{mz}{ChromBackend}(object) <- value

\S4method{mzMax}{ChromBackend}(object)

\S4method{mzMax}{ChromBackend}(object) <- value

\S4method{mzMin}{ChromBackend}(object)

\S4method{mzMin}{ChromBackend}(object) <- value

\S4method{peaksVariables}{ChromBackend}(object)

\S4method{precursorMz}{ChromBackend}(object)

\S4method{precursorMz}{ChromBackend}(object) <- value

\S4method{precursorMzMax}{ChromBackend}(object)

\S4method{precursorMzMax}{ChromBackend}(object) <- value

\S4method{precursorMzMin}{ChromBackend}(object)

\S4method{precursorMzMin}{ChromBackend}(object) <- value

\S4method{productMz}{ChromBackend}(object)

\S4method{productMz}{ChromBackend}(object) <- value

\S4method{productMzMax}{ChromBackend}(object)

\S4method{productMzMax}{ChromBackend}(object) <- value

\S4method{productMzMin}{ChromBackend}(object)

\S4method{productMzMin}{ChromBackend}(object) <- value

\S4method{reset}{ChromBackend}(object)

\S4method{rtime}{ChromBackend}(object)

\S4method{rtime}{ChromBackend}(object) <- value

\S4method{split}{ChromBackend,ANY}(x, f, drop = FALSE, ...)

\S4method{filterChromData}{ChromBackend}(
  object,
  variables = character(),
  ranges = numeric(),
  match = c("any", "all"),
  keep = TRUE
)

\S4method{filterEmptyChromatograms}{ChromBackend}(object, ...)

\S4method{filterPeaksData}{ChromBackend}(
  object,
  variables = character(),
  ranges = numeric(),
  match = c("any", "all"),
  keep = TRUE
)

\S4method{supportsSetBackend}{ChromBackend}(object, ...)

\S4method{imputePeaksData}{ChromBackend}(
  object,
  method = c("linear", "spline", "gaussian", "loess"),
  span = 0.3,
  sd = 1,
  window = 2,
  extrapolate = FALSE,
  ...
)
}
\arguments{
\item{x}{Object extending \code{ChromBackend}.}

\item{name}{For \code{$} and \verb{$<-}: the name of the chromatogram variable to
return or set.}

\item{value}{Replacement value for \verb{<-} methods. See individual
method description or expected data type.}

\item{object}{Object extending \code{ChromBackend}.}

\item{...}{Additional arguments.}

\item{columns}{For \code{chromData()} accessor: optional \code{character} with column
names (chromatogram variables) that should be included in the
returned \code{data.frame}. By default, all columns are returned.}

\item{drop}{For \code{chromData()} and \code{peaksData()}: \code{logical(1)} default to
\code{FALSE}. If \code{TRUE}, and one column is requested by the user, the
method should return a vector (or list of vector for \code{peaksData()})
of the single column requested.}

\item{peak.table}{For \code{chromExtract()} A \verb{data frame} containing the
following minimum columns:
- rtMin: Minimum retention time for each peak. Cannot be NA.
- rtMax: Maximum retention time for each peak. Cannot be NA.
- mzMin: Minimum m/z value for each peak.
- mzMax: Maximum m/z value for each peak.
Additionally, the \code{peak.table} must include columns that uniquely
identify chromatograms in the \code{object}. Common choices are
"msLevel" and/or "dataOrigin". These columns must also be present
in the \code{chromData} of the \code{object}. Any extra columns in
\code{peak.table} will be added to the \code{chromData} of the newly created
object.}

\item{by}{for \code{chromExtract() }A \code{character} vector specifying one or more
column names that are present in both \code{peak.table} and
\code{chromData(object)}. These columns uniquely identify chromatograms.
The combination of these columns must be unique in
\code{chromData(object)}. Can be of length 1 or greater.}

\item{i}{For \code{[}: \code{integer}, \code{logical} or \code{character} to subset the object.}

\item{j}{For \code{[} and \code{[[}: ignored.}

\item{BPPARAM}{Parallel setup configuration. See \code{\link[BiocParallel:register]{BiocParallel::bpparam()}}
for more information.}

\item{f}{\code{factor} defining the grouping to split \code{x}. See \code{\link[=split]{split()}}.}

\item{variables}{For \code{filterChromData()}: \code{character} vector with the names
of the chromatogram variables to filter for. The list of available
chromatogram variables can be obtained with \code{chromVariables()}.}

\item{ranges}{For \code{filterChromData()} : a \code{numeric}
vector of paired values (upper and lower boundary) that define the
ranges to filter the \code{object}. These paired values need to be in the
same order as the \code{variables} parameter (see below).}

\item{match}{For \code{filterChromData()} : \code{character(1) }
defining whether the condition has to match for all provided
\code{ranges} (\code{match = "all"}; the default), or for any of them
(\code{match = "any"}) for chromatogram data to be retained.}

\item{keep}{For \code{filterChromData()}: \code{logical(1)}
defining whether to keep (\code{keep = TRUE}) or remove (\code{keep = FALSE})
the chromatogram data that match the condition.}

\item{method}{For \code{imputePeaksData()}: \code{character(1)}: Imputation
method ("linear", "spline", "gaussian", "loess")}

\item{span}{For \code{imputePeaksData}: \code{numeric(1)}, for the loess method:
Smoothing parameter (only used if method == "loess")}

\item{sd}{For \code{imputePeaksData}: \code{numeric(1)}, for the gaussian method:
Standard deviation for Gaussian kernel
(only used if method == "gaussian")}

\item{window}{For \code{imputePeaksData}: \code{integer}, for the gaussian method:
Half-width of Gaussian kernel window (e.g., 2 gives window size 5)}

\item{extrapolate}{For \code{imputePeaksData}: \code{logical(1)} (default \code{FALSE}).
If \code{TRUE}, missing values at the beginning and end of a chromatogram
(outside the range of observed values) will be extrapolated. If
\code{FALSE}, only interpolation is performed and leading/trailing \code{NA}
values remain \code{NA}.}
}
\value{
Refer to the individual function description for information on the
return value.
}
\description{
\code{ChromBackend} is a virtual class that defines what different backends need
to provide to be used by the \code{Chromatograms} package and classes.

The backend should provide access to the chromatographic data which mainly
consists of (paired) intensity and retention time values. Additional
chromatographic metadata such as MS level and precursor and product m/z
should also be provided.

Through their implementation different backends can be either optimized for
minimal memory requirements or performance. Each backend needs to implement
data access methods listed in section \emph{Backend functions:} below.

And example implementation and more details and descriptions are provided
in the \emph{Creating new \code{ChromBackend} classes for Chromatograms} vignette.

Currently available backends are:
\itemize{
\item \code{ChromBackendMemory}: This backend stores chromatographic data directly
in memory, making it ideal for small datasets or testing. It can be
initialized with a \code{data.frame} of chromatographic data via the
\code{chromData} parameter and a \code{list} of \code{data.frame} entries for peaks data
using the \code{peaksData} parameter. These data can be accessed with the
\code{chromData()} and \code{peaksData()} functions.
\item \code{ChromBackendMzR}: The \code{ChromBackendMzR} inherits all slots and methods
from the base \code{ChromBackendMemory} backend, providing additional
functionality for reading chromatographic data from mzML files.
\item \code{ChromBackendSpectra}: The \code{ChromBackendSpectra} inherits all slots and
methods from the base \code{ChromBackendMemory} backend, providing additional
functionality for reading chromatographic data from \code{Spectra} objects.
}

Filter the peak data based on the provided ranges for the given variables.
}
\section{Core chromatogram variables}{


The \emph{core} chromatogram variables are variables (metadata) that can/should
be provided by a backend. For each of these variables a value needs to be
returned, if none is defined, a missing value (of the correct data type)
should be returned. The names of the chromatogram variables in your current
chromatogram object are returned with the \code{chromVariables()} function.

For each core chromatogram variable a dedicated access method exists. In
contrast to the peaks data described below, a single value should be
returned for each chromatogram.

The \code{coreChromVariables()} function returns the core chromatogram variables
along with their expected (defined) data type.

The core chromatogram variables (in alphabetical order) are:
\itemize{
\item \code{chromIndex}: an \code{integer} with the index of the chromatogram in the
original source file (e.g. \emph{mzML} file). In backedn with no original
source file, this variable should be set to \code{NA_integer_}.
\item \code{collisionEnergy}: for SRM data, \code{numeric} with the collision energy of
the precursor.
\item \code{dataOrigin}: optional \code{character} with the origin of a chromatogram.
\item \code{msLevel}: \code{integer} defining the MS level of the data.
\item \code{mz}: optional \code{numeric} with the (target) m/z value for the
chromatographic data.
\item \code{mzMin}: optional \code{numeric} with the lower m/z value of the m/z range in
case the data (e.g. an extracted ion chromatogram EIC) was extracted from
a \code{Spectra} object.
\item \code{mzMax}: optional \code{numeric} with the upper m/z value of the m/z range.
\item \code{precursorMz}: for SRM data, \code{numeric} with the target m/z of the
precursor (parent).
\item \code{precursorMzMin}: for SRM data, optional \code{numeric} with the lower m/z of
the precursor's isolation window.
\item \code{precursorMzMax}: for SRM data, optional \code{numeric} with the upper m/z of
the precursor's isolation window.
\item \code{productMz} for SRM data, \code{numeric} with the target m/z of the
product ion.
\item \code{productMzMin}: for SRM data, optional \code{numeric} with the lower m/z of
the product's isolation window.
\item \code{productMzMax}: for SRM data, optional \code{numeric} with the upper m/z of
the product's isolation window.
}
}

\section{Core Peaks variables}{


Similar to the \emph{core} chromatogram variables, \emph{core} peaks variables
represent  metadata that should be provided by a backend. Each of these
variables should return a value, and if undefined, a missing value (with the
appropriate data type) is returned. The number of values for a peaks
variable in a single chromatogram can vary, from none to multiple, and may
differ between chromatograms.

The names of peaks variables in the current chromatogram object can be
obtained with the \code{peaksVariables()} function.

Each core peaks variable has a dedicated accessor method.

The \code{corePeaksVariables()} function returns the core peaks variables along
with their expected (defined) data type.

The core peaks variables, listed in the required order for \code{peaksData}, are:
\itemize{
\item \code{rtime}: A \code{numeric} vector containing retention time values.
\item \code{intensity}: A \code{numeric} vector containing intensity values.
}

They should be provided for each chromatogram in the backend,
\strong{in this order}, No NAs are allowed for the \code{rtime} values. These
characteristics will be checked with the \code{validPeaksData()} function.
}

\section{Mandatory methods}{


New backend classes \strong{must} extend the base \code{ChromBackend} class and
implement the following mandatory methods:
\itemize{
\item \code{backendInitialize()}: initialises the backend. This method is
supposed to be called right after creating an instance of the
backend class and should prepare the backend.
Parameters can be defined freely for each backend, depending on what is
needed to initialize the backend.
This method has to ensure to set the chromatogram variable \code{dataOrigin}
correctly.
\item \code{backendBpparam()}: returns the parallel processing setup supported by
the backend class. This function can be used by any higher
level function to evaluate whether the provided parallel processing
setup (or the default one returned by \code{bpparam()}) is supported
by the backend. Backends not supporting parallel processing (e.g.
because they contain a connection to a database that can not be
shared across processes) should extend this method to return only
\code{SerialParam()} and hence disable parallel processing for (most)
methods and functions. See also \code{backendParallelFactor()} for a
function to provide a preferred splitting of the backend for parallel
processing.
\item \code{backendParallelFactor()}: returns a \code{factor} defining an optimal
(preferred) way how the backend can be split for parallel processing
used for all peak data accessor or data manipulation functions.
The default implementation returns a factor of length 0 (\code{factor()})
providing thus no default splitting. \code{backendParallelFactor()} for
\code{ChromBackendMzR} on the other hand returns \code{factor(dataOrigin(object))}
hence suggesting to split the object by data file.
\item \code{chromData()}, \verb{chromData<-}: gets or sets general chromatogram metadata
(annotation). \code{chromData()} returns a \code{data.frame}, \verb{chromData<-} expects
a \code{data.frame} with the same number of rows as there are chromatograms in
\code{object}. Read-only backends might not need to implement the
replacement method \verb{chromData<-} (unless some internal caching mechanism
could be used). \code{chromData()} should be implemented with the parameter
\code{drop} set to \code{FALSE} as default. With \code{drop = FALSE} the method should
return a \code{data.frame} even if one column is requested. If \code{drop = TRUE}
is specified, the output will be a vector of the single column requested.
New backends should be implemented such as if empty, the method returns a
\code{data.frame} with 0 rows and the columns defined by \code{chromVariables()}.
By default, the function \emph{should} return at minimum the
coreChromVariables, even if NAs.
\item \code{chromExtract()}: return A new \code{Chrombackend} object containing separated
chromatographic area as individual chromatograms. The chromatographic areas
are defined by the \code{peak.table} parameter. The new object will contain
chromatograms that match the conditions defined in \code{peak.table}. If no
chromatograms match the conditions, an empty \code{ChromBackend} object
should be returned.
\item \code{extractByIndex()}: function to subset a backend to selected elements
defined by the provided index. Similar to \code{[}, this method should allow
extracting (or to subset) the data in any order. In contrast to \code{[},
however, \code{i} is expected to be an \code{integer} (while \code{[} should also
support \code{logical} and eventually \code{character}). While being apparently
redundant to \code{[}, this methods avoids package namespace errors/problems
that can result in implementations of \code{[} being not found by R (which
can happen sometimes in parallel processing using the
\code{\link[BiocParallel:SnowParam-class]{BiocParallel::SnowParam()}}). This method is used internally to
extract/subset its backend. Implementation of this method is mandatory.
\item \code{peaksData()}: returns a \code{list} of \code{data.frame} with the data
(e.g. retention time - intensity pairs) from each chromatogram. The length
of the \code{list} is equal to the number of chromatograms in \code{object}. For an
empty chromatogram a \code{data.frame} with 0 rows and two columns (named
\code{"rtime"} and \code{"intensity"}) has to be returned. The optional parameter
\code{columns}, if supported by the backend allows to define which peak
variables should be returned in each array. As default (minimum) columns
\code{"rtime"} and \code{"intensity"} have to be provided. \code{peaksData()} should be
implemented with the parameter \code{drop} set to \code{FALSE} as default.  With
\code{drop = FALSE} the method should return a \code{data.frame} even if only one
column is requested. If \code{drop = TRUE}  is specified, the output will be a
vector of the single column requested.
\item \verb{peaksData<-} replaces the peak data (retention time and intensity values)
of the backend. This method expects a \code{list} of two-dimensional arrays
(\code{data.frame}) with columns representing the peak variables.
All existing peaks data are expected to be replaced with these new values.
The length of the \code{list} has to match the number of chromatogram of
\code{object}. Note that only writeable backends need to support this method.
\item \code{[}: subset the backend. Only subsetting by element (\emph{row}/\code{i}) is
allowed. This method should be implemented as to support empty integer.
\item \code{$}, \verb{$<-}: access or set/add a single chromatogram variable (column) in
the backend.
\item \code{backendMerge()}: merges (combines) \code{ChromBackend} objects into a single
instance. All objects to be merged have to be of the same type.
}
}

\section{Optional methods with default implementations}{


Additional methods that might be implemented, but for which default
implementations are already present are:
\itemize{
\item \code{[[}
\item \code{backendParallelFactor()}: returns a \code{factor} defining an optimal
(preferred) way how the backend can be split for parallel processing
used for all \emph{peak} data accessor or data manipulation functions.
The default implementation returns a factor of length 0 (\code{factor()})
providing thus no default splitting.
\item \code{chromIndex()}: returns an \code{integer} vector with the index of the
chromatograms in the original source file.
\item \code{chromVariables()}: returns a \code{character} vector with the
available chromatogram variables (columns, fields or attributes)
available in \code{object}. Variables listed by this function are expected to
be returned (if requested) by the \code{chromData()} function.
\item \code{collisionEnergy()}, \verb{collisionEnergy<-}: gets or sets the collision
energy for the precursor (for SRM data). \code{collisionEnergy()} returns a
\code{numeric} of length equal to the number of chromatograms in \code{object}.
\item \code{dataOrigin()}, \verb{dataOrigin<-}: gets or sets the \emph{data origin} variable.
\code{dataOrigin()} returns a \code{character} of length equal to the number of
chromatograms, \verb{dataOrigin<-} expects a \code{character} of length equal
\code{length(object)}.
\item \code{filterChromData()}: filters any numerical chromatographic data variables
based on the provided numerical \code{ranges}. The method should return a
\code{ChromBackend} object with the chromatograms that match the condition.
This function will results in an object with less chromatogram than the
original.
\item \code{filterEmptyChromatograms()}: removes empty chromatograms (i.e.
chromatograms without peaks). Implementation of this method is optional
since a default implementation for \code{ChromBackend} is available.
\item \code{intensity()}: gets the intensity values from the chromatograms. Returns
a \code{list} of \code{numeric} vectors (intensity values for each
chromatogram). The length of the list is equal to the number of
chromatograms in \code{object}.
\item \verb{intensity<-}: replaces the intensity values. \code{value} has to be a \code{list}
of length equal to the number of chromatograms and the number of values
within each list element identical to the number of data pairs in each
chromatogram. Note that just writeable backends need to support this
method.
\item \code{imputePeaksData()}: Imputes missing intensity values in the
chromatographic peaks data using various methods such as linear
interpolation, spline interpolation, Gaussian kernel smoothing, or LOESS
smoothing. This method modifies the peaks data in place and returns the
same \code{ChromBackend} object with imputed values.
\item \code{isReadOnly()}: returns a \code{logical(1)} whether the backend is \emph{read
only} or does allow also to write/update data. Defaults to FALSE.
\item \code{isEmpty()}: returns a \code{logical} of length equal to the number of
chromatograms with \code{TRUE} for chromatograms without any data pairs.
\item \code{length()}: returns the number of chromatograms in the object.
\item \code{lengths()}: returns the number of data pairs (retention time and
intensity values) per chromatogram.
\item \code{msLevel()}: gets the chromatogram's MS level. Returns an \code{integer}
vector (of length equal to the number of chromatograms) with the MS
level for each chromatogram (or \code{NA_integer_} if not available).
\item \code{mz()},\verb{mz<-}: gets or sets the m/z value of the chromatograms. \code{mz()}
returns a \code{numeric} of length equal to the number of chromatograms in \code{ object}, \verb{mz<-} expects a \code{numeric} of length \code{length(object)}.
\item \code{mzMax()},\verb{mzMax<-}: gets or sets the upper m/z of the mass-to-charge
range from which a chromatogram contains signal (e.g. if the chromatogram
was extracted from MS data in spectra format and a m/z range was
provided). \code{mzMax()} returns a \code{numeric} of length equal to the number of
chromatograms in \code{object}, \verb{mzMax<-} expects a \code{numeric} of length equal
to the number of chromatograms in \code{object}.
\item \code{mzMin()},\verb{mzMin<-}: gets or sets the lower m/z of the mass-to-charge
range from which a chromatogram contains signal (e.g. if the chromatogram
was extracted from MS data in spectra format and a m/z range was
provided). \code{mzMin()} returns a \code{numeric} of length equal to the number of
chromatograms in \code{object}, \verb{mzMin<-} expects a \code{numeric} of length equal
to the number of chromatograms in \code{object}.
\item \code{peaksVariables()}: lists the available data variables for the
chromatograms. Default peak variables are \code{"rtime"} and \code{"intensity"}
(which all backends need to support and provide), but some backends
might provide additional variables.
Variables listed by this function are expected to be returned (if
requested) by the \code{peaksData()} function.
\item \code{precursorMz()},\verb{precursorMz<-}: gets or sets the (target) m/z of the
precursor (for SRM data). \code{precursorMz()} returns a \code{numeric} of length
equal to the number of chromatograms in \code{object}. \verb{precursorMz<-} expects
a \code{numeric} of length equal to the number of chromatograms.
\item \code{precursorMzMin()},\code{precursorMzMax()},\code{productMzMin()}, \code{productMzMax()}:
gets the lower and upper margin for the precursor or product isolation
windows. These functions might return the value of \code{productMz()} if the
respective minimal or maximal m/z values are not defined in \code{object}.
\item \code{productMz()},\verb{productMz<-}: gets or sets the (target) m/z of the
product (for SRM data). \code{productMz()} returns a \code{numeric} of length
equal to the number of chromatograms in \code{object}. \verb{productMz<-} expects
a \code{numeric} of length equal to the number of chromatograms.
\item \code{rtime()}: gets the retention times from the chromatograms. returns a
\code{list} of \code{numeric} vectors (retention times for each
chromatogram). The length of the returned list is equal to the number of
chromatograms in \code{object}.
\item \verb{rtime<-}: replaces the retention times. \code{value} has to be a \code{list} of
length equal to the number of chromatograms and the
number of values within each list element identical to the number of
data pairs in each chromatogram. Note that just writeable backends support
this method.
\item \code{split()}: splits the backend into a \code{list} of backends (depending on
parameter \code{f}). The default method for \code{ChromBackend} uses
\code{\link[=split.default]{split.default()}}, thus backends extending \code{ChromBackend} don't
necessarily need to implement this method.
\item \code{supportsSetBackend()}: whether a \code{ChromBackend} supports the
\code{Chromatograms} \code{setBackend()} function. The default function will
take the \code{peaksData()} and \code{chromData()} of the user's backend and pass it
to the new backend. If the backend does not support this function, it
should return \code{FALSE}. Therefore both backend in question should have a
adequate \code{peaksData()} and \code{chromData()} method as well as their
respective replacement method.
}
}

\section{Implementation notes}{


Backends extending \code{ChromBackend} \strong{must} implement all of its methods
(listed above). A guide to create new backend classes is provided as a
dedicated vignette. Additional information and an example for a backend
implementation is provided in the respective vignette.
}

\examples{

## Create a simple backend implementation
ChromBackendDummy <- setClass("ChromBackendDummy",
    contains = "ChromBackend"
)

## We will show examples on a `ChromBackendMemory` backend.
be <- ChromBackendMemory()

## The `backendInitialize()` method initializes the backend filling it with
## data. This method can take any parameters needed for the backend to
## get loaded with the data.
cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    mz = c(112.2, 123.3, 134.4),
    dataOrigin = c("mem1", "mem2", "mem3")
)

pdata <- list(
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    ),
    data.frame(
        rtime = c(45.1, 46.2),
        intensity = c(100, 80.1)
    ),
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    )
)
be <- backendInitialize(be, chromData = cdata, peaksData = pdata)

be

## Data can be accessed with the accessor methods
msLevel(be)

rtime(be)

## Even if no data was provided for all chromatogram variables, its accessor
## methods are supposed to return a value.
precursorMz(be)

## The `peaksData()` method is supposed to return data/frames of rtime and
## intensity pairs as a `list`.
peaksData(be)

## Use columns to extract specific peaks variables. Below we extract rtime
## and intensity values, but in reversed order to the default.
peaksData(be, columns = c("intensity", "rtime"))

## List available chromatographic variables
chromVariables(be)

## List available peak variables
peaksVariables(be)

## Extract multiple chromatographic variables
chromData(be, c("dataOrigin", "mz", "msLevel"))

## Single variables can also be accessed and replaced
mz(be)
mz(be) <- c(123.4, 134.5, 145.6)

be$msLevel
be$msLevel <- c(2L, 2L, 2L)

be[["rtime"]]
be[["rtime"]] <- list(
    c(12.4, 12.8, 13.2, 14.6),
    c(45.1, 46.2),
    c(12.4, 12.8, 13.2, 14.6)
)
}
\author{
Johannes Rainer, Philippine Louail
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Chromatograms.R
\name{concatenateChromatograms}
\alias{concatenateChromatograms}
\alias{c,Chromatograms-method}
\alias{split,Chromatograms,ANY-method}
\title{Merging, combining and splitting Chromatograms}
\usage{
concatenateChromatograms(x, ...)

\S4method{c}{Chromatograms}(x, ...)

\S4method{split}{Chromatograms,ANY}(x, f, drop = FALSE, ...)
}
\arguments{
\item{x}{A \code{Chromatograms} object.}

\item{...}{For \code{c()} and \code{concatenateChromatograms()}: \code{Chromatograms}
objects or a \code{list} of \code{Chromatograms} objects.}

\item{f}{\code{factor} defining the grouping to split the \code{Chromatograms} object.}

\item{drop}{For \code{split()}: not considered.}
}
\value{
\itemize{
\item \code{c()} and \code{concatenateChromatograms()}: a single \code{Chromatograms} object
containing the data from all input objects.
\item \code{split()}: a \code{list} of \code{Chromatograms} objects.
}
}
\description{
Various functions are available to combine or split data from one or more
\code{Chromatograms} objects. These are:
\itemize{
\item \code{c()} and \code{concatenateChromatograms()}: combines several \code{Chromatograms}
objects into a single object. The resulting \code{Chromatograms} contains all
data from all individual \code{Chromatograms}, i.e. the union of all their
chromatograms variables. Concatenation will fail if the processing queue
of any of the \code{Chromatograms} objects is not empty or if different backends
are used for the \code{Chromatograms} objects. In such cases it is suggested to
first change the backends of all \code{Chromatograms} to the same type of
backend (using the \code{\link[ProtGenerics:backendInitialize]{ProtGenerics::setBackend()}} function) and to eventually (if needed)
apply the processing queue using the \code{\link[ProtGenerics:processingQueue]{ProtGenerics::applyProcessing()}} function.
\item \code{split()}: splits the \code{Chromatograms} object based on a provided grouping
factor returning a \code{list} of \code{Chromatograms} objects.
}
}
\examples{

## Create two Chromatograms objects
cdata1 <- data.frame(
    msLevel = c(1L, 1L),
    mz = c(112.2, 123.3),
    dataOrigin = c("file1", "file1")
)
pdata1 <- list(
    data.frame(rtime = c(1.0, 2.0, 3.0), intensity = c(100, 200, 150)),
    data.frame(rtime = c(1.0, 2.0, 3.0), intensity = c(80, 120, 90))
)
chr1 <- Chromatograms(
    ChromBackendMemory(),
    chromData = cdata1,
    peaksData = pdata1
)

cdata2 <- data.frame(
    msLevel = c(2L, 2L),
    mz = c(134.4, 145.5),
    dataOrigin = c("file2", "file2")
)
pdata2 <- list(
    data.frame(rtime = c(4.0, 5.0, 6.0), intensity = c(300, 400, 350)),
    data.frame(rtime = c(4.0, 5.0, 6.0), intensity = c(200, 250, 180))
)
chr2 <- Chromatograms(
    ChromBackendMemory(),
    chromData = cdata2,
    peaksData = pdata2
)

## Combine using c()
chr_combined <- c(chr1, chr2)
chr_combined

## Combine using concatenateChromatograms
chr_combined2 <- concatenateChromatograms(chr1, chr2)

## Combine a list of Chromatograms
chr_list <- list(chr1, chr2)
chr_combined3 <- concatenateChromatograms(chr_list)

## Split by msLevel
chr_split <- split(chr_combined, f = chr_combined$msLevel)
chr_split

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ChromBackendMzR.R
\name{ChromBackendMzR}
\alias{ChromBackendMzR}
\alias{backendInitialize,ChromBackendMzR-method}
\title{Chromatographic Data Backend for Reading mzML Files}
\usage{
ChromBackendMzR()

\S4method{backendInitialize}{ChromBackendMzR}(object, files = character(), BPPARAM = bpparam(), ...)
}
\arguments{
\item{object}{A \code{ChromBackendMzR} object.}

\item{files}{A character vector of file paths to mzML files.}

\item{BPPARAM}{Parallel setup configuration. See \code{\link[BiocParallel:register]{BiocParallel::bpparam()}}
for more information.}

\item{...}{Additional parameters to be passed.}
}
\value{
Refer to the individual function description for information on the
return value.
}
\description{
The \code{ChromBackendMzR} inherits all slots and methods from the base
\code{ChromBackendMemory} backend, providing additional functionality for reading
chromatographic data from mzML files.

Unlike the \code{ChromBackendMemory} backend, the \code{ChromBackendMzR} backend
should have the \emph{dataOrigin} chromatographic variables populated with the
file path of the mzML file from which the chromatographic data was read.

Note that the \code{ChromBackendMzR} backend is read-only and does not support
direct modification of chromatographic data. However, it does support
\code{peaksData} slot replacement, which will modify the \verb{@peaksData} slot but not
the local mzML files. This is indicated by the "inMemory" slot being set to
TRUE.

Implementing functionalities with the \code{ChromBackendMzR} backend should be
simplified as much as possible and reuse the methods already implemented for
\code{ChromBackendMemory} when possible.
}
\examples{
library(mzR)
library(msdata)

## Load an mzML file
MRM_file <- system.file("proteomics", "MRM-standmix-5.mzML.gz",
    package = "msdata"
)

## Initialize the ChromBackendMzR object
be_empty <- ChromBackendMzR()
be <- backendInitialize(be_empty, files = MRM_file, BPPARAM = SerialParam())

}
\author{
Philippine Louail
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ChromBackendSpectra.R
\name{ChromBackendSpectra}
\alias{ChromBackendSpectra}
\alias{backendInitialize,ChromBackendSpectra-method}
\alias{chromSpectraIndex}
\title{Chromatographic Data Backend for Spectra Objects}
\usage{
ChromBackendSpectra()

\S4method{backendInitialize}{ChromBackendSpectra}(
  object,
  spectra = Spectra(),
  factorize.by = c("msLevel", "dataOrigin"),
  summarize.method = c("sum", "max"),
  chromData = fillCoreChromVariables(),
  spectraVariables = character(),
  ...
)

chromSpectraIndex(object)
}
\arguments{
\item{object}{A \code{ChromBackendSpectra} object.}

\item{spectra}{A \code{Spectra} object.}

\item{factorize.by}{A \code{character} vector of \code{spectraVariables} for grouping
\code{Spectra} data into chromatographic data (i.e., creating separate
chromatograms for each unique combination of these variables).
Default: \code{c("msLevel", "dataOrigin")}, which creates one chromatogram
per MS level per data file.
If \code{chromData} is provided, it \strong{must} also contain these columns.}

\item{summarize.method}{A \code{character} string specifying intensity summary:
\code{"sum"} (default) or \code{"max"}.}

\item{chromData}{A \code{data.frame} with chromatographic data for use in
\code{backendInitialize()}. If missing, a default is generated. Columns
like \code{rtMin}, \code{rtMax}, \code{mzMin}, and \code{mzMax} must be provided and not
contain \code{NA} values. Use \code{-Inf/Inf} for unspecified values. The
\code{"dataOrigin"} column must match the \code{Spectra} object's
\code{"dataOrigin"}.}

\item{spectraVariables}{A \code{character} vector specifying which variables
from the \code{Spectra} object should be added to the chromData. These
will be mapped using the \code{chromSpectraIndex} variable.}

\item{...}{Additional parameters.}
}
\value{
Refer to the individual function description for information on the
return value.
}
\description{
The \code{ChromBackendSpectra} class extends \code{ChromBackendMemory}, inheriting
all its slots and methods while providing additional functionality for
summarizing chromatographic data from \code{\link[Spectra:Spectra]{Spectra::Spectra()}} objects.

It can be initialized with a \code{Spectra} object, which is stored in the
\code{spectra} slot of the backend. Users can also provide a \code{data.frame}
containing chromatographic metadata, stored in \verb{@chromData}. This metadata
filters the \code{Spectra} object and generates \code{peaksData}. If \code{chromData} is
not provided, a default \code{data.frame} is created from the \code{Spectra} data.
An "rtMin", "rtMax", "mzMin", and "mzMax" column will be created by
condensing the \code{Spectra} data corresponding to each unique combination of
the \code{factorize.by} variables.

By "factorization" we mean the process of grouping spectra
into chromatograms based on specified variables. For example, using
\code{factorize.by = c("msLevel", "dataOrigin")} means that all MS1 spectra from
file "A" form one chromatogram, all MS2 spectra from file "A" form another,
and so on. Each unique combination of the factorization variables creates
a separate chromatogram. This is essential for organizing spectral data into
meaningful chromatographic traces that can be visualized and analyzed.

The \emph{dataOrigin} core chromatogram variable should reflect the \emph{dataOrigin}
of the \code{Spectra} object. The \code{factorize.by} parameter defines the variables
for grouping \code{Spectra} data into chromatographic data. The default is
\code{c("msLevel", "dataOrigin")}, which will define separate chromatograms for
each combination of \code{msLevel} and \code{dataOrigin}. These variables must be in
both the \code{spectraData()} of the \code{Spectra} and \code{chromData} (if provided).

The \code{summarize.method} parameter defines how spectral data intensity is
summarized:
\itemize{
\item \strong{"sum"}: Sums intensity to create a Total Ion Chromatogram (TIC).
\item \strong{"max"}: Takes max intensity for a Base Peak Chromatogram (BPC).
}

If \code{chromData} or its factorization columns are modified, the \code{factorize()}
method must be called to update \code{chromSpectraIndex}.
}
\details{
No \code{peaksData} is stored until the user calls a function that generates it
(e.g., \code{rtime()}, \code{peaksData()}, \code{intensity()}). The \verb{@peaksData} slot
replacement is unsupported — modifications are temporary to optimize memory.
The \verb{@inMemory} slot indicates this with \code{TRUE}.

\strong{Spectra Sort Index}: The \code{ChromBackendSpectra} backend maintains a
\code{spectraSortIndex} slot that stores a sort order for the internal \code{Spectra}
object based on \code{dataOrigin} and \code{rtime}. To optimize performance, the sort
index is only computed and stored when the spectra are unsorted; if already
sorted (which is typical for most real-world data), \code{spectraSortIndex} remains
empty (\code{integer()}). This avoids unnecessary subsetting operations. The sort
index is automatically recalculated whenever the \code{factorize()} method is called,
ensuring it remains valid and consistent. This approach avoids the need to
physically reorder disk-backed \code{Spectra} objects, which would require loading
all data into memory.

\strong{Factorize and Subsetting}: The \code{factorize()} method updates the
\code{chromSpectraIndex} in both \code{chromData} and the \verb{@spectra} to reflect
the current grouping, and recalculates \code{spectraSortIndex} to maintain the
correct sort order. The \code{[} subsetting operator properly handles subsetting
of both \verb{@chromData}, \verb{@peaksData}, and \verb{@spectra}, while updating the
\code{spectraSortIndex} to reference valid positions in the subsetted data.

\code{ChromBackendSpectra} should reuse \code{ChromBackendMemory} methods whenever
possible to keep implementations simple.
}
\examples{
library(Spectra)
library(MsBackendMetaboLights)

## Get Spectra data from MetaboLights
be <- backendInitialize(MsBackendMetaboLights(),
    mtblsId = "MTBLS39",
    filePattern = c("63B.cdf")
)
s <- Spectra(be)

s <- setBackend(s, MsBackendMemory())

## Initialize ChromBackendSpectra
be_empty <- new("ChromBackendSpectra")
be <- backendInitialize(be_empty, s)

## replace the msLevel data
msLevel(be) <- c(1L, 2L, 3L)

## re-factorize the data
be <- factorize(be)

## Create BPC : we summarize the intensity present in the Spectra object
## by the maximum value, thus creating a Base Peak Chromatogram.
be <- backendInitialize(be_empty, s, summarize.method = "max")

## Can now see the details of this bpc by looking at the chromData of our
## object
chromData(be)

## Another possibilities is to create eics from the Spectra object.
## Here we create an EIC with a specific m/z and retention time window.
df <- data.frame(mzMin = 100.01, mzMax = 100.02 , rtMin = 50, rtMax = 100)
be <- backendInitialize(be_empty, s, summarize.method = "sum")
chromData(be) <- cbind(chromData(be), df)

## now when we call the peaksData function, we will get the intensity
## of the spectra object that are in the m/z and retention time window
## defined in the chromData.
peaksData(be)

}
\author{
Philippine Louail, Johannes Rainer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Chromatograms-plot.R
\name{plotChromatograms}
\alias{plotChromatograms}
\alias{plotChromatogramsOverlay}
\title{Plot chromatograms}
\usage{
plotChromatograms(
  x,
  xlab = "rtime (s)",
  ylab = "intensity",
  type = "o",
  pch = 20,
  cex = 0.6,
  lwd = 1.5,
  xlim = numeric(),
  ylim = numeric(),
  main = character(),
  col = "#00000080",
  asp = 1,
  ...
)

plotChromatogramsOverlay(
  x,
  xlab = "rtime (s)",
  ylab = "intensity",
  type = "o",
  pch = 20,
  cex = 0.6,
  lwd = 1.5,
  xlim = numeric(),
  ylim = numeric(),
  main = paste(length(x), "chromatograms"),
  col = "#00000080",
  axes = TRUE,
  frame.plot = axes,
  ...
)
}
\arguments{
\item{x}{A \link{Chromatograms} object.}

\item{xlab}{\code{character(1)} with the label for the x-axis (by default
\code{xlab = "rtime (s)"}).}

\item{ylab}{\code{character(1)} with the label for the y-axis (by default
\code{ylab = "intensity"}).}

\item{type}{\code{character(1)} specifying the type of plot. See \code{\link[=plot.default]{plot.default()}}
for details. Defaults to \code{type = "l"} which draws each peak as a line.}

\item{pch}{\code{integer(1)} or \code{character(1)} specifying the plotting symbol
(see \code{\link[=plot.default]{plot.default()}}).}

\item{cex}{\code{numeric(1)} specifying the size of the plotting symbol (see
\code{\link[=plot.default]{plot.default()}}).}

\item{lwd}{\code{numeric(1)} specifying the line width (see \code{\link[=plot.default]{plot.default()}}).}

\item{xlim}{\code{numeric(2)} defining the x-axis limits. The range of m/z values
are used by default.}

\item{ylim}{\code{numeric(2)} defining the y-axis limits. The range of intensity
values are used by default.}

\item{main}{\code{character(1)} with the title for the plot. By default the
spectrum's MS level and retention time (in seconds) is used.}

\item{col}{color to be used to draw the peaks. Should be either of length 1,
or equal to the number of chromatograms (to plot each chromatograms
in a different color) or be a \code{list} with colors for each individual
peak in each spectrum.}

\item{asp}{\code{numeric(1)} the aspect ratio of the plot, i.e. the ratio of
the y-axis to the x-axis. Defaults to 1.}

\item{...}{Additional arguments to be passed to \code{\link[=plot.default]{plot.default()}}.}

\item{axes}{\code{logical(1)} whether (x and y) axes should be drawn.}

\item{frame.plot}{\code{logical(1)} whether a box should be drawn around the
plotting area.}
}
\value{
These functions create a plot.

Refer to the individual function description for information on the
return value.
}
\description{
\code{\link[=Chromatograms]{Chromatograms()}} can be plotted with the following functions:

The \code{plotChromatograms()}: plots each chromatogram in its separate plot by
splitting the plot area into as many panels as there are spectra.
}
\examples{

## Create a Chromatograms object
cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    mz = c(112.2, 123.3, 134.4),
    chromIndex = c(1L, 2L, 3L)
)
pdata <- list(
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    ),
    data.frame(
        rtime = c(45.1, 46.2),
        intensity = c(100, 80.1)
    ),
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    )
)
chr <- backendInitialize(ChromBackendMemory(),
    chromData = cdata,
    peaksData = pdata
) |> Chromatograms()

## Plot one chromatogram
plotChromatograms(chr[1])

## Plot the full Chromatograms object
plotChromatograms(chr)

## Define a color for each peak in each chromatogram
plotChromatograms(chr[1:2], col = c("green", "blue"))

## Overlay all chromatograms
plotChromatogramsOverlay(chr[1:2], col = c("green", "blue"))

}
\author{
Philippine Louail, Johannes Rainer.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Chromatograms-peaksData.R
\name{peaksData}
\alias{peaksData}
\alias{peaksVariables}
\alias{filterPeaksData}
\alias{imputePeaksData}
\alias{peakBoundary}
\alias{compareChromatograms}
\alias{matchRtime}
\alias{imputePeaksData,Chromatograms-method}
\alias{filterPeaksData,Chromatograms-method}
\alias{intensity,Chromatograms-method}
\alias{intensity<-,Chromatograms-method}
\alias{peaksData,Chromatograms-method}
\alias{peaksData<-,Chromatograms-method}
\alias{peaksVariables,Chromatograms-method}
\alias{rtime,Chromatograms-method}
\alias{rtime<-,Chromatograms-method}
\alias{lengths,Chromatograms-method}
\alias{compareChromatograms,Chromatograms,Chromatograms-method}
\alias{compareChromatograms,Chromatograms,missing-method}
\alias{peakBoundary,Chromatograms-method}
\title{Chromatographic peaks data}
\usage{
\S4method{imputePeaksData}{Chromatograms}(
  object,
  method = c("linear", "spline", "gaussian", "loess"),
  span = 0.3,
  sd = 1,
  window = 2,
  extrapolate = FALSE,
  ...
)

\S4method{filterPeaksData}{Chromatograms}(
  object,
  variables = character(),
  ranges = numeric(),
  match = c("any", "all"),
  keep = TRUE
)

\S4method{intensity}{Chromatograms}(object, ...)

\S4method{intensity}{Chromatograms}(object) <- value

\S4method{peaksData}{Chromatograms}(
  object,
  columns = peaksVariables(object),
  f = processingChunkFactor(object),
  BPPARAM = bpparam(),
  drop = FALSE,
  ...
)

\S4method{peaksData}{Chromatograms}(object) <- value

\S4method{peaksVariables}{Chromatograms}(object, ...)

\S4method{rtime}{Chromatograms}(object, ...)

\S4method{rtime}{Chromatograms}(object) <- value

\S4method{lengths}{Chromatograms}(x)

matchRtime(x, y, tolerance = Inf, ...)

\S4method{compareChromatograms}{Chromatograms,Chromatograms}(
  x,
  y,
  MAPFUN = matchRtime,
  FUN = cor,
  ...,
  minPeaks = 4L,
  BPPARAM = SerialParam()
)

\S4method{compareChromatograms}{Chromatograms,missing}(
  x,
  y = NULL,
  MAPFUN = matchRtime,
  FUN = cor,
  ...,
  minPeaks = 4L,
  labelsColumn = NULL,
  BPPARAM = SerialParam()
)

\S4method{peakBoundary}{Chromatograms}(
  object,
  threshold = 0.1,
  baselineThreshold = 0.1,
  baselineQuantile = 0.1,
  ...
)
}
\arguments{
\item{object}{A \link{Chromatograms} object.}

\item{method}{For \code{imputePeaksData()}: \code{character(1)}: Imputation
method ("linear", "spline", "gaussian", "loess").}

\item{span}{For \code{imputePeaksData}: \code{numeric(1)}, for the loess method:
Smoothing parameter (only used if method == "loess")}

\item{sd}{For \code{imputePeaksData}: \code{numeric(1)}, for the gaussian method:
Standard deviation for Gaussian kernel
(only used if method == "gaussian")}

\item{window}{For \code{imputePeaksData}: \code{integer}, for the gaussian method:
Half-width of Gaussian kernel window (e.g., 2 gives window size 5)}

\item{extrapolate}{For \code{imputePeaksData}: \code{logical(1)} (default \code{FALSE}).
If \code{TRUE}, missing values at the beginning and end of a chromatogram
(outside the range of observed values) will be extrapolated. If
\code{FALSE}, only interpolation is performed and leading/trailing \code{NA}
values remain \code{NA}.}

\item{...}{Additional arguments passed to the method.}

\item{variables}{For \code{filterPeaksData()}: \code{character} vector with the names
of the peaks data variables to filter for. The list of available
peaks data variables can be obtained with \code{peaksVariables()}.}

\item{ranges}{For \code{filterPeaksData()} : a \code{numeric} vector of paired values
(upper and lower boundary) that define the ranges to filter the
\code{object}. These paired values need to be in the same order as the
\code{variables} parameter (see below).}

\item{match}{For \code{filterPeaksData()} : \code{character(1) } defining whether the
condition has to match for all provided \code{ranges} (\code{match = "all"};
the default), or for any of them (\code{match = "any"}).}

\item{keep}{For \code{filterPeaksData()}: \code{logical(1)} defining whether to
keep (\code{keep = TRUE}) or remove (\code{keep = FALSE}) the chromatographic
peaks data that match the condition.}

\item{value}{For \code{rtime()} and \code{intensity()}: \code{numeric} vector with the
values to replace the current values. The length of the vector must
match the number of peaks data pairs in the \code{Chromatograms} object.}

\item{columns}{For \code{peaksData()}: optional \code{character} with column
names (peaks variables) that should be included in the
returned \code{list} of \code{data.frame}. By default, all columns are returned.
Available variables can be found by calling \code{peaksVariables()} on the
object.}

\item{f}{\code{factor} defining the grouping to split the \code{Chromatograms} object.}

\item{BPPARAM}{Parallel setup configuration. See \code{\link[BiocParallel:register]{BiocParallel::bpparam()}}
for more information.}

\item{drop}{\code{logical(1)} For \code{peaksData()}, default to \code{FALSE}. If \code{TRUE},
and one column is called by the user, the method returns a list of
vector of the single column requested.}

\item{x}{For \code{lengths()} and \code{compareChromatograms()}: A \link{Chromatograms}
object. For \code{matchRtime()}: a \code{data.frame} with columns \code{rtime} and
\code{intensity} representing the first chromatogram.}

\item{y}{For \code{compareChromatograms()}: A \link{Chromatograms} object against
which \code{x} is compared. If missing, each chromatogram in \code{x} is
compared with each other chromatogram in \code{x}. For \code{matchRtime()}:
a \code{data.frame} with columns \code{rtime} and \code{intensity} representing
the second chromatogram.}

\item{tolerance}{For \code{matchRtime()}: \code{numeric(1)} (default \code{Inf}). Maximum
RT difference between two measured points to be considered a match.
Controls both the overlap detection and the shared RT grid. Lower
values prevent a peak from being compared against a long interpolated
gap in the other chromatogram. Use \code{Inf} (the default) to consider
all RT points as matching. Can be forwarded via \code{...} in
\code{compareChromatograms()}.}

\item{MAPFUN}{For \code{compareChromatograms()}: \code{function} to align the
retention times of two chromatograms before computing similarity.
Must accept two \code{data.frame}s (with columns \code{rtime} and \code{intensity})
and return a \code{list} with elements \code{x} and \code{y}: numeric vectors of
equal length containing the aligned intensities of the first and
second chromatogram respectively, interpolated onto a common
retention-time grid. Defaults to \code{\link[=matchRtime]{matchRtime()}}. Additional
arguments can be passed via \code{...}.}

\item{FUN}{For \code{compareChromatograms()}: \code{function} to compute the
similarity between two chromatograms from their aligned intensity
vectors (as returned by \code{MAPFUN}). Must accept two numeric vectors
as the first two arguments and return a single numeric value.
Defaults to \code{\link[stats:cor]{stats::cor()}} (Pearson correlation). Additional
arguments can be passed via \code{...} (e.g., \code{method = "spearman"} for
\code{\link[stats:cor]{stats::cor()}}).}

\item{minPeaks}{For \code{compareChromatograms()}: \code{integer(1)} (default \code{4L}).
Minimum number of overlapping retention-time points (as returned by
\code{MAPFUN}) required to compute a similarity score. Pairs whose
retention-time overlap contains fewer than \code{minPeaks} points return
\code{NA} in the score layer; the actual overlap count is still recorded
in the \code{n_peaks} layer. Setting \code{minPeaks = 2L} recovers the
behaviour of always computing a score whenever at least two points
overlap.}

\item{labelsColumn}{For \code{compareChromatograms()}: optional \code{character(1)}
giving the name of a chromatogram variable (column in \code{chromData()})
whose values should be used as row and column names of the returned
array. The column must contain unique values. If \code{NULL} (the
default), the array dimensions are unnamed. Only used when \code{y} is
missing.}

\item{threshold}{For \code{peakBoundary()}: \code{numeric(1)} (default \code{0.1}).
Fraction of the peak height above baseline used as a fallback
cut-off when valley-based boundaries are not suitable. Must be
\verb{>= 0} and \verb{< 1}.}

\item{baselineThreshold}{For \code{peakBoundary()}: \code{numeric(1)} (default
\code{0.1}). Fraction of the peak height above the baseline. Valley
positions returned by \code{MsCoreUtils::valleys()} are accepted only if
the intensity at the valley is at or below
\code{baseline + peak_height * baselineThreshold}. Must be \verb{>= 0} and
\verb{< 1}.}

\item{baselineQuantile}{For \code{peakBoundary()}: \code{numeric(1)} (default
\code{0.1}). Quantile of the intensity distribution used as the
baseline estimate. Must be \verb{>= 0} and \verb{<= 1}.}
}
\value{
Refer to the individual function description for information on the
return value.
}
\description{
As explained in the \link{Chromatograms} class documentation, the \code{Chromatograms}
object is a container for chromatographic data that includes chromatographic
peaks data (\emph{retention time} and related intensity values, also referred to
as \emph{peaks data variables} in the context of \code{Chromatograms}) and metadata of
individual chromatograms (so called \emph{chromatograms variables}).

The \emph{peaks data variables} information can be accessed using the
\code{peaksData()} function. It is also possible to access specific peaks
variables using \code{$}.

The peaks data can be accessed, replaced but also filtered/subsetted. Refer
to the sections below for more details.
}
\section{Filter Peaks Variables}{


Functions that filter a \code{Chromatograms}'s peaks data (i.e., \verb{@peaksData}).
These functions remove peaks data that do not meet the
specified conditions. If a chromatogram in a \code{Chromatograms} object is
filtered, only the corresponding peaks variable pairs (i.e., rows) in the
\code{peaksData} are removed, while the chromatogram itself remains in the object.

The available functions to filter chromatographic peaks data include:
\itemize{
\item \code{filterPeaksData()}: Filters numerical peaks data variables based on the
specified numerical \code{ranges} parameter. This method returns the same input
\code{Chromatograms} object, but the filtering step is added to the processing
queue. The filtered data will be reflected when the user accesses
\code{peaksData}. This function does \emph{not} reduce the number of chromatograms
in the object, but it removes the specified peaks data (e.g., "rtime" and
"intensity" pairs) from the \code{peaksData}.
}

In the case of a read-only backend, (such as the \link{ChromBackendMzR}), the
replacement of the peaks data is not possible. The peaks data can be
filtered, but the filtered data will not be saved in the backend. This means
the original mzML files will not be affected by computations performed on
the \link{Chromatograms}.
}

\section{Impute Peaks Variables}{


\code{imputePeaksData} will impute missing values in a \code{Chromatograms}'s peaks data
(i.e., \verb{@peaksData}). This functions replace missing peaks data values with
specified imputation methods using various methods such as linear
interpolation, spline interpolation, Gaussian kernel smoothing, or LOESS
smoothing. This method modifies the peaks data in place and returns the
same \code{Chromatograms} object with imputed values.
}

\section{Peak Boundary Detection}{


\code{peakBoundary()} determines the retention time boundaries of the tallest
peak in each chromatogram. The function uses \code{MsCoreUtils::valleys()} to
locate the valleys (local minima) flanking the apex. If the valley
intensities exceed a baseline-relative threshold (controlled by
\code{baselineThreshold}), it falls back to a threshold-based boundary search
using \code{threshold}. The baseline is estimated as the \code{baselineQuantile}
quantile of the chromatogram's intensity values.
The result is a \code{matrix} with one row per
chromatogram and columns \code{left_boundary} and \code{right_boundary}
(retention times). Chromatograms that are empty, have fewer than 3 data
points, contain only \code{NA} or all-zero intensities return \code{NA} for both
boundaries.
}

\section{Compare Chromatograms}{


\code{compareChromatograms()} compares chromatograms in two steps:
\enumerate{
\item \strong{Align} – \code{MAPFUN} (default \code{\link[=matchRtime]{matchRtime()}}) maps two chromatograms
onto a common retention-time grid and returns \code{list(x, y)}, where
\code{x} and \code{y} are numeric vectors of equal length containing the aligned
intensities of the first and second chromatogram respectively.
\item \strong{Score} – \code{FUN} (default \code{\link[stats:cor]{stats::cor()}}, Pearson correlation)
computes a single similarity value from those aligned intensity vectors.
}

If \code{y} is missing, each chromatogram in \code{x} is compared against every
other chromatogram in \code{x}; otherwise, each in \code{x} is compared with
each in \code{y}.

The result is a 3-dimensional numeric array with dimensions
\code{length(x)} x \code{length(y)} x 2 (or symmetric \verb{n x n x 2} for
self-comparison). Layer \verb{[, , 1]} (named \code{"score"}) contains pairwise
similarity scores; layer \verb{[, , 2]} (named \code{"n_peaks"}) contains the
number of overlapping retention-time points used to compute each score.
Pairs with fewer overlapping retention-time points than \code{minPeaks} (default
4) return \code{NA} in the score layer; the actual overlap count is still
recorded in the \code{n_peaks} layer. The diagonal of a self-comparison is
always 1 (score) and the number of data points in that chromatogram
(count).

\code{matchRtime()} is the default \code{MAPFUN}. Given two chromatograms as
\code{data.frame}s with \code{rtime} and \code{intensity} columns, it aligns their RT axes
and returns a named list with elements \code{x} and \code{y}: equal-length intensity
vectors evaluated on a shared RT grid, ready for similarity scoring.

The alignment works as follows: \code{matchRtime()} first identifies the RT range
where both chromatograms have measured points within \code{tolerance} of each
other (the \emph{overlap}). Within that range, it builds a shared RT grid from
all of \code{x}'s RT points, adding any RT points from \code{y} that have no close
match in \code{x} (within \code{tolerance}). Both intensity vectors are then linearly
interpolated at grid positions they do not natively cover, using
\code{\link[stats:approxfun]{stats::approx()}}. If either chromatogram has fewer than 2 data points, or
the two chromatograms do not overlap, empty vectors are returned.

The \code{tolerance} parameter (default \code{Inf}, meaning all RT points are
considered matching) controls the strictness of the matching. Lowering it
prevents comparing a measured peak against a long interpolated gap in the
other chromatogram. Pass \code{tolerance} via \code{...} in \code{compareChromatograms()}.

When \code{y} is missing, the \code{labelsColumn} parameter assigns meaningful
row/column names to the output from a \code{chromData()} column (e.g., \code{"mz"}
or a user-defined feature identifier). The column must contain unique
values. To compare groups of chromatograms separately, split the object
with \code{split()} beforehand and apply \code{compareChromatograms()} to each subset.
}

\examples{

# Create a Chromatograms object
cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    mz = c(112.2, 123.3, 134.4),
    dataOrigin = c("mem1", "mem2", "mem3")
)

pdata <- list(
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    ),
    data.frame(
        rtime = c(45.1, 46.2),
        intensity = c(100, 80.1)
    ),
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    )
)

be <- backendInitialize(new("ChromBackendMemory"),
    chromData = cdata,
    peaksData = pdata
)

chr <- Chromatograms(be)

# Access peaks data
peaksData(chr)

# Access specific peaks data variables
peaksData(chr, columns = "rtime")
rtime(chr)

# Replace peaks data
rtime(chr)[[1]] <- c(1, 2, 3, 4)

# Filter peaks data
filterPeaksData(chr, variables = "rtime", ranges = c(12.5, 13.5))

# Pairwise similarity: returns a 3D array [i, j, layer]
res <- compareChromatograms(chr)
res[, , "score"]   ## similarity scores
res[, , "n_peaks"] ## number of overlapping RT points

## Use Spearman correlation (passed to cor() via ...)
compareChromatograms(chr, method = "spearman")[, , "score"]

# Use a chromData column as row/column labels
compareChromatograms(chr, labelsColumn = "mz")[, , "score"]

# Compare two Chromatograms objects
compareChromatograms(chr[1:2], chr[3])

}
\seealso{
\link{Chromatograms} for a general description of the \code{Chromatograms}
object, and \link{chromData} for accessing,substituting and filtering
chromatographic variables. For more information on the queuing
of processings and parallelization for larger dataset processing
see \link{processingQueue}.
}
\author{
Philippine Louail
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ChromBackendMemory.R
\name{ChromBackendMemory}
\alias{ChromBackendMemory}
\alias{backendInitialize,ChromBackendMemory-method}
\title{Improved in-memory Chromatographic data backend}
\usage{
ChromBackendMemory()

\S4method{backendInitialize}{ChromBackendMemory}(
  object,
  chromData = fillCoreChromVariables(data.frame()),
  peaksData = list(.EMPTY_PEAKS_DATA),
  ...
)
}
\arguments{
\item{object}{A \code{ChromBackendMemory} object.}

\item{chromData}{For \code{backendInitialize()} of a \code{ChromBackendMemory}
backend, a \code{data.frame} with the chromatographic data. If not
provided (or if empty), a default \code{data.frame} with the core
chromatographicvariables will be created.}

\item{peaksData}{For \code{backendInitialize()} of a \code{ChromBackendMemory}
backend, a \code{list} of \code{data.frame} with the peaks data. If not
provided (or if empty), a default \code{list} of empty \code{data.frame} with
the core peaks variables will be created. The length of the list
should match the number of chromatograms in the \code{chromData}
parameter.}

\item{...}{Additional parameters to be passed.}
}
\value{
Refer to the individual function description for information on the
return value.
}
\description{
\code{ChromBackendMemory}: This backend stores chromatographic data directly
in memory, making it ideal for small datasets or testing. It can be
initialized with a \code{data.frame} of chromatographic data via the \code{chromData}
parameter and a \code{list} of \code{data.frame} entries for peaks data using the
\code{peaksData} parameter. These data can be accessed with the \code{chromData()} and
\code{peaksData()} functions.
}
\examples{

## Method 1: Initialize backend directly
cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    mz = c(112.2, 123.3, 134.4),
    dataOrigin = c("mem1", "mem2", "mem3")
)

pdata <- list(
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    ),
    data.frame(
        rtime = c(45.1, 46.2),
        intensity = c(100, 80.1)
    ),
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    )
)

cbm <- ChromBackendMemory()
cbm <- backendInitialize(cbm, chromData = cdata, peaksData = pdata)
cbm

## Method 2: Use Chromatograms constructor (recommended)
chr <- Chromatograms(ChromBackendMemory(), chromData = cdata, peaksData = pdata)
chr

}
\author{
Philippine Louail
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Chromatograms-chromData.R
\name{chromData}
\alias{chromData}
\alias{msLevel}
\alias{chromIndex}
\alias{collisionEnergy}
\alias{dataOrigin}
\alias{mz}
\alias{mzMin}
\alias{mzMax}
\alias{precursorMz}
\alias{precursorMzMin}
\alias{precursorMzMax}
\alias{productMz}
\alias{productMzMin}
\alias{productMzMax}
\alias{filterChromData}
\alias{chromVariables}
\alias{chromData<-}
\alias{msLevel<-}
\alias{chromIndex<-}
\alias{collisionEnergy<-}
\alias{dataOrigin<-}
\alias{mz<-}
\alias{mzMin<-}
\alias{mzMax<-}
\alias{precursorMz<-}
\alias{precursorMzMin<-}
\alias{precursorMzMax<-}
\alias{productMz<-}
\alias{productMzMin<-}
\alias{productMzMax<-}
\alias{chromVariables<-}
\alias{chromData,Chromatograms-method}
\alias{chromData<-,Chromatograms-method}
\alias{chromVariables,Chromatograms-method}
\alias{chromIndex,Chromatograms-method}
\alias{chromIndex<-,Chromatograms-method}
\alias{collisionEnergy,Chromatograms-method}
\alias{collisionEnergy<-,Chromatograms-method}
\alias{dataOrigin,Chromatograms-method}
\alias{dataOrigin<-,Chromatograms-method}
\alias{msLevel,Chromatograms-method}
\alias{msLevel<-,Chromatograms-method}
\alias{mz,Chromatograms-method}
\alias{mz<-,Chromatograms-method}
\alias{mzMax,Chromatograms-method}
\alias{mzMax<-,Chromatograms-method}
\alias{mzMin,Chromatograms-method}
\alias{mzMin<-,Chromatograms-method}
\alias{length,Chromatograms-method}
\alias{precursorMz,Chromatograms-method}
\alias{precursorMz<-,Chromatograms-method}
\alias{precursorMzMin,Chromatograms-method}
\alias{precursorMzMin<-,Chromatograms-method}
\alias{precursorMzMax,Chromatograms-method}
\alias{precursorMzMax<-,Chromatograms-method}
\alias{productMz,Chromatograms-method}
\alias{productMz<-,Chromatograms-method}
\alias{productMzMin,Chromatograms-method}
\alias{productMzMin<-,Chromatograms-method}
\alias{productMzMax,Chromatograms-method}
\alias{productMzMax<-,Chromatograms-method}
\alias{filterChromData,Chromatograms-method}
\title{Chromatographic Peaks Metadata.}
\usage{
\S4method{chromData}{Chromatograms}(object, columns = chromVariables(object), drop = FALSE)

\S4method{chromData}{Chromatograms}(object) <- value

\S4method{chromVariables}{Chromatograms}(object)

\S4method{chromIndex}{Chromatograms}(object)

\S4method{chromIndex}{Chromatograms}(object) <- value

\S4method{collisionEnergy}{Chromatograms}(object)

\S4method{collisionEnergy}{Chromatograms}(object) <- value

\S4method{dataOrigin}{Chromatograms}(object)

\S4method{dataOrigin}{Chromatograms}(object) <- value

\S4method{msLevel}{Chromatograms}(object)

\S4method{msLevel}{Chromatograms}(object) <- value

\S4method{mz}{Chromatograms}(object)

\S4method{mz}{Chromatograms}(object) <- value

\S4method{mzMax}{Chromatograms}(object)

\S4method{mzMax}{Chromatograms}(object) <- value

\S4method{mzMin}{Chromatograms}(object)

\S4method{mzMin}{Chromatograms}(object) <- value

\S4method{length}{Chromatograms}(x)

\S4method{precursorMz}{Chromatograms}(object)

\S4method{precursorMz}{Chromatograms}(object) <- value

\S4method{precursorMzMin}{Chromatograms}(object)

\S4method{precursorMzMin}{Chromatograms}(object) <- value

\S4method{precursorMzMax}{Chromatograms}(object)

\S4method{precursorMzMax}{Chromatograms}(object) <- value

\S4method{productMz}{Chromatograms}(object)

\S4method{productMz}{Chromatograms}(object) <- value

\S4method{productMzMin}{Chromatograms}(object)

\S4method{productMzMin}{Chromatograms}(object) <- value

\S4method{productMzMax}{Chromatograms}(object)

\S4method{productMzMax}{Chromatograms}(object) <- value

\S4method{filterChromData}{Chromatograms}(
  object,
  variables = character(),
  ranges = numeric(),
  match = c("any", "all"),
  keep = TRUE
)
}
\arguments{
\item{object}{A \link{Chromatograms} object.}

\item{columns}{A \code{character} vector of chromatograms variables to extract.}

\item{drop}{A \code{logical} indicating whether to drop dimensions when
extracting a single variable.}

\item{value}{replacement value for \verb{<-} methods. See individual
method description or expected data type.}

\item{x}{A \link{Chromatograms} object.}

\item{variables}{For \code{filterChromData()}: \code{character} vector with the names
of the chromatogram variables to filter for. The list of available
chromatogram variables can be obtained with \code{chromVariables()}.}

\item{ranges}{For \code{filterChromData()} : a \code{numeric}
vector of paired values (upper and lower boundary) that define the
ranges to filter the \code{object}. These paired values need to be in the
same order as the \code{variables} parameter (see below).}

\item{match}{For \code{filterChromData()} : \code{character(1) }
defining whether the condition has to match for all provided
\code{ranges} (\code{match = "all"}; the default), or for any of them
(\code{match = "any"}) for chromatogram data to be retained.}

\item{keep}{For \code{filterChromData()}: \code{logical(1)}
defining whether to keep (\code{keep = TRUE}) or remove (\code{keep = FALSE})
the chromatogram data that match the condition.}
}
\value{
Refer to the individual function description for information on the
return value.
}
\description{
As explained in the \code{\link{Chromatograms}} class documentation, the
\code{Chromatograms} object is a container for chromatogram data that includes
chromatographic peaks data (\emph{retention time} and related intensity values,
also referred to as \emph{peaks data variables} in the context of
\code{Chromatograms}) and metadata of individual chromatograms (so called
\emph{chromatograms variables}).

The \emph{chromatograms variables} information can be accessed using the
\code{chromData()} function. it is also possible to access specific
chromatograms variables using \code{$}.

\verb{@chromData} can be accessed, replaced but also filtered/subsetted. Refer to
the sections below for more details.
}
\section{Chromatograms variables and accessor functions}{


The following chromatograms variables are guaranteed to be provided by a
\code{Chromatograms} object and to be accessible with either the \code{chromData()} or
a specific function named after the variables names:
\itemize{
\item \code{chromIndex}: an \code{integer} with the index of the chromatogram in the
original source file (e.g. \emph{mzML} file).
\item \code{collisionEnergy}: for SRM data, \code{numeric} with the collision energy of
the precursor.
\item \code{dataOrigin}: optional \code{character} with the origin of the data.
\item \code{msLevel}: \code{integer} defining the MS level of the data.
\item \code{mz}: optional \code{numeric} with the (target) m/z value for the
chromatographic data.
\item \code{mzMin}: optional \code{numeric} with the lower m/z value of the m/z range in
case the data (e.g. an extracted ion chromatogram EIC) was extracted from
a \code{Spectra} object.
\item \code{mzMax}: optional \code{numeric} with the upper m/z value of the m/z range.
\item \code{precursorMz}: for SRM data, \code{numeric} with the target m/z of the
precursor (parent).
\item \code{precursorMzMin}: for SRM data, optional \code{numeric} with the lower m/z of
the precursor's isolation window.
\item \code{precursorMzMax}: for SRM data, optional \code{numeric} with the upper m/z of
the precursor's isolation window.
\item \code{productMz} for SRM data, \code{numeric} with the target m/z of the
product ion.
\item \code{productMzMin}: for SRM data, optional \code{numeric} with the lower m/z of
the product's isolation window.
\item \code{productMzMax}: for SRM data, optional \code{numeric} with the upper m/z of
the product's isolation window.
}
}

\section{Filter Chromatograms variables}{


Functions that filter \code{Chromatograms} based on chromatograms variables
(i.e, \verb{@chromData} ) will remove chromatographic data that do not meet the
specified conditions. This means that if a chromatogram is filtered out, its
corresponding \verb{@chromData} and \verb{@peaksData} will be removed from the object
immediately.

The available functions to filter chromatogram data are:
\itemize{
\item \code{filterChromData()}: Filters numerical chromatographic data variables
based on the provided numerical \code{ranges}. The method returns a
\code{Chromatograms} object containing only the chromatograms that match the
specified conditions. This function results in an object with fewer
chromatograms than the original.
}
}

\examples{

# Create a Chromatograms object
cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    mz = c(112.2, 123.3, 134.4),
    chromIndex = c(1L, 2L, 3L)
)

be <- backendInitialize(new("ChromBackendMemory"), chromData = cdata)

chr <- Chromatograms(be)

# Access chromatograms variables
chromData(chr)

# Access specific chromatograms variables
chromData(chr, columns = "msLevel")

msLevel(chr)

# Replace chromatograms variables
msLevel(chr) <- c(1L, 2L, 2L)

# Filter chromatograms variables
filterChromData(chr,
    variables = "msLevel", ranges = c(1L, 1L),
    keep = FALSE
)

}
\seealso{
\link{Chromatograms} for a general description of the \code{Chromatograms}
object.
\link{peaksData} for a general description of the chromatographic peaks
data available in the object, as well as how to access, replace and
subset them.
\link{processingQueue} for more information on the queuing
of processings and parallelization for larger dataset processing.
}
\author{
Philippine Louail
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Chromatograms-processingQueue.R
\name{processingQueue}
\alias{processingQueue}
\alias{applyProcessing,Chromatograms-method}
\alias{addProcessing,Chromatograms-method}
\alias{processingChunkSize,Chromatograms-method}
\alias{processingChunkSize<-,Chromatograms-method}
\alias{processingChunkFactor,Chromatograms-method}
\title{Efficiently processing \code{Chromatograms} objects.}
\usage{
\S4method{applyProcessing}{Chromatograms}(
  object,
  f = processingChunkFactor(object),
  BPPARAM = bpparam(),
  ...
)

\S4method{addProcessing}{Chromatograms}(object, FUN, ...)

\S4method{processingChunkSize}{Chromatograms}(object, ...)

\S4method{processingChunkSize}{Chromatograms}(object) <- value

\S4method{processingChunkFactor}{Chromatograms}(object, chunkSize = processingChunkSize(object), ...)
}
\arguments{
\item{object}{A \code{Chromatograms} object.}

\item{f}{\code{factor} defining the grouping to split the \code{Chromatograms} object.}

\item{BPPARAM}{Parallel setup configuration. See \code{\link[BiocParallel:register]{BiocParallel::bpparam()}}
for more information.}

\item{...}{Additional arguments passed to the methods.}

\item{FUN}{For \code{addProcessing()}, a function to be added to the
\code{Chromatograms} object's processing queue.}

\item{value}{\code{integer(1)} defining the chunk size.}

\item{chunkSize}{\code{integer(1)} for \code{processingChunkFactor} defining the chunk
size. The default is the value stored in the \code{Chromatograms} object's
\code{processingChunkSize} slot.}
}
\value{
\code{processingChunkSize()} returns the currently defined processing
chunk size (or \code{Inf} if it is not defined). \code{processingChunkFactor()}
returns a \code{factor} defining the chunks into which \code{object} will be
split for (parallel) chunk-wise processing or a \code{factor} of length 0
if no splitting is defined.

Refer to the individual function description for information on the
return value.
}
\description{
The \code{processingQueue} of a \code{Chromatograms} object is a list of processing
steps (i.e., functions) that are stored within the object and applied only
when needed. This design allows data to be processed in a single step,
which is particularly useful for larger datasets. The processing queue
enables functions to be applied in a chunk-wise manner, facilitating
parallel processing and reducing memory demand.

Since the peaks data can be quite large, a processing queue is used to
ensure efficiency. Generally, the processing queue is applied either
temporarily when calling \code{peaksData()} or permanently when calling
\code{applyProcessing()}. As explained below the processing efficiency can be
further improved by enabling chunk-wise processing.
}
\note{
Some backends might not support parallel processing. For these, the
\code{backendBpparam()} function will always return a \code{SerialParam()} regardless
of how parallel processing was defined.
}
\section{Apply Processing}{


The \code{applyProcessing()} function applies the processing queue to the backend
and returns the updated \code{Chromatograms} object. The processing queue is a
list of processing steps applied to the chromatograms data. Each element in
the list is a function that processes the chromatograms data. To apply
processing to the peaks data, the backend must be set to a non-read-only
backend using the \code{setBackend()} function.
}

\section{Parallel and Chunk-wise Processing of \code{Chromatograms}}{


Many operations on \code{Chromatograms} objects, especially those involving the
actual peaks data (see \link{peaksData}), support chunk-wise processing. This
involves splitting the \code{Chromatograms} into smaller parts (chunks) that are
processed iteratively. This enables parallel processing by data chunk and
reduces memory demand since only the peak data of the currently processed
subset is loaded into memory. Chunk-wise processing, which is disabled by
default, can be enabled by setting the processing chunk size of a
\code{Chromatograms} object using the \code{processingChunkSize()} function to a value
smaller than the length of the \code{Chromatograms} object. For example, setting
\code{processingChunkSize(chr) <- 1000} will cause any data manipulation
operation on \code{chr}, such as \code{filterPeaksData()}, to be performed in parallel
for sets of 1000 chromatograms in each iteration.

Chunk-wise processing is particularly useful for \code{Chromatograms} objects
using an \emph{on-disk} backend or for very large experiments. For small datasets
or \code{Chromatograms} using an in-memory backend, direct processing might be
more efficient. Setting the chunk size to \code{Inf} will disable chunk-wise
processing.

Some backends may prefer a specific type of splitting and chunk-wise
processing. For example, the \code{ChromBackendMzR} backend needs to load MS data
from the original (mzML) files, so chunk-wise processing on a per-file basis
is ideal. The \link[=ChromBackend]{backendParallelFactor()} function for
\code{ChromBackend} allows backends to suggest a preferred data chunking by
returning a \code{factor} defining the respective data chunks. The
\code{ChromBackendMzR} returns a \code{factor} based on the \emph{dataOrigin}
chromatograms variable. A \code{factor} of length 0 is returned if no particular
preferred splitting is needed. The suggested chunk definition will be used
if no finite \code{processingChunkSize()} is defined. Setting the
\code{processingChunkSize} overrides \code{backendParallelFactor}.

Functions to configure parallel or chunk-wise processing:
\itemize{
\item \code{processingChunkSize()}: Gets or sets the size of the chunks for parallel
or chunk-wise processing of a \code{Chromatograms} object. With a value of
\code{Inf} (the default), no chunk-wise processing will be performed.
\item \code{processingChunkFactor()}: Returns a \code{factor} defining the chunks into
which a \code{Chromatograms} object will be split for chunk-wise (parallel)
processing. A \code{factor} of length 0 indicates that no chunk-wise processing
will be performed.
}
}

\examples{
# Create a Chromatograms object
cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    mz = c(112.2, 123.3, 134.4),
    chromIndex = c(1L, 2L, 3L)
)

pdata <- list(
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    ),
    data.frame(
        rtime = c(45.1, 46.2),
        intensity = c(100, 80.1)
    ),
    data.frame(
        rtime = c(12.4, 12.8, 13.2, 14.6),
        intensity = c(123.3, 153.6, 2354.3, 243.4)
    )
)

be <- backendInitialize(new("ChromBackendMemory"),
    chromData = cdata,
    peaksData = pdata
)

chr <- Chromatograms(be)

divide_intensities <- function(x, y, ...) {
    intensity(x) <- lapply(intensity(x), `/`, y)
    x
}

## Add the function to the procesing queue
chr <- addProcessing(chr, divide_intensities, y = 2)
chr

# Apply the processing queue
chr <- applyProcessing(chr)

}
\author{
Johannes Rainer, Philippine Louail
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/AllGenerics.R, R/ChromBackend-functions.R,
%   R/hidden_aliases.R, R/ChromBackendMemory.R, R/ChromBackendMzR.R,
%   R/ChromBackendSpectra.R, R/Chromatograms.R
\docType{methods}
\name{reset}
\alias{reset}
\alias{fillCoreChromVariables}
\alias{validChromData}
\alias{validPeaksData}
\alias{hidden_aliases}
\alias{backendMerge,ChromBackendMemory-method}
\alias{chromData,ChromBackendMemory-method}
\alias{chromData<-,ChromBackendMemory-method}
\alias{chromVariables,ChromBackendMemory-method}
\alias{peaksData,ChromBackendMemory-method}
\alias{peaksData<-,ChromBackendMemory-method}
\alias{peaksVariables,ChromBackendMemory-method}
\alias{backendParallelFactor,ChromBackendMemory-method}
\alias{isReadOnly,ChromBackendMemory-method}
\alias{lengths,ChromBackendMemory-method}
\alias{intensity,ChromBackendMemory-method}
\alias{rtime,ChromBackendMemory-method}
\alias{show,ChromBackendMemory-method}
\alias{supportsSetBackend,ChromBackendMemory-method}
\alias{[,ChromBackendMemory-method}
\alias{$,ChromBackendMemory-method}
\alias{$<-,ChromBackendMemory-method}
\alias{chromExtract,ChromBackendMemory-method}
\alias{show,ChromBackendMzR-method}
\alias{backendParallelFactor,ChromBackendMzR-method}
\alias{isReadOnly,ChromBackendMzR-method}
\alias{peaksData,ChromBackendMzR-method}
\alias{peaksData<-,ChromBackendMzR-method}
\alias{chromData<-,ChromBackendMzR-method}
\alias{supportsSetBackend,ChromBackendMzR-method}
\alias{intensity,ChromBackendMzR-method}
\alias{rtime,ChromBackendMzR-method}
\alias{lengths,ChromBackendMzR-method}
\alias{[,ChromBackendMzR-method}
\alias{chromExtract,ChromBackendMzR-method}
\alias{show,ChromBackendSpectra-method}
\alias{factorize,ChromBackendSpectra-method}
\alias{backendParallelFactor,ChromBackendSpectra-method}
\alias{isReadOnly,ChromBackendSpectra-method}
\alias{peaksData,ChromBackendSpectra-method}
\alias{peaksData<-,ChromBackendSpectra-method}
\alias{supportsSetBackend,ChromBackendSpectra-method}
\alias{intensity,ChromBackendSpectra-method}
\alias{rtime,ChromBackendSpectra-method}
\alias{lengths,ChromBackendSpectra-method}
\alias{[,ChromBackendSpectra-method}
\alias{chromExtract,ChromBackendSpectra-method}
\alias{show,Chromatograms-method}
\title{Fill data.frame with columns for missing core chromatogram variables.}
\usage{
reset(object, ...)

fillCoreChromVariables(x = data.frame())

validChromData(x = data.frame(), error = TRUE)

validPeaksData(x = list(), error = TRUE)

\S4method{backendMerge}{ChromBackendMemory}(object, ...)

\S4method{chromData}{ChromBackendMemory}(object, columns = chromVariables(object), drop = FALSE)

\S4method{chromData}{ChromBackendMemory}(object) <- value

\S4method{chromVariables}{ChromBackendMemory}(object)

\S4method{peaksData}{ChromBackendMemory}(object, columns = peaksVariables(object), drop = FALSE, ...)

\S4method{peaksData}{ChromBackendMemory}(object) <- value

\S4method{peaksVariables}{ChromBackendMemory}(object)

\S4method{backendParallelFactor}{ChromBackendMemory}(object, ...)

\S4method{isReadOnly}{ChromBackendMemory}(object)

\S4method{lengths}{ChromBackendMemory}(x)

\S4method{intensity}{ChromBackendMemory}(object)

\S4method{rtime}{ChromBackendMemory}(object)

\S4method{show}{ChromBackendMemory}(object)

\S4method{supportsSetBackend}{ChromBackendMemory}(object, ...)

\S4method{[}{ChromBackendMemory}(x, i, j, ..., drop = FALSE)

\S4method{$}{ChromBackendMemory}(x, name)

\S4method{$}{ChromBackendMemory}(x, name) <- value

\S4method{chromExtract}{ChromBackendMemory}(object, peak.table, by)

\S4method{show}{ChromBackendMzR}(object)

\S4method{backendParallelFactor}{ChromBackendMzR}(object, ...)

\S4method{isReadOnly}{ChromBackendMzR}(object)

\S4method{peaksData}{ChromBackendMzR}(
  object,
  columns = peaksVariables(object),
  drop = FALSE,
  BPPARAM = SerialParam(),
  ...
)

\S4method{peaksData}{ChromBackendMzR}(object) <- value

\S4method{chromData}{ChromBackendMzR}(object) <- value

\S4method{supportsSetBackend}{ChromBackendMzR}(object, ...)

\S4method{intensity}{ChromBackendMzR}(object)

\S4method{rtime}{ChromBackendMzR}(object)

\S4method{lengths}{ChromBackendMzR}(x)

\S4method{[}{ChromBackendMzR}(x, i, j, ..., drop = TRUE)

\S4method{chromExtract}{ChromBackendMzR}(object, peak.table, by, ...)

\S4method{show}{ChromBackendSpectra}(object)

\S4method{factorize}{ChromBackendSpectra}(object, factorize.by = c("msLevel", "dataOrigin"), ...)

\S4method{backendParallelFactor}{ChromBackendSpectra}(object, ...)

\S4method{isReadOnly}{ChromBackendSpectra}(object)

\S4method{peaksData}{ChromBackendSpectra}(object, columns = peaksVariables(object), drop = FALSE, ...)

\S4method{peaksData}{ChromBackendSpectra}(object) <- value

\S4method{supportsSetBackend}{ChromBackendSpectra}(object, ...)

\S4method{intensity}{ChromBackendSpectra}(object)

\S4method{rtime}{ChromBackendSpectra}(object)

\S4method{lengths}{ChromBackendSpectra}(x)

\S4method{[}{ChromBackendSpectra}(x, i, j, ..., drop = TRUE)

\S4method{chromExtract}{ChromBackendSpectra}(object, peak.table, by, ...)

\S4method{show}{Chromatograms}(object)
}
\arguments{
\item{object}{A \link{Chromatograms} object.}

\item{x}{\code{list} representing the peaks data of a \code{Chromatograms}}

\item{error}{\code{logical(1)} whether an error should be thrown (the default)
if one or more columns don't have the correct data type.}
}
\value{
input data frame \code{x} with missing core variables added (with the
correct data type).

If the core variables have all the correct data type: an empty character.
If one or more core variables (columns) have the wrong data type the
function either throws an error (with \code{error = TRUE}) or returns a
\code{character} specifying which variables/columns don't have the correct
type (for \code{error = FALSE}).

Not applicable
}
\description{
\code{fillCoreChromVariables()} fills a provided \code{data.frame}
with columns for eventually missing \emph{core} chromatogram variables.
The missing core variables are added as new columns with missing values
(\code{NA}) of the correct data type.
Use \code{\link[=coreChromVariables]{coreChromVariables()}} to list the set of core variables and their data
types.

\code{validChromData()} checks that columns, representing \emph{core} chromatogram
variables are of the correct data type.

For S4 methods that require a documentation entry but only clutter the
index.

This method returns the chromatographic data stored in the
backend. If not specified otherwise it will return all defined columns in
the chromData slot as well as adding the \code{coreChromVariables} missing with
NA values.
}
\examples{

## Define a data frame
a <- data.frame(msLevel = c(1L, 1L, 2L), other_column = "b")

## Add missing core chromatogram variables to this data frame
fillCoreChromVariables(a)

## The data.frame thus contains columns for all core chromatogram
## variables in the respective expected data type (but filled with
## missing values).

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/Chromatograms.R
\name{Chromatograms}
\alias{Chromatograms}
\alias{Chromatograms-class}
\alias{[,Chromatograms-method}
\alias{[<-,Chromatograms-method}
\alias{[[,Chromatograms-method}
\alias{[[<-,Chromatograms-method}
\alias{chromExtract}
\alias{filterEmptyChromatograms}
\alias{Chromatograms,ChromBackendOrMissing-method}
\alias{Chromatograms,Spectra-method}
\alias{setBackend,Chromatograms,ChromBackend-method}
\alias{$,Chromatograms-method}
\alias{$<-,Chromatograms-method}
\alias{factorize,Chromatograms-method}
\alias{chromExtract,Chromatograms-method}
\alias{filterEmptyChromatograms,Chromatograms-method}
\title{The Chromatograms class to manage and access chromatographic data}
\usage{
\S4method{Chromatograms}{ChromBackendOrMissing}(object = ChromBackendMemory(), processingQueue = list(), ...)

\S4method{Chromatograms}{Spectra}(
  object,
  summarize.method = c("sum", "max"),
  chromData = data.frame(),
  factorize.by = c("msLevel", "dataOrigin"),
  spectraVariables = character(),
  ...
)

\S4method{setBackend}{Chromatograms,ChromBackend}(
  object,
  backend,
  f = processingChunkFactor(object),
  BPPARAM = SerialParam(),
  ...
)

\S4method{$}{Chromatograms}(x, name)

\S4method{$}{Chromatograms}(x, name) <- value

\S4method{[}{Chromatograms}(x, i, j, ..., drop = FALSE)

\S4method{[[}{Chromatograms}(x, i, j, ...)

\S4method{[[}{Chromatograms}(x, i, j, ...) <- value

\S4method{factorize}{Chromatograms}(object, factorize.by = c("msLevel", "dataOrigin"), ...)

\S4method{chromExtract}{Chromatograms}(object, peak.table, by, ...)

\S4method{filterEmptyChromatograms}{Chromatograms}(object, ...)
}
\arguments{
\item{object}{A \link{Chromatograms} object.}

\item{processingQueue}{\link{list} a list of processing steps (i.e. functions) to
be applied to the chromatographic data. The processing steps are
applied in the order they are listed in the \code{processingQueue}.}

\item{...}{Additional arguments.}

\item{summarize.method}{For \code{Chromatograms} created with a \code{Spectra} object:
A \code{character} vector with the name of the function to be used to
summaries the spectra data intensity. The available methods are "sum"
and "max". The default is "sum".}

\item{chromData}{For \code{Chromatograms()} build from a \code{Spectra} object backend,
a \code{data.frame} with the chromatographic data. If not provided
(or if empty), a default \code{data.frame} with the core chromatographic
variables will be created.}

\item{factorize.by}{A \code{character} vector with the names of the variables in
the \code{Spectra} object and the \code{chromData} slot that should be used
to factorize the \code{Spectra} object data to generate the
chromatographic data.}

\item{spectraVariables}{A \code{character} vector specifying which variables
from the \code{Spectra} object should be added to the chromData. These
will be mapped using the \code{chromSpectraIndex} variable.}

\item{backend}{\link{ChromBackend} object providing the raw data for the
\code{Chromatograms} object.}

\item{f}{\code{factor} defining the grouping to split the \code{Chromatograms} object.}

\item{BPPARAM}{Parallel setup configuration. See \code{\link[BiocParallel:register]{BiocParallel::bpparam()}}
for more information.}

\item{x}{A \link{Chromatograms} object.}

\item{name}{A \code{character} string specifying the name of the variable to
access.}

\item{value}{The value to replace the variable with.}

\item{i}{For \code{[}: \code{integer}, \code{logical} or \code{character} to subset the object.}

\item{j}{For \code{[} and \code{[[}: ignored.}

\item{drop}{For \code{[}: \code{logical(1)} default to \code{FALSE}.}

\item{peak.table}{For \code{chromExtract()} A \verb{data frame} containing the
following minimum columns:
- rtMin: Minimum retention time for each peak. Cannot be NA.
- rtMax: Maximum retention time for each peak. Cannot be NA.
- mzMin: Minimum m/z value for each peak.
- mzMax: Maximum m/z value for each peak.
Additionally, the \code{peak.table} must include columns that uniquely
identify chromatograms in the \code{object}. Common choices are
"msLevel" and/or "dataOrigin". These columns must also be present
in the \code{chromData} of the \code{object}. Any extra columns in
\code{peak.table} will be added to the \code{chromData} of the newly created
object.}

\item{by}{A \code{character} vector naming one or more columns that uniquely
identify chromatograms in both \code{peak.table} and
\code{chromData(object)}. The combination of these columns must be unique
within \code{chromData(object)}. Typically includes \code{"dataOrigin"},
\code{"msLevel"}, or both.}
}
\value{
Refer to the individual function description for information on
the return value.
}
\description{
The \code{Chromatograms} class encapsules chromatographic data and related
metadata. The chromatographic data is represented by a \emph{backend} extending
the virtual \link{ChromBackend} class which provides the raw data to the
\code{Chromatograms} object. Different backends and their properties are
described in the \link{ChromBackend} class documentation.

\strong{Available Backends}: The package provides several backends:
\itemize{
\item \code{ChromBackendMemory}: Stores data in memory (default, ideal for small datasets).
\item \code{ChromBackendMzR}: Reads peaks data from raw MS files on demand.
\item \code{ChromBackendSpectra}: Generates chromatographic data from a \code{Spectra} object.
This backend supports both in-memory and file-backed \code{Spectra} objects, using
an internal \code{spectraSortIndex} to avoid physically reordering the spectra.
}
}
\note{
This needs to be discussed, if we want for example to be able to set a
a backend to \code{ChromBackendMzR} we need to implement backendInitialize()
better. = Support peaksData and chromData as arguments AND have a way to
write .mzml files (which we do not have for chromatographic data).
}
\section{Creation of objects}{


\code{Chromatograms} objects can be created using the \code{Chromatograms()}
construction function. Either by providing a \code{ChromBackend} object or by
providing a \code{Spectra} object. The \code{Spectra} object will be used to generate
a \code{Chromatograms} object with a backend of class \code{\link{ChromBackendSpectra}}.
}

\section{Data stored in a \code{Chromatograms} object}{


The \code{Chromatograms} object is a container for chromatographic data, which
includes peaks data (\emph{retention time} and related intensity values, also
referred to as \emph{peaks data variables} in the context of \code{Chromatograms}) and
metadata of individual chromatogram (so called \emph{chromatograms variables}).
While a core set of chromatograms variables (the
\code{coreChromatogramsVariables()}) and peaks data variables (the
\code{corePeaksVariables()}) are guaranteed to be provided by a \code{Chromatograms},
it is possible to add arbitrary variables to a \code{Chromatograms} object.

The \code{Chromatograms} object is designed to contain chromatographic data of a
(large) set of chromatograms. The data is organized \emph{linearly} and can be
thought of a list of chromatograms, i.e. each element in the \code{Chromatograms}
is one chromatogram.

The \emph{chromatograms variables} information in the \code{Chromatograms} object can
be accessed using the \code{chromData()} function. Specific chromatograms
variables can be accessed by either precising the \code{"columns"} parameter in
\code{chromData()} or using \code{$}. \verb{@chromData} can be accessed, replaced but
also filtered/subsetted. Refer to the \link{chromData} documentation for more
details.

The \emph{peaks data variables} information in the \code{Chromatograms} object can be
accessed using the \code{peaksData()} function. Specific peaks variables can be
accessed by either precising the \code{"columns"} parameter in \code{peaksData()} or
using \code{$}. \verb{@peaksData} can be accessed, replaced but also
filtered/subsetted. Refer to the \link{peaksData} documentation for more details.
}

\section{Processing of \code{Chromatograms} objects}{


Functions that process the chromatograms data in some ways can be applied to
the object either directly or by using the \code{processingQueue} mechanism. The
\verb{@processingQueue} is a list of processing steps that are stored within the
object and only applied when needed. This was created so that the data can be
processed in a single step and is very useful for larger datasets. This is
even more true as this processing queue will call function that can be
applied on the data in a chunk-wise manner. This allows for parallel
processing of the data and reduces the memory demand. To read more about the
\code{processingQueue}, and how to parallelize your processes, see the
\link{processingQueue} documentation.
}

\section{Subsetting and accessing data}{

The \code{Chromatograms} class supports subsetting by chromatogram (i.e. rows) using
the \code{[} operator. The \code{[} operator does not support subsetting by columns.
Specific chromatograms or peaks variables can be accessed using the \code{[[}
operator or the \code{$} operator. The \code{[[} operator can also be used to
replace specific chromatograms or peaks variables.
}

\section{Changing the backend}{

The \code{setBackend()} function can be used to change the backend of a
\code{Chromatograms} object. This can be useful to switch to a backend that
better suits the needs of the user, for example switching to a memory-based
backend for smaller datasets or to a file-based backend for larger datasets.
The \code{setBackend()} function supports parallelization of the backend
conversion using the \code{BPPARAM} parameter. Note that any queued processing
steps are applied during the backend switch (since \code{peaksData()} is called
to transfer the data) and the processing queue is emptied afterwards.
}

\section{Filtering chromatograms}{

\itemize{
\item \code{filterEmptyChromatograms()}: removes empty chromatograms (i.e.
chromatograms without peaks) from the object. Returns the filtered
\code{Chromatograms} object (with chromatograms in their original order).
}
}

\section{Extracting chromatograms based on a peak table}{


The \code{chromExtract()} function allows users to extract specific regions of
interest from a \code{Chromatograms} object based on a user-provided peak table.
Each row in the \code{peak.table} defines a region to extract, using minimum and
maximum retention time (and m/z in the case of \code{chromBackendSpectra})
boundaries, and identifiers that uniquely match chromatograms in the object.

The resulting \strong{new} \code{Chromatograms} object contains only chromatograms
overlapping the specified regions, with updated metadata reflecting the
extracted boundaries.

This function is most commonly used to subset chromatographic data around
detected peaks or predefined time/mass ranges, for example to reprocess,
visualize, or quantify extracted chromatograms corresponding to known
features. It's important to notes that filtering by m/z is only supported
when using a \code{ChromBackendSpectra} backend. if the \code{mzMin} and \code{mzMax}
columns are provided when using other backends, they will be ignored.
}

\examples{

## Create a Chromatograms object with ChromBackendMemory
cdata <- data.frame(
    msLevel = c(1L, 1L, 1L),
    mz = c(112.2, 123.3, 134.4),
    dataOrigin = c("mem1", "mem2", "mem3")
)
pdata <- list(
    data.frame(rtime = c(2.1, 2.5, 3.0, 3.4, 3.9),
               intensity = c(100, 250, 400, 300, 150)),
    data.frame(rtime = c(3.5, 4.0, 4.5),
               intensity = c(80, 120, 90)),
    data.frame(rtime = c(5.1, 5.8, 6.3, 6.9, 7.5),
               intensity = c(80, 500, 1200, 600, 120))
)
chr <- Chromatograms(ChromBackendMemory(), chromData = cdata, peaksData = pdata)
chr

## Create a Chromatograms object from a Spectra object
library(MsBackendMetaboLights)
library(Spectra)

be <- backendInitialize(MsBackendMetaboLights(),
    mtblsId = "MTBLS39",
    filePattern = c("63B.cdf")
)
s <- Spectra(be)
s <- setBackend(s, MsBackendMemory())
chr <- Chromatograms(s)

## Subset
chr[1:2]

## Access a specific variable
chr[["msLevel"]]
chr$msLevel

## Replace data of a specific variable
chr$msLevel <- c(2L, 2L, 2L)

## Re-factorize the data
chr <- factorize(chr)

## Change the backend to memory
chr <- setBackend(chr, ChromBackendMemory())

}
\seealso{
\link{chromData} for a general description of the chromatographic
metadata available in the object, as well as how to access, replace
and subset them.
\link{peaksData} for a general description of the chromatographic peaks
data available in the object, as well as how to access, replace and
subset them.
\link{processingQueue} for more information on the queuing of
processings and parallelization for larger dataset.
}
